#ifndef _RXCIO_H_WRAP_
#define _RXCIO_H_WRAP_

#include <stdio.h>
#include <iostream.h>
#include <RxPrintSpec.h>
#include <RxHashTab.h>

class RxCoutput
{
	int type_;
	int nchars_;
	union {
	  FILE *fp;
	  char *buf;
	  ostream *osp;
	  } u;
	int nmax;
public:
	enum {
	  OUT_FILE,
	  STRING,
	  STRING_N,
	  OSTREAM	  
	  } ;
	int type() const
	 { return type_;
	 }
	const int &nchars() const
	 { return nchars_;
	 }

	void setStream (FILE *fp);
	void setStream (char *buf, int n=0);
	void setStream (ostream *osp);
	void *getStream () const;

	RxCoutput ()
	 {
	 }
	RxCoutput (FILE *fp)
	 { setStream (fp);
	 }
	RxCoutput (char *buf)
	 { setStream (buf);
	 }
	RxCoutput (char *buf, int n)
	 { setStream (buf, n);
	 }
	RxCoutput (ostream *osp)
	 { setStream (osp);
	 }
	
	int output_double (const char *fmt, double val);
	int output_float (const char *fmt, float val);
	int output_int (const char *fmt, int val);
	int output_str (const char *fmt, const char *val);
	int output_str (const char *s);
	int output_char (char c, int n=1);
	int output_newline (int leadspacing=0);
	int output_doubleArray (const char *fmt0, const char *fmt1,
				const char *fmt2, double *vec, int len);
};

class RxCinput
{
	int type_;
	union {
	  FILE *fp;
	  const char *buf;
	  istream *isp;
	} u;
	int nchars_;
	int status_;
	int *cntp_;

	RxStrHashTab *htab;		// used by RxParse

	void init();
public:
	enum {
	  IN_FILE,
	  IN_FILE_X,
	  STRING,
	  ISTREAM
	  } ;
	int type() const
	 { return type_;
	 }
	const int &nchars() const
	 { return nchars_;
	 }
	void setCnt (int val)
	 { if (cntp_)
	    { *cntp_ = val;
	    }
	 }
	const int &status() const
	 { return status_;
	 }
	void setStatus (int val)
	 { status_ = val;
	 }
	void setStream (FILE *fp, int useXscan=0);
	void setStream (const char *buf, int *cntp=NULL);
	void setStream (istream *isp);
	void *getStream () const;

	RxCinput()
	 { init();
	 }
	RxCinput (FILE *fp, int useXscan=0)
	 { init();
	   setStream (fp, useXscan);
	 }
	RxCinput (const char *buf, int *cntp=NULL)
	 { init();
	   setStream (buf, cntp);
	 }
	RxCinput (istream *isp)
	 { init();
	   setStream (isp);
	 }

	int input_double (double *val);
	int input_float (float *val);
	int input_int (int *val);
	int input_str (char *val);
	int input_str (char *val, int size);
	int getChar ();
	int ungetChar (int c); 
	int peekChar ()
	 { int c = getChar();
	   return ungetChar(c);
	 }
	int eatWhiteSpace ();

	int setHashTab  (int size);
	RxStrHashTab *getHashTab()
	 { return htab;
	 }
};

#endif

#if 0
print (FILE *fp, fmt)
{
	RxCoutput co (RxCio::fprintf, fp);
	RxCfmtList flist (fmt);
	return print (co, flist);
}

print (char *buf, fmt)
{
	RxCio co (RxCio::sprintf, buf);
	RxCfmtList flist (fmt);
	return print (co, flist);
}

for vector:

print (RxCoutput &co, RxCfmtList &flist)
{
	RxCfmtNode *cf, *cf0;

	co->doPrefix(flist);
	cf0 = flist.middle.head();
	if (cf0 == NULL)
	 { cf0 = new RxCfmtNode ("%g");
	 }
	cf = cf0;
	for (each element)
	 { co->doDouble (cf, elem);
	   cf = cf->next();
	   if (not last element)
	    { if (cf && cf->type() == str)
	       { co->doStr (cf);
		 cf = cf->next();
	       }
	      else
	       { co->doStr (" ");
	       }
	    }
	   if (cf == NULL)
	    { cf = cf0;
	    }
	 }
	co->doSuffix(flist);
	return co->nchars();
}

for edge:

print (RxCoutput *co, RxCfmtList &flist)
{
	co->doPrefix();
	lse.v->print (co);
	rse.v->print (co);
	co->doSuffix();
	return co->nchars();
}

for face:

print (RxCoutput *co)
{
	co->doPrefix();
	for (each vertex)
	 { vrtx->print (co);
	 }
	co->doSuffix();
	return co->nchars();
}


we have a format list

S F S F F F S F S
  ^ 

doPrim:
	
	if (S)
	 doStr (S)
	if (
NULL

	print ("%g");

S

	printf S, %g

S F

	printf (
S F

S F F S F S F S F F F S F 
F F F F F F F
S F F F S F F S F
S F S
S F F S F S F S
S F F F S F F F S

doPrim:
	if (S)

S F F F S F S F S F 

#endif
