#ifndef _RXDISTREPORT3_H_WRAP_
#define _RXDISTREPORT3_H_WRAP_

#include <RxVecd3.h>
#include <RxBuffer.h>
#include <RxSlist.h>
#include <RxDlist.h>
#include <RxPrintable.h>
#include <RxSolidType3.h>
#include <RxCvxPrim3.h>

typedef RxSnode<RxVecd3> RxVecd3Snode;
typedef RxSlist<RxVecd3Snode> RxVecd3Slist;

class RxSolidPrim3;

class RxDistRecord3 : public RxPrintable
{
	friend class RxSolidPrim3;
	friend class RxCvxPolyhed3;
	friend class RxDistReport3;
	friend class RxVrtx3;
	friend class RxEdge3;
private:
	RxFastAlloc **availp;
	RxVecd3Snode pnode0;
	RxVecd3Slist plist;
	int npts_;
	RxVecd3 n_;
	RxSolidType3Pair objs;
	RxCvxPrim3Pair feats;
	double d_;
	double atol;
	RxVecd3Snode *curNode;

	void init(RxFastAlloc **avp);
public:
	// access

	static const double EPS;

	int npts()
	 { return npts_;
	 }
	RxVecd3 *getPnt(int k);
	RxVecd3 *firstPnt()
	 { curNode = plist.head();
	   return &curNode->val;
	 }
	RxVecd3 *nextPnt()
	 { if (curNode)
	    { curNode = curNode->next();
	    }
	   return curNode ? &curNode->val : NULL;
	 }
	RxVecd3 *addPnt(const RxVecd3 &p);
	RxVecd3 *addPnt();
	void clearPnts();

	// construction

	RxDistRecord3()
	 { init(NULL);
	 }
	// used internally
	RxDistRecord3(RxFastAlloc **avp)
	 { init(avp);
	 }
	~RxDistRecord3();

	// assignment

	void set (const RxDistRecord3 &rec1);

	RxVecd3 *nrm()
	 { return &n_;
	 }
	const RxCvxPrim3 *const &f(int i) const
	 { return feats.p[i];
	 }
	const RxCvxPrim3 *&f(int i)
	 { return feats.p[i];
	 }
	const RxSolidType3 *const &obj(int i) const
	 { return objs.obj[i];
	 }
	const RxSolidType3 *&obj(int i)
	 { return objs.obj[i];
	 }
	double &d()
	 { return d_;
	 }
	const double &d() const
	 { return d_;
	 }
	double &angtol()
	 { return atol;
	 }

	void clear();

	virtual int oscan (class RxParseObj *obj);
	virtual int cprint (class RxCoutput *co, class RxPrintSpec *fs) const;

	int equal (const RxDistRecord3 &rec1, double eps=EPS) const;
};

typedef RxDnode<RxDistRecord3> RxDistRecord3Dnode;
typedef RxDlist<RxDistRecord3Dnode> RxDistRecord3Dlist;

#if 0
class RxDistReport3
{
	friend class RxSolidPrim3;

private:
	int maxrecs;
	int maxpnts;
	int nrecs_;
	RxDistRecord3 *recs;
	RxVecd3 *pntbuf;
	RxVecd3 *freepnts;

	double dtol;
	double atol;

	void init();
	void reset();
public:
	// access	
	
	RxDistRecord3 &operator()(int i)
	 { return recs[i];
	 }

	// setting and query

	int nrecs()
	 { return nrecs_;
	 }
	int overflow()
	 { return overflow_;
	 }
	void setDistTol (double dt)
	 { dtol = dt;
	 }
	double getDistTol()
	 { return dtol;
	 }
	void setAngTol (double at)
	 { atol = at;
	 }
	double getAngTol()
	 { return atol;
	 }
	void setBufs (int maxr, int maxp);

	RxDistReport3()
	 { init();
	   setBufs (1, 0);
	 }
	RxDistReport3()
	 { init();
	   setBufs (maxr, maxp);
	 }
	~RxDistReport3();

	// access to the first record

	int npts()
	 { return recs->npts();
	 }
	RxVecd3 &p()
	 { return recs->p();
	 }
	RxVecd3 &n()
	 { return recs->n();
	 }
	RxVecd3 &p(int n)
	 { return recs->p(n);
	 }
	RxSolidPrim3 *f1() const
	 { return recs->f1();
	 }
	RxSolidPrim3 *f2() const
	 { return recs->f2();
	 }
	double d() const
	 { return recs->d();
	 }


}
#endif

#endif
