#ifndef _RXEDGE3_H_WRAP_
#define _RXEDGE3_H_WRAP_

#include <RxCvxPrim3.h>
#include <RxDistReport3.h>
#include <RxVrtx3.h>
#include <RxDlist.h>
#include <RxClist.h>
#include <RxPlane3.h>

class RxSedg3 : public RxCnodeBase
{
	friend class RxCvxPolyhed3;
	friend class PolyTree;
	friend class RxEdge3;

	RxPlane3 vplane;	// plane in direction of u at vertex
	RxPlane3 pplane;	// plane perp to both 

	class RxFace3 *f;
	RxVrtx3 *v;
	RxSedg3 *adj;
	RxSedg3 *vtxlink;
	RxEdge3 *e;

	int idx;
	RxVecd3 u_;
	double len;

public:
	RxSedg3()
	 {
	 }
	RxSedg3 *nextAboutVrtx() const
	 { return next()->adj;
	 }
	RxVrtx3 *vrtx()
	 { return v;
	 }
	const RxVrtx3 *vrtx() const
	 { return v;
	 }
	RxFace3 *face()
	 { return f;
	 }
	const RxFace3 *face() const
	 { return f;
	 }
	RxEdge3 *edge()
	 { return e;
	 }
	const RxEdge3 *edge() const
	 { return e;
	 }
	RX_CNODE_DECLS (RxSedg3);
};

typedef RxClist<RxSedg3> RxSedg3List;

class RxEdge3 : public RxCvxPrim3
{
	friend class RxCvxPolyhed3;
	friend class PolyTree;
	int soft_;	/* implies that the edge is really smooth, or rounded */
public:
#define RX_EDGE3_ID	2
	static const int ID;

	RxSedg3 lse;
	RxSedg3 rse;

	void vinit();
	int parallel (const RxEdge3 &e2, const RxTrsf &T01, double atol) const;

	RxEdge3()
	 { type_ = ID;
	   soft_ = 0;
	 }
	const char *name() const;
	const char *typeName() const
	 { return "Edge3";
	 }
	
	void projectToPlane (class RxEdge2 *poly, const RxTrsf &TP) const;
	double project (RxVecd3 *qr, const RxVecd3 &q1);

	double dist (RxDistRecord3 *rec, RxVrtx3 &v1, const RxTrsf *T01=NULL);
	double dist (RxDistRecord3 *rec, RxEdge3 &e1, const RxTrsf *T01=NULL);

	int isSoft() const
	 { return soft_;
	 }
	int setSoft(int soft)
	 { soft_ = soft;
	 }

	int scan (FILE *fp);
	int xscan (FILE *fp);
	int scan (const char *buf, int *cntp=NULL);
	int scan (class RxCinput *ci);
	int scan ()
	 { return this->scan (stdin);
	 }
	int print (FILE *fp, const char *fmt) const;
	int print (char *buf, const char *fmt) const;
	int print (char *buf, int n, const char *fmt) const;
	int print (class RxCoutput *co, class RxPrintSpec *ps) const;
	int print(const char *fmt) const
	 { return this->print(stdout, fmt);
	 }

	ostream &print(ostream &os) const;
	istream &read(istream &is);
};

class RxEdge3ListNode : public RxDnodeBase
{
public:
	RxEdge3 e;
	RX_DNODE_DECLS (RxEdge3ListNode);
};

typedef RxDlist<RxEdge3ListNode> RxEdge3List;

inline ostream &operator<<(ostream &os, const RxEdge3 &e)
{
	return e.print(os);
}

inline istream &operator>>(istream &is, RxEdge3 &e)
{
	return e.read(is);
}

#endif
