#ifndef _RXLINK_H_WRAP_
#define _RXLINK_H_WRAP_

#include <RxPrintable.h>
#include <RxVecd.h>
#include <RxVecd6.h>
#include <RxTrsf.h>
#include <RxMatd.h>
#include <RxTwist.h>
#include <RxSpatialAccel.h>

class RxLink : public RxPrintable
{
public:
	int dof_;
	int type_;
	RxTrsf A_;
public:
	virtual RxLink *clone() const = 0;
	void copy (const RxLink &link)
	 { dof_ = link.dof_;
	   type_ = link.type_;
	   A_ = link.A_;
	 }

	virtual void getA (RxTrsf *Ap, double val) const = 0;
	virtual void getA (RxTrsf *Ap, const RxVecd &vec, int j=0) const
	 { getA (Ap, vec(j));
	 }

	virtual void setJnt (double val)
	 { getA (&A_, val);
	 }
	virtual void setJnt (const RxVecd &vec, int j=0)
	 { getA (&A_, vec, j);
	 }
	
	const RxTrsf &getA () const
	 { return A_;
	 }
	void getA (RxTrsf *Ap) const
	 { *Ap = A_;
	 }
	virtual void mulA (RxTrsf *Tp) const
	 { Tp->mul (A_);
	 }
	virtual void mulA (RxTrsf *Tp, double val) const
	 { RxTrsf A;
	   getA (&A, val);
	   Tp->mul (A);
	 }
	virtual void mulA (RxTrsf *Tp, const RxVecd &vec, int j=0) const
	 { RxTrsf A;
	   getA (&A, vec, j);
	   Tp->mul (A);
	 }
	
	virtual void Amul (RxTrsf *Tp) const
	 { Tp->mul (A_, *Tp);
	 }
	virtual void Amul (RxTrsf *Tp, double val) const
	 { RxTrsf A;
	   getA (&A, val);
	   Tp->mul (A, *Tp);
	 }
	virtual void Amul (RxTrsf *Tp, const RxVecd &vec, int j=0) const
	 { RxTrsf A;
	   getA (&A, vec, j);
	   Tp->mul (A, *Tp);
	 }

	virtual void getJcol (RxMatd *cols, const RxTrsf &U) const = 0;
	virtual void getJcol (RxMatd *cols, const RxTrsf &U, 
			      double val) const = 0; 
	virtual void getJcol (RxMatd *cols, const RxTrsf &U, 
			      const RxVecd &vec, int j=0) const
	 { getJcol (cols, U, vec(j));
	 }

	virtual void addVel (RxTwist *vel, double jvel) = 0;
	virtual void addAcc (RxSpatialAccel *acc, const RxTwist &vel,
			     double jvel, double jacc) = 0;

	int type() const
	 { return type_;
	 }
	int dof() const
	 { return dof_;
	 }
	
};

#endif
