#ifndef _RXLINKEDBODY_H_WRAP_
#define _RXLINKEDBODY_H_WRAP_

#include <RxCompoundBody.h>
#include <RxLink.h>
#include <RxVecd.h>
#include <RxMatd.h>
#include <RxTwist.h>
#include <RxAccel.h>

class RxLinkedBody : public RxCompoundBody
{
	RxLink *link;
	void init();

	virtual void copyBody (const RxCompoundBody &cmpb);

	virtual RxCompoundBody *newBody();
	virtual int oscanBody (const char *fname, class RxParseObj *&lobj, 
			       class RxParseObj *obj);
	virtual int cprintBody (class RxCoutput *co,
				class RxPrintSpec *ps) const;

	const RxLinkedBody **getBodyPath (const RxLinkedBody **buf, int n,
				    const RxLinkedBody &endlnk) const;

	int doGetVel (RxTwist *tw, const RxLinkedBody &endlnk,
		      const RxVecd &jvel, const RxVecd *jnts=NULL) const;
	int doGetAcc (RxAccel *acc, RxTwist *tw, const RxLinkedBody &endlnk, 
		      const RxVecd &jvel, const RxVecd &jacc, 
		      const RxVecd *jnts=NULL) const;
public:
#define RX_LINKED_BODY_ID	6
	static const int ID;

	RxLinkedBody()
	 { init();
	 }
	RxLinkedBody (RxCvxPolyhed3 *poly)
	 { init();
	   set (poly);
	 }
	~RxLinkedBody();

	void setLink (const RxLink *lnkp)
	 { link = lnkp->clone();
	 }

	RxLinkedBody *getLinkedBody (int i, int j=0);

	void setJnt (double jval);
	int setJnts (const RxVecd &jnts, int j=0);

	/** Multiply together all the link transforms from the current body
	  * to the body indicated by 'endlnk', inclusive. 
	  *
	  * @param	T	Returns the computed transform
	  * @param	endlnk	Denotes the final link in the sequence.
	  * @param	jnts	Optional argument, which if supplied
	  *			gives the joint values for all the links
				in the sequence. Otherwise, the internal
				joint settings are used.
	  * @return	1 if successful, or 0 if the current body is not an 
			ancestor of endlnk.
	  */
	int fwdK (RxTrsf *T, const RxLinkedBody &endlnk) const;
	int fwdK (RxTrsf *T, const RxLinkedBody &endlnk, 
			     const RxVecd &jnts) const;

	int getJ (RxMatd *J, const RxLinkedBody &endlnk) const;
	int getJ (RxMatd *J, const RxLinkedBody &endlnk, 
			     const RxVecd &jnts) const;

	int getVel (RxTwist *tw, const RxLinkedBody &endlnk, 
				 const RxVecd &jvel) const
	 { doGetVel (tw, endlnk, jvel);
	 }
	int getVel (RxTwist *tw, const RxLinkedBody &endlnk,
		    const RxVecd &jvel, const RxVecd &jnts) const
	 { doGetVel (tw, endlnk, jvel, &jnts);
	 }

	int getAcc (RxAccel *acc, RxTwist *tw, const RxLinkedBody &endlnk, 
		    const RxVecd &jvel, const RxVecd &jacc) const
	 { doGetAcc (acc, tw, endlnk, jvel, jacc);
	 }
	int getAcc (RxAccel *acc, RxTwist *tw, const RxLinkedBody &endlnk,
		    const RxVecd &jvel, const RxVecd &jacc,
		    const RxVecd &jnts) const
	 { doGetAcc (acc, tw, endlnk, jvel, jacc, &jnts);
	 }

	int numJnts() const;
};

#endif
