#ifndef _RXMATDN3X3_H_WRAP_
#define _RXMATDN3X3_H_WRAP_

#include <RxMatdBase.h>
#include <RxVecd3.h>
#include <stdarg.h>
#include <RxRot.h>

class RxMatd3x3 : public RxMatdBase
{
	friend class RxMatdLu;
	friend class RxVecd3;
	friend class RxMatdSvd;

	double store[9];
	static const int width;
	static const int nr;
	static const int nc;

	void doCopy (const RxMatd3x3 &m1);

    public:

	double &xx() { return store[0]; }
	const double &xx() const { return store[0]; }
	double &xy() { return store[1]; }
	const double &xy() const { return store[1]; }
	double &xz() { return store[2]; }
	const double &xz() const { return store[2]; }

	double &yx() { return store[3]; }
	const double &yx() const { return store[3]; }
	double &yy() { return store[4]; }
	const double &yy() const { return store[4]; }
	double &yz() { return store[5]; }
	const double &yz() const { return store[5]; }

	double &zx() { return store[6]; }
	const double &zx() const { return store[6]; }
	double &zy() { return store[7]; }
	const double &zy() const { return store[7]; }
	double &zz() { return store[8]; }
	const double &zz() const { return store[8]; }

	// assignment

	RxMatd3x3& set (const RxMatd3x3& m2)
	 { doCopy (m2);
	   return (*this);
	 }
	RxMatd3x3& set (const double *dp);
	RxMatd3x3& set (double e00, double e01, double e02, 
			 double e10, double e11, double e12,
			 double e20, double e21, double e22);
	RxMatd3x3& set (const RxMatdBase &m1);
	RxMatd3x3& setDiag (const RxVecd3 &d);
	RxMatd3x3& setDiag (double *buf);
	RxMatd3x3& setSymm (const RxVecd3 &d, const RxVecd3 &o);
	RxMatd3x3& setSymm (double *buf);

	RxMatd3x3& operator=(const RxMatd3x3& m2)
	 { return (set (m2));
	 }	   
	void get (double *dp) const;
	void getDiag (RxVecd3 *d) const;
	void getDiag (double *buf) const;
	void getSymm (RxVecd3 *d, RxVecd3 *o) const;
	void getSymm (double *buf) const;

	// constructors

	RxMatd3x3 ()
	 { 
	 }
	RxMatd3x3 (const RxMatd3x3 &m1)
	 { doCopy (m1);
	 }
	RxMatd3x3 (const RxMatdBase &m1)
	 { set (m1);
	 }
	RxMatd3x3 (const double *dp)
	 { set (dp);
	 }
	RxMatd3x3 (double e00, double e01, double e02, 
		    double e10, double e11, double e12,
		    double e20, double e21, double e22)
	 { set (e00, e01, e02, e10, e11, e12, e20, e21, e22);
	 }

	// query and access

	int nrows() const
	 { return nr;
	 }
	int ncols() const
	 { return nc;
	 }
	double *buf()
	 { return store;
	 }
	double *buf(int *wp)
	 { *wp = width;
	   return store;
	 }
	double& operator()(int i, int j)
	 { return store[width*i+j];
	 };
	double operator()(int i, int j) const
	 { return store[width*i+j];
	 };
	
	// T <- T,T operations

	RxMatd3x3& add (const RxMatd3x3 &m1, const RxMatd3x3 &m2);
	RxMatd3x3& add (const RxMatd3x3 &m2);
	RxMatd3x3& operator+= (const RxMatd3x3 &m2)
	 { return (add (m2));
	 }
	friend RxMatd3x3 operator+(const RxMatd3x3 &m1, const RxMatd3x3 &m2);

	RxMatd3x3& sub (const RxMatd3x3 &m1, const RxMatd3x3 &m2);
	RxMatd3x3& sub (const RxMatd3x3 &m2);
	RxMatd3x3& operator-= (const RxMatd3x3 &m2)
	 { return (sub (m2));
	 }
	friend RxMatd3x3 operator-(const RxMatd3x3 &m1, const RxMatd3x3 &m2);

	RxMatd3x3& mul (const RxMatd3x3 &m1, const RxMatd3x3 &m2);
	RxMatd3x3& mul (const RxMatd3x3 &m2)
	 { return (mul (*this, m2));
	 }
	RxMatd3x3& Tmul (const RxMatd3x3 &m1, const RxMatd3x3 &m2);
	RxMatd3x3& Tmul (const RxMatd3x3 &m2)
	 { return (Tmul (*this, m2));
	 }
	RxMatd3x3& mulT (const RxMatd3x3 &m1, const RxMatd3x3 &m2);
	RxMatd3x3& mulT (const RxMatd3x3 &m2)
	 { return (mulT (*this, m2));
	 }
	RxMatd3x3& operator*= (const RxMatd3x3 &m2)
	 { return (mul (*this, m2));
	 }
	friend RxMatd3x3 operator*(const RxMatd3x3 &m1, const RxMatd3x3 &m2);

	// T <- T,O operations

	RxMatd3x3& scale (double a, const RxMatd3x3 &m1);
	RxMatd3x3& scale (double a);
	RxMatd3x3& operator *= (double a)
	 { return (scale (a));
	 }
	RxMatd3x3& operator /= (double a)
	 { return (scale (1/a));
	 }
	friend RxMatd3x3 operator*(const RxMatd3x3 &m1, double a);	   
	friend RxMatd3x3 operator*(double a, const RxMatd3x3 &m1);
	friend RxMatd3x3 operator/(const RxMatd3x3 &m1, double a);

	RxMatd3x3& xform (const RxMatd3x3 &m1, const RxRot& R);
	RxMatd3x3& xform (const RxRot& R)
	 { return (xform (*this, R));
	 }
	RxMatd3x3& Ixform (const RxMatd3x3 &m1, const RxRot& R);
	RxMatd3x3& Ixform (const RxRot& R)
	 { return (Ixform (*this, R));
	 }

	RxMatd3x3& cross (const RxVecd3 &p, const RxMatd3x3 &m1);
	RxMatd3x3& cross (const RxMatd3x3 &m1, const RxVecd3 &p);

	// O <- T,O operations

	RxVecd3 &Vmul (RxVecd3 *vr, const RxVecd3 &v1) const;
	RxVecd3 &mulV (RxVecd3 *vr, const RxVecd3 &v1) const;
	friend RxVecd3 operator*(const RxMatd3x3 &m1, const RxVecd3 &v1);
	friend RxVecd3 operator*(const RxVecd3 &v1, const RxMatd3x3 &m1);
	   
	// T <- T operations

	RxMatd3x3 &tsp(const RxMatd3x3 &m1);
	RxMatd3x3 &tsp();
	RxMatd3x3 T()
	 { RxMatd3x3 mr;
	   return (mr.tsp(*this));
	 }

	RxMatd3x3 &inv(const RxMatd3x3 &m1);
	RxMatd3x3 &inv()
	 { return (inv (*this));
	 }
	RxMatd3x3 I()
	 { RxMatd3x3 mr;
	   return (mr.inv(*this));
	 }

	RxMatd3x3 &neg(const RxMatd3x3 &m1);
	RxMatd3x3 &neg()
	 { return (neg (*this));
	 }
	RxMatd3x3 operator-()
	 { RxMatd3x3 mr;
	   return (mr.neg(*this));
	 }

	// T <- O operations

	RxMatd3x3& setZero();
	RxMatd3x3& setIdent();

	// O <- T operations

	double det() const;
	double norm1() const;
	double normInf() const;
	double normF() const;
	double trace() const;	
	double condest() const;	

	// constants

	static const RxMatd3x3 UNIT;
	static const RxMatd3x3 ZERO;
};

inline RxVecd3 operator*(const RxMatd3x3 &m1, const RxVecd3 &v1)
{
	RxVecd3 vr;
	return (m1.mulV(&vr,v1));
}

inline RxVecd3 operator*(const RxVecd3 &v1, const RxMatd3x3 &m1)
{
	RxVecd3 vr;
	return (m1.Vmul(&vr,v1));
}

#endif
