#ifndef _RXREFMANAGER_H_WRAP_
#define _RXREFMANAGER_H_WRAP_

#include <stdlib.h>
#include <stddef.h>

/* This class provides a compact way to maintain control over
   the allocation of mulitply referenced objects, based on
   the standard method of keeping a reference count for each
   object. */

class RxRefManager
{
    private:
	int cnt;
    public:
	RxRefManager()
	 { cnt = 0;
	 }
	static void *alloc (size_t n)
	 { return malloc(n);
	 }
	static int dealloc (RxRefManager &ref, void *p, size_t)
	 { if (ref.cnt > 0)
	    { ref.cnt--;
	      return 0;
	    }
	   else
	    { free (p);
	      return 1;
	    }	   
	 }
	int inc()
	 { return cnt++;
	 }
	int get() const
	 { return cnt;
	 }
	int set(int val)
	 { int old = cnt;
	   cnt = val;
	   return old;
	 }
	int deleteOK()
	 { return cnt==0;
	 }
};

/* To use the reference manager within a particular class,
   add the following definitions to the class header:

class Foo
{
   private:
	RxRefManager ref;
   public:
	void *operator new (size_t n)
	 { return (Foo*)RxRefManager::alloc(n);
	 }
	void operator delete (void *p, size_t n)
	 { RxRefManager::dealloc (((Foo*)p)->ref, p, n);
	 }
	int reference()
	 { return ref.inc();
	 }
}

   Also, add the following code to the destructor:

Foo::~Foo()
{
	if (ref.deleteOK())
	 { // do stuff
	 }
}
*/

/* Note that at the moment, this code is *NOT* MT safe */

#endif
