#ifndef _RXTWIST_H_
#define _RXTWIST_H_

#include <RxSpatialVec.h>

class RxTwist: public RxSpatialVec
{
public:
	RxVecd3 &v()
	 { return b_;
	 }
	const RxVecd3 &v() const
	 { return b_;
	 }
	RxVecd3 &w()
	 { return a_;
	 }
	const RxVecd3 &w() const
	 { return a_;
	 }

	// assignment

	RxTwist& set (double vx, double vy, double vz,
		      double wx, double wy, double wz);
	RxTwist& set (const double *dp);
	void get (double *vxp, double *vyp, double *vzp,
		  double *wxp, double *wyp, double *wzp);
	void get (double *dp) const;

	// constructors

	RxTwist () : RxSpatialVec()
	 {
	 }
	RxTwist (double vx, double vy, double vz, 
		 double wx, double wy, double wz) : 
		RxSpatialVec()
	 { set (vx, vy, vz, wx, wy, wz);
	 }
	RxTwist (RxVecd3 &vv, RxVecd3 &ww) : RxSpatialVec()
	 { w() = ww;
	   v() = vv;
	 }
	RxTwist (const double *dp) : RxSpatialVec()
	 { set (dp);
	 }
#if 0
	RxTwist (const RxTwist& t2) : RxSpatialVec(t2)
	 {
	 }
#endif
	// query and access

	double& operator()(int i);
	double operator()(int i) const;

	// T <- T,T operations

	RxTwist& add (const RxTwist &t1, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::add (t1, t2);
	 }
	RxTwist& add (const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::add (t2);
	 }
	RxTwist& operator+=(const RxTwist &t2)
	 { return (add(t2));
	 }
	friend RxTwist operator+(const RxTwist &t1, const RxTwist &t2);

	RxTwist& sub (const RxTwist &t1, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::sub (t1, t2);
	 }
	RxTwist& sub (const RxTwist &t1)
	 { return (RxTwist&) RxSpatialVec::sub (t1);
	 }
	RxTwist& operator-=(const RxTwist &t2)
	 { return (sub(t2));
	 }

	friend RxTwist operator-(const RxTwist &t1, const RxTwist &t2);

	RxTwist& extrap (const RxTwist &t1, double a, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::extrap (t1, a, t2);
	 }
	RxTwist& extrap (double a, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::extrap (a, t2);
	 }
	RxTwist& interp (const RxTwist &t1, double a, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::interp (t1, a, t2);
	 }
	RxTwist& interp (double a, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::interp (a, t2);
	 }
#if 0
	RxTwist& cross (const RxTwist &t1, const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::cross (t1, t2);
	 }
	RxTwist& cross (const RxTwist &t2)
	 { return (RxTwist&) RxSpatialVec::cross (t2);
	 }
#endif
	// T <- T,O operations

	RxTwist& scale (double a, const RxTwist &t1)
	 { return (RxTwist&) RxSpatialVec::scale (a, t1);
	 }
	RxTwist& scale (double a)
	 { return (RxTwist&) RxSpatialVec::scale (a);
	 }
	RxTwist& operator*=(double a)
	 { return (scale(a));
	 }
	RxTwist& operator/=(double a)
	 { return (scale(1/a));
	 }

	friend RxTwist operator*(const RxTwist &t1, double a);
	friend RxTwist operator*(double a, const RxTwist &t1);
	friend RxTwist operator/(const RxTwist &t1, double a);

	RxTwist& xform (const RxTwist &t1, const RxRot& R)
	 { return (RxTwist&) RxSpatialVec::xform (t1, R);
	 }
	RxTwist& xform (const RxRot& R)
	 { return (RxTwist&) RxSpatialVec::xform (R);
	 }
	RxTwist& Ixform (const RxTwist &t1, const RxRot& R)
	 { return (RxTwist&) RxSpatialVec::Ixform (t1, R);
	 }
	RxTwist& Ixform (const RxRot& R)
	 { return (RxTwist&) RxSpatialVec::Ixform (R);
	 }
	RxTwist& xform (const RxTwist &t1, const RxTrsf& T)
	 { return (RxTwist&) RxSpatialVec::xform (t1, T);
	 }
	RxTwist& xform (const RxTrsf& T)
	 { return (RxTwist&) RxSpatialVec::xform (T);
	 }
	RxTwist& Ixform (const RxTwist &t1, const RxTrsf& T)
	 { return (RxTwist&) RxSpatialVec::Ixform (t1, T);
	 }
	RxTwist& Ixform (const RxTrsf& T)
	 { return (RxTwist&) RxSpatialVec::Ixform (T);
	 }

	// O <- T,T operations

	// T <- T operations

	RxTwist &unit(const RxTwist &t1)
	 { return (RxTwist&) RxSpatialVec::unit(t1);
	 }
	RxTwist &unit()
	 { return (RxTwist&) RxSpatialVec::unit();
	 }
	RxTwist &neg(const RxTwist &t1)
	 { return (RxTwist&) RxSpatialVec::neg(t1);
	 }
	RxTwist &neg()
	 { return (RxTwist&) RxSpatialVec::neg();
	 }
	friend RxTwist operator-(const RxTwist &t1);

	// T <- O operations

	RxTwist &setZero()
	 { return (RxTwist&) RxSpatialVec::setZero();
	 }

	// O <- T operations

	double norm1(double cv, double cw) const
	 { return RxSpatialVec::norm1 (cw, cv);
	 }
	double norm2(double cv, double cw) const
	 { return RxSpatialVec::norm2 (cw, cv);
	 }
	double normInf(double cv, double cw) const
	 { return RxSpatialVec::normInf (cw, cv);
	 }

	// constants

	static const RxTwist ZERO;
	static const RxTwist UNIT_VX;
	static const RxTwist UNIT_VY;
	static const RxTwist UNIT_VZ;
	static const RxTwist UNIT_WX;
	static const RxTwist UNIT_WY;
	static const RxTwist UNIT_WZ;
};

inline RxTwist operator+(const RxTwist &t1, const RxTwist &t2)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->add (*(RxSpatialVec*)&t1, *(RxSpatialVec*)&t2);
	return vr;
}

inline RxTwist operator-(const RxTwist &t1, const RxTwist &t2)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->sub (*(RxSpatialVec*)&t1, *(RxSpatialVec*)&t2);
	return vr;
}

inline RxTwist operator*(const RxTwist &t1, double a)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxTwist operator*(double a, const RxTwist &t1)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxTwist operator/(const RxTwist &t1, double a)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->scale (1/a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxTwist operator-(const RxTwist &t1)
{
	RxTwist vr;
	((RxSpatialVec*)&vr)->neg (*(RxSpatialVec*)&t1);
	return vr;
}

#endif
