

#ifndef _RXVECDN_H_
#define _RXVECDN_H_

#include <RxVecdBase.h>
#include <string.h>

class RxVecd: public RxVecdBase
{
	friend class RxMatd;
	friend class RxMatdLu;
	friend class RxMatdSvd;
	friend class RxMatdChol;

    protected:
	int len;
	int malloced;

    public:

	// constructors	

	RxVecd ()
	 { len = 0;
	   v = NULL;
	   malloced = 0;
	 }
	RxVecd (int l, double *buf)
	 { len = l;
	   v = buf;
	   malloced = 0;
	 }
	RxVecd (int l)
	 { len = l;
	   v = new double[l];
	   malloced = 1;
	 }
	RxVecd (const RxVecd &v2)
	 { len = v2.len;
	   v = new double[len];
	   malloced = 1;
	   doCopy (v2, len);
	 }
	RxVecd (const RxVecdBase &v2);
	RxVecd (double *dp, int l);
	~RxVecd ()
	 { if (malloced)
	    { delete[] v;
	    }
	 }

	void init (int l, double *buf)
	 { len = l;
	   v = buf;
	   malloced = 0;
	 }

	// query and access

	int length() const
	 { return (len);
	 }
	double *buf() const
	 { return v;
	 }
	double& operator()(int i)
	 { return (v[i]);
	 } 
	double operator()(int i) const
	 { return (v[i]);
	 } 

	// assignment

	RxVecd& operator=(const RxVecd &v2) 
	 { doCopy (v2, len);
	   return (*this);
	 }
	RxVecd& set (const double *dp);
	void get (double *dp);
	RxVecd& operator=(const RxMatd &m2);

	// T <- T,T operations

	RxVecd& add (const RxVecd &v1, const RxVecd &v2)
	 { doAdd (v1, v2, len); return (*this);
	 }
	RxVecd& add (const RxVecd &v2)
	 { doAdd (v2, len); return (*this);
	 }
	RxVecd& operator+=(const RxVecd &v2)
	 { doAdd (v2, len); return (*this);
	 }
	friend RxVecd operator+(const RxVecd &v1, const RxVecd &v2);

	RxVecd& sub (const RxVecd &v1, const RxVecd &v2)
	 { doSub (v1, v2, len); return (*this);
	 }
	RxVecd& sub (const RxVecd &v2)
	 { doSub (v2, len); return (*this);
	 }
	RxVecd& operator-=(const RxVecd &v2)
	 { doSub (v2, len); return (*this);
	 }
	friend RxVecd operator-(const RxVecd &v1, const RxVecd &v2);

	RxVecd& extrap (const RxVecd &v1, double a, const RxVecd &v2)
	 { doExtrap (v1, a, v2, len); return (*this);
	 }
	RxVecd& extrap (double a, const RxVecd &v2)
	 { doExtrap (*this, a, v2, len); return (*this);
	 }

	// T <- T,O operations

	RxVecd& scale (double a, const RxVecd &v1)
	 { doScale (a, v1, len); return (*this);
	 }
	RxVecd& scale (double a)
	 { doScale (a, len); return (*this);
	 }
	RxVecd& operator*=(double a)
	 { doScale (a, len); return (*this);
	 }
	RxVecd& operator/=(double a)
	 { doScale (1/a, len); return (*this);
	 }
	friend RxVecd operator*(const RxVecd &v1, double a);
	friend RxVecd operator*(double a, const RxVecd &v1);
	friend RxVecd operator/(const RxVecd &v1, double a);

	// O <- T,T operations

	double dot(const RxVecd &v2) const
	 { return (doDot (v2, len));
	 }

	// T <- T operations

	RxVecd &unit(const RxVecd &v1)
	 { doUnit (v1, len); return (*this);
	 }
	RxVecd &unit()
	 { doUnit (*this, len); return (*this);
	 }
	RxVecd &neg(const RxVecd &v1)
	 { doNeg (v1, len); return (*this);
	 }
	RxVecd &neg()
	 { doNeg (*this, len); return (*this);
	 }
	friend RxVecd operator-(const RxVecd &v1);

	// T <- O operations

	RxVecd &setZero()
	 { doZero (len); return (*this);
	 }

	// O <- T operations

	double norm1() const
	 { return (doNorm1(len));
	 }
	double norm2() const
	 { return (doNorm2(len));
	 }
	double normInf() const
	 { return (doNormInf(len));
	 }

	friend RxVecd operator*(const RxMatd &m1, const RxVecd &v1);
	friend RxVecd operator*(const RxVecd &v1, const RxMatd &m1);
};

inline RxVecd operator+(const RxVecd &v1, const RxVecd &v2)
{
	RxVecd vr(v1.len);
	vr.doAdd (v1, v2, vr.len);
	return (vr);
}

inline RxVecd operator-(const RxVecd &v1, const RxVecd &v2)
{
	RxVecd vr(v1.len);
	vr.doSub (v1, v2, vr.len);
	return (vr);
}

inline RxVecd operator*(const RxVecd &v1, double a)
{
	RxVecd vr(v1.len);
	vr.doScale (a, v1, vr.len);
	return (vr);
}

inline RxVecd operator*(double a, const RxVecd &v1)
{
	RxVecd vr(v1.len);
	vr.doScale (a, v1, vr.len);
	return (vr);
}

inline RxVecd operator/(const RxVecd &v1, double a)
{
	RxVecd vr(v1.len);
	vr.doScale (1/a, v1, vr.len);
	return (vr);
}

inline RxVecd operator-(const RxVecd &v1)
{
	RxVecd vr(v1.len);
	vr.doNeg (v1, vr.len);
	return (vr);
}

#endif
