#ifndef _RXVECDN3_H_WRAP_
#define _RXVECDN3_H_WRAP_

#include <RxVecd.h>

class RxTrsf;
class RxRot;

class RxVecd3: public RxVecd
{
	friend class RxMatd3x3;

	double vbuf[3];

    public:

	// access

	double &x()
	 { return vbuf[0];
	 }
	const double &x() const
	 { return vbuf[0];
	 }
	double &y()
	 { return vbuf[1];
	 }
	const double &y() const
	 { return vbuf[1];
	 }
	double &z()
	 { return vbuf[2];
	 }
	const double &z() const
	 { return vbuf[2];
	 }

	// query 
	
	int length()
	 { return 3;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 

	// assignment

	RxVecd3& set(const RxVecd3 &v2)
	 { x()=v2.x(); y()=v2.y(); z()=v2.z();
	   return (*this);
	 }
	RxVecd3& operator=(const RxVecd3 &v2) 
	 { return (set(v2));
	 }
	RxVecd3& set (double xx, double yy, double zz)
	 { x() = xx; y() = yy; z() = zz;
	   return (*this);
	 }
	void get (double *xp, double *yp, double *zp) const
	 { *xp = x(); *yp = y(); *zp = z();
	 }
	RxVecd3& set (const double *dp)
	 { x() = dp[0]; y() = dp[1]; z() = dp[2];
	   return (*this);
	 }
	void get (double *dp) const
	 { dp[0] = x(); dp[1] = y(); dp[2] = z();
	 }

	// constructors	

	RxVecd3 () : RxVecd (3, vbuf) 
	 { 
	 }
	RxVecd3 (double x, double y, double z) : RxVecd (3, vbuf)
	 { set (x, y, z);
	 }
	RxVecd3 (const RxVecd3 &v2) : RxVecd (3, vbuf)
	 { set (v2);
	 }
	RxVecd3 (double *dp) : RxVecd (3, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd3& add (const RxVecd3 &v1, const RxVecd3 &v2);
	RxVecd3& add (const RxVecd3 &v2);
	RxVecd3& operator+=(const RxVecd3 &v2)
	 { return (add (v2));
	 }
	friend RxVecd3 operator+(const RxVecd3 &v1, const RxVecd3 &v2);
	RxVecd3& addQ (const RxVecd3 &v1, const RxVecd3 &v2)
	 { x() = v1.x()+v2.x(); y() = v1.y()+v2.y(); z() = v1.z()+v2.z(); 
	   return (*this);
	 }

	RxVecd3& sub (const RxVecd3 &v1, const RxVecd3 &v2);
	RxVecd3& sub (const RxVecd3 &v2);
	RxVecd3& operator-=(const RxVecd3 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd3 operator-(const RxVecd3 &v1, const RxVecd3 &v2);
	RxVecd3& subQ (const RxVecd3 &v1, const RxVecd3 &v2)
	 { x() = v1.x()-v2.x(); y() = v1.y()-v2.y(); z() = v1.z()-v2.z(); 
	   return (*this);
	 }


	RxVecd3& cross (const RxVecd3 &v1, const RxVecd3 &v2);
	RxVecd3& cross (const RxVecd3 &v2)
	 { return (cross(*this,v2));
	 }

	RxVecd3& extrap (const RxVecd3 &v1, double a, const RxVecd3 &v2);
	RxVecd3& extrap (double a, const RxVecd3 &v2);
	RxVecd3& extrapQ (const RxVecd3 &v1, double a, const RxVecd3 &v2)
	 { x() = v1.x()+a*v2.x(); y() = v1.y()+a*v2.y(); z() = v1.z()+a*v2.z(); 
	   return (*this);
	 }

	RxVecd3& interp (const RxVecd3 &v1, double a, const RxVecd3 &v2);
	RxVecd3& interp (double a, const RxVecd3 &v2);
	RxVecd3& interpQ (const RxVecd3 &v1, double a, const RxVecd3 &v2)
	 { x() = (1-a)*v1.x()+a*v2.x(); y() = (1-a)*v1.y()+a*v2.y(); z() = (1-a)*v1.z()+a*v2.z(); 
	   return (*this);
	 }

	// T <- T,O operations

	RxVecd3& scale (double a, const RxVecd3 &v1);
	RxVecd3& scale (double a);
	RxVecd3& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd3& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd3 operator*(const RxVecd3 &v1, double a);
	friend RxVecd3 operator*(double a, const RxVecd3 &v1);
	friend RxVecd3 operator/(const RxVecd3 &v1, double a);
	RxVecd3& scaleQ (double a, const RxVecd3 &v1)
	 { x() = v1.x()*a; y() = v1.y()*a; z() = v1.z()*a;
	   return (*this);
	 }

	RxVecd3& xform (const RxVecd3 &v1, const RxRot& R);
	RxVecd3& xform (const RxRot& R)
	 { return (xform (*this, R));
	 }
	RxVecd3& Ixform (const RxVecd3 &v1, const RxRot& R);
	RxVecd3& Ixform (const RxRot& R)
	 { return (Ixform (*this, R));
	 }
	RxVecd3& xform (const RxVecd3 &v1, const RxTrsf& T);
	RxVecd3& xform (const RxTrsf& T)
	 { return (xform (*this, T));
	 }
	RxVecd3& Ixform (const RxVecd3 &v1, const RxTrsf& T);
	RxVecd3& Ixform (const RxTrsf& T)
	 { return (Ixform (*this, T));
	 }

	// O <- T,T operations

	double dot(const RxVecd3 &v2) const;
	double dotQ(const RxVecd3 &v2) const
	 { return (x()*v2.x() + y()*v2.y() + z()*v2.z());
	 }
	double dist(const RxVecd3 &v2) const
	 { RxVecd3 diff;
	   diff.sub(*this, v2);
	   return diff.norm2();
	 }
	int operator==(const RxVecd3 &v2) const
	 { return x() == v2.x() && y() == v2.y() && z() == v2.z();
	 }


	// T <- T operations

	RxVecd3 &unit(const RxVecd3 &v1);
	RxVecd3 &unit()
	 { return (unit(*this));
	 }
	RxVecd3 &neg(const RxVecd3 &v1);
	RxVecd3 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd3 operator-(const RxVecd3 &v1);

	// T <- O operations

	RxVecd3 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;

	// constants

	static const RxVecd3 ZERO;
	static const RxVecd3 I;
	static const RxVecd3 J;
	static const RxVecd3 K;
	static const RxVecd3 Ineg;
	static const RxVecd3 Jneg;
	static const RxVecd3 Kneg;
};

inline RxVecd3 operator+(const RxVecd3 &v1, const RxVecd3 &v2)
{
	RxVecd3 vr;
	return (vr.add (v1, v2));
}

inline RxVecd3 operator-(const RxVecd3 &v1, const RxVecd3 &v2)
{
	RxVecd3 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd3 operator*(const RxVecd3 &v1, double a)
{
	RxVecd3 vr;
	return (vr.scale (a, v1));
}

inline RxVecd3 operator*(double a, const RxVecd3 &v1)
{
	RxVecd3 vr;
	return (vr.scale (a, v1));
}

inline RxVecd3 operator/(const RxVecd3 &v1, double a)
{
	RxVecd3 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd3 operator-(const RxVecd3 &v1)
{
	RxVecd3 vr;
	return (vr.neg (v1));
}

#endif
