#ifndef _RXVECDN4_H_WRAP_
#define _RXVECDN4_H_WRAP_

#include <RxVecd.h>

class RxVecd4: public RxVecd
{
	double vbuf[4];

    public:

	// access

	double &x()
	 { return vbuf[0];
	 }
	const double &x() const
	 { return vbuf[0];
	 }
	double &y()
	 { return vbuf[1];
	 }
	const double &y() const
	 { return vbuf[1];
	 }
	double &z()
	 { return vbuf[2];
	 }
	const double &z() const
	 { return vbuf[2];
	 }
	double &w()
	 { return vbuf[3];
	 }
	const double &w() const
	 { return vbuf[3];
	 }

	// query 
	
	int length() const
	 { return 4;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 


	// assignment

	RxVecd4& operator=(const RxVecd4 &v2)
	 { set (v2.v);
	   return (*this);
	 }
	RxVecd4& set (const double *dp);
	void get (double *dp);

	// constructors	

	RxVecd4 () : RxVecd (4, vbuf) 
	 { 
	 }
	RxVecd4 (const RxVecd4 &v2) : RxVecd (4, vbuf)
	 { set (v2.v);
	 }
	RxVecd4 (double *dp) : RxVecd (4, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd4& add (const RxVecd4 &v1, const RxVecd4 &v2);
	RxVecd4& add (const RxVecd4 &v2);

	RxVecd4& operator+=(const RxVecd4 &v2)
	 { return (add (v2));
	 }
	friend RxVecd4 operator+(const RxVecd4 &v1, const RxVecd4 &v2);

	RxVecd4& sub (const RxVecd4 &v1, const RxVecd4 &v2);
	RxVecd4& sub (const RxVecd4 &v2);
	RxVecd4& operator-=(const RxVecd4 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd4 operator-(const RxVecd4 &v1, const RxVecd4 &v2);

	RxVecd4& extrap (const RxVecd4 &v1, double a, const RxVecd4 &v2);
	RxVecd4& extrap (double a, const RxVecd4 &v2)
	 { return (extrap (*this, a, v2));
	 }

	// T <- T,O operations

	RxVecd4& scale (double a, const RxVecd4 &v1);
	RxVecd4& scale (double a);
	RxVecd4& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd4& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd4 operator*(const RxVecd4 &v1, double a);
	friend RxVecd4 operator*(double a, const RxVecd4 &v1);
	friend RxVecd4 operator/(const RxVecd4 &v1, double a);

	// O <- T,T operations

	double dot(const RxVecd4 &v2) const;

	// T <- T operations

	RxVecd4 &unit(const RxVecd4 &v1);
	RxVecd4 &unit()
	 { return (unit(*this));
	 }
	RxVecd4 &neg(const RxVecd4 &v1);
	RxVecd4 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd4 operator-(const RxVecd4 &v1);

	// T <- O operations

	RxVecd4 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;
};

inline RxVecd4 operator+(const RxVecd4 &v1, const RxVecd4 &v2)
{
	RxVecd4 vr;
	return (vr.add (v1, v2));
}

inline RxVecd4 operator-(const RxVecd4 &v1, const RxVecd4 &v2)
{
	RxVecd4 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd4 operator*(const RxVecd4 &v1, double a)
{
	RxVecd4 vr;
	return (vr.scale (a, v1));
}

inline RxVecd4 operator*(double a, const RxVecd4 &v1)
{
	RxVecd4 vr;
	return (vr.scale (a, v1));
}

inline RxVecd4 operator/(const RxVecd4 &v1, double a)
{
	RxVecd4 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd4 operator-(const RxVecd4 &v1)
{
	RxVecd4 vr;
	return (vr.neg (v1));
}

#endif
