#ifndef _PARSEARGS_H_
#define _PARSEARGS_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifndef PROTO
#ifndef NO_PROTOTYPES
#define PROTO(x) x
#else
#define PROTO(x) ()
#endif
#endif

#include <stdio.h>

#define ADEF_INT	1
#define ADEF_HEX	2
#define ADEF_OCTAL	3
#define ADEF_FLOAT	4
#define ADEF_STRING	5
#define ADEF_VALUE	6

#define ADEF_LOW_RANGE	0x01
#define ADEF_LOW_OPEN	0x02
#define ADEF_HIGH_RANGE	0x04
#define ADEF_HIGH_OPEN	0x08
#define ADEF_ONE_WORD	0x10
#define ADEF_NEEDS_ARG  0x20
#define ADEF_LONG	0x40

typedef struct argDef {
	struct argDef *next;

	int type;
	int flags;
	int num;
	char *name;
	void *addr;
	char *comment;

	union {
	  struct {
	    long high;
	    long low;
	    } i;

	  struct {
	    unsigned long high;
	    unsigned long low;
	    } u;

	  struct {
	    double high;
	    double low;
	    } f;

	  struct {
	    long val;
	    } v;
	  
	  } u;
} ArgDef;

typedef struct argDefList {
	struct argDef *head;
	struct argDef *tail;
} ArgDefList;

ArgDefList *argDefCreate();
char *argDefErrorStr PROTO(());
int argDefMatch PROTO((ArgDefList*, int, char**));
int argDefPrintUsage PROTO((FILE*, ArgDefList*, int));
int argDefPrintHelp PROTO((FILE*, ArgDefList*));
int argDefAdd PROTO((ArgDefList* list, ...));
void argDefArgvPrependFile PROTO((int *argcp, char ***argvp,const char *name));
void argDefParseArgs PROTO((int argc, char **argv, ArgDefList *argDefs,
			   const char *argFile, const char *helparg));
void buildArgList PROTO((char *arg0, ...));

#ifdef __cplusplus
}
#endif

#endif
