
/* some routines for working with 3 x 3 matrices */
	
solve3x3 (rv, iv, matrix)
double rv[];
double iv[];
double matrix[3][3];
{
	static int callnum = 0;
	double work[3][4];
	int i, j, k;

	/* map stuff into work area */

	callnum++;

	for (i=0; i<3; i++)
	 { for (j=0; j<3; j++)
	    { work[i][j] = matrix[i][j];
	    }
	   work[i][3] = iv[i];
	 }

	/* solve using Gaussian elimination */
	
	for (i=0; i<3; i++)
	 { double x = work[i][i];

	   if (x==0.)
	    { printf ("x = 0 at call %d\n", callnum);
	    }

	   for (j=i+1; j<4; j++)
	    { work[i][j] /= x;
	    }
	   for (k=i+1; k<3; k++)
	    { x = work[k][i];
	      for (j=i+1; j<4; j++)
	       { work[k][j] -= x * work[i][j];
	       }
	    }
	 }

	rv[2] = work [2][3];
	for (i=2-1; i>=0; i--)
	 { double result = work [i][3];
	   for (j=2; j>i; j--)
	    { result -= work [i][j] * rv[j];
	    }
	   rv[i] = result;
	 }
}

invert3x3 (imat, matrix)
double imat[3][3];
double matrix[3][3];
{
	static int callnum = 0;
	double work[3][6];
	int i, j, k;

	/* map stuff into work area */

	callnum++;

	for (i=0; i<3; i++)
	 { for (j=0; j<3; j++)
	    { work[i][j] = matrix[i][j];
	      work[i][j+3] = 0.;
	    }
	   work[i][i+3] = 1.;
	 }

	/* solve using Gaussian elimination */
	
	for (i=0; i<3; i++)
	 { double x = work[i][i];

	   if (x==0.)
	    { printf ("x = 0 at call %d\n", callnum);
	    }

	   for (j=i+1; j<6; j++)
	    { work[i][j] /= x;
	    }
	   for (k=i+1; k<3; k++)
	    { x = work[k][i];
	      for (j=i+1; j<6; j++)
	       { work[k][j] -= x * work[i][j];
	       }
	    }
	 }

	for (i=2-1; i>=0; i--)
	 { 
	   for (j=2; j>i; j--)
	    { double x = work[i][j];
	      for (k=3; k<6; k++)
	       { work [i][k] -= x * work [j][k];
	       }
	    }
	 }

	for (i=0;i<3;i++)
	 { for (j=0;j<3;j++)
	    { imat[i][j] = work [i][j+3];
	    }
	 }
}
	
mult3x3 (r, m1, m2)
double r[3][3];
double m1[3][3];
double m2[3][3];
{
	register int i, j, k;

	for (i=0;i<3;i++)
	 { for (j=0;j<3;j++)
	    { double x = 0.;
	      for (k=0;k<3;k++)
	       { x += m1[i][k] * m2[k][j];
	       }
	      r[i][j] = x;
	    }
	 }
}

xform3 (rv, iv, matrix)
double rv[3];
double iv[3];
double matrix[3][3];
{
	int i, j;
	double x;

	for (i=0;i<3;i++)
	 { x = 0.;
	   for (j=0;j<3;j++)
	    { x += iv[j] * matrix[i][j];
	    }
	   rv[i] = x;
	 }
}

dot (rv, v1, v2)
double rv[3];
double v1[3];
double v2[3];
{
	int i;
	for (i=0;i<3;i++)
	 { rv[i] = v1[i] * v2[i];
	 }
}
	
transpose3x3 (out, in)
double out[3][3];
double in[3][3];
{
	int i, j;

	for (i=0;i<3;i++)
	 { for (j=0;j<3;j++)
	    { out[j][i] = in[i][j];
	    }
	 }
}


