#include <ctype.h>
#include <math.h>
#include "../include/useful.h"

#define YES 1
#define NO  0

static unsigned int maxUnsignedInt;
static unsigned int maxSignedInt;
static init = NO;

initialize()
{
	maxUnsignedInt = -1;
	maxSignedInt = maxUnsignedInt >> 1;
	init = YES;
}

scan_int (string, number, base)
char *string;
int *number, base;
{
      {
	int increment, temp, i, start;
	int overflow;
	int signed;
	int negative;

	if (!init)
	 { initialize();
	 }

	overflow = NO;
	negative = NO;

	if (base < 0)
	 { base = -base;
	   signed = YES;
	 }
	else
	 { signed = NO;
	 }

	if (base > 16 || base < 2)
	 { return (S_SYNTAX);
	 }
	i = 0;
	while (isspace (string[i]))
	 { i++;
	 }
	if (string[i] == '-')
	 { negative = YES;
	   i++;
	 }
	if (negative && !signed)      /* cannot have negative unsigned int */
	 { return (S_SYNTAX);
	 }

	start = i;
	temp = 0;
	increment = evalchar (string [i]);
	while (increment >= 0 && increment < base)
	 { 
	   /* test for overflow at this point */

	   if ( (signed && temp > (maxSignedInt - increment) / base ) ||
	        (!signed && temp > (maxUnsignedInt - increment) / base ) )
	    { overflow = YES;
	    }
	    
	   temp = base * temp + increment;
	   increment = evalchar (string[++i]);
	 }

	if (i ==  start)
	 { return (S_SYNTAX);
	 }
	if (!isspace (string[i]) && string[i] != 0)
	 { return (S_SYNTAX);
	 }

	*number = negative ? -temp : temp;
	return ( overflow ? S_OVERFLOW : 0 );
      }
}

static evalchar (c)
char c;
{
	if ('0' <= c && c <= '9')
	 { return (c - '0');
	 }
	else if ('a' <= c <= 'f')
	 { return (c - 'a' + 10);
	 }
	else if ('A' <= c <= 'F')
	 { return (c - 'A' + 10);
	 }
	else
 	 { return (-1);
	 }
}

#define STACK_SIZE 256

print_int (string, number, base)
char *string;
int number, base;
{
      {
	unsigned int output;
	char stack [STACK_SIZE];
	int sp, j;
	int signed;

	if (base < 0)
	 { base = -base;
	   signed = YES;
	 }
	else
	 { signed = NO;
	 }
  
	if (base > 16 || base < 2)
	 { return (-1);
	 }

	if (signed)
	 { output = (unsigned int) ((number < 0) ? -number : number);
	 }
	else
	 { output = (unsigned int) (number);
	 }

	sp = STACK_SIZE;
	stack [--sp] = 0;
	while (output != 0)
	 { 
	   if ((output % base) < 10)
	    { stack [--sp] = '0' + (output % base);
	    }
	   else
	    { stack [--sp] = 'a' + (output % base) - 10;
	    }
	   output = output / base;
	 }
	if (sp == STACK_SIZE-1 || evalchar (stack[sp]) >= 10)
	 { stack [--sp] = '0';
	 }
	if (signed && number < 0)
	 { stack [--sp] = '-';
	 }
	strcpy (string, &stack[sp]);
      }

     return (0);
}


#ifdef TEST

#define COMMAND(x)	(strncmp(x,cbuf,255) == 0)

main()
{
	int base = 10;
	char cbuf[256];

	do
	 { printf ("> ");
	   scanf ("%255s", cbuf);
	   while (getchar() != '\n')
	      ;
	   
	   if (COMMAND ("?"))
	    { printf (" 'sb'  : set base\n");
	      printf (" 'i'   : scan integer\n");
	      printf (" 'q'   : quit\n");
	    }
	   else if (COMMAND ("sb"))
	    { printf ("New base> ");
	      scanf ("%d", &base);
	    }
	   else if (COMMAND ("i"))
	    { int status;
	      int number;
	      printf ("Integer> ");

	      gets (cbuf);
	      status = scan_int (cbuf, &number, base);
	      printf ("number scanned = %d with status %d\n", number, status);
	      status = print_int (cbuf, number, base);
	      printf ("number printed = %s with status %d\n", cbuf, status);
	      cbuf [0] = 0;
	    }

	 }
	while (!COMMAND ("q"));
}

#endif TEST
