#include <RxClist.h>
#include <string.h>

int RxClistBase::removeAndDelete(RxCnodeBase* ptr1, RxCnodeBase* ptr2)
{
	RxCnodeBase *node;
	RxCnodeBase *nodeBefore;
	RxCnodeBase *nodeAfter;
	RxCnodeBase *nodeLast;
	RxCnodeBase *nodeNext = NULL;
	int cnt = 0;

	if (ptr2->next_ == ptr1)
	 { return (removeAndDeleteAll());
	 }
	else
	 { nodeBefore = ptr1->prev_;
	   nodeAfter = ptr2->next_;

	   nodeBefore->next_ = nodeAfter;
	   nodeAfter->prev_ = nodeBefore;

	   nodeLast = ptr2->next_;
	   for (node=ptr1; node!=nodeLast; node=nodeNext)
	    { nodeNext = node->next_;
	      if (node == head_)
	       { head_ = nodeAfter;
	       }
	      if (node == tail_)
	       { tail_ = nodeBefore;
	       }
	      freeNode (node);
	      cnt++;
	    }
	   return cnt;
	 }
}

int RxClistBase::remove(RxCnodeBase* ptr1, RxCnodeBase* ptr2)
{
	RxCnodeBase *node;
	RxCnodeBase *nodeBefore;
	RxCnodeBase *nodeAfter;
	RxCnodeBase *nodeLast;
	int cnt = 0;

	if (ptr2->next_ == ptr1)
	 { return (removeAll());
	 }
	else
	 { nodeBefore = ptr1->prev_;
	   nodeAfter = ptr2->next_;

	   nodeBefore->next_ = nodeAfter;
	   nodeAfter->prev_ = nodeBefore;

	   nodeLast = ptr2->next_;
	   for (node=ptr1; node!=nodeLast; node=node->next_)
	    { if (node == head_)
	       { head_ = nodeAfter;
	       }
	      if (node == tail_)
	       { tail_ = nodeBefore;
	       }
	      cnt++;
	    }
	   return cnt;
	 }
}

int RxClistBase::removeAll()
{
	int notEmpty;

	notEmpty = (head_ != NULL);
	head_ = NULL;
	tail_ = NULL;
	return notEmpty;
}

int RxClistBase::removeAndDeleteAll()
{
	RxCnodeBase *node, *nextNode = NULL;
	int cnt = 0;

	if (head_)
	 { tail_->next_ = NULL;
	   for (node=head_; node; node=nextNode)
	    { nextNode = node->next_;
	      freeNode (node);
	      cnt++;
	    }
	   head_ = NULL;
	   tail_ = NULL;
	 }
	return cnt;
}

RxCnodeBase *RxClistBase::remove(RxCnodeBase* ptr)
{
	if (ptr == ptr->next_)
	 { head_ = NULL;
	   tail_ = NULL;
	 }
	else
	 { ptr->next_->prev_ = ptr->prev_;
	   ptr->prev_->next_ = ptr->next_;
	   if (ptr == head_)
	    { head_ = ptr->next_;
	    }
	   else if (ptr == tail_)
	    { tail_ = ptr->prev_;
	    }
	 }
	return (ptr);
}

#if 0
RxClistBase::~RxClistBase()
{
	if (autoDelete) removeAndDeleteAll ();
}
#endif

RxCnodeBase *RxClistBase::addHead(RxCnodeBase* ptr)
{
	if (head_ == NULL)
	 { ptr->next_ = ptr;
	   ptr->prev_ = ptr;
	   tail_ = ptr;
	 }
	else
	 { ptr->next_ = head_;
	   ptr->prev_ = tail_;
	   head_->prev_ = ptr;
	   tail_->next_ = ptr;
	 }
	head_ = ptr;
	return (ptr);
}

RxCnodeBase *RxClistBase::addTail(RxCnodeBase* ptr)
{
	if (tail_ == NULL)
	 { ptr->next_ = ptr;
	   ptr->prev_ = ptr;
	   head_ = ptr;
	 }
	else
	 { ptr->next_ = head_;
	   ptr->prev_ = tail_;
	   head_->prev_ = ptr;
	   tail_->next_ = ptr;
	 }
	tail_ = ptr;
	return (ptr);
}

RxCnodeBase *RxClistBase::addAfter(RxCnodeBase* ref, RxCnodeBase* ptr)
{
	if (ref == NULL)
	 { return addHead (ptr);
	 }
	else
	 { ptr->prev_ = ref;
	   ptr->next_ = ref->next_;
	   ref->next_->prev_ = ptr;
	   ref->next_ = ptr;
	   if (ref == tail_)
	    { tail_ = ptr;
	    }
	   return (ptr);
	 }
}

RxCnodeBase *RxClistBase::addBefore(RxCnodeBase* ref, RxCnodeBase* ptr)
{
	if (ref == NULL)
	 { return addTail (ptr);
	 }
	else
	 { ptr->next_ = ref;
	   ptr->prev_ = ref->prev_;
	   ref->prev_->next_ = ptr;
	   ref->prev_ = ptr;
	   if (ref == head_)
	    { head_ = ptr;
	    }
	   return (ptr);
	 }
}

RxCnodeBase *RxClistBase::findNodeByPtr (const void *ptr) const
{
	RxCnodeBase *node;

	if ((node = head_))
	 { do
	    { if (node->compare (ptr))
	       { return (node);
	       }
	      node = node->next_;
	    }
	   while (node != head_);
	 }
	return (NULL);
}

RxCnodeBase *RxClistBase::findNodeByName (const char *name) const
{
	RxCnodeBase *node;

	if ((node = head_))
	 { do
	    { if (strcmp (node->getName(), name) == 0)
	       { return (node);
	       }
	      node = node->next_;
	    }
	   while (node != head_);
	 }
	return (NULL);
}

void RxClistBase::shift (int n)
{
	RxCnodeBase *node;

	if ((node = head_))
	 { 
	   if (n >= 0)
	    { while (n--)
	       { node = node->next_;
	       }
	    }
	   else
	    { while (n++)
	       { node = node->prev_;
	       }
	    }
	   head_ = node;
	   tail_ = node->prev_;
	 }
}

int RxClistBase::length () const
{
	RxCnodeBase *node;
	int cnt = 0;

	if ((node = head_))
	 { do
	    { cnt++;
	      node = node->next_;
	    }
	   while (node != head_);
	 }
	return (cnt);
}

RxCnodeBase *RxCnodeBase::getNext(int n)
{
	RxCnodeBase *node;

	if (n < 0)
	 { return (getPrev(-n));
	 }
	else
	 { node = this;
	   while (n--)
	    { node = node->next_;
	    }
	   return (node);
	 }
}

RxCnodeBase *RxCnodeBase::getPrev(int n)
{
	RxCnodeBase *node;

	if (n < 0)
	 { return (getNext(-n));
	 }
	else
	 { node = this;
	   while (n--)
	    { node = node->prev_;
	    }
	   return (node);
	 }
}
