#include <RxHashTab.h>
#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <RxTimeval.h>

class StrHashNode : public RxHashNodeBase
{
public:
	char *str;
	void *ptr;
	StrHashNode (char *s, void *p)
	 { str = strdup(s);
	   ptr = p;
	 }
	int match(void *s) const
	 { return (strcmp(str,(char*)s) == 0);
	 }
};

class StrHashTab : public RxHashTabBase
{
public:
	StrHashTab(int size=100) : RxHashTabBase(size) { }
	int hash(const char *s) const
	 { return (hashStr(s));
	 }
	RX_HASH_TAB_DEFS(StrHashNode, char *, void *);
};

class IntHashNode : public RxHashNodeBase
{
public:
	int key;
	void *ptr;
	IntHashNode (int i, void *p)
	 { key = i;
	   ptr = p;
	 }
	int match(void *s) const
	 { return (key == (int)s);
	 }
};

class IntHashTab : public RxHashTabBase
{
public:
	IntHashTab(int size=100) : RxHashTabBase(size) { }
	int hash(int s) const
	 { return (s % size_);
	 }
	RX_HASH_TAB_DEFS(IntHashNode, int, void *);
};

int randAlnumChar()
{
	int c;
	do
	 { c = (int)(drand48()*128);
	 }
	while (!isalnum(c));
	return (c);
}

char *randStr (char *buf, int maxlen)
{
	char *cp;
	int len;

	do
	 { len = (int)((maxlen+1)*drand48());
	 }
	while (len == 0 || len > maxlen);
	for (cp=buf; len; len--)
	 { *cp++ = randAlnumChar();
	 }
	*cp = '\0';
}

#define TEST_SIZE	200

typedef struct {
	char *str;
	int val;
} TestEntry;

TestEntry listMain[TEST_SIZE];
TestEntry listBack[TEST_SIZE];

int findStr (const TestEntry *list, int n, const char *s)
{
	int k;

	for (k=0; k<n; k++)
	 { if (strcmp (s, list[k].str) == 0)
	    { return k;
	    }
	 }
	return -1;
}

int findInt (const TestEntry *list, int n, int key)
{
	int k;

	for (k=0; k<n; k++)
	 { if (key == list[k].val)
	    { return k;
	    }
	 }
	return -1;
}

fillStrTable (StrHashTab *htab, TestEntry *list, int n)
{
	TestEntry *test;
	int i;

	for (i=0; i<n; i++)
	 { test = &list[i];
	   htab->insert (test->str, &test->val);
	   htab->find (test->str);
	 }
	for (i=0; i<n; i++)
	 { test = &list[i];
	   htab->remove (test->str);
	 }
}

lookStrTable (StrHashTab *htab, TestEntry *list, int n)
{
	int i;

	for (i=0; i<n; i++)
	 { htab->find (list[i].str);
	 }
}

fillIntTable (IntHashTab *htab, TestEntry *list, int n)
{
	TestEntry *test;
	int i;

	for (i=0; i<n; i++)
	 { test = &list[i];
	   htab->insert (test->val, &test->val);
	   htab->find (test->val);
	 }
	for (i=0; i<n; i++)
	 { test = &list[i];
	   htab->remove (test->val);
	 }
}

lookIntTable (IntHashTab *htab, TestEntry *list, int n)
{
	int i;

	for (i=0; i<n; i++)
	 { htab->find (list[i].val);
	 }
}

main()
{
	StrHashTab *stab;
	StrHashNode *shp0, *shp1;
	IntHashTab *itab;
	IntHashNode *ihp0, *ihp1;
	TestEntry *test;
	char cbuf[256];
	int i, k;

	for (i=0; i<TEST_SIZE; i++)
	 { do
	    { randStr (cbuf, 8);
	    }
	   while (findStr (listMain, i, cbuf) != -1);
	   listMain[i].str = strdup(cbuf);
	   listMain[i].val = i;
	 }

	stab = new StrHashTab(101);

	assert (stab->numEntries()==0);
	assert (stab->size() == 101);

	stab->clear();

	assert (stab->numEntries()==0);

	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   assert (stab->find (test->str) == NULL);
	   assert (stab->remove (test->str) == 0);
	   shp0 = stab->insert (test->str, &test->val);
	   assert (shp0 != NULL && 
		   strcmp(shp0->str,test->str) == 0 &&
		   shp0->ptr == &test->val);
	   assert (stab->numEntries() == i+1);
	   assert ((shp1 = stab->find (test->str)) == shp0);
	 }
	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   assert (stab->remove (test->str) == 1);
	   assert (stab->find (test->str) == NULL);
	   assert (stab->remove (test->str) == 0);
	   assert (stab->numEntries() == TEST_SIZE-(i+1));
	 }
	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   shp0 = stab->get (test->str, &test->val);
	   assert (shp0 != NULL && 
		   strcmp(shp0->str,test->str) == 0 &&
		   shp0->ptr == &test->val);
	   assert (stab->numEntries() == i+1);
	   assert ((shp1 = stab->find (test->str)) == shp0);
	 }
	i = 0;
	for (shp0=stab->first(); shp0; shp0=stab->next())
	 { assert (findStr (listBack, i, shp0->str) == -1);
	   listBack[i].str = shp0->str;
	   i++;
	 }
	assert (i == TEST_SIZE);
	assert (stab->next() == NULL);
	stab->clear();
	assert (stab->numEntries() == 0);
	assert (stab->first() == NULL);
	for (i=0; i<TEST_SIZE; i++)
	 { assert (stab->find (test->str) == NULL);
	   assert (stab->remove (test->str) == 0);
	 }

	TIMEIT (fillStrTable(stab, listMain, TEST_SIZE), 1000);
	for (i=0; i<TEST_SIZE; i++)
	 { stab->insert (listMain[i].str, &listMain[i].val);
	 }
	TIMEIT (lookStrTable(stab, listMain, TEST_SIZE), 1000);

	delete stab;

	itab = new IntHashTab(101);

	assert (itab->numEntries()==0);
	assert (itab->size() == 101);

	itab->clear();

	assert (itab->numEntries()==0);

	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   assert (itab->find (test->val) == NULL);
	   assert (itab->remove (test->val) == 0);
	   ihp0 = itab->insert (test->val, &test->val);
	   assert (ihp0 != NULL && 
		   ihp0->key == test->val &&
		   ihp0->ptr == &test->val);
	   assert (itab->numEntries() == i+1);
	   assert ((ihp1 = itab->find (test->val)) == ihp0);
	 }
	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   assert (itab->remove (test->val) == 1);
	   assert (itab->find (test->val) == NULL);
	   assert (itab->remove (test->val) == 0);
	   assert (itab->numEntries() == TEST_SIZE-(i+1));
	 }
	for (i=0; i<TEST_SIZE; i++)
	 { test = &listMain[i];
	   ihp0 = itab->get (test->val, &test->val);
	   assert (ihp0 != NULL && 
		   ihp0->key == test->val &&
		   ihp0->ptr == &test->val);
	   assert (itab->numEntries() == i+1);
	   assert ((ihp1 = itab->find (test->val)) == ihp0);
	 }
	i = 0;
	for (ihp0=itab->first(); ihp0; ihp0=itab->next())
	 { assert (findInt (listBack, i, ihp0->key) == -1);
	   listBack[i].val = ihp0->key;
	   i++;
	 }
	assert (i == TEST_SIZE);
	assert (itab->next() == NULL);
	itab->clear();
	assert (itab->numEntries() == 0);
	assert (itab->first() == NULL);
	for (i=0; i<TEST_SIZE; i++)
	 { assert (itab->find (test->val) == NULL);
	   assert (itab->remove (test->val) == 0);
	 }

	TIMEIT (fillIntTable(itab, listMain, TEST_SIZE), 1000);
	for (i=0; i<TEST_SIZE; i++)
	 { itab->insert (listMain[i].val, &listMain[i].val);
	 }
	TIMEIT (lookIntTable(itab, listMain, TEST_SIZE), 1000);

	delete itab;


	printf ("\nROUTINES OK\n\n");
}
