#include <RxMuxedNameTable.h>

void RxMuxedNameTable::createTable(int size)
{
	MUTEX_LOCK(&mux);
	if (htab)
	 { delete htab;
	 }
	htab = new RxStrHashTab (size);
	tabsize = size;
	MUTEX_UNLOCK(&mux);
}

int RxMuxedNameTable::insertEntry (const char *name, void *ptr)
{
	MUTEX_LOCK(&mux);
	if (htab == NULL)
	 { htab = new RxStrHashTab (tabsize);
	 }
	if (htab->find (name))
	 { MUTEX_UNLOCK(&mux);
	   return 0;
	 }
	else
	 { htab->insert (name, ptr);
	   MUTEX_UNLOCK(&mux);
	   return 1;
	 }
}

int RxMuxedNameTable::deleteEntry (const char *name)
{
	int retval;
	MUTEX_LOCK(&mux);
	// probably safe to assume we have a hash table, but whatever
	if (htab == NULL)
	 { htab = new RxStrHashTab (tabsize);
	 }
	retval = htab->remove (name);
	MUTEX_UNLOCK(&mux);
	return retval;
}

void *RxMuxedNameTable::getEntry(const char *name)
{
	RxStrHashNode *node;
	MUTEX_LOCK(&mux);
	// probably safe to assume we have a hash table, but whatever
	if (htab == NULL)
	 { htab = new RxStrHashTab (tabsize);
	 }
	node = htab->find(name);
	MUTEX_UNLOCK(&mux);
	return node ? node->ptr() : NULL;
}
