#include <RxSysUtil.h>
#include <RxSysError.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

static FILE *pathSearch (char **fullnamep, const char *pathStr, const char *name)
{
	/* Returns a FILE * to the first occurance of the file
	   'name' in the directories 'pathStr' which can be opened
	   read-only. If no file can be openned, NULL is returned. */

	const char *pathPtr;
	char *nameBuf;
	char *namePtr;

	FILE *fp = NULL;

	nameBuf = (char*)malloc(strlen(name)+strlen(pathStr)+2);

	pathPtr = pathStr;
	do
	 { namePtr = nameBuf;
	   while (*pathPtr != ':' && *pathPtr != '\0')
	    { *namePtr++ = *pathPtr++;
	    }
	   if (namePtr != nameBuf)
	    { *namePtr++ = '/';
	      *namePtr = '\0';
	      strcat (nameBuf, name);
	      fp = fopen (nameBuf, "r");
	    }
	   else
	    { if (*pathPtr == '\0')
	       { free (nameBuf);
		 return (NULL);
	       }
	    }
	   if (*pathPtr == ':')
	    { pathPtr++;
	    }
	 }
	while (fp == NULL);

	if (fullnamep)
	 { *fullnamep = strdup(nameBuf);
	 }
	free (nameBuf);
	return (fp);	
}

FILE *locateFile (char **fullnamep, const char *listName, const char *name)
{
	/* Driver routine for pathSearch. Checks environment variables,
	   and prints appropriate error messages. */

	FILE *fp;

	if (getenv(listName) == NULL)
	 { RxSysError(RxSysError::UndefEnviron, 
		      "locateFile %K '%s'\n", listName).raise();
	   return (NULL);
	 }
	fp = pathSearch (fullnamep, getenv(listName), name);
	if (fp)
	 { if (getenv ("RX_PATH_PRINT"))
	    { printf ("reading %s ...\n", *fullnamep);
	    }
	 }
	else
	 { RxSysError(RxSysError::CantFindFile,
"locateFile %K '%s' in path '%s\n", name, listName).raise();
	 }
	return (fp);
}
