#include <RxTimeval.h>
#include <string.h>
#include <stringx.h>
#include <cdefs.h>
#include <stdlib.h>

void RxTimeval::doset (long s, long u)
{
	RxTimeval tv;

	if (ABS(u) > 1000000)
	 { s += (u/1000000);
	   u = u % 1000000;
	 }
	if (s > 0 && u < 0)
	 { sec = s - 1;
	   usec = 1000000 + u;
	 }
	else if (s < 0 && u > 0)
	 { sec = s + 1;
	   usec = -1000000 + u;
	 }
	else
	 { sec = s;
	   usec = u;
	 }
}

int RxTimeval::docmp (const RxTimeval &tv2) const
{
	RxTimeval diff;

	diff = *this - tv2;
	if (diff.sec > 0 || diff.usec > 0)
	 { return (1);
	 }
	else if (diff.sec < 0 || diff.usec < 0)
	 { return (-1);
	 }
	else
	 { return (0);
	 }
}

RxTimeval operator+(const RxTimeval &tv1, const RxTimeval &tv2)
{
	RxTimeval tr;

	tr.doset (tv1.sec + tv2.sec, tv1.usec + tv2.usec);
	return (tr);
}

RxTimeval operator-(const RxTimeval &tv1, const RxTimeval &tv2)
{
	RxTimeval tr;

	tr.doset (tv1.sec - tv2.sec, tv1.usec - tv2.usec);
	return (tr);
}

RxTimeval operator*(const RxTimeval &tv1, double a)
{
	RxTimeval tr;

	tr.doset ((long)(tv1.sec*a), (long)(tv1.usec*a));
	return (tr);
}

RxTimeval operator*(double a, const RxTimeval &tv1)
{
	RxTimeval tr;

	tr.doset ((long)(a*tv1.sec), (long)(a*tv1.usec));
	return (tr);
}

RxTimeval operator/(const RxTimeval &tv1, double a)
{
	RxTimeval tr;

	tr.doset ((long)(tv1.sec/a), (long)(tv1.usec/a));
	return (tr);
}

RxTimeval operator-(const RxTimeval &tv1)
{
	RxTimeval tr;

	tr.doset (-tv1.sec, -tv1.usec);
	return (tr);
}

RxTimeval& RxTimeval::operator+=(const RxTimeval &tv2)
{
	doset (sec+tv2.sec, usec+tv2.usec);
	return (*this);
}

RxTimeval& RxTimeval::operator-=(const RxTimeval &tv2)
{
	doset (sec-tv2.sec, usec-tv2.usec);
	return (*this);
}

RxTimeval& RxTimeval::operator*=(double a)
{
	doset ((long)(sec*a), (long)(usec*a));
	return (*this);
}

RxTimeval& RxTimeval::operator/=(double a)
{
	doset ((long)(sec/a), (long)(usec/a));
	return (*this);
}

int RxTimeval::print (FILE *fp, const char *s)
{
	char *prefix;
	const char *cp;
	int numchars = 0;
	int nc;

	if ((cp = strstr(s, "%T")) == NULL)
	 { return (fprintf (fp, "%s%ld.%06ld\n", sec, usec));
	 }
	if ((cp-s) > 0)
	 { prefix = strndup (s, cp-s);
	   nc = fprintf (fp, prefix);
	   free (prefix);
	   if (nc < 0)
	    { return (nc);
	    }
	   numchars += nc;
	 }
	if ((nc = fprintf (fp, "%ld.%06ld", sec, usec)) < 0)
	 { return (nc);
	 }
	numchars += nc;
	cp += 2;
	if (*cp != '\0')
	 { if ((nc = fprintf (fp, cp)) < 0)
	    { return (nc);
	    }
	   numchars += nc;
	 }
	return (numchars);
}

int RxTimeval::scan (FILE *fp)
{
	return (fscanf (fp, "%ld.%ld", &sec, &usec));
}

int RxTimeval::sysSetTod ()
{
	struct timeval tv;

	if (gettimeofday (&tv, (struct timezone*)0) < 0)
	 { return -1;
	 }
	sec = tv.tv_sec;
	usec = tv.tv_usec;
	return (1);
}
