#ifndef _RXTIMEVAL_H_
#define _RXTIMEVAL_H_

#include <sys/time.h>
#include <stdio.h>

class RxTimeval
{
private:
	long sec;
	long usec;

	void doset (long s, long u);
	int docmp (const RxTimeval &tv2) const;
public:
	RxTimeval()
	 { sec = 0; usec = 0;
	 };
	RxTimeval(const long s, const long u)
	 { this->doset(s,u);
	 };
	RxTimeval(const struct timeval &tv)
	 { this->doset(tv.tv_sec, tv.tv_usec);
	 };
	RxTimeval(const double t)
	 { sec = (long)(t+0.5); usec = (long)((t - sec)*1e6+0.5);
	 };
	~RxTimeval() { };

	RxTimeval& set(const long s, const long u)
	 { this->doset(s, u); return (*this);
	 };
	void get(long &s, long &u)
	 { s = sec; u = usec;
	 };
	void setZero()
	 { sec = usec = 0;
	 };

	RxTimeval& set(double t)
	 { sec = (long)(t+0.5); usec = (long)((t - sec)*1e6+0.5);
	   return (*this);
	 }
	RxTimeval& operator=(const struct timeval &tv)
	 { this->doset(tv.tv_sec, tv.tv_usec);
	   return (*this);
	 }
#if 0
	RxTimeval& operator=(double t)
	 { sec = (long)(t+0.5); usec = (long)((t - sec)*1e6+0.5);
	   return (*this);
	 }
#endif
	operator double()
	 { return (sec + usec*1e-6);
	 };
	operator struct timeval()
	 { struct timeval tv; tv.tv_sec = sec; tv.tv_usec = usec; return (tv);
	 };

	friend RxTimeval operator+(const RxTimeval &tv1, const RxTimeval &tv2);
	friend RxTimeval operator-(const RxTimeval &tv1, const RxTimeval &tv2);
	friend RxTimeval operator*(const RxTimeval &tv1, double a);
	friend RxTimeval operator*(double a, const RxTimeval &tv1);
	friend RxTimeval operator/(const RxTimeval &tv1, double a);
	friend RxTimeval operator-(const RxTimeval &tv1);

	RxTimeval& RxTimeval::operator+=(const RxTimeval &tv2);
	RxTimeval& RxTimeval::operator-=(const RxTimeval &tv2);
	RxTimeval& RxTimeval::operator/=(double a);
	RxTimeval& RxTimeval::operator*=(double a);

	friend int operator==(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) == 0);
	 };
	friend int operator!=(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) != 0);
	 };
	friend int operator>(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) > 0);
	 };
	friend int operator<(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) < 0);
	 };
	friend int operator>=(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) >= 0);
	 };
	friend int operator<=(const RxTimeval &tv1, const RxTimeval &tv2)
	 { return (tv1.docmp (tv2) <= 0);
	 };

	int print (FILE *fp, const char *s);
	int print (const char *s)
	 { return (print (stdout, s));
	 }
	int scan (FILE *fp);
	int scan (void)
	 { return (scan (stdin));
	 }

	int sysSetTod (void);
};

#define TIMEIT(code,cnt)		\
	{ int _i;			\
	  RxTimeval t0, t1;		\
	  t0.sysSetTod();		\
	  for (_i=0; _i<(cnt); _i++)	\
	   { code;			\
	   }				\
	  t1.sysSetTod();		\
	  printf ("%s: %g usec\n", #code, (double)(t1-t0)*1e6/(cnt)); \
	}

#endif
