#include <RxAccel.h>
#include <RxRot.h>
#include <RxTrsf.h>
#include <RxTwist.h>
#include <RxMathError.h>
#include <math.h>
#include <cdefs.h>

const RxAccel RxAccel::ZERO    (0,0,0, 0,0,0);
const RxAccel RxAccel::UNIT_VX (1,0,0, 0,0,0);
const RxAccel RxAccel::UNIT_VY (0,1,0, 0,0,0);
const RxAccel RxAccel::UNIT_VZ (0,0,1, 0,0,0);
const RxAccel RxAccel::UNIT_WX (0,0,0, 1,0,0);
const RxAccel RxAccel::UNIT_WY (0,0,0, 0,1,0);
const RxAccel RxAccel::UNIT_WZ (0,0,0, 0,0,1);

RxAccel &RxAccel::xform (const RxAccel &a1, const RxTrsf &T,
			 const RxVecd3 &wa)
{
	/*         [ R  p ]
	   Let T = [ 0  1 ] be the transform from frame A to frame B.

	   Then accelertion a1 is converted from A into B via by

				 B    B
	   B              A     [ w X  w X p ]
	    a1  =  adj(T)  a1 - [            ]
				[     0      ]

	   where
			      T
			 [   R     0   ]       [  0  -pz  py ]
			 [             ]       [ 	     ]
		adj(T) = [   T      T  ],  P = [  pz  0  -px ]
			 [ -R  P   R   ]       [ 	     ]
					       [ -py  px  0  ]

	*/
	RxVecd3 wb;
	RxVecd3 x;

	wb.xform (wa, T.R);
	a.xform (a1.a, T);	// this does the adjoint multiply 
	
	x.cross (wb, T.p);	
	x.cross (wb, x);
	a.v().sub (x);

	return (*this);
}

RxAccel &RxAccel::Ixform (const RxAccel &a1, const RxTrsf &T, 
			  const RxVecd3 &wa)
{
	/* The inverse of the above transformation */

	RxVecd3 x;

	a.Ixform (a1.a, T);	// this does the adjoint multiply 
	
	x.cross (wa, T.p);
	x.cross (wa, x);
	x.Ixform (T.R);
	a.v().add (x);

	return (*this);
}

RxAccel& RxAccel::set (const RxSpatialAccel &sa, const RxTwist &tw)
{
	RxVecd3 x;

	x.cross (tw.w(), tw.v());
	a.v().add (sa.v(), x);
	a.w() = sa.w();
	return *this;
}

