#ifndef _RXACCEL_H_
#define _RXACCEL_H_

#include <RxVecdBase.h>
#include <RxSpatialAccel.h>
#include <RxTrsf.h>
#include <RxVecd3.h>
#include <RxTwist.h>

class RxAccel: public RxVecdBase
{
private:
	RxSpatialAccel a;
public:

	RxVecd3 &v()
	 { return a.v();
	 }
	const RxVecd3 &v() const
	 { return a.v();
	 }
	RxVecd3 &w()
	 { return a.w();
	 }
	const RxVecd3 &w() const
	 { return a.w();
	 }

	// assignment

	RxAccel& set (double vx, double vy, double vz,
		      double wx, double wy, double wz)
	 { a.set (vx, vy, vz, wx, wy, wz);
	   return *this;
	 }
	RxAccel& set (const double *dp)
	 { a.set (dp);
	   return *this;
	 }

	RxAccel& set (const RxSpatialAccel &sa, const RxTwist &tw);

	void get (double *vxp, double *vyp, double *vzp,
		  double *wxp, double *wyp, double *wzp)
	 { a.get (vxp, vyp, vzp, wxp, wyp, wzp);
	 }
	void get (double *dp) const
	 { a.get (dp);
	 }

	// constructors

	RxAccel ()
	 {
	 }
	RxAccel (double vx, double vy, double vz, 
		 double wx, double wy, double wz)
	 { set (vx, vy, vz, wx, wy, wz);
	 }
	RxAccel (RxVecd3 &vv, RxVecd3 &ww)
	 { a.w() = ww;
	   a.v() = vv;
	 }
	RxAccel (const double *dp)
	 { set (dp);
	 }

	// query and access

	int length() const
	 { return 6;
	 }
	double& operator()(int i)
	 { return a(i);
	 }
	double operator()(int i) const
	 { return a(i);
	 }

	// T <- T,T operations

	RxAccel& add (const RxAccel &t1, const RxAccel &t2)
	 { a.add (t1.a, t2.a);
	   return *this;
	 }
	RxAccel& add (const RxAccel &t2)
	 { a.add (t2.a);
	   return *this;
	 }
	RxAccel& operator+=(const RxAccel &t2)
	 { a.add (t2.a);
	   return *this;
	 }
	friend RxAccel operator+(const RxAccel &t1, const RxAccel &t2);

	RxAccel& sub (const RxAccel &t1, const RxAccel &t2)
	 { a.sub (t1.a, t2.a);
	   return *this;
	 }
	RxAccel& sub (const RxAccel &t1)
	 { a.sub (t1.a);
	   return *this;
	 }
	RxAccel& operator-=(const RxAccel &t2)
	 { a.sub (t2.a);
	   return *this;
	 }

	friend RxAccel operator-(const RxAccel &t1, const RxAccel &t2);

	RxAccel& extrap (const RxAccel &t1, double aa, const RxAccel &t2)
	 { a.extrap (t1.a, aa, t2.a);
	   return *this;
	 }
	RxAccel& extrap (double aa, const RxAccel &t2)
	 { a.extrap (aa, t2.a);
	   return *this;
	 }
	RxAccel& interp (const RxAccel &t1, double aa, const RxAccel &t2)
	 { a.interp (t1.a, aa, t2.a);
	   return *this;
	 }
	RxAccel& interp (double aa, const RxAccel &t2)
	 { a.interp (aa, t2.a);
	   return *this;
	 }
#if 0
	RxAccel& cross (const RxAccel &t1, const RxAccel &t2)
	 { return (RxAccel&) RxSpatialVec::cross (t1, t2);
	 }
	RxAccel& cross (const RxAccel &t2)
	 { return (RxAccel&) RxSpatialVec::cross (t2);
	 }
#endif
	// T <- T,O operations

	RxAccel& scale (double aa, const RxAccel &t1)
	 { a.scale (aa, t1.a);
	   return *this;	   
	 }
	RxAccel& scale (double aa)
	 { a.scale (aa);
	   return *this;
	 }
	RxAccel& operator*=(double aa)
	 { a.scale (aa);
	   return *this;
	 }
	RxAccel& operator/=(double aa)
	 { a.scale (1/aa);
	   return *this;
	 }

	friend RxAccel operator*(const RxAccel &t1, double aa);
	friend RxAccel operator*(double aa, const RxAccel &t1);
	friend RxAccel operator/(const RxAccel &t1, double aa);

	RxAccel& xform (const RxAccel &t1, const RxRot& R)
	 { a.xform (t1.a, R);
	   return *this;
	 }
	RxAccel& xform (const RxRot& R)
	 { a.xform (R);
	   return *this;
	 }
	RxAccel& Ixform (const RxAccel &t1, const RxRot& R)
	 { a.Ixform (t1.a, R);
	   return *this;
	 }
	RxAccel& Ixform (const RxRot& R)
	 { a.Ixform (R);
	   return *this;
	 }
	RxAccel& xform (const RxAccel &t1, const RxTrsf& T, const RxVecd3 &w);
	RxAccel& xform (const RxTrsf& T, const RxVecd3 &w)
	 { return xform (*this, T, w);
	 }
	RxAccel& Ixform (const RxAccel &t1, const RxTrsf& T, const RxVecd3 &w);
	RxAccel& Ixform (const RxTrsf& T, const RxVecd3 &w)
	 { return Ixform (*this, T, w);
	 }

	// O <- T,T operations

	// T <- T operations

	RxAccel &unit(const RxAccel &t1)
	 { a.unit (t1.a);
	   return *this;
	 }
	RxAccel &unit()
	 { a.unit ();
	   return *this;
	 }
	RxAccel &neg(const RxAccel &t1)
	 { a.neg (t1.a);
	   return *this;
	 }
	RxAccel &neg()
	 { a.neg ();
	   return *this;
	 }
	friend RxAccel operator-(const RxAccel &t1);

	// T <- O operations

	RxAccel &setZero()
	 { a.setZero();
	   return *this;
	 }

	// O <- T operations

	double norm1(double cv, double cw) const
	 { return a.norm1 (cw, cv);
	 }
	double norm2(double cv, double cw) const
	 { return a.norm2 (cw, cv);
	 }
	double normInf(double cv, double cw) const
	 { return a.normInf (cw, cv);
	 }

	// constants

	static const RxAccel ZERO;
	static const RxAccel UNIT_VX;
	static const RxAccel UNIT_VY;
	static const RxAccel UNIT_VZ;
	static const RxAccel UNIT_WX;
	static const RxAccel UNIT_WY;
	static const RxAccel UNIT_WZ;
};

inline RxAccel operator+(const RxAccel &t1, const RxAccel &t2)
{
	RxAccel vr;
	vr.a.add (t1.a, t2.a);
	return vr;
}

inline RxAccel operator-(const RxAccel &t1, const RxAccel &t2)
{
	RxAccel vr;
	vr.a.sub (t1.a, t2.a);
	return vr;
}

inline RxAccel operator*(const RxAccel &t1, double aa)
{
	RxAccel vr;
	vr.a.scale (aa, t1.a);
	return vr;
}

inline RxAccel operator*(double aa, const RxAccel &t1)
{
	RxAccel vr;
	vr.a.scale (aa, t1.a);
	return vr;
}

inline RxAccel operator/(const RxAccel &t1, double aa)
{
	RxAccel vr;
	vr.a.scale (1/aa, t1.a);
	return vr;
}

inline RxAccel operator-(const RxAccel &t1)
{
	RxAccel vr;
	vr.a.neg (t1.a);
	return vr;
}

#endif
