#include <RxQtrn.h>
#include <RxRot.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>
#include <stdlib.h>
#include <RxTimeval.h>

#define EPS	(1e-10)
#define EQ(a,b)	(ABS(a-b) < EPS)

char buf[1024];

char *identStr = "[ 1.000  0.000  0.000  0.000]\n";

double identVec[] = { 1.0, 0.0, 0.0, 0.0 };

double angList[] = {   0.0,   30.0,   45.0,   90.0,   71.0,
		     180.0,  -30.0,  -90.0, -180.0 };

#define NRANDOM	1000

double randomAng()
{
	return (2*M_PI*(drand48() - 0.5));
}

int testMul (const RxQtrn &QR, const RxQtrn &Q1, const RxQtrn &Q2, double eps)
{
	RxQtrn QT;

	QT.c = Q1.c*Q2.c - Q1.v.dot(Q2.v);
	QT.v.cross (Q1.v, Q2.v);
	QT.v.extrap (Q1.c, Q2.v);
	QT.v.extrap (Q2.c, Q1.v);
	if (QT.v.equal (QR.v, eps) && ABS(QR.c-QT.c) < eps)
	 { return (1);
	 }
	else
	 { QT.print ("QT %8.3f\n");
	   QR.print ("QR %8.3f\n");
	   return (0);
	 }
}

char **argvec;
int doTiming = 0;

main(int argc, char **argv)
{
	RxQtrn Q1(identVec);
	RxQtrn Q2;
	int i, j, k;
	RxQtrn QA, QB, QC;
	double a, b, c;
	RxVecd3 axis;
	RxRot R1;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-time"))
	    { doTiming = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-time[\n", argvec[0]);
	   exit (1);
	 }

	if (doTiming)
	 { int count = 10000;

	   Q1.setRpy (randomAng(), randomAng(), randomAng());
	   Q2.setRpy (randomAng(), randomAng(), randomAng());

	   TIMEIT (QC.mul (Q1, Q2), count);
	   TIMEIT (QC.slerp (Q1, Q2, 0.5), count);
	   exit (0);
	 }

	Q1.print (buf, "%6.3f\n");

	assert (strcmp (buf, identStr) == 0);
	assert (Q1.equal (Q2.setIdent(), EPS));

	assert (Q1.length() == 4);

	assert (EQ(Q1(0), 1.0));
	assert (EQ(Q1(1), 0.0));
	assert (EQ(Q1(2), 0.0));
	assert (EQ(Q1(3), 0.0));

	for (i=0; i<4; i++)
	 { Q2(i) = (i+1)*10;
	   assert (EQ(Q2(i),(double)((i+1)*10)));
	 }
	assert (!Q2.equal (Q1, EPS));
	RxQtrn Q3(Q2);
	assert (Q2.equal (Q3, EPS));

	Q1 = Q2;
	assert (Q2.equal (Q3, EPS));

	Q1.setIdent();
	assert (EQ(Q1(0), 1.0));
	assert (EQ(Q1(1), 0.0));
	assert (EQ(Q1(2), 0.0));
	assert (EQ(Q1(3), 0.0));

	Q2.set (identVec);
	assert (Q2.equal(Q1, EPS));

	for (i=0; i<ARRAY_SIZE(angList); i++)
	 { for (j=0; j<ARRAY_SIZE(angList); j++)
	    { for (k=0; k<ARRAY_SIZE(angList); k++)
	       { Q3 = Q1.setEulD (angList[i], angList[j], angList[k]);
		 
		 QA.setAxis (RxVecd3(0,0,1), DTOR*angList[i]);
		 QB.setAxis (RxVecd3(0,1,0), DTOR*angList[j]);
		 QC.setAxis (RxVecd3(0,0,1), DTOR*angList[k]);

		 assert (Q3.equal(QA*QB*QC, EPS));  
		 assert (Q3.equal(Q1, EPS));

		 Q3 = Q1.setEul (DTOR*angList[i],DTOR*angList[j],DTOR*angList[k]);

		 assert (Q3.equal(QA*QB*QC, EPS));  
		 assert (Q3.equal(Q1, EPS));
	
		 Q1.getEul (&a, &b, &c);

		 QA.setAxis (RxVecd3(0,0,1), a);
		 QB.setAxis (RxVecd3(0,1,0), b);
		 QC.setAxis (RxVecd3(0,0,1), c);

		 assert (Q1.equal(QA*QB*QC, EPS));

		 Q1.getEulD (&a, &b, &c);

		 QA.setAxis (RxVecd3(0,0,1), DTOR*a);
		 QB.setAxis (RxVecd3(0,1,0), DTOR*b);
		 QC.setAxis (RxVecd3(0,0,1), DTOR*c);

		 assert (Q1.equal(QA*QB*QC, EPS));
		 QC.setAngsD ("zyz", angList[i], angList[j], angList[k]);

		 assert (Q1.equal(QC, EPS));
		 QC.setAngs ("zyz", DTOR*angList[i], 
				    DTOR*angList[j], DTOR*angList[k]);
		 assert (Q1.equal(QC, EPS));

		 Q3 = Q1.setRpyD (angList[i], angList[j], angList[k]);
		 
		 QA.setAxis (RxVecd3(0,0,1), DTOR*angList[i]);
		 QB.setAxis (RxVecd3(0,1,0), DTOR*angList[j]);
		 QC.setAxis (RxVecd3(1,0,0), DTOR*angList[k]);

		 assert (Q3.equal(QA*QB*QC, EPS));  
		 assert (Q1.equal(Q3, EPS));  

		 Q3 = Q1.setRpy (DTOR*angList[i],DTOR*angList[j],DTOR*angList[k]);

		 assert (Q3.equal(QA*QB*QC, EPS));  
		 assert (Q1.equal(Q3, EPS));  
	
		 Q1.getRpy (&a, &b, &c);

		 QA.setAxis (RxVecd3(0,0,1), a);
		 QB.setAxis (RxVecd3(0,1,0), b);
		 QC.setAxis (RxVecd3(1,0,0), c);

		 assert (Q1.equal(QA*QB*QC, EPS));

		 Q1.getRpyD (&a, &b, &c);

		 QA.setAxis (RxVecd3(0,0,1), DTOR*a);
		 QB.setAxis (RxVecd3(0,1,0), DTOR*b);
		 QC.setAxis (RxVecd3(1,0,0), DTOR*c);

		 assert (Q1.equal(QA*QB*QC, EPS));

		 assert (Q1.equal(QC.setAngsD ("zyx",
		    angList[i], angList[j], angList[k]), EPS));
		 assert (Q1.equal(QC.setAngs ("zyx", 
		    DTOR*angList[i], DTOR*angList[j], DTOR*angList[k]), EPS));

		 Q1.getAxis (&axis, &a);
	 	 Q1.getAxisD (&axis, &b);
		 assert (EQ(a, DTOR*b));

		 Q3 = QC.setAxis (axis, a);
		 assert (Q1.equal(QC, EPS));
		 assert (Q1.equal(Q3, EPS));

		 QC.setAxis (axis, DTOR*b);
		 assert (Q1.equal(QC, EPS));

		 Q3 = QC.setAxisD (axis, b);
		 assert (Q1.equal(QC, EPS));
		 assert (Q1.equal(Q3, EPS));

		 Q1.getRot (&R1);

		 QC.setVecs ("xy", R1.n, R1.o);
		 assert (Q1.equal(QC, EPS));
		 QC.setVecs ("yz", R1.o, R1.a);
		 assert (Q1.equal(QC, EPS));
		 QC.setVecs ("zx", R1.a, R1.n);
		 assert (Q1.equal(QC, EPS));

		 QC.setVecs ("yx", R1.o, R1.n);
		 assert (Q1.equal(QC, EPS));
		 QC.setVecs ("zy", R1.a, R1.o);
		 assert (Q1.equal(QC, EPS));
		 QC.setVecs ("xz", R1.n, R1.a);
		 assert (Q1.equal(QC, EPS));

		 RxQtrn QX(angList[i], angList[j], angList[k]);
		 assert (Q1.equal (QX, EPS));
	       }
	    }
	 }

	Q1.setRpyD (11.0, 22.0, 33.0);
	assert (Q1.equal (RxQtrn(11.0, 22.0, 33.0), EPS));

	for (i=0; i<NRANDOM; i++)
	 { RxVecd3 axis0;
	   double ang0;
	   RxVecd3 axisS;
	   double angS;
	   double u;
		   
	   Q1.setRpy (randomAng(), randomAng(), randomAng());
	   Q2.setRpy (randomAng(), randomAng(), randomAng());

	   assert (testMul (Q3.mul(Q1,Q2), Q1, Q2, EPS));
	   Q3 = Q1;
	   assert (testMul (Q3.mul(Q2), Q1, Q2, EPS));
	   Q3 = Q1;
	   assert (testMul (Q3 *= Q2, Q1, Q2, EPS));
	   assert (testMul (Q1*Q2, Q1, Q2, EPS));

	   assert (testMul (Q3.mulI(Q1,Q2), Q1, Q2.I(), EPS));
	   Q3 = Q1;
	   assert (testMul (Q3.mulI(Q2), Q1, Q2.I(), EPS));

	   assert (testMul (Q3.Imul(Q1,Q2), Q1.I(), Q2, EPS));
	   Q3 = Q1;
	   assert (testMul (Q3.Imul(Q2), Q1.I(), Q2, EPS));
	   
	   assert (testMul (Q3.ImulI(Q1,Q2), Q1.I(), Q2.I(), EPS));
	   Q3 = Q1;
	   assert (testMul (Q3.ImulI(Q2), Q1.I(), Q2.I(), EPS));
	
	   QA.inv(Q1);
	   QB.setIdent();
	   assert (QB.equal (QA*Q1, EPS));
	   QC = Q1;
	   assert (QA.equal (QC.inv(), EPS));
	   assert (QA.equal (Q1.I(), EPS));

	   QC = Q1;
	   QC *= 5.4;
	   QC.unit();
	   assert (Q1.equal (QC, EPS));

#if 1
	   QC.slerp (Q1, Q2, 1.0);
	   assert (QC.equal (Q2, EPS));
	   QC = Q1;
	   QC.slerp (Q2, 0.0);
	   assert (QC.equal (Q1, EPS));
	   QC.slerp (Q1, Q2, 0.5);
	   QB.Imul (Q1, QC);
	   QA.mul (QC, QB);
	   assert (QA.equal (Q2, EPS));

	   QC.Imul (Q1, Q2);
	   QC.getAxis (&axis0, &ang0);
	   for (u=0.1; u<=1.0; u+=0.1)
	    { QB.slerp (Q1, Q2, u);
	      QC.Imul (Q1, QB);
	      QC.getAxis (&axisS, &angS);
	      assert (axisS.equal(axis0,EPS));
	      assert (ABS(ang0*u-angS) < EPS);
	    }

	   QC.slerp (Q1, Q1, 0.5);
	   assert (QC.equal (Q1, EPS));
#endif
	 }

	printf ("\nROUTINES OK\n\n");
	return 0;
}
