#include <RxRot.h>
#include <RxMatdBaseTest.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>
#include <stdlib.h>
#include <RxTimeval.h>
#include <RxError.h>

#define EPS	(1e-10)
#define EQ(a,b)	(ABS(a-b) < EPS)

char buf0[1024];
char buf1[1024];

char *identStr = "\
[ 1.000  0.000  0.000\n  0.000  1.000  0.000\n  0.000  0.000  1.000]\n";

double identVec[] = { 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 };

double angList[] = {   0.0,   30.0,   45.0,   90.0,   71.0,
		     180.0,  -30.0,  -90.0, -180.0 };

#define NRANDOM	1000

double randomAng()
{
	return (2*M_PI*(drand48() - 0.5));
}

char **argvec;
int doTiming = 0;

main(int argc, char **argv)
{
	RxRot R1(identVec);
	RxRot R2;
	int i, j, k;
	RxRot RA, RB, RC;
	double a, b, c;
	RxVecd3 axis;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-time"))
	    { doTiming = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-time[\n", argvec[0]);
	   exit (1);
	 }

	if (doTiming)
	 { int count = 10000;

	   R1.setRpy (randomAng(), randomAng(), randomAng());
	   R2.setRpy (randomAng(), randomAng(), randomAng());

	   TIMEIT (RC.mul (R1, R2), count);
	   TIMEIT (RC.slerp (R1, R2, 0.5), count);
	   exit (0);
	 }

	R1.print (buf0, "%M%6.3f\n");

	assert (strcmp (buf0, identStr) == 0);
	assert (R1.equal (R2.setIdent(), EPS));

	assert (R1.nrows() == 3);
	assert (R1.ncols() == 3);

	for (i=0; i<3; i++)
	 { for (j=0; j<3; j++)
	    { if (i==j)
	       { assert (EQ(R1(i,j),1.0));
	       }
	      else
	       { assert (EQ(R1(i,j),0.0));
	       }
	    }
	 }

	for (i=0; i<3; i++)
	 { for (j=0; j<3; j++)
	    { R2(i,j) = (i+1)*10 + j+1;
	      assert (EQ(R2(i,j),(double)((i+1)*10 + j+1)));
	    }
	 }
	assert (!R2.equal (R1, EPS));
	RxRot R3(R2);
	assert (R2.equal (R3, EPS));

	R1 = R2;
	assert (R2.equal (R3, EPS));

	R1.setIdent();
	for (i=0; i<3; i++)
	 { for (j=0; j<3; j++)
	    { if (i==j)
	       { assert (EQ(R1(i,j),1.0));
	       }
	      else
	       { assert (EQ(R1(i,j),0.0));
	       }
	    }
	 }
	R2.set (identVec);
	assert (R2.equal(R1, EPS));

	for (i=0; i<ARRAY_SIZE(angList); i++)
	 { for (j=0; j<ARRAY_SIZE(angList); j++)
	    { for (k=0; k<ARRAY_SIZE(angList); k++)
	       { R3 = R1.setEulD (angList[i], angList[j], angList[k]);
		 
		 RA.setAxis (RxVecd3(0,0,1), DTOR*angList[i]);
		 RB.setAxis (RxVecd3(0,1,0), DTOR*angList[j]);
		 RC.setAxis (RxVecd3(0,0,1), DTOR*angList[k]);

		 assert (R3.equal(RA*RB*RC, EPS));  
		 assert (R3.equal(R1, EPS));

		 R3 = R1.setEul (DTOR*angList[i],DTOR*angList[j],DTOR*angList[k]);

		 assert (R3.equal(RA*RB*RC, EPS));  
		 assert (R3.equal(R1, EPS));
	
		 R1.getEul (&a, &b, &c);

		 RA.setAxis (RxVecd3(0,0,1), a);
		 RB.setAxis (RxVecd3(0,1,0), b);
		 RC.setAxis (RxVecd3(0,0,1), c);

		 assert (R1.equal(RA*RB*RC, EPS));

		 R1.getEulD (&a, &b, &c);

		 RA.setAxis (RxVecd3(0,0,1), DTOR*a);
		 RB.setAxis (RxVecd3(0,1,0), DTOR*b);
		 RC.setAxis (RxVecd3(0,0,1), DTOR*c);

		 assert (R1.equal(RA*RB*RC, EPS));
		 assert (R1.equal(RC.setAngsD ("zyz",
		    angList[i], angList[j], angList[k]), EPS));
		 assert (R1.equal(RC.setAngs ("zyz", 
		    DTOR*angList[i], DTOR*angList[j], DTOR*angList[k]), EPS));

		 R3 = R1.setRpyD (angList[i], angList[j], angList[k]);
		 
		 RA.setAxis (RxVecd3(0,0,1), DTOR*angList[i]);
		 RB.setAxis (RxVecd3(0,1,0), DTOR*angList[j]);
		 RC.setAxis (RxVecd3(1,0,0), DTOR*angList[k]);

		 assert (R3.equal(RA*RB*RC, EPS));  
		 assert (R1.equal(R3, EPS));  

		 R3 = R1.setRpy (DTOR*angList[i],DTOR*angList[j],DTOR*angList[k]);

		 assert (R3.equal(RA*RB*RC, EPS));  
		 assert (R1.equal(R3, EPS));  
	
		 R1.getRpy (&a, &b, &c);

		 RA.setAxis (RxVecd3(0,0,1), a);
		 RB.setAxis (RxVecd3(0,1,0), b);
		 RC.setAxis (RxVecd3(1,0,0), c);

		 assert (R1.equal(RA*RB*RC, EPS));

		 R1.getRpyD (&a, &b, &c);

		 RA.setAxis (RxVecd3(0,0,1), DTOR*a);
		 RB.setAxis (RxVecd3(0,1,0), DTOR*b);
		 RC.setAxis (RxVecd3(1,0,0), DTOR*c);

		 assert (R1.equal(RA*RB*RC, EPS));

		 assert (R1.equal(RC.setAngsD ("zyx",
		    angList[i], angList[j], angList[k]), EPS));
		 assert (R1.equal(RC.setAngs ("zyx", 
		    DTOR*angList[i], DTOR*angList[j], DTOR*angList[k]), EPS));

		 R1.getAxis (&axis, &a);
	 	 R1.getAxisD (&axis, &b);
		 assert (EQ(a, DTOR*b));

		 R3 = RC.setAxis (axis, a);
		 assert (R1.equal(RC, EPS));
		 assert (R1.equal(R3, EPS));

		 RC.setAxis (axis, DTOR*b);
		 assert (R1.equal(RC, EPS));

		 R3 = RC.setAxisD (axis, b);
		 assert (R1.equal(RC, EPS));
		 assert (R1.equal(R3, EPS));

		 RC.setVecs ("xy", R1.n, R1.o);
		 assert (R1.equal(RC, EPS));
		 RC.setVecs ("yz", R1.o, R1.a);
		 assert (R1.equal(RC, EPS));
		 RC.setVecs ("zx", R1.a, R1.n);
		 assert (R1.equal(RC, EPS));

		 RC.setVecs ("yx", R1.o, R1.n);
		 assert (R1.equal(RC, EPS));
		 RC.setVecs ("zy", R1.a, R1.o);
		 assert (R1.equal(RC, EPS));
		 RC.setVecs ("xz", R1.n, R1.a);
		 assert (R1.equal(RC, EPS));

		 RxRot RX(angList[i], angList[j], angList[k]);
		 assert (R1.equal (RX, EPS));
	       }
	    }
	 }

	R1.setRpyD (11.0, 22.0, 33.0);
	assert (R1.equal (RxRot(11.0, 22.0, 33.0), EPS));

	for (i=0; i<NRANDOM; i++)
	 { R1.setRpy (randomAng(), randomAng(), randomAng());
	   R2.setRpy (randomAng(), randomAng(), randomAng());

	   assert (testMul (R3.mul(R1,R2), R1, R2, EPS));
	   R3 = R1;
	   assert (testMul (R3.mul(R2), R1, R2, EPS));
	   R3 = R1;
	   assert (testMul (R3 *= R2, R1, R2, EPS));
	   assert (testMul (R1*R2, R1, R2, EPS));

	   assert (testTsp (RB.tsp(R2), R2, EPS));
	   RB = R2;
	   assert (testTsp (RB.tsp(), R2, EPS));
	   assert (testTsp (R2.T(), R2, EPS));

	   assert (testMul (R3.mulI(R1,R2), R1, R2.T(), EPS));
	   R3 = R1;
	   assert (testMul (R3.mulI(R2), R1, R2.T(), EPS));

	   assert (testMul (R3.Imul(R1,R2), R1.T(), R2, EPS));
	   R3 = R1;
	   assert (testMul (R3.Imul(R2), R1.T(), R2, EPS));
	   
	   assert (testMul (R3.ImulI(R1,R2), R1.T(), R2.T(), EPS));
	   R3 = R1;
	   assert (testMul (R3.ImulI(R2), R1.T(), R2.T(), EPS));
	
	   RA.tsp(R1);
	   assert (RA.equal (RC.inv(R1), EPS));
	   RC = R1;
	   assert (RA.equal (RC.inv(), EPS));
	   assert (RA.equal (R1.I(), EPS));

	   RC = R1;
	   RC.a.scale (2.1);
	   RC.o.extrap (1.0, RC.a);
	   RC.n.setZero();
	   assert (R1.equal (R3.normalize(RC), EPS));
	   assert (R1.equal (RC.normalize(), EPS));

	   RC.slerp (R1, R2, 1.0);
	   assert (RC.equal (R2, EPS));
	   RC = R1;
	   RC.slerp (R2, 0.0);
	   assert (RC.equal (R1, EPS));
	   RC.slerp (R1, R2, 0.5);
	   RB.Imul (R1, RC);
	   RA.mul (RC, RB);
	   assert (RA.equal (R2, EPS));
	 }

	assert (RA.setIdent().equal(RxRot::IDENT, EPS));

	RxRot R;
	
	char *Rdef1 = "\
R:[   0.354   -0.707    0.612\n\
      0.354    0.707    0.612\n\
     -0.866    0.000    0.500]";

	if (R.scan (Rdef1) != 1)
	 { RxErrCtx::printCur(stdout);
	   exit (1);
	 }
	assert (R.scan (Rdef1) == 1);
	R.print (buf0, "R:%M%8.3f");
	assert (strcmp (buf0, Rdef1) == 0);

	R.print (buf0, "R=%M%8.3f,%7.2f");

	assert (strcmp (buf0, "\
R=[   0.354,  -0.71,   0.61\n\
      0.354,   0.71,   0.61\n\
     -0.866,   0.00,   0.50]") == 0);

	char *Rdef2 = "\
R: [-0.31882  0.76996  0.55273   73.709]";
	assert (R.scan (Rdef2) == 1);
	R.print (buf0, "R: %8.3f%8.5f");
	assert (strcmp (buf0, Rdef2) == 0);

	R.print (buf0, "Roo: ");
	R.print (buf1, "Roo: %8.3f\n");
	assert (strcmp (buf0, buf1) ==0);


	printf ("\nROUTINES OK\n\n");
	return 0;
}
