#include <RxSpatialAccel.h>
#include <RxVecdBaseTest.h>
#include <RxTrsf.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	6

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 },
 { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0 },
 {-3.0, -2.0, -1.0, -0.5, 1.1, 2.2 }};

double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};

int testSvUnit (const RxSpatialAccel &au, const RxSpatialAccel &a1,
		double eps)
{
	RxSpatialAccel chk;

	chk.v().unit(a1.v());
	chk.w().unit(a1.w());
	return (chk.equal(au, eps));	
}

int testSvNorm1 (double nc, const RxSpatialAccel &a1,
		 double cv, double cw, double eps)
{
	double chk;

	chk = cv*a1.v().norm1() + cw*a1.w().norm1();
	return (ABS(chk-nc) <= eps);
}

int testSvNorm2 (double nc, const RxSpatialAccel &a1,
		 double cv, double cw, double eps)
{
	double chk;

	chk = cv*a1.v().norm2() + cw*a1.w().norm2();
	return (ABS(chk-nc) <= eps);
}

int testSvNormInf (double nc, const RxSpatialAccel &a1,
		 double cv, double cw, double eps)
{
	double chk, amax, bmax;

	amax = cv*a1.v().normInf();
	bmax = cw*a1.w().normInf();
	return (ABS(MAX(amax,bmax)-nc) <= eps);
}

main()
{
	double vec[LEN];
	RxSpatialAccel a1(10,20,30,40,50,60);
	RxSpatialAccel a2(100,200,300,400,500,600);
	int i, j, k;

	assert (EQ(a1.v().x(),10) && EQ(a1.v().y(),20) && EQ(a1.v().z(),30));
	assert (EQ(a1.w().x(),40) && EQ(a1.w().y(),50) && EQ(a1.w().z(),60));
	
	RxSpatialAccel a3(a2);

	assert (EQ(a3.v().x(),100) && EQ(a3.v().y(),200) && EQ(a3.v().z(),300));
	assert (EQ(a3.w().x(),400) && EQ(a3.w().y(),500) && EQ(a3.w().z(),600));

	assert (a3.length() == LEN);

	assert (a3.equal(a2,EPS));
	assert (!a3.equal(a1,EPS));
	for (i=0;i<LEN;i++)
	 { a1(i) = i*10+1;
	 }
	assert (a1.equal(RxSpatialAccel(1,11,21,31,41,51),EPS));
	
	a1 = a3;

	assert (a3.equal(a1,EPS));
	
	a1.set(11,22,33,44,55,66);

	assert (a1.equal(RxSpatialAccel(11,22,33,44,55,66),EPS));
	
	a1.get(vec);
	for (i=0; i<LEN; i++)
	 { assert (EQ(vec[i],(i+1)*11));
	 }
	for (i=0; i<LEN; i++)
	 { vec[i] = (i+1)*111;
	 }
	assert ((a3.set(vec)).equal(RxSpatialAccel(111,222,333,444,555,666),EPS));

	a3 = a1;
	assert (testAdd (a3.add(a3.add(a1)), 2*a1, 2*a1, EPS));
	
	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { a1.set(testVals[i]);
	      a2.set(testVals[j]);

	      assert (testAdd (a1+a2, a1, a2, EPS));
	      assert (testAdd (a3.add(a1,a2), a1, a2, EPS));
	      a3 = a1;
	      assert (testAdd (a3.add(a2), a1, a2, EPS));
	      a3 = a1;					    
	      assert (testAdd (a3 += a2, a1, a2, EPS));

	      assert (testSub (a1-a2, a1, a2, EPS));
	      assert (testSub (a3.sub(a1,a2), a1, a2, EPS));
	      a3 = a1;
	      assert (testSub (a3.sub(a2), a1, a2, EPS));
	      a3 = a1;					    
	      assert (testSub (a3 -= a2, a1, a2, EPS));

	      assert (testExtrap (a3.extrap(a1,11.1,a2), a1, 11.1, a2, EPS));
	      a3 = a1;
	      assert (testExtrap (a3.extrap(11.1,a2), a1, 11.1, a2, EPS));

#if 0
	      a3 = a1;
	      assert (testDot (a3.dot(a2), a1, a2, EPS));
#endif
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { a1.set(testVals[i]);

	   assert (testScale (a3.scale(M_PI,a1), M_PI, a1, EPS));
	   a3 = a1;
	   assert (testScale (a3.scale(M_PI), M_PI, a1, EPS));
	   a3 = a1;
	   assert (testScale (a3 *= M_PI, M_PI, a1, EPS));
	   a3 = a1;
	   assert (testScale (a3 /= M_PI, 1/M_PI, a1, EPS));
	   assert (testScale (M_PI*a1, M_PI, a1, EPS));
	   assert (testScale (a1*M_PI, M_PI, a1, EPS));
	   assert (testScale (a1/M_PI, 1/M_PI, a1, EPS));

	   assert (testSvUnit (a3.unit(a1), a1, EPS));
	   a3 = a1;
	   assert (testSvUnit (a3.unit(), a1, EPS));
	   assert (testScale (a3.neg(a1), -1.0, a1, EPS));
	   a3 = a1;
	   assert (testScale (a3.neg(), -1.0, a1, EPS));
	   assert (testScale (-a1, -1.0, a1, EPS));

	   assert (testSvNorm1 (a1.norm1(1, M_PI), a1, 1, M_PI, EPS));
	   assert (testSvNorm2 (a1.norm2(1, M_PI), a1, 1, M_PI, EPS));
	   assert (testSvNormInf (a1.normInf(1, M_PI), a1, 1, M_PI, EPS));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { a1.set(testVals[i]);
 
	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxTrsf T;

	      T.R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      a3.xform (a1, T.R);
	      a3.Ixform (a3, T.R);
	      assert (a3.equal(a1,EPS));
	      a3 = a1;
	      a3.xform (T.R);
	      a3.Ixform (T.R);
	      assert (a3.equal(a1,EPS));

	      for (k=0; k<ARRAY_SIZE(testVals); k++)
	       { T.p.set (testVals[k][0], testVals[k][1], testVals[k][2]);

		 a3.xform (a1, T);
		 a3.Ixform (a3, T);
		 assert (a3.equal(a1,EPS));
		 a3 = a1;
		 a3.xform (T);
		 a3.Ixform (T);
		 assert (a3.equal(a1,EPS));
	       }
	    }
	 }

	a1.set(0,0,0,0,0,0);
	assert ((a2.setZero()).equal(a1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}

