#include <RxSpatialMat.h>
#include <RxTrsf.h>
#include <RxMatd.h>
#include <RxMatdLu.h>
#include <RxMathError.h>

static double zero[36];
static double ident[36] =
 { 1,0,0,0,0,0, 0,1,0,0,0,0, 0,0,1,0,0,0, 
   0,0,0,1,0,0, 0,0,0,0,1,0, 0,0,0,0,0,1 };

const RxSpatialMat RxSpatialMat::IDENT (ident);
const RxSpatialMat RxSpatialMat::ZERO (zero);

const int RxSpatialMat::width = 6;
const int RxSpatialMat::nr = 6;
const int RxSpatialMat::nc = 6;

void RxSpatialMat::doCopy (const RxSpatialMat &m1)
{
	A_ = m1.A_;
	B_ = m1.B_;
	C_ = m1.C_;
	D_ = m1.D_;
}

RxSpatialMat& RxSpatialMat::set(const RxSpatialMat &m2)
{
	A_ = m2.A_;
	B_ = m2.B_;
	C_ = m2.C_;
	D_ = m2.D_;
	return *this;
}

RxSpatialMat& RxSpatialMat::set (const double *dp)
{
	// Ugh!! Should do this with more finesse!

	A_.xx() = dp[ 0];
	A_.xy() = dp[ 1];
	A_.xz() = dp[ 2];
	A_.yx() = dp[ 6];
	A_.yy() = dp[ 7];
	A_.yz() = dp[ 8];
	A_.zx() = dp[12];
	A_.zy() = dp[13];
	A_.zz() = dp[14];

	B_.xx() = dp[ 3];
	B_.xy() = dp[ 4];
	B_.xz() = dp[ 5];
	B_.yx() = dp[ 9];
	B_.yy() = dp[10];
	B_.yz() = dp[11];
	B_.zx() = dp[15];
	B_.zy() = dp[16];
	B_.zz() = dp[17];

	C_.xx() = dp[18];
	C_.xy() = dp[19];
	C_.xz() = dp[20];
	C_.yx() = dp[24];
	C_.yy() = dp[25];
	C_.yz() = dp[26];
	C_.zx() = dp[30];
	C_.zy() = dp[31];
	C_.zz() = dp[32];

	D_.xx() = dp[21];
	D_.xy() = dp[22];
	D_.xz() = dp[23];
	D_.yx() = dp[27];
	D_.yy() = dp[28];
	D_.yz() = dp[29];
	D_.zx() = dp[33];
	D_.zy() = dp[34];
	D_.zz() = dp[35];

	return *this;
}

void RxSpatialMat::get (double *dp) const
{
	// Ugh!! Should do this with more finesse!

	dp[ 0] = A_.xx();
	dp[ 1] = A_.xy();
	dp[ 2] = A_.xz();
	dp[ 6] = A_.yx();
	dp[ 7] = A_.yy();
	dp[ 8] = A_.yz();
	dp[12] = A_.zx();
	dp[13] = A_.zy();
	dp[14] = A_.zz();

	dp[ 3] = B_.xx();
	dp[ 4] = B_.xy();
	dp[ 5] = B_.xz();
	dp[ 9] = B_.yx();
	dp[10] = B_.yy();
	dp[11] = B_.yz();
	dp[15] = B_.zx();
	dp[16] = B_.zy();
	dp[17] = B_.zz();

	dp[18] = C_.xx();
	dp[19] = C_.xy();
	dp[20] = C_.xz();
	dp[24] = C_.yx();
	dp[25] = C_.yy();
	dp[26] = C_.yz();
	dp[30] = C_.zx();
	dp[31] = C_.zy();
	dp[32] = C_.zz();

	dp[21] = D_.xx();
	dp[22] = D_.xy();
	dp[23] = D_.xz();
	dp[27] = D_.yx();
	dp[28] = D_.yy();
	dp[29] = D_.yz();
	dp[33] = D_.zx();
	dp[34] = D_.zy();
	dp[35] = D_.zz();
}

RxSpatialMat& RxSpatialMat::set (const RxMatdBase &m1)
{
	if (nr != m1.nrows() || nc != m1.ncols())
	 { RxMathError (RxMathError::BadSize,
			"RxMatd3x3::operator=: %K\n").raise();
	   return (*this);
	 }
	A_.xx() = m1(0,0);
	A_.xy() = m1(0,1);
	A_.xz() = m1(0,2);
	A_.yx() = m1(1,0);
	A_.yy() = m1(1,1);
	A_.yz() = m1(1,2);
	A_.zx() = m1(2,0);
	A_.zy() = m1(2,1);
	A_.zz() = m1(2,2);

	B_.xx() = m1(0,3);
	B_.xy() = m1(0,4);
	B_.xz() = m1(0,5);
	B_.yx() = m1(1,3);
	B_.yy() = m1(1,4);
	B_.yz() = m1(1,5);
	B_.zx() = m1(2,3);
	B_.zy() = m1(2,4);
	B_.zz() = m1(2,5);

	C_.xx() = m1(3,0);
	C_.xy() = m1(3,1);
	C_.xz() = m1(3,2);
	C_.yx() = m1(4,0);
	C_.yy() = m1(4,1);
	C_.yz() = m1(4,2);
	C_.zx() = m1(5,0);
	C_.zy() = m1(5,1);
	C_.zz() = m1(5,2);

	D_.xx() = m1(3,3);
	D_.xy() = m1(3,4);
	D_.xz() = m1(3,5);
	D_.yx() = m1(4,3);
	D_.yy() = m1(4,4);
	D_.yz() = m1(4,5);
	D_.zx() = m1(5,3);
	D_.zy() = m1(5,4);
	D_.zz() = m1(5,5);
}

RxSpatialMat& RxSpatialMat::set (const RxMatd &m1)
{
	if (nr != m1.nrows() || nc != m1.ncols())
	 { RxMathError (RxMathError::BadSize,
			"RxMatd3x3::operator=: %K\n").raise();
	   return (*this);
	 }
	double *dp1 = m1.buf();

	A_.xx() = *dp1++;
	A_.xy() = *dp1++;
	A_.xz() = *dp1++;
	B_.xx() = *dp1++;
	B_.xy() = *dp1++;
	B_.xz() = *dp1++;
	dp1 += (m1.width-m1.nc);

	A_.yx() = *dp1++;
	A_.yy() = *dp1++;
	A_.yz() = *dp1++;
	B_.yx() = *dp1++;
	B_.yy() = *dp1++;
	B_.yz() = *dp1++;
	dp1 += (m1.width-m1.nc);

	A_.zx() = *dp1++;
	A_.zy() = *dp1++;
	A_.zz() = *dp1++;
	B_.zx() = *dp1++;
	B_.zy() = *dp1++;
	B_.zz() = *dp1++;
	dp1 += (m1.width-m1.nc);

	C_.xx() = *dp1++;
	C_.xy() = *dp1++;
	C_.xz() = *dp1++;
	D_.xx() = *dp1++;
	D_.xy() = *dp1++;
	D_.xz() = *dp1++;
	dp1 += (m1.width-m1.nc);

	C_.yx() = *dp1++;
	C_.yy() = *dp1++;
	C_.yz() = *dp1++;
	D_.yx() = *dp1++;
	D_.yy() = *dp1++;
	D_.yz() = *dp1++;
	dp1 += (m1.width-m1.nc);

	C_.zx() = *dp1++;
	C_.zy() = *dp1++;
	C_.zz() = *dp1++;
	D_.zx() = *dp1++;
	D_.zy() = *dp1++;
	D_.zz() = *dp1++;
	dp1 += (m1.width-m1.nc);
}

double& RxSpatialMat::operator()(int i, int j)
{
	if (i < 3)
	 { if (j < 3)
	    { return A_(i,j);
	    }
	   else
	    { return B_(i,j-3);
	    }
	 }
	else
	 { if (j < 3)
	    { return C_(i-3,j);
	    }
	   else
	    { return D_(i-3,j-3);
	    }
	 }
}

double RxSpatialMat::operator()(int i, int j) const
{
	if (i < 3)
	 { if (j < 3)
	    { return A_(i,j);
	    }
	   else
	    { return B_(i,j-3);
	    }
	 }
	else
	 { if (j < 3)
	    { return C_(i-3,j);
	    }
	   else
	    { return D_(i-3,j-3);
	    }
	 }
}

RxSpatialMat& RxSpatialMat::add(const RxSpatialMat &m1, const RxSpatialMat &m2)
{
	A_.add (m1.A_, m2.A_);
	B_.add (m1.B_, m2.B_);
	C_.add (m1.C_, m2.C_);
	D_.add (m1.D_, m2.D_);
}
	
RxSpatialMat& RxSpatialMat::sub(const RxSpatialMat &m1, const RxSpatialMat &m2)
{
	A_.sub (m1.A_, m2.A_);
	B_.sub (m1.B_, m2.B_);
	C_.sub (m1.C_, m2.C_);
	D_.sub (m1.D_, m2.D_);
	return *this;
}

RxSpatialMat& RxSpatialMat::mul (const RxSpatialMat &m1, 
				 const RxSpatialMat &m2)
{
	RxMatd3x3 X_;

	A_.mul (m1.A_, m2.A_);
	X_.mul (m1.B_, m2.C_);
	A_.add (X_);

	B_.mul (m1.A_, m2.B_);
	X_.mul (m1.B_, m2.D_);
	B_.add (X_);

	C_.mul (m1.C_, m2.A_);
	X_.mul (m1.D_, m2.C_);
	C_.add (X_);

	D_.mul (m1.C_, m2.B_);
	X_.mul (m1.D_, m2.D_);
	D_.add (X_);
	return *this;
}

RxSpatialMat& RxSpatialMat::scale (double a, const RxSpatialMat &m1)
{
	A_.scale (a, m1.A_);
	B_.scale (a, m1.B_);
	C_.scale (a, m1.C_);
	D_.scale (a, m1.D_);
	return *this;
}

RxSpatialMat& RxSpatialMat::xform (const RxSpatialMat &m1, const RxRot& R)
{
	A_.xform (m1.A_, R);
	B_.xform (m1.B_, R);
	C_.xform (m1.C_, R);
	D_.xform (m1.D_, R);
	return *this;
}

RxSpatialMat& RxSpatialMat::Ixform (const RxSpatialMat &m1, const RxRot& R)
{
	A_.Ixform (m1.A_, R);
	B_.Ixform (m1.B_, R);
	C_.Ixform (m1.C_, R);
	D_.Ixform (m1.D_, R);
	return *this;
}

RxSpatialMat& RxSpatialMat::xform (const RxSpatialMat &m1, const RxTrsf& T)
{
	RxMatd3x3 X1;
	RxMatd3x3 X2;

	X1.cross (m1.B_, T.p);
	X1.add (m1.A_);
	A_.xform (X1, T.R);

	X1.cross (T.p, X1);
	X2.cross (m1.D_, T.p);
	X2.sub (X1);
	X2.add (m1.C_);
	C_.xform (X2, T.R);

	X1.cross (T.p, m1.B_);
	X2.sub (m1.D_, X1);
	D_.xform (X2, T.R);

	B_.xform (m1.B_, T.R);
	return *this;
}

RxSpatialMat& RxSpatialMat::Ixform (const RxSpatialMat &m1, const RxTrsf& T)
{
	RxMatd3x3 X1;
	RxMatd3x3 X2;

	B_.Ixform (m1.B_, T.R);
	X1.cross (B_, T.p);
	A_.Ixform (m1.A_, T.R);
	A_.sub (X1);

	X1.Ixform (m1.D_, T.R);
	D_.cross (T.p, B_);
	D_.add (X1);

	X1.cross (X1, T.p);
	X2.cross (T.p, A_);
	C_.Ixform (m1.C_, T.R);
	C_.sub (X1);
	C_.add (X2);

	return *this;
}

RxSpatialVec &RxSpatialMat::Vmul (RxSpatialVec &vr, 
				  const RxSpatialVec &v1) const
{
	double ax, ay, az;
	double bx, by, bz;

	ax = v1.a().x();
	ay = v1.a().y();
	az = v1.a().z();
	bx = v1.b().x();
	by = v1.b().y();
	bz = v1.b().z();

	vr.a().x() = A_.xx()*bx + A_.yx()*by + A_.zx()*bz +
		     C_.xx()*ax + C_.yx()*ay + C_.zx()*az;
	vr.a().y() = A_.xy()*bx + A_.yy()*by + A_.zy()*bz +
		     C_.xy()*ax + C_.yy()*ay + C_.zy()*az;
	vr.a().z() = A_.xz()*bx + A_.yz()*by + A_.zz()*bz +
		     C_.xz()*ax + C_.yz()*ay + C_.zz()*az;

	vr.b().x() = B_.xx()*bx + B_.yx()*by + B_.zx()*bz +
		     D_.xx()*ax + D_.yx()*ay + D_.zx()*az;
	vr.b().y() = B_.xy()*bx + B_.yy()*by + B_.zy()*bz +
		     D_.xy()*ax + D_.yy()*ay + D_.zy()*az;
	vr.b().z() = B_.xz()*bx + B_.yz()*by + B_.zz()*bz +
		     D_.xz()*ax + D_.yz()*ay + D_.zz()*az;

	return vr;
}

RxSpatialVec &RxSpatialMat::mulV (RxSpatialVec &vr, 
				  const RxSpatialVec &v1) const
{
	double ax, ay, az;
	double bx, by, bz;

	ax = v1.a().x();
	ay = v1.a().y();
	az = v1.a().z();
	bx = v1.b().x();
	by = v1.b().y();
	bz = v1.b().z();

	vr.a().x() = A_.xx()*ax + A_.xy()*ay + A_.xz()*az +
		     B_.xx()*bx + B_.xy()*by + B_.xz()*bz;
	vr.a().y() = A_.yx()*ax + A_.yy()*ay + A_.yz()*az +
		     B_.yx()*bx + B_.yy()*by + B_.yz()*bz;
	vr.a().z() = A_.zx()*ax + A_.zy()*ay + A_.zz()*az +
		     B_.zx()*bx + B_.zy()*by + B_.zz()*bz;

	vr.b().x() = C_.xx()*ax + C_.xy()*ay + C_.xz()*az +
		     D_.xx()*bx + D_.xy()*by + D_.xz()*bz;
	vr.b().y() = C_.yx()*ax + C_.yy()*ay + C_.yz()*az +
		     D_.yx()*bx + D_.yy()*by + D_.yz()*bz;
	vr.b().z() = C_.zx()*ax + C_.zy()*ay + C_.zz()*az +
		     D_.zx()*bx + D_.zy()*by + D_.zz()*bz;

	return vr;
}

RxSpatialMat &RxSpatialMat::stp(const RxSpatialMat &m1)
{
	// spatial transpose
	RxMatd3x3 XX;

	XX = m1.A_;
	A_.tsp (m1.D_);
	D_.tsp (XX);
	B_.tsp (m1.B_);
	C_.tsp (m1.C_);
}

RxSpatialMat &RxSpatialMat::stp()
{
	// spatial transpose
	RxMatd3x3 XX;

	XX = A_;
	A_.tsp (D_);
	D_.tsp (XX);
	B_.tsp ();
	C_.tsp ();
}

RxSpatialMat &RxSpatialMat::inv(const RxSpatialMat &m1)
{
	RxMatdLu lu;
	RxMatd mtmp(m1);

	lu.calc (mtmp);
	lu.inv (&mtmp);

	this->set(mtmp);
	return *this;
}

RxSpatialMat &RxSpatialMat::neg(const RxSpatialMat &m1)
{
	A_.neg (m1.A_);
	B_.neg (m1.B_);
	C_.neg (m1.C_);
	D_.neg (m1.D_);
	return *this;
}

RxSpatialMat& RxSpatialMat::setZero()
{
	set (zero);
	return *this;
}

RxSpatialMat& RxSpatialMat::setIdent()
{
	set (ident);
	return *this;
}

double RxSpatialMat::det () const
{
	RxMatd mtmp(*this);

	return mtmp.det();	
}

double RxSpatialMat::norm1() const
{
	RxMatd mtmp(*this);

	return mtmp.norm1();
}

double RxSpatialMat::normInf() const
{
	RxMatd mtmp(*this);

	return mtmp.normInf();
}

double RxSpatialMat::normF() const
{
	// Ugh! Not sure why I decided to expand this when I didn't
	// expand the other calls. Oh well :) 

	double sum;

	sum = A_.xx()*A_.xx() + A_.xy()*A_.xy() + A_.xz()*A_.xz() + 
	      A_.yx()*A_.yx() + A_.yy()*A_.yy() + A_.yz()*A_.yz() + 
	      A_.zx()*A_.zx() + A_.zy()*A_.zy() + A_.zz()*A_.zz() + 

	      B_.xx()*B_.xx() + B_.xy()*B_.xy() + B_.xz()*B_.xz() + 
	      B_.yx()*B_.yx() + B_.yy()*B_.yy() + B_.yz()*B_.yz() + 
	      B_.zx()*B_.zx() + B_.zy()*B_.zy() + B_.zz()*B_.zz() + 

	      C_.xx()*C_.xx() + C_.xy()*C_.xy() + C_.xz()*C_.xz() + 
	      C_.yx()*C_.yx() + C_.yy()*C_.yy() + C_.yz()*C_.yz() + 
	      C_.zx()*C_.zx() + C_.zy()*C_.zy() + C_.zz()*C_.zz() + 

	      D_.xx()*D_.xx() + D_.xy()*D_.xy() + D_.xz()*D_.xz() + 
	      D_.yx()*D_.yx() + D_.yy()*D_.yy() + D_.yz()*D_.yz() + 
	      D_.zx()*D_.zx() + D_.zy()*D_.zy() + D_.zz()*D_.zz();

	return (sqrt(sum));
}

double RxSpatialMat::trace() const
{
	return (A_.xx()+A_.yy()+A_.zz()+D_.xx()+D_.yy()+D_.zz());
}

double RxSpatialMat::condest() const
{
	RxMatd mtmp(*this);

	return mtmp.condest();
}

RxSpatialMat operator+(const RxSpatialMat &m1, const RxSpatialMat &m2)
{
	RxSpatialMat mr;
	return (mr.add (m1, m2));
}

RxSpatialMat operator-(const RxSpatialMat &m1, const RxSpatialMat &m2)
{
	RxSpatialMat mr;
	return (mr.sub (m1, m2));
}

RxSpatialMat operator*(const RxSpatialMat &m1, const RxSpatialMat &m2)
{
	RxSpatialMat mr;
	return (mr.mul (m1, m2));
}

RxSpatialMat operator*(const RxSpatialMat &m1, double a)
{
	RxSpatialMat mr;
	return (mr.scale (a, m1));
}

RxSpatialMat operator*(double a, const RxSpatialMat &m1)
{
	RxSpatialMat mr;
	return (mr.scale (a, m1));
}

RxSpatialMat operator/(const RxSpatialMat &m1, double a)
{
	RxSpatialMat mr;
	return (mr.scale (1/a, m1));
}
