#ifndef _RXSPATIALVEC_H_WRAP_
#define _RXSPATIALVEC_H_WRAP_

#include <RxVecdBase.h>
#include <RxVecd3.h>

class RxRot;
class RxTrsf;

class RxSpatialVec: public RxVecdBase
{
public:
	RxVecd3 a_;
	RxVecd3 b_;

    public:

	RxVecd3 &a()
	 { return (a_);
	 }
	const RxVecd3 &a() const
	 { return (a_);
	 }
	RxVecd3 &b()
	 { return (b_);
	 }
	const RxVecd3 &b() const
	 { return (b_);
	 }

	// assignment

	RxSpatialVec& set(const RxSpatialVec &v2);
	RxSpatialVec& operator=(const RxSpatialVec &v2)
	 { return (set(v2));
	 }
	RxSpatialVec& set (double ax, double ay, double az,
		      double bx, double by, double bz);
	RxSpatialVec& set (const double *dp);
	void get (double *axp, double *ayp, double *azp,
		  double *bxp, double *byp, double *bzp);
	void get (double *dp) const;

	// constructors

	RxSpatialVec ()
	 {
	 }
	RxSpatialVec (double ax, double ay, double az, 
		 double bx, double by, double bz)
	 { set (ax, ay, az, bx, by, bz);
	 }
	RxSpatialVec (RxVecd3 &av, RxVecd3 &bv)
	 { a_ = av;
	   b_ = bv;
	 }
	RxSpatialVec (const double *dp)
	 { set (dp);
	 }
	RxSpatialVec (const RxSpatialVec& v2)
	 { set (v2);
	 }

	// query and access

	int length() const
	 { return 6;
	 }
	double& operator()(int i);
	double operator()(int i) const;

	// T <- T,T operations

	RxSpatialVec& add (const RxSpatialVec &v1, const RxSpatialVec &v2);
	RxSpatialVec& add (const RxSpatialVec &v2);
	RxSpatialVec& operator+=(const RxSpatialVec &v2)
	 { return (add(v2));
	 }
	friend RxSpatialVec operator+(const RxSpatialVec &v1, const RxSpatialVec &v2);

	RxSpatialVec& sub (const RxSpatialVec &v1, const RxSpatialVec &v2);
	RxSpatialVec& sub (const RxSpatialVec &v1);
	RxSpatialVec& operator-=(const RxSpatialVec &v2)
	 { return (sub(v2));
	 }
	friend RxSpatialVec operator-(const RxSpatialVec &v1, const RxSpatialVec &v2);

	RxSpatialVec& extrap (const RxSpatialVec &v1, double a, const RxSpatialVec &v2);
	RxSpatialVec& extrap (double a, const RxSpatialVec &v2)
	 { return (extrap (*this, a, v2));
	 }
	RxSpatialVec& interp (const RxSpatialVec &v1, double a, const RxSpatialVec &v2);
	RxSpatialVec& interp (double a, const RxSpatialVec &v2)
	 { return (interp (*this, a, v2));
	 }
	RxSpatialVec& cross (const RxSpatialVec &v1, const RxSpatialVec &v2);
	RxSpatialVec& cross (const RxSpatialVec &v2)
	 { return cross (*this, v2);
	 }

	// T <- T,O operations

	RxSpatialVec& scale (double a, const RxSpatialVec &v1);
	RxSpatialVec& scale (double a);
	RxSpatialVec& operator*=(double a)
	 { return (scale(a));
	 }
	RxSpatialVec& operator/=(double a)
	 { return (scale(1/a));
	 }
	friend RxSpatialVec operator*(const RxSpatialVec &v1, double a);
	friend RxSpatialVec operator*(double a, const RxSpatialVec &v1);
	friend RxSpatialVec operator/(const RxSpatialVec &v1, double a);

	RxSpatialVec& xform (const RxSpatialVec &v1, const RxRot& R);
	RxSpatialVec& xform (const RxRot& R)
	 { return (xform (*this, R));
	 }
	RxSpatialVec& Ixform (const RxSpatialVec &v1, const RxRot& R);
	RxSpatialVec& Ixform (const RxRot& R)
	 { return (Ixform (*this, R));
	 }
	RxSpatialVec& xform (const RxSpatialVec &v1, const RxTrsf& T);
	RxSpatialVec& xform (const RxTrsf& T)
	 { return (xform (*this, T));
	 }
	RxSpatialVec& Ixform (const RxSpatialVec &v1, const RxTrsf& T);
	RxSpatialVec& Ixform (const RxTrsf& T)
	 { return (Ixform (*this, T));
	 }

	// O <- T,T operations

	double dot(const RxSpatialVec &v2) const;

	// T <- T operations

	RxSpatialVec &unit(const RxSpatialVec &v1);
	RxSpatialVec &unit()
	 { return (unit(*this));
	 }
	RxSpatialVec &neg(const RxSpatialVec &v1);
	RxSpatialVec &neg()
	 { return (neg(*this));
	 }
	friend RxSpatialVec operator-(const RxSpatialVec &v1);

	// T <- O operations

	RxSpatialVec &setZero();

	// O <- T operations

	double norm1(double ca, double cb) const;
	double norm2(double ca, double cb) const;
	double normInf(double ca, double cb) const;

	// constants

	static const RxSpatialVec ZERO;
};

inline RxSpatialVec operator+(const RxSpatialVec &v1, const RxSpatialVec &v2)
{
	RxSpatialVec vr;
	return (vr.add(v1, v2));
}

inline RxSpatialVec operator-(const RxSpatialVec &v1, const RxSpatialVec &v2)
{
	RxSpatialVec vr;
	return (vr.sub(v1, v2));
}

inline RxSpatialVec operator*(const RxSpatialVec &v1, double a)
{
	RxSpatialVec vr;
	return (vr.scale(a, v1));
}

inline RxSpatialVec operator*(double a, const RxSpatialVec &v1)
{
	RxSpatialVec vr;
	return (vr.scale(a, v1));
}

inline RxSpatialVec operator/(const RxSpatialVec &v1, double a)
{
	RxSpatialVec vr;
	return (vr.scale(1/a, v1));
}

inline RxSpatialVec operator-(const RxSpatialVec &v1)
{
	RxSpatialVec vr;
	return (vr.neg(v1));
}

#endif
