#include <RxSpatialVec.h>
#include <RxVecdBaseTest.h>
#include <RxTrsf.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	6

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 },
 { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0 },
 {-3.0, -2.0, -1.0, -0.5, 1.1, 2.2 }};

double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};

int testSvDot (double dot, const RxSpatialVec &v1, 
	       const RxSpatialVec &v2, double eps)
{
	double chk;

	chk = v1.a().dot(v2.b()) + v1.b().dot(v2.a());
	if (ABS(chk-dot) > eps)
	 { return 0;
	 }
	else
	 { return 1;
	 }
}

int testSvCross (const RxSpatialVec &vc, const RxSpatialVec &v1, 
	     const RxSpatialVec &v2, double eps)
{
	RxSpatialVec chk;
	RxVecd3 tmp;

	chk.a().cross(v1.a(), v2.a());
	tmp.cross(v1.a(), v2.b());
	chk.b().cross(v1.b(), v2.a());
	chk.b().add (tmp);
	return (chk.equal(vc, eps));	
}

int testSvUnit (const RxSpatialVec &vu, const RxSpatialVec &v1,
		double eps)
{
	RxSpatialVec chk;

	chk.a().unit(v1.a());
	chk.b().unit(v1.b());
	return (chk.equal(vu, eps));	
}

int testSvNorm1 (double nc, const RxSpatialVec &v1, 
		 double ca, double cb, double eps)
{
	double chk;

	chk = ca*v1.a().norm1() + cb*v1.b().norm1();
	return (ABS(chk-nc) <= eps);
}

int testSvNorm2 (double nc, const RxSpatialVec &v1,
		 double ca, double cb, double eps)
{
	double chk;

	chk = ca*v1.a().norm2() + cb*v1.b().norm2();
	return (ABS(chk-nc) <= eps);
}

int testSvNormInf (double nc, const RxSpatialVec &v1,
		 double ca, double cb, double eps)
{
	double chk, amax, bmax;

	amax = ca*v1.a().normInf();
	bmax = cb*v1.b().normInf();
	return (ABS(MAX(amax,bmax)-nc) <= eps);
}

main()
{
	double vec[LEN];
	RxSpatialVec v1(10,20,30,40,50,60);
	RxSpatialVec v2(100,200,300,400,500,600);
	int i, j, k;

	assert (EQ(v1.a().x(),10) && EQ(v1.a().y(),20) && EQ(v1.a().z(),30));
	assert (EQ(v1.b().x(),40) && EQ(v1.b().y(),50) && EQ(v1.b().z(),60));
	
	RxSpatialVec v3(v2);

	assert (EQ(v3.a().x(),100) && EQ(v3.a().y(),200) && EQ(v3.a().z(),300));
	assert (EQ(v3.b().x(),400) && EQ(v3.b().y(),500) && EQ(v3.b().z(),600));

	assert (v3.length() == LEN);

	assert (v3.equal(v2,EPS));
	assert (!v3.equal(v1,EPS));
	for (i=0;i<LEN;i++)
	 { v1(i) = i*10+1;
	 }
	assert (v1.equal(RxSpatialVec(1,11,21,31,41,51),EPS));
	
	v1 = v3;

	assert (v3.equal(v1,EPS));
	
	v1.set(11,22,33,44,55,66);

	assert (v1.equal(RxSpatialVec(11,22,33,44,55,66),EPS));
	
	v1.get(vec);
	for (i=0; i<LEN; i++)
	 { assert (EQ(vec[i],(i+1)*11));
	 }
	for (i=0; i<LEN; i++)
	 { vec[i] = (i+1)*111;
	 }
	assert ((v3.set(vec)).equal(RxSpatialVec(111,222,333,444,555,666),EPS));

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { v1.set(testVals[i]);
	      v2.set(testVals[j]);

	      assert (testAdd (v1+v2, v1, v2, EPS));
	      assert (testAdd (v3.add(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testAdd (v3.add(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testAdd (v3 += v2, v1, v2, EPS));

	      assert (testSub (v1-v2, v1, v2, EPS));
	      assert (testSub (v3.sub(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testSub (v3.sub(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testSub (v3 -= v2, v1, v2, EPS));

	      assert (testExtrap (v3.extrap(v1,11.1,v2), v1, 11.1, v2, EPS));
	      v3 = v1;
	      assert (testExtrap (v3.extrap(11.1,v2), v1, 11.1, v2, EPS));

	      v3 = v1;
	      assert (testSvDot (v3.dot(v2), v1, v2, EPS));

	      assert (testSvCross (v3.cross(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testSvCross (v3.cross(v2), v1, v2, EPS));
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);

	   assert (testScale (v3.scale(M_PI,v1), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.scale(M_PI), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 *= M_PI, M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 /= M_PI, 1/M_PI, v1, EPS));
	   assert (testScale (M_PI*v1, M_PI, v1, EPS));
	   assert (testScale (v1*M_PI, M_PI, v1, EPS));
	   assert (testScale (v1/M_PI, 1/M_PI, v1, EPS));

	   assert (testSvUnit (v3.unit(v1), v1, EPS));
	   v3 = v1;
	   assert (testSvUnit (v3.unit(), v1, EPS));
	   assert (testScale (v3.neg(v1), -1.0, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.neg(), -1.0, v1, EPS));
	   assert (testScale (-v1, -1.0, v1, EPS));

	   assert (testSvNorm1 (v1.norm1(1, M_PI), v1, 1, M_PI, EPS));
	   assert (testSvNorm2 (v1.norm2(1, M_PI), v1, 1, M_PI, EPS));
	   assert (testSvNormInf (v1.normInf(1, M_PI), v1, 1, M_PI, EPS));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);
 
	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxTrsf T;

	      T.R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      v3.xform (v1, T.R);
	      v3.Ixform (v3, T.R);
	      assert (v3.equal(v1,EPS));
	      v3 = v1;
	      v3.xform (T.R);
	      v3.Ixform (T.R);
	      assert (v3.equal(v1,EPS));

	      for (k=0; k<ARRAY_SIZE(testVals); k++)
	       { T.p.set (testVals[k][0], testVals[k][1], testVals[k][2]);

		 v3.xform (v1, T);
		 v3.Ixform (v3, T);
		 assert (v3.equal(v1,EPS));
		 v3 = v1;
		 v3.xform (T);
		 v3.Ixform (T);
		 assert (v3.equal(v1,EPS));
	       }
	    }
	 }

	v1.set(0,0,0,0,0,0);
	assert ((v2.setZero()).equal(v1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}
