#include <RxTrsf.h>
#include <RxTimeval.h>
#include <assert.h>

typedef struct {
	double nx, ny, nz;
	double ox, oy, oz;
	double ax, ay, az;
	double px, py, pz;
} TRSFD;

assign(RxTrsf &A, RxTrsf &B)
{
	A.R.n.x() = B.R.n.x();
	A.R.n.y() = B.R.n.y();
	A.R.n.z() = B.R.n.z();
	A.R.o.x() = B.R.o.x();
	A.R.o.y() = B.R.o.y();
	A.R.o.z() = B.R.o.z();
	A.R.a.x() = B.R.a.x();
	A.R.a.y() = B.R.a.y();
	A.R.a.z() = B.R.a.z();
	A.p.x() = B.p.x();
	A.p.y() = B.p.y();
	A.p.z() = B.p.z();
}

#define EPS	(1e-10)

volatile int ii;

bar()
{
	ii = 1;
	ii = 2;
	ii = 3;
}

main()
{
	char buf0[1024];
	char buf1[1024];
	RxTrsf T, M, A;
	TRSFD T1, T2;
	RxTimeval t0, t1;
	int count = 10000;
	int i;

	M.setIdent();
	RxTrsf::IDENT.print("TrsfI %8.3f\n");
	RxRot::IDENT.print("RotI %8.3f\n");
	assert (M.equal(RxTrsf::IDENT, EPS));

	char *Tdef1 = "\
T:[   0.354   -0.707    0.612   11.200\n\
      0.354    0.707    0.612   22.300\n\
     -0.866    0.000    0.500   33.400]";

	assert (T.scan (Tdef1) == 1);
	T.print (buf0, "T:%M%8.3f");

	assert (strcmp (buf0, Tdef1) == 0);

	T.print (buf0, "T=%M%8.3f,%7.2f");

	assert (strcmp (buf0, "\
T=[   0.35,  -0.71,   0.61,  11.200\n\
      0.35,   0.71,   0.61,  22.300\n\
     -0.87,   0.00,   0.50,  33.400]") == 0);

	char *Tdef2 = "\
T: [  11.200   22.300   33.400  -0.31882  0.76996  0.55273   73.709]";
	assert (T.scan (Tdef2) == 1);
	T.print (buf0, "T: %8.3f %8.5f");
	assert (strcmp (buf0, Tdef2) == 0);

	T.print (buf0, "Too: ");
	T.print (buf1, "Too: %8.3f %8.3f\n");
	assert (strcmp (buf0, buf1) ==0);

	T.setIdent();
	T.print ("T:");
	T.p.set (1, 2, 3);
	T.R.setRpyD (45.0, 90.0, 0.0);
	T.print ("N:");

	M = T;

	M = T.inv();
	M.print ("M:");
	(M * T).print ("I:");

	printf ("size = %d\n", sizeof(RxTrsf));

	t0.sysSetTod();
	M.R.setRpy (10, 20, 30);
	M.p.set (1, 2,3);
	T = M;
	for (i=0; i<count; i++)
	 { A.mul (M, T);
	 }
	t1.sysSetTod();
	printf ("A.mul (M, T): %g usec\n", (double)(t1-t0)*1e6/count);

	TIMEIT (A = M * T, count)

	TIMEIT (A = M, count)

	TIMEIT (RxTrsf M(A), count)

	TIMEIT(assign(M,A), count);
	TIMEIT(T1 = T2, count);
	TIMEIT(memcpy(&M, &A, sizeof(RxTrsf)), count);
	TIMEIT(RxVecd3 v, count);
	TIMEIT(RxMatdBase b, count);
	TIMEIT(RxRot R, count);
	TIMEIT(RxTrsf F, count);

	TIMEIT(ii=1; ii=2; ii=3, count);

	printf ("\nROUTINES OK\n\n");
	return 0;
}
