#ifndef _RXWRENCH_H_
#define _RXWRENCH_H_

#include <RxSpatialVec.h>

class RxWrench: public RxSpatialVec
{
public:
	RxVecd3 &f()
	 { return a_;
	 }
	const RxVecd3 &f() const
	 { return a_;
	 }
	RxVecd3 &m()
	 { return b_;
	 }
	const RxVecd3 &m() const
	 { return b_;
	 }

	// assignment

	RxWrench& set (double fx, double fy, double fz,
		      double mx, double my, double mz);
	RxWrench& set (const double *dp);
	void get (double *fxp, double *fyp, double *fzp,
		  double *mxp, double *myp, double *mzp);
	void get (double *dp) const;

	// constructors

	RxWrench () : RxSpatialVec()
	 {
	 }
	RxWrench (double fx, double fy, double fz, 
		 double mx, double my, double mz) : 
		RxSpatialVec()
	 { set (fx, fy, fz, mx, my, mz);
	 }
	RxWrench (RxVecd3 &fv, RxVecd3 &mv) : RxSpatialVec()
	 { m() = mv;
	   f() = fv;
	 }
	RxWrench (const double *dp) : RxSpatialVec()
	 { set (dp);
	 }
#if 0
	RxWrench (const RxWrench& t2) : RxSpatialVec(t2)
	 {
	 }
#endif
	// query and access

	double& operator()(int i);
	double operator()(int i) const;

	// T <- T,T operations

	RxWrench& add (const RxWrench &t1, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::add (t1, t2);
	 }
	RxWrench& add (const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::add (t2);
	 }
	RxWrench& operator+=(const RxWrench &t2)
	 { return (add(t2));
	 }
	friend RxWrench operator+(const RxWrench &t1, const RxWrench &t2);

	RxWrench& sub (const RxWrench &t1, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::sub (t1, t2);
	 }
	RxWrench& sub (const RxWrench &t1)
	 { return (RxWrench&) RxSpatialVec::sub (t1);
	 }
	RxWrench& operator-=(const RxWrench &t2)
	 { return (sub(t2));
	 }

	friend RxWrench operator-(const RxWrench &t1, const RxWrench &t2);

	RxWrench& extrap (const RxWrench &t1, double a, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::extrap (t1, a, t2);
	 }
	RxWrench& extrap (double a, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::extrap (a, t2);
	 }
	RxWrench& interp (const RxWrench &t1, double a, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::interp (t1, a, t2);
	 }
	RxWrench& interp (double a, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::interp (a, t2);
	 }
#if 0
	RxWrench& cross (const RxWrench &t1, const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::cross (t1, t2);
	 }
	RxWrench& cross (const RxWrench &t2)
	 { return (RxWrench&) RxSpatialVec::cross (t2);
	 }
#endif
	// T <- T,O operations

	RxWrench& scale (double a, const RxWrench &t1)
	 { return (RxWrench&) RxSpatialVec::scale (a, t1);
	 }
	RxWrench& scale (double a)
	 { return (RxWrench&) RxSpatialVec::scale (a);
	 }
	RxWrench& operator*=(double a)
	 { return (scale(a));
	 }
	RxWrench& operator/=(double a)
	 { return (scale(1/a));
	 }

	friend RxWrench operator*(const RxWrench &t1, double a);
	friend RxWrench operator*(double a, const RxWrench &t1);
	friend RxWrench operator/(const RxWrench &t1, double a);

	RxWrench& xform (const RxWrench &t1, const RxRot& R)
	 { return (RxWrench&) RxSpatialVec::xform (t1, R);
	 }
	RxWrench& xform (const RxRot& R)
	 { return (RxWrench&) RxSpatialVec::xform (R);
	 }
	RxWrench& Ixform (const RxWrench &t1, const RxRot& R)
	 { return (RxWrench&) RxSpatialVec::Ixform (t1, R);
	 }
	RxWrench& Ixform (const RxRot& R)
	 { return (RxWrench&) RxSpatialVec::Ixform (R);
	 }
	RxWrench& xform (const RxWrench &t1, const RxTrsf& T)
	 { return (RxWrench&) RxSpatialVec::xform (t1, T);
	 }
	RxWrench& xform (const RxTrsf& T)
	 { return (RxWrench&) RxSpatialVec::xform (T);
	 }
	RxWrench& Ixform (const RxWrench &t1, const RxTrsf& T)
	 { return (RxWrench&) RxSpatialVec::Ixform (t1, T);
	 }
	RxWrench& Ixform (const RxTrsf& T)
	 { return (RxWrench&) RxSpatialVec::Ixform (T);
	 }

	// O <- T,T operations

	// T <- T operations

	RxWrench &unit(const RxWrench &t1)
	 { return (RxWrench&) RxSpatialVec::unit(t1);
	 }
	RxWrench &unit()
	 { return (RxWrench&) RxSpatialVec::unit();
	 }
	RxWrench &neg(const RxWrench &t1)
	 { return (RxWrench&) RxSpatialVec::neg(t1);
	 }
	RxWrench &neg()
	 { return (RxWrench&) RxSpatialVec::neg();
	 }
	friend RxWrench operator-(const RxWrench &t1);

	// T <- O operations

	RxWrench &setZero()
	 { return (RxWrench&) RxSpatialVec::setZero();
	 }

	// O <- T operations

	double norm1(double cf, double cm) const
	 { return RxSpatialVec::norm1 (cf, cm);
	 }
	double norm2(double cf, double cm) const
	 { return RxSpatialVec::norm2 (cf, cm);
	 }
	double normInf(double cf, double cm) const
	 { return RxSpatialVec::normInf (cf, cm);
	 }

	// constants

	static const RxWrench ZERO;
	static const RxWrench UNIT_FX;
	static const RxWrench UNIT_FY;
	static const RxWrench UNIT_FZ;
	static const RxWrench UNIT_MX;
	static const RxWrench UNIT_MY;
	static const RxWrench UNIT_MZ;
};

inline RxWrench operator+(const RxWrench &t1, const RxWrench &t2)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->add (*(RxSpatialVec*)&t1, *(RxSpatialVec*)&t2);
	return vr;
}

inline RxWrench operator-(const RxWrench &t1, const RxWrench &t2)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->sub (*(RxSpatialVec*)&t1, *(RxSpatialVec*)&t2);
	return vr;
}

inline RxWrench operator*(const RxWrench &t1, double a)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxWrench operator*(double a, const RxWrench &t1)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxWrench operator/(const RxWrench &t1, double a)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->scale (1/a, *(RxSpatialVec*)&t1);
	return vr;
}

inline RxWrench operator-(const RxWrench &t1)
{
	RxWrench vr;
	((RxSpatialVec*)&vr)->neg (*(RxSpatialVec*)&t1);
	return vr;
}

#endif

