#ifdef HAS_OPEN_GL

#include <stdio.h>
#include <math.h>

#include <GL/gl.h>
//#include <GL/glu.h> 

#include <string.h>
#include <cdefs.h>
#include <grf4.h>
#include "grf4Internal.h"
#include "grf4GenObject.h"
#include "grf4GlCommon.h"

extern TRSF* iGLTransPtr;
extern int iGLPerspectiveMode;
extern int grf4GeneralFlags;
extern float iGLNearPlane;
extern void *malloc();

static TRSF Identity = { 1.0, 0.0, 0.0, 
			 0.0, 1.0, 0.0,
			 0.0, 0.0, 1.0,
			 0.0, 0.0, 0.0 };

static void multV3fMatrix(float *vr, float *v1, GLfloat *m1)
{
	float tmp[3];

	tmp[0] = v1[0]*m1[0*4+0] + v1[1]*m1[1*4+0] + v1[2]*m1[2*4+0] + m1[3*4+0];
	tmp[1] = v1[0]*m1[0*4+1] + v1[1]*m1[1*4+1] + v1[2]*m1[2*4+1] + m1[3*4+1];
	tmp[2] = v1[0]*m1[0*4+2] + v1[1]*m1[1*4+2] + v1[2]*m1[2*4+2] + m1[3*4+2];
	vr[0] = tmp[0];
	vr[1] = tmp[1];
	vr[2] = tmp[2];
}

static float _v2f[2];
static float _v3f[3];
static float _n3f[3];
#define V2F(x,y)	(_v2f[0]=(x), _v2f[1]=(y), glVertex2fv(_v2f))
#define V3F(x,y,z)	(_v3f[0]=(x), _v3f[1]=(y), _v3f[2]=(z), glVertex3fv(_v3f))
#define N3F(x,y,z)	(_n3f[0]=(x), _n3f[1]=(y), _n3f[2]=(z), glNormal3fv(_n3f))

static void buildNormals();

#define NEXT_IDX(k,num)	(((k)==((num)-1)) ? 0 : (k)+1)
#define PREV_IDX(k,num)	((k)==0 ? ((num)-1) : (k)-1)

static void buildNormals(PRISM *cylin)
{
	CEDGE *ce;
	float v1[3], v2[3];
	int i, k, n, p;
	float *nrmls;
	float *fnk, *fnn, *fnp;
	float *vn1, *vn2;

	ce = cylin->edges;

	nrmls = (float*)malloc(sizeof(float)*3*3*cylin->size);
	
	for (k=0; k<cylin->size; k++)
	 { n = NEXT_IDX(k,cylin->size);

	   v1[0] = ce[n].x_b - ce[k].x_b;
	   v1[1] = ce[n].y_b - ce[k].y_b;
	   v1[2] = 0.0;

	   v2[0] = ce[k].x_t - ce[k].x_b;
	   v2[1] = ce[k].y_t - ce[k].y_b;
	   v2[2] = cylin->depth;

	   fnk = nrmls; 
	   nrmls += 3;
	   _grf4NormalizeV3f (fnk, _grf4CrossV3f(fnk, v1, v2));
	   ce[k].faceNrml = fnk;
	 }

	for (k=0; k<cylin->size; k++)
	 { n = NEXT_IDX(k,cylin->size);
	   p = PREV_IDX(k,cylin->size);

	   vn1 = nrmls; 
	   nrmls += 3;
	   vn2 = nrmls;
	   nrmls += 3;
	   fnk = ce[k].faceNrml;
	   fnn = ce[n].faceNrml;
	   fnp = ce[p].faceNrml;

	   for (i=0; i<3; i++)
	    { if (ce[k].hard && ce[n].hard)
	       { vn1[i] = vn2[i] = fnk[i];
	       }
	      else if (!ce[k].hard && !ce[n].hard)
	       { vn1[i] = (fnp[i] + fnk[i])/2;
		 vn2[i] = (fnk[i] + fnn[i])/2;
	       }
	      else if (ce[k].hard && !ce[n].hard)
	       { vn1[i] = (3*fnk[i] - fnn[i])/2;
		 vn2[i] = (fnk[i] + fnn[i])/2;
	       }
	      else
	       { vn1[i] = (fnp[i] + fnk[i])/2;
		 vn2[i] = (3*fnk[i] - fnp[i])/2;
	       }
	    }
	   _grf4NormalizeV3f (vn1, vn1);
	   _grf4NormalizeV3f (vn2, vn2);
	   ce[k].vrtxNrml1 = vn1;
	   ce[k].vrtxNrml2 = vn2;
	 }
}

PRISM *iGLGenDisc(int npts, double radius, double depth)
{
	PRISM *cylin;

	cylin = gdrGenDisc (npts, radius, depth);
	buildNormals (cylin);
	return(cylin);
}

PRISM *iGLGenCone(int npts, double radius1, double radius2, double depth)
{
	PRISM *cylin;

	cylin = gdrGenCone(npts, radius1, radius2, depth);
	buildNormals (cylin);
	return(cylin);
}

PRISM *iGLGenPrism (int npts, float *bot, float *top, char *hidden,double depth)
{
	PRISM *cylin;

	cylin = gdrGenPrism (npts, bot, top, hidden, depth);
	buildNormals (cylin);
	return(cylin);
}

POLYGON *iGLGenPolygon (int npts, float *vtxs)
{
	POLYGON *pgn;

	pgn = gdrGenPolygon (npts, vtxs);
	return(pgn);
}

POLYGON *iGLGenRegPolygon (int npts, double size)
{
	POLYGON *pgn;

	pgn = gdrGenRegPolygon (npts, size);
	return(pgn);
}

LINESET *iGLGenLineSet (int num, float *lines)
{
	LINESET *lset;

	lset = gdrGenLineSet (num, lines);
	return(lset);
}

double iGLSetDepth (Grf4Object *obj, double depth)
{
	float old;

	old = obj->ptr.prism->depth;
	obj->ptr.prism->depth = depth;
	return (old);
}

#define NEXT_IDX(k,num)	(((k)==((num)-1)) ? 0 : (k)+1)
#define PREV_IDX(k,num)	((k)==0 ? ((num)-1) : (k)-1)

iGLDrawPrism (Grf4Object *obj)
{
	PRISM *prism = obj->ptr.prism;

	TRSF *T;
	int topVisible;
	int bottomVisible;
	int visible[256];
	float nrml[3];
	float vpnt[3];
	float depth;
	float dex, dey, dez;

	register CEDGE *ep;
	register int k, n;

	if (prism->size < 2)
	 { fprintf (stderr, "drawPrism: less than 2 points\n");
	   exit (-1);
	 }
	T = (iGLTransPtr ? iGLTransPtr : &Identity);
	ep = prism->edges;
	depth = prism->depth;

	dex = T->a.x * depth;
	dey = T->a.y * depth;
	dez = T->a.z * depth;

	if (!iGLPerspectiveMode)
	 { bottomVisible = (T->a.z < 0.0);
	   topVisible = !bottomVisible;
	 }
	else
	 { float w;
	   w = T->p.x*T->a.x + T->p.y*T->a.y + T->p.z*T->a.z;
	   bottomVisible = (w > 0.0);
	   topVisible = (w + dex*T->a.x + dey*T->a.y + dez*T->a.z <= 0.0);
	 }

	for (k=0; k<prism->size; k++)
	 { 
	   _grf4MultRotV3f (nrml, T, ep[k].faceNrml);
	   if (!iGLPerspectiveMode)
	    { visible[k] = (nrml[2] > 0.0);
	    }
	   else
	    { vpnt[0] = ep[k].x_b;
	      vpnt[1] = ep[k].y_b;
	      vpnt[2] = 0.0;
	      _grf4MultTrsfV3f (vpnt, T, vpnt);
	      visible[k] = (_grf4DotV3f (vpnt, nrml) < 0.0);
	    }
	 }
	for (k=0; k<prism->size; k++)
	 { int n;
	   
	   n = NEXT_IDX(k,prism->size);

	   if (visible[k])
	/* OGLXXX
	 * special cases for polygons:
	 * 	independant quads: use GL_QUADS
	 * 	independent triangles: use GL_TRIANGLES
	 */
	    { glBegin(GL_POLYGON);
	      glNormal3fv(ep[k].vrtxNrml1);
	      V3F (ep[k].x_t, ep[k].y_t, depth);
	      V3F (ep[k].x_b, ep[k].y_b, 0.0);
	      glNormal3fv(ep[k].vrtxNrml2);
	      V3F (ep[n].x_b, ep[n].y_b, 0.0);
	      V3F (ep[n].x_t, ep[n].y_t, depth);
	      glEnd();
	    }
	 }
	if (topVisible)
	/* OGLXXX
	 * special cases for polygons:
	 * 	independant quads: use GL_QUADS
	 * 	independent triangles: use GL_TRIANGLES
	 */
	 { glBegin(GL_POLYGON);
	   N3F (0.0, 0.0, 1.0);
	   for (k=0; k<prism->size; k++)
	    { V3F (ep[k].x_t, ep[k].y_t, depth);
	    }
	   glEnd();
	 }
	else if (bottomVisible)
	/* OGLXXX
	 * special cases for polygons:
	 * 	independant quads: use GL_QUADS
	 * 	independent triangles: use GL_TRIANGLES
	 */
	 { glBegin(GL_POLYGON);
	   N3F (0.0, 0.0, -1.0);
	   for (k=prism->size-1; k>=0; k--)
	    { V3F (ep[k].x_b, ep[k].y_b, 0.0);
	    }
	   glEnd();
	 }
}

int iGLsolidsphere(SPHERE *s, int smooth, int n, double r)
{
	GLfloat m[16];
	int i;
	int phi;
	int theta;
	float v1[3], nrml[3];
	float *vp;

	glGetFloatv(GL_MODELVIEW_MATRIX, m);

	for (theta=0;theta<s->nTheta;theta++)
	 { for (phi=0;phi<s->nPhi;phi++)
	    { 
	      i = SPHERE_IDX(s,theta,phi);

	      multV3fMatrix (v1, s->pnts[i], m);
	      vp = s->ctrNormals[i];
	      nrml[0] = vp[0]*m[0*4+0] + vp[1]*m[1*4+0] + vp[2]*m[2*4+0];
	      nrml[1] = vp[0]*m[0*4+1] + vp[1]*m[1*4+1] + vp[2]*m[2*4+1];
	      nrml[2] = vp[0]*m[0*4+2] + vp[1]*m[1*4+2] + vp[2]*m[2*4+2];

	      if (_grf4DotV3f(nrml, v1) < 0.0)
	       { 
	/* OGLXXX
	 * special cases for polygons:
	 * 	independant quads: use GL_QUADS
	 * 	independent triangles: use GL_TRIANGLES
	 */
		 glBegin(GL_POLYGON);
		 if (!smooth)
		  { glNormal3fv(s->ctrNormals[i]);
		  }
		 else
		  { glNormal3fv(s->pntNormals[i]);
		  }
		 V3F(r*s->pnts[i][0], r*s->pnts[i][1], r*s->pnts[i][2]);
		 i += SPHERE_IDX(s,0,1);
		 if (smooth)
		  { glNormal3fv(s->pntNormals[i]);
		  }
		 V3F(r*s->pnts[i][0], r*s->pnts[i][1], r*s->pnts[i][2]);
		 i += SPHERE_IDX(s,1,0);
		 if (smooth)
		  { glNormal3fv(s->pntNormals[i]);
		  }
		 V3F(r*s->pnts[i][0], r*s->pnts[i][1], r*s->pnts[i][2]);
		 i -= SPHERE_IDX(s,0,1);
		 if (smooth)
		  { glNormal3fv(s->pntNormals[i]);
		  }
		 V3F(r*s->pnts[i][0], r*s->pnts[i][1], r*s->pnts[i][2]);
		 glEnd();
	       }
	    }
	 }
	return(0);
}

SPHERE *iGLGenSphere (int npts, double radius)
{
	SPHERE *s;
	SPHERE_TEMPLATE *t;

	s = ALLOC(SPHERE);
	memset (s, 0, sizeof(SPHERE));
	s->radius = radius;
	s->npts = npts;
	t = _grf4GetSphereTemplate(npts);
	s->nTheta = t->nTheta;
	s->nPhi = t->nPhi;
	s->pnts = t->pnts;
	s->pntNormals = t->pntNormals;
	s->ctrNormals = t->ctrNormals;	

	return(s);
}

void iGLDrawSphere (Grf4Object *obj)
{
	SPHERE *s = obj->ptr.sphere;

	iGLsolidsphere (s, /*smooth=*/1, s->npts, s->radius);
}

void iGLDrawPolygon(Grf4Object *obj)
{
	POLYGON *pgn = obj->ptr.polygon;
	float *ep;
	int k;

	/* OGLXXX
	 * special cases for polygons:
	 * 	independant quads: use GL_QUADS
	 * 	independent triangles: use GL_TRIANGLES
	 */
	glBegin(GL_POLYGON);
	if (!(grf4GeneralFlags & GRF_2DONLY))
	 { 
	   N3F (0.0, 0.0, 1.0);
	   for (k=0, ep=pgn->edges; k<pgn->size; k++, ep+=2)
	    { V3F (ep[0], ep[1], 0.0);
	    }
	 }
	else
	 { for (k=0, ep=pgn->edges; k<pgn->size; k++, ep+=2)
	    { V2F (ep[0], ep[1]);
	    }
	 }
	glEnd();
}

void iGLDrawLineSet(Grf4Object *obj)
{
	LINESET *lset = obj->ptr.lineSet;
	register float *l;
	float *last_l;

	for (l=lset->lines, last_l=l+lset->num*4; l<last_l; l+=4)
	 { glBegin(GL_LINE_STRIP);
	   glVertex2f(l[0],  l[1]);
	   glVertex2f(l[2],  l[3]);
	   glEnd();
	 }
}

#endif /* HAS_GL_WINDOWS */
