#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <cdefs.h>
#include <grf4.h>

void drawIt()
{
	grf4Clear ();
	grf4DrawJack();
	grf4Flush ();
}

printUpDown(char *s, int n)
{
	printf ("%s %s\n", s, n ? "down" : "up");
}

void onCursor (Grf4Event *ev, int arg)
{
	drawIt();
}

void onButton (Grf4Event *ev, int arg)
{
	switch (ev->code)
	 { case GRF_LEFT_MOUSE:
	    { printUpDown ("Left mouse", ev->val1);
	      break;
	    }
	   case GRF_MIDDLE_MOUSE:
	    { printUpDown ("Middle mouse", ev->val1);
	      break;
	    }
	   case GRF_RIGHT_MOUSE:
	    { printUpDown ("Right mouse", ev->val1);
	      break;
	    }
	   default:
	    { grf4Fatal ("onButton(): bogus code %d\n", ev->code);
	    }
	 }
}

void onKey (Grf4Event *ev, int arg)
{
	if (!GRF_IS_KEY(ev->code))
	 { grf4Fatal ("onKey(): code %d is not a key code\n", ev->code);
	 }
}

void onConfig (Grf4Event *ev, int arg)
{
	if (ev->code == GRF_RESIZE)
	 { printf ("Resize %d %d\n", ev->val1, ev->val2);
	 }
	else if (ev->code == GRF_REDRAW)
	 { printf ("Redraw %d %d\n", ev->val1, ev->val2);
	 }
	else if (ev->code == GRF_FOCUS_CHANGE)
	 { printf ("Focus change %d\n", ev->val1);
	 }
	else
	 { grf4Fatal ("onConfig(): bogus code %d\n", ev->code);
	 }
}


#define IGNORE	0
#define QUEUE	1
#define ATTACH	2

int eventSel = IGNORE;

void onAB (int n, int code, long mode)
{
}

void onEventSelect (int n, int code, long mode)
{
	if (code == BF_INFOCUS)
	 { grf4LightButton(n);
	 }
	else if (code == BF_OUTFOCUS)
	 { grf4ClearButton(n);
	 }
	else if (code == BF_PRESSED)
	 { eventSel = (eventSel+1)%3;
	   switch (eventSel)
	    { case IGNORE:
	       { grf4ReleaseEvent (GRF_MOTION_EVENT);
		 grf4ReleaseEvent (GRF_BUTTON_EVENT);
		 grf4ReleaseEvent (GRF_RESIZE_EVENT);
		 grf4ReleaseEvent (GRF_REDRAW_EVENT);
		 grf4ReleaseEvent (GRF_KEY_EVENT);
		 grf4ReleaseEvent (GRF_FOCUS_EVENT);
		 grf4SetButtonName (n, "ignore");
		 break;
	       }
	      case QUEUE:
	       { grf4FlushQueue();
		 grf4QueueEvent (GRF_MOTION_EVENT);
		 grf4QueueEvent (GRF_BUTTON_EVENT);
		 grf4QueueEvent (GRF_RESIZE_EVENT);
		 grf4QueueEvent (GRF_REDRAW_EVENT);
		 grf4QueueEvent (GRF_KEY_EVENT);
		 grf4QueueEvent (GRF_FOCUS_EVENT);
		 grf4SetButtonName (n, "ignore");
		 grf4SetButtonName (n, "queue");
		 break;
	       }
	      case ATTACH:
	       { grf4AttachEvent (GRF_MOTION_EVENT, onCursor, 1);
		 grf4AttachEvent (GRF_BUTTON_EVENT, onButton, 1);
		 grf4AttachEvent (GRF_RESIZE_EVENT, onConfig, 1);
		 grf4AttachEvent (GRF_REDRAW_EVENT, onConfig, 1);
		 grf4AttachEvent (GRF_KEY_EVENT, onKey, 1);
		 grf4AttachEvent (GRF_FOCUS_EVENT, onConfig, 1);
		 grf4SetButtonName (n, "attach");
		 break;
	       }
	    }
	 }
}

int winw;
int winh;

char **argvec;

int noPanel;

main(int argc, char **argv)
{
	float dw, dh;
	void *mat;
	void *light;
	double t0, t1;
	long flags;
	int i;

	flags = GRF_VIEW_CONTROL;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 {
	   if (ARGUMENT("-X"))
	    { flags |= GRF_X11;
	    }
	   else if (ARGUMENT("-oldgl"))
	    { flags |= GRF_OLD_GL;
	    }
	   else if (ARGUMENT("-wf"))
	    { flags |= GRF_WIRE_FRAME;
	    }
	   else if (ARGUMENT("-nopan"))
	    { flags &= ~GRF_VIEW_CONTROL;
	      noPanel = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-wf] [-X] [-nopan] [-oldgl] [cross|cone|poly]\n", *argvec);
	   exit (1);
	 }

	if (!noPanel)
	 { grf4AddButton ("a", 0, 1, onAB, 0);
	   grf4AddButton ("b", 0, 1, onAB, 0);
	 }
/*	grf4PlaceWindow (300, 300, 400, 400); */
	if (grf4Init ("foo", GRF_WHITE, GRF_BLACK, flags) < 0)
	 { printf ("Can't init graphics\n");
	   exit (-1);
	 }
	grf4SetDrawFxn (drawIt, 0);

	grf4Color (0xffffff);
	grf4GetWindowSize (&winw, &winh);
	dw = 1/(float)winw;
	dh = 1/(float)winh;

	grf4Window (-75.0, 75.0, -75.0, 75.0, 300.0, 500.0, 1.0);
	grf4SetView (0.0, 0.0, -400.0, 0.0, 0.0, 0.0);

	mat = grf4DefMaterial(0xffffff, 0, 0x400000, 0, 0.0, 1.0);
	light = grf4DefLight (0xffffff, 0, 1.0, 1.0, 1.0, 0.0);

	grf4SetLight (0, light);
	grf4EnableJack (1);
	grf4SetJackSize (40.0);
	 { TRSF T;
	   rpyToTrsf (&T, 45.0, 45.0, 0);
	   grf4SetJackTrsf (&T);
	 }
	drawIt();
	grf4QueueEvent (GRF_BUTTON_EVENT);
	grf4QueueEvent (GRF_KEY_EVENT);
	grf4QueueEvent (GRF_MOTION_EVENT);

	while (1)
	 { Grf4Event ev;

	   grf4Wait (-1);
	   while (grf4TestQueue())
	    { grf4ReadQueue (&ev);
	      if (GRF_IS_BUTTON(ev.code))
	       { onButton (&ev, 100);
	       }
	      else if (GRF_IS_KEY(ev.code))
	       { onKey (&ev, 100);
	       }
	      else
	       { switch (ev.code)
		  { case GRF_CURSOR:
		     { onCursor (&ev, 100);
		       break;
		     }
		    case GRF_RESIZE:
		     { onConfig (&ev, 100);
		       break;
		     }
		    case GRF_REDRAW:
		     { onConfig (&ev, 100);
		       break;
		     }
		    case GRF_FOCUS_CHANGE:
		     { onConfig (&ev, 100);
		       break;
		     }
		  }
	       }
	    }
	 }
}
