#include <grf4.h>
#include <RxCvxPolyhed3.h>
#include <RxCompoundBody.h>
#include <RxFrameNode.h>
#include <RxRigidBody.h>
#include "grf4Support.h"
#include "grf4Internal.h"
#include <assert.h>

extern int grf4FxnInitDone;
extern int grf4HostType;

Grf4Object *grf4GenPolyhed3 (RxCvxPolyhed3 &poly, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));

	/* generate the bottom face */

	obj->ptr.gen = &poly;
	obj->type = GRF_POLYHED3;
	if (material)
	 { obj->material = material;
	 }
	return (obj);	
}

Grf4Object *grf4GenCompoundBody (RxCompoundBody &body, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));

	obj->ptr.gen = &body;
	obj->type = GRF_RIGIDBODY;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenFrameNode (RxFrameNode &node, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));

	obj->ptr.gen = &node;
	obj->type = GRF_FRAMENODE;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

static void RxTrsfToTRSF (TRSF *TR, const RxTrsf &T)
{
	TR->n.x = T.R.n.x();
	TR->n.y = T.R.n.y();
	TR->n.z = T.R.n.z();

	TR->o.x = T.R.o.x();
	TR->o.y = T.R.o.y();
	TR->o.z = T.R.o.z();

	TR->a.x = T.R.a.x();
	TR->a.y = T.R.a.y();
	TR->a.z = T.R.a.z();

	TR->p.x = T.p.x();
	TR->p.y = T.p.y();
	TR->p.z = T.p.z();
}

static void TRSFToRxTrsf (RxTrsf *TR, const TRSF *T)
{
	TR->R.n.x() = T->n.x;
	TR->R.n.y() = T->n.y;
	TR->R.n.z() = T->n.z;
	TR->R.o.x() = T->o.x;
	TR->R.o.y() = T->o.y;
	TR->R.o.z() = T->o.z;
	TR->R.a.x() = T->a.x;
	TR->R.a.y() = T->a.y;
	TR->R.a.z() = T->a.z;
	TR->p.x() = T->p.x;
	TR->p.y() = T->p.y;
	TR->p.z() = T->p.z;
}

static void *doDefMaterial (const RxMaterial &mat)
{
	MATERIAL_DEF mdef;
	float alpha;

	mdef.diffuse = mat.getDiffuse();
	mdef.ambient = mat.getAmbient();
	mdef.emission = mat.getEmission();
	mdef.specular = mat.getSpecular();
	mdef.shine = mat.getShine();
printf ("%08x %08x %08x %08x\n", mdef.diffuse,
	mdef.ambient, mdef.emission, mdef.specular);
	mat.getDiffuse (NULL, NULL, NULL, &alpha); /* XXX */
	mdef.alpha = alpha;
	return grf4DefMaterialD (&mdef);
}

static void drawCompoundBody (RxCompoundBody *body, void *mptr, void *&mptrSet)
{
	RxCompoundBody *child;
	RxMaterial *mat;

	if ((mat = body->getMat()))
	 { if (mat->getGraphicObj() == NULL)
	    { void *mptr;
	      mptr = doDefMaterial (*mat);
	      assert (mptr != NULL);
	      mat->setGraphicObj (mptr);
	    }
	   mptr = mat->getGraphicObj();
	 }
	if (body->poly())
	 { TRSF TRL;
	   RxTrsfToTRSF (&TRL, *body->getTRL());
	   grf4PushTrans ();
	   grf4MultTrans (&TRL);
	   if (mptr && mptr != mptrSet)
	    { grf4SetMaterial (mptr);
	      mptrSet = mptr;
	    }
	   grf4IncDrawObjCnt();
	   grf4DrawPolyhed ((void*)body->poly());
	   grf4PopTrans ();
	 }
	else
	 { for (child=body->getChild(0); child; child=child->next())
	    { drawCompoundBody (child, mptr, mptrSet);
	    }
	 }
}

void grf4DrawCompoundBody (Grf4Object *obj)
{
	void *mptrSet = grf4GetMaterial();

	drawCompoundBody ((RxCompoundBody*)obj->ptr.gen, mptrSet, mptrSet);
}

static void drawFrameNode (const RxFrameNode *node, const TRSF &TWP,
			   void *mptr, void *&mptrSet)
{
	const RxFrameNode *child;
	RxMaterial *mat;
	TRSF TWL;
	TRSF TPL;

	RxTrsfToTRSF (&TPL, node->getTrsf());
	multTrsf (&TWL, &TWP, &TPL);

	if ((mat = node->getMat()))
	 { if (mat->getGraphicObj() == NULL)
	    { void *mptr;
	      mptr = doDefMaterial (*mat);
	      assert (mptr != NULL);
	      mat->setGraphicObj (mptr);
	    }
	   mptr = mat->getGraphicObj();
	 }
	if (node->subframe())
	 { drawFrameNode (node->subframe(), TWL, mptr, mptrSet);
	 }
	else if (node->type() == RxRigidBody::ID)
	 { RxRigidBody *body = (RxRigidBody*)node;
	   if (body->poly())
	    { grf4PushTrans ();
	      grf4MultTrans (&TWL);
	      if (mptr && mptr != mptrSet)
	       { grf4SetMaterial (mptr);
		 mptrSet = mptr;
	       }
	      grf4IncDrawObjCnt();
	      grf4DrawPolyhed ((void*)body->poly());
	      grf4PopTrans ();
	    }
	 }
	for (child=node->getChild(0); child; child=child->next())
	 { drawFrameNode (child, TWL, mptr, mptrSet);
	 }
}

void grf4DrawFrameNode (Grf4Object *obj)
{
	void *mptrSet = grf4GetMaterial();
	extern TRSF grf4IdentTrsf;

	drawFrameNode ((RxFrameNode*)obj->ptr.gen, 
			grf4IdentTrsf, mptrSet, mptrSet);
}

void grf4SetJackTrsf (const RxTrsf &T)
{
	TRSF TT;

	RxTrsfToTRSF (&TT, T);
	grf4SetJackTrsf (&TT);
}

void grf4GetJackTrsf (RxTrsf *Tp)
{
	TRSFToRxTrsf (Tp, grf4GetJackTrsfPtr());
}

void grf4SetViewTrsf (const class RxTrsf &T)
{
	RxTrsfToTRSF (grf4GetViewTrsfPtr(), T);
}

void grf4SetTrans (const class RxTrsf &T)
{
	RxTrsfToTRSF (grf4GetTransPtr(), T);
}

void grf4GetTrans (RxTrsf *Tp)
{
	TRSFToRxTrsf (Tp, grf4GetTransPtr());
}

void grf4MultTrans (const class RxTrsf &T)
{
	TRSF TM;
	RxTrsfToTRSF (&TM, T);
	grf4MultTrans (&TM);	
}

void grf4GetViewTrsf (RxTrsf *Tp)
{
	TRSFToRxTrsf (Tp, grf4GetViewTrsfPtr());	
}
