#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <cdefs.h>
#include <grf4.h>

#include <GL/gl.h>

#define PI        3.14159265358979320
#define DTOR	  (PI/180.0)
#define RTOD	  (180.0/PI)
	
#define CROSS	1
#define POLY    2
#define CONE	3
#define TEST_QUAD 4 // hack to test GL_QUAD arrays

int mode = CONE;
Grf4Object *cone;
Grf4Object *disk;
Grf4Object *sphere;
Grf4Object *p1, *p2, *p3;

float p1_edges[] = { 0.0, -20.0,   0.0, 50.0,   -50.0, -20.0 };

float roll, pitch, yaw;

extern double frandom();

extern void* theFace;
void* spherePoly;
void* conePoly;
void* discPoly;

static double cputime();
	
void drawIt()
{
	float a;

	grf4Clear ();
	switch (mode)
	 { case CROSS:
	    { 
	      grf4Move2 (0.0, 0.0);
	      grf4Draw2 (1.0, 1.0);
	      grf4Move2 (1.0, 0.0);
	      grf4Draw2 (0.0, 1.0);
	      break;
	    }
	   case CONE:
	    { 
#if !TEST_QUAD
	      grf4DrawObject (cone);
#endif
	      conePoly = theFace;
	      grf4PushTrans();
#if !TEST_QUAD
	      grf4Translate (0.0, 0.0, 95.0);
#endif
#if 0
	      grf4DrawObject (sphere);
	      spherePoly = theFace;
#endif
#if !TEST_QUAD
	      grf4Translate (0.0, 25.0, -120.0);
	      grf4DrawObject (disk);
	      grf4Translate (0.0, 25.0, -120.0);
	      grf4DrawObject (disk);
	      discPoly = theFace;
#endif
#if TEST_QUAD
// this was a hack to see if GL_QUADS worked properly under Mesa
	       { double z1 = 10.0;
		 double w = 40.0;
		 double h = 60.0;
		 double z2 = -10.0;

		 glBegin(GL_QUADS);
		 glNormal3f (0.0, 0.0, 2);
		 glVertex3f (w/2, -h/2, z1);
		 glVertex3f (w/2,  h/2, z1);
		 glVertex3f (-w/2,  h/2, z1);
		 glVertex3f (-w/2, -h/2, z1);

		 glNormal3f (0.0, 0.0, -2);
		 glVertex3f (w/2, -h/2, z2);
		 glVertex3f (-w/2, -h/2, z2);
		 glVertex3f (-w/2,  h/2, z2);
		 glVertex3f (w/2,  h/2, z2);

		 glEnd();
	       }
#endif
	      grf4PopTrans();
	      break;
	    }
	   case POLY:
	    { 
	      grf4PushTrans();
	      grf4Rotate (roll, pitch, yaw);
	      grf4Translate (40.0, 0.0, 0.0);
	      grf4Rotate (45.0, 0.0, 0.00);
	      grf4DrawObject (p2);
	      a = 20.0*(sqrt(2.0));
	      grf4Translate (-a, a, -40.0);
	      grf4Rotate (45.0, -45.0, -90.0);
	      grf4DrawObject (p3);
#if 0
	      grf4Translate (30.0, 40.0, -20.0);
	      grf4DrawObject (p3);
#endif
	      grf4PopTrans();
	      break;
	    }
	 }
	grf4Flush ();
}

static double cputime()
{
	struct timeval tv;
	gettimeofday (&tv, (struct timezone*)0);
	return (tv.tv_sec + tv.tv_usec*1e-6);
}

int xButton;

void onHideLines(int n, int code, long m)
{
	if (code == BF_INFOCUS)
	 { grf4LightButton(n);
	 }
	else if (code == BF_OUTFOCUS)
	 { grf4ClearButton(n);
	 }
	else if (code == BF_RELEASED)
	 { grf4HideLines(1);
	   drawIt();
	   grf4HideLines(0);
	   printf ("Released\n");
	   if (n == xButton)
	    { grf4SetButtonName (n, "x");
	    }
	 }
	else if (code == BF_PRESSED)
	 { printf ("Pressed\n");
	   if (n == xButton)
	    { grf4SetButtonName (n, "foo");
	    }
	 }
}

printUpDown(char *s, int n)
{
	printf ("%s %s\n", s, n ? "down" : "up");
}

void onCursor (Grf4Event *ev, int arg)
{
	printf ("Cursor %d %d\n", ev->val1, ev->val2);
}

void onButton (Grf4Event *ev, int arg)
{
	switch (ev->code)
	 { case GRF_LEFT_MOUSE:
	    { printUpDown ("Left mouse", ev->val1);
	      break;
	    }
	   case GRF_MIDDLE_MOUSE:
	    { printUpDown ("Middle mouse", ev->val1);
	      break;
	    }
	   case GRF_RIGHT_MOUSE:
	    { printUpDown ("Right mouse", ev->val1);
	      break;
	    }
	   default:
	    { grf4Fatal ("onButton(): bogus code %d\n", ev->code);
	    }
	 }
}

void onKey (Grf4Event *ev, int arg)
{
	if (!GRF_IS_KEY(ev->code))
	 { grf4Fatal ("onKey(): code %d is not a key code\n", ev->code);
	 }
	printf ("%s %s\n", grf4DevName(ev->code), 
			   ev->val1 ? "down" : "up");
}

void onConfig (Grf4Event *ev, int arg)
{
	if (ev->code == GRF_RESIZE)
	 { printf ("Resize %d %d\n", ev->val1, ev->val2);
	 }
	else if (ev->code == GRF_REDRAW)
	 { printf ("Redraw %d %d\n", ev->val1, ev->val2);
	 }
	else if (ev->code == GRF_FOCUS_CHANGE)
	 { printf ("Focus change %d\n", ev->val1);
	 }
	else
	 { grf4Fatal ("onConfig(): bogus code %d\n", ev->code);
	 }
}


#define IGNORE	0
#define QUEUE	1
#define ATTACH	2

int eventSel = IGNORE;

void onEventSelect (int n, int code, long mode)
{
	if (code == BF_INFOCUS)
	 { grf4LightButton(n);
	 }
	else if (code == BF_OUTFOCUS)
	 { grf4ClearButton(n);
	 }
	else if (code == BF_PRESSED)
	 { eventSel = (eventSel+1)%3;
	   switch (eventSel)
	    { case IGNORE:
	       { grf4ReleaseEvent (GRF_MOTION_EVENT);
		 grf4ReleaseEvent (GRF_BUTTON_EVENT);
		 grf4ReleaseEvent (GRF_RESIZE_EVENT);
		 grf4ReleaseEvent (GRF_REDRAW_EVENT);
		 grf4ReleaseEvent (GRF_KEY_EVENT);
		 grf4ReleaseEvent (GRF_FOCUS_EVENT);
		 grf4SetButtonName (n, "ignore");
		 break;
	       }
	      case QUEUE:
	       { grf4FlushQueue();
		 grf4QueueEvent (GRF_MOTION_EVENT);
		 grf4QueueEvent (GRF_BUTTON_EVENT);
		 grf4QueueEvent (GRF_RESIZE_EVENT);
		 grf4QueueEvent (GRF_REDRAW_EVENT);
		 grf4QueueEvent (GRF_KEY_EVENT);
		 grf4QueueEvent (GRF_FOCUS_EVENT);
		 grf4SetButtonName (n, "ignore");
		 grf4SetButtonName (n, "queue");
		 break;
	       }
	      case ATTACH:
	       { grf4AttachEvent (GRF_MOTION_EVENT, onCursor, 1);
		 grf4AttachEvent (GRF_BUTTON_EVENT, onButton, 1);
		 grf4AttachEvent (GRF_RESIZE_EVENT, onConfig, 1);
		 grf4AttachEvent (GRF_REDRAW_EVENT, onConfig, 1);
		 grf4AttachEvent (GRF_KEY_EVENT, onKey, 1);
		 grf4AttachEvent (GRF_FOCUS_EVENT, onConfig, 1);
		 grf4SetButtonName (n, "attach");
		 break;
	       }
	    }
	 }
}

int winw;
int winh;

char **argvec;

int noPanel;

main(int argc, char **argv)
{
	float dw, dh;
	void *mat;
	void *light;
	double t0, t1;
	long flags;
	int i;

	flags = GRF_VIEW_CONTROL;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 {
	   if (ARGUMENT("-X"))
	    { flags |= GRF_X11;
	    }
	   else if (ARGUMENT("-oldgl"))
	    { flags |= GRF_OLD_GL;
	    }
	   else if (ARGUMENT("-wf"))
	    { flags |= GRF_WIRE_FRAME;
	    }
	   else if (ARGUMENT("-nopan"))
	    { flags &= ~GRF_VIEW_CONTROL;
	      noPanel = 1;
	    }
	   else if (ARGUMENT("cross"))
	    { mode = CROSS;
	    }
	   else if (ARGUMENT("cone"))
	    { mode = CONE;
	    }
	   else if (ARGUMENT("poly"))
	    { mode = POLY;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-wf] [-X] [-nopan] [-oldgl] [cross|cone|poly]\n", *argvec);
	   exit (1);
	 }

	if (!noPanel)
	 { grf4AddButton ("hide lines", 0, 1, onHideLines, 0);
	   xButton = grf4AddButton ("x", 0, 1, onHideLines, 0);
	   grf4AddButton ("y", BF_SAMEROW, 1, onHideLines, 0);
	   grf4AddButton ("z", BF_SAMEROW, 1, onHideLines, 0);
	   grf4AddButton ("ignore", 0, 1, onEventSelect, 0);
	 }
/*	grf4PlaceWindow (300, 300, 400, 400); */
	if (grf4Init ("foo", GRF_WHITE, GRF_BLACK, flags) < 0)
	 { printf ("Can't init graphics\n");
	   exit (-1);
	 }
	grf4SetDrawFxn (drawIt, 0);

	grf4Color (0xffffff);
	grf4GetWindowSize (&winw, &winh);
	dw = 1/(float)winw;
	dh = 1/(float)winh;
	switch (mode)
	 { case CROSS:
	    { 
	      grf4Ortho (-dw, 1+dw, -dh, 1+dh, -1.0, 1.0, 1.0);
	      break;
	    }
	   case CONE:
	    { 
	      grf4Window (-75.0, 75.0, -75.0, 75.0, 300.0, 500.0, 1.0);
	      grf4SetView (0.0, 0.0, -400.0, 0.0, 0.0, 0.0);

	      mat = grf4DefMaterial(0xffffff, 0, 0x400000, 0, 0.0, 1.0);
	      light = grf4DefLight (0xffffff, 0, 1.0, 1.0, 1.0, 0.0);

	      grf4SetLight (0, light);
#if TEST_QUAD
	      iGLSetMaterialHook (mat);
#endif
	      cone = grf4GenCone (4, 25.0, 25.0, 75.0, mat);
	      disk = grf4GenDisc (4, 50.0, 25.0, mat);
	      sphere = grf4GenSphere (12, 20.0, mat);
	      break;
	    }
	   case POLY:
	    { 
	      grf4Ortho (-75.0, 75.0, -75.0, 75.0, 300.0, 500.0, 1.0);
	      grf4SetView (0.0, 0.0, -400.0, 0.0, 0.0, 0.0);

	      light = grf4DefLight (0xffffff, 0, 1.0, 1.0, 1.0, 0.0);
	      grf4SetLight (0, light);

	      p1 = grf4GenPolygon (3, p1_edges, 0);
	      p2 = grf4GenRegPolygon (4, 80.0, 0);
	      p3 = grf4GenRegPolygon (4, 80.0, 0);
	      break;
	    }
	 }
#if 0
#define NTIMES	5000

	t0 = cputime(User);
	for (i=0; i<NTIMES; i++)
	 { grf4DrawObject (cone);
	 }
	t1 = cputime(User);
	printf ("time = %g msec.\n", (t1-t0)*1000/NTIMES);
#endif
#if 0
	while (1)
	 { roll = frandom (-1.0, 1.0);
	   pitch = frandom (-10.0, 0.0);
	   yaw = frandom (-0.0, 10.0);
	   printf ("%11.8f %11.8f %11.8f\n", roll, pitch, yaw);
	   roll = 0.704318;
	   pitch = -0.790037;
	   yaw = -6.625;
	   drawIt();
	 }
#endif
	drawIt();
	grf4QueueEvent (GRF_BUTTON_EVENT);
	grf4QueueEvent (GRF_KEY_EVENT);

	while (1)
	 { Grf4Event ev;

	   grf4Wait (-1);
	   while (grf4TestQueue())
	    { grf4ReadQueue (&ev);
	      printf ("Q: ");		    
	      if (GRF_IS_BUTTON(ev.code))
	       { onButton (&ev, 100);
	       }
	      else if (GRF_IS_KEY(ev.code))
	       { onKey (&ev, 100);
	       }
	      else
	       { switch (ev.code)
		  { case GRF_CURSOR:
		     { onCursor (&ev, 100);
		       break;
		     }
		    case GRF_RESIZE:
		     { onConfig (&ev, 100);
		       break;
		     }
		    case GRF_REDRAW:
		     { onConfig (&ev, 100);
		       break;
		     }
		    case GRF_FOCUS_CHANGE:
		     { onConfig (&ev, 100);
		       break;
		     }
		  }
	       }
	    }
	 }
}
