#include <RxMatd.h>
#include <RxMathError.h>
#include <RxMatdLu.h>
#include <stdlib.h>
#include <cdefs.h>
#include <math.h>

void RxMatd::doCopy (const RxMatd &m1)
{
	double *dpr;
	double *dp1;
	int i, j;

	dpr = store;
	dp1 = m1.store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dpr++ = *dp1++;
	    }
	   dpr += (width-nc);
	   dp1 += (m1.width-m1.nc);
	 }
}

RxMatd::RxMatd (const RxMatd &m1)
{
	nr = m1.nr;
	nc = m1.nc;
	store = new double [nr*nc];
	width = nc;
	malloced = 1;
	doCopy (m1);
}

RxMatd::RxMatd (const RxMatdBase &m1)
{
	int i, j;

	nr = m1.nrows();
	nc = m1.ncols();
	store = new double [nr*nc];
	width = nc;
	malloced = 1;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { store[i*nc+j] = m1(i,j);
	    }
	 }
}

RxMatd::RxMatd (const double *dp, int nrow, int ncol)
{
	nr = nrow;
	nc = ncol;
	store = new double [nr*nc];
	width = nc;
	malloced = 1;
	set (dp);
}

RxMatd::RxMatd (const RxVecd &v, int column)
{
	int i;

	if (column)
	 { nr = v.len;
	   nc = 1;
	 }
	else
	 { nr = 1;
	   nc = v.len;
	 }
	store = new double [v.len];
	for (i=0; i<v.len; i++)
	 { store[i] = v.v[i];
	 }
	width = nc;
	malloced = 1;
}

RxMatd& RxMatd::set(const RxMatd& m1)
{
	if (nr*nc != m1.nr*m1.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::operator=: %K\n").raise();
	   return (*this);
	 }
	if (nr != m1.nr)
	 { if (width != nc)
	    { RxMathError (RxMathError::BadSize,
			"RxMatd::operator=: %K\n").raise();
	      return (*this);
	    }
	   nr = m1.nr;
	   nc = m1.nc;
	   width = nc;
	 }
	doCopy (m1);
	return (*this);
}

RxMatd& RxMatd::set (const double *dp)
{
	double *dpr;
	int i, j;

	dpr = store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dpr++ = *dp++;
	    }
	   dpr += (width-nc);
	 }
	return (*this);
}

RxMatd& RxMatd::set (const RxVecd &v1)
{
	if (nr*nc != v1.len)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::set %K\n").raise();
	 }
	else
	 { double *dpr = store;
	   double *dpv = v1.v;
	   int i, j;
	   
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dpv++;
	       }
	      dpr += (width-nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::set (double e1, ...)
{
	va_list ap;
	int i, j;
	int j0 = 1;
	
	va_start(ap, e1);
	store[0] = e1;
	for (i=0; i<nr; i++)
	 { for (j=j0; j<nc; j++)
	    { store[i*width+j] = va_arg(ap, double);
	    }
	   j0 = 0;
	 }
	va_end(ap);
	return (*this);
}

RxMatd& RxMatd::setDiag (const RxVecd &v1)
{
	if (v1.len != MIN(nr,nc))
	 { RxMathError (RxMathError::BadSize, 
			"RxMatd::setDiag: %K\n").raise();
	 }
	else
	 { int i, j;

	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { store[i*width+j] = (i==j) ? v1.v[i] : 0;
	       }
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::setDiag (double *buf)
{
	int i, j;

	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { store[i*width+j] = (i==j) ? buf[i] : 0;
	    }
	 }
	return (*this);
}

void RxMatd::get (double *dp)
{
	double *dpr;
	int i, j;

	dpr = store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dp++ = *dpr++;
	    }
	   dpr += (width-nc);
	 }
}

RxVecd& RxMatd::get (RxVecd *vr)
{
	if (nr*nc != vr->len)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::get(): %K\n").raise();
	 }
	else
	 { double *dpr = store;
	   double *dpv = vr->v;
	   int i, j;
	   
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpv++ = *dpr++;
	       }
	      dpr += (width-nc);
	    }
	 }
	return (*vr);
}

RxMatd& RxMatd::add(const RxMatd& m1, const RxMatd& m2)
{
	if (nr != m1.nr || nr != m2.nr || nc != m1.nc || nc != m2.nc)
	 { RxMathError (RxMathError::BadSize, "RxMatd::add: %K\n").raise();
	 }
	else
	 { double *dpr, *dp1, *dp2;
	   int i, j;

	   dpr = store;
	   dp1 = m1.store;
	   dp2 = m2.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dp1++ + *dp2++;
	       }
	      dpr += (width-nc);
	      dp1 += (m1.width-m1.nc);
	      dp2 += (m2.width-m2.nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::add(const RxMatd& m2)
{
	if (nr != m2.nr || nc != m2.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::add: %K\n").raise();
	 }
	else
	 { double *dpr, *dp2;
	   int i, j;

	   dpr = store;
	   dp2 = m2.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ += *dp2++;
	       }
	      dpr += (width-nc);
	      dp2 += (m2.width-m2.nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::sub(const RxMatd& m1, const RxMatd& m2)
{
	if (nr != m1.nr || nr != m2.nr || nc != m1.nc || nc != m2.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::sub: %K\n").raise();
	 }
	else
	 { double *dpr, *dp1, *dp2;
	   int i, j;

	   dpr = store;
	   dp1 = m1.store;
	   dp2 = m2.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dp1++ - *dp2++;
	       }
	      dpr += (width-nc);
	      dp1 += (m1.width-m1.nc);
	      dp2 += (m2.width-m2.nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::sub(const RxMatd& m2)
{
	if (nr != m2.nr || nc != m2.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::sub: %K\n").raise();
	 }
	else
	 { double *dpr, *dp2;
	   int i, j;

	   dpr = store;
	   dp2 = m2.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ -= *dp2++;
	       }
	      dpr += (width-nc);
	      dp2 += (m2.width-m2.nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::mul(const RxMatd& m1, const RxMatd& m2)
{
	double *dpr, *dpt;
	double *dp1, *dp2;
	int i, j, k, w;
	double *tmp;
	double accum;

	if (nr != m1.nr || nc != m2.nc || m1.nc != m2.nr)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::mul: %K\n").raise();
	   return (*this);
	 }
	if (m1.store == store || m2.store == store)
	 { tmp = (double*)malloc(sizeof(double)*nr*nc);
	   w = nc;
	 }
	else
	 { tmp = store;
	   w = width;
	 }

	for (i=0; i<nr; i++)
	 { dpt = tmp + i*w;
	   for (j=0; j<nc; j++)
	    { accum = 0;
	      dp2 = m2.store+j;
	      dp1 = m1.store+(i*m1.width);
	      for (k=0; k<m1.nc; k++)
	       { accum += *dp1++ * (*dp2);
		 dp2 += m2.width;
	       }
	      *dpt++ = accum;
	    }
	 }
	if (tmp != store)
	 { dpt = tmp;
	   dpr = store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dpt++;
	       }
	      dpr += (width-nc);
	    }
	   free (tmp);
	 }
	return (*this);
}

RxMatd& RxMatd::Tmul(const RxMatd& m1, const RxMatd& m2)
{
	double *dpr, *dpt;
	double *dp1, *dp2;
	int i, j, k, w;
	double *tmp;
	double accum;

	if (nr != m1.nc || nc != m2.nc || m1.nr != m2.nr)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::Tmul: %K\n").raise();
	   return (*this);
	 }
	if (m1.store == store || m2.store == store)
	 { tmp = (double*)malloc(sizeof(double)*nr*nc);
	   w = nc;
	 }
	else
	 { tmp = store;
	   w = width;
	 }

	for (i=0; i<nr; i++)
	 { dpt = tmp + i*w;
	   for (j=0; j<nc; j++)
	    { accum = 0;
	      dp2 = m2.store+j;
	      dp1 = m1.store+i;
	      for (k=0; k<m1.nr; k++)
	       { accum += *dp1 * (*dp2);
		 dp2 += m2.width;
		 dp1 += m1.width;
	       }
	      *dpt++ = accum;
	    }
	 }
	if (tmp != store)
	 { dpt = tmp;
	   dpr = store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dpt++;
	       }
	      dpr += (width-nc);
	    }
	   free (tmp);
	 }
	return (*this);
}

RxMatd& RxMatd::mulT(const RxMatd& m1, const RxMatd& m2)
{
	double *dpr, *dpt;
	double *dp1, *dp2;
	int i, j, k, w;
	double *tmp;
	double accum;

	if (nr != m1.nr || nc != m2.nr || m1.nc != m2.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::Tmul: %K\n").raise();
	   return (*this);
	 }
	if (m1.store == store || m2.store == store)
	 { tmp = (double*)malloc(sizeof(double)*nr*nc);
	   w = nc;
	 }
	else
	 { tmp = store;
	   w = width;
	 }

	for (i=0; i<nr; i++)
	 { dpt = tmp + i*w;
	   for (j=0; j<nc; j++)
	    { accum = 0;
	      dp1 = m1.store+i*m1.width;
	      dp2 = m2.store+j*m2.width;
	      for (k=0; k<m1.nc; k++)
	       { accum += (*dp1++) * (*dp2++);
	       }
	      *dpt++ = accum;
	    }
	 }
	if (tmp != store)
	 { dpt = tmp;
	   dpr = store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = *dpt++;
	       }
	      dpr += (width-nc);
	    }
	   free (tmp);
	 }
	return (*this);
}

RxMatd& RxMatd::scale(double a, const RxMatd& m1)
{
	if (nr != m1.nr || nc != m1.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::add: %K\n").raise();
	 }
	else
	 { double *dpr, *dp1;
	   int i, j;

	   dpr = store;
	   dp1 = m1.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = a*(*dp1++);
	       }
	      dpr += (width-nc);
	      dp1 += (m1.width-m1.nc);
	    }
	 }
	return (*this);
}

RxMatd &RxMatd::scale(double a)
{
	double *dpr;
	int i, j;
	
	dpr = store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dpr++ *= a;
	    }
	   dpr += (width-nc);
	 }
	return (*this);
}

#define MAXT	16

RxVecd &RxMatd::Vmul (RxVecd *vr, const RxVecd &v1) const
{
	if (nr != v1.len || nc != vr->len)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::Vmul: %K\n").raise();
	 }
	else
	 { int i, j;
	   double sum;
	   double buf[MAXT];
	   double *tmp;

	   if (vr->v == v1.v)
	    { if (nc > MAXT)
	       { tmp = (double*)malloc(sizeof(double)*nc);
	       }
	      else
	       { tmp = buf;
	       }
	    }
	   else
	    { tmp = vr->v;
	    }
	
	   for (j=0; j<nc; j++)
	    { sum = 0;
	      for (i=0; i<nr; i++)
	       { sum += store[i*width+j]*v1.v[i];
	       }
	      tmp[j] = sum;
	    }

	   if (vr->v == v1.v)
	    { for (j=0; j<nc; j++)
	       { vr->v[j] = tmp[j];
	       }
	      if (tmp != buf)
	       { free (buf);
	       }
	    }
	 }
	return (*vr);
}

RxVecd &RxMatd::mulV (RxVecd *vr, const RxVecd &v1) const
{
	if (nc != v1.len || nr != vr->len)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::mulV: %K\n").raise();
	 }
	else
	 { int i, j;
	   double sum;
	   double buf[MAXT];
	   double *tmp;

	   if (vr->v == v1.v)
	    { if (nr > MAXT)
	       { tmp = (double*)malloc(sizeof(double)*nr);
	       }
	      else
	       { tmp = buf;
	       }
	    }
	   else
	    { tmp = vr->v;
	    }
	
	   for (i=0; i<nr; i++)
	    { sum = 0;
	      for (j=0; j<nc; j++)
	       { sum += store[i*width+j]*v1.v[j];
	       }
	      tmp[i] = sum;
	    }

	   if (vr->v == v1.v)
	    { for (i=0; i<nr; i++)
	       { vr->v[i] = tmp[i];
	       }
	      if (tmp != buf)
	       { free (buf);
	       }
	    }
	 }
	return (*vr);
}

RxMatd& RxMatd::tsp(const RxMatd& m1)
{
	double *dpr;
	double *dp1;
	int i, j;

	if (&m1 == this)
	 { return (tsp());
	 }
	if (nr*nc != m1.nc*m1.nr || width != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::tsp: %K\n").raise();
	   return (*this);
	 }
	nr = m1.nc;
	nc = m1.nr;
	width = nc;

	dpr = store;
	for (i=0; i<nr; i++)
	 { dp1 = m1.store+i;
	   for (j=0; j<nc; j++)
	    { *dpr++ = *dp1;
	      dp1 += m1.width;
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::tsp()
{
	int i, j;

	if (nr == nc)
	 { double dtmp;

	   for (i=0; i<nr; i++)
	    { for (j=i+1; j<nc; j++)
	       { dtmp = store[i*width+j];
		 store[i*width+j] = store[j*width+i];
		 store[j*width+i] = dtmp;
	       }
	    }
	 }
	else if (width != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::tsp: %K\n").raise();
	 }
	else
	 { double *tmp = (double*)malloc(sizeof(double)*nr*nc);
	   double *dpr, *dpt;

	   i = nr;
	   nr = nc;
	   nc = i;

	   dpt = tmp;
	   for (i=0; i<nr; i++)
	    { dpr = store+i;
	      for (j=0; j<nc; j++)
	       { *dpt++ = *dpr;
		 dpr += width;
	       }
	    }
	   width = nc;
	   for (i=0; i<nr*nc; i++)
	    { store[i] = tmp[i];
	    }
	   free (tmp);
	 }
	return (*this);
}

RxMatd& RxMatd::inv(const RxMatd& m1)
{
	RxMatdLu lu;

	if (m1.nr != m1.nc || m1.nr != nr || m1.nc != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::inv(): %K\n").raise();
	   return (*this);
	 }
	lu.calc (m1);
	lu.inv (this);
	return (*this);
}

RxMatd& RxMatd::neg(const RxMatd& m1)
{
	if (nr != m1.nr || nc != m1.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::neg: %K\n").raise();
	 }
	else
	 { double *dpr, *dp1;
	   int i, j;

	   dpr = store;
	   dp1 = m1.store;
	   for (i=0; i<nr; i++)
	    { for (j=0; j<nc; j++)
	       { *dpr++ = -*dp1++;
	       }
	      dpr += (width-nc);
	      dp1 += (m1.width-m1.nc);
	    }
	 }
	return (*this);
}

RxMatd& RxMatd::setZero()
{
	double *dpr;
	int i, j;
	
	dpr = store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dpr++ = 0;
	    }
	   dpr += (width-nc);
	 }
	return (*this);
}

RxMatd& RxMatd::setIdent()
{
	double *dpr;
	int i, j;
	
	dpr = store;
	for (i=0; i<nr; i++)
	 { for (j=0; j<nc; j++)
	    { *dpr++ = (i==j) ? 1 : 0;
	    }
	   dpr += (width-nc);
	 }
	return (*this);
}

double RxMatd::det() const
{
	RxMatdLu lu;

	if (nr != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::det: %K\n").raise();
	   return (0);
	 }
	lu.calc (*this);
	return (lu.det());
}

double RxMatd::norm1() const
{
	double *mp;
	double max = 0;
	double sum;
	int i, j;

	for (j=0; j<nc; j++)
	 { mp = store + j;
	   sum = 0;
	   for (i=0; i<nr; i++)
	    { sum += ABS(*mp);
	      mp += width;
	    }
	   if (sum > max)
	    { max = sum;
	    }
	 }
	return (max);	
}

double RxMatd::normInf() const
{
	double *mp;
	double max = 0;
	double sum;
	int i, j;

	for (i=0; i<nr; i++)
	 { mp = store + i*width;
	   sum = 0;
	   for (j=0; j<nc; j++)
	    { sum += ABS(*mp);
	      mp++;
	    }
	   if (sum > max)
	    { max = sum;
	    }
	 }
	return (max);	
}

double RxMatd::normF() const
{
	double *mp;
	double sum = 0;
	int i, j;

	for (i=0; i<nr; i++)
	 { mp = store + i*width;
	   for (j=0; j<nc; j++)
	    { sum += (*mp)*(*mp);
	      mp++;
	    }
	 }
	return (sqrt(sum));
}

double RxMatd::trace() const
{
	double sum = 0;
	int i;

	for (i=0; i<MIN(nr,nc); i++)
	 { sum += store[i*width+i];
	 }
	return (sum);
}

double RxMatd::condest() const
{
	RxMatdLu lu;

	if (nr != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd::condest: %K\n").raise();
	   return -1;
	 }
	lu.calc (*this);
	return (lu.condest(*this));
}

RxMatd operator+(const RxMatd &m1, const RxMatd &m2)
{
	RxMatd mr(m1.nr, m1.nc);
	return (mr.add (m1, m2));
}

RxMatd operator-(const RxMatd &m1, const RxMatd &m2)
{
	RxMatd mr(m1.nr, m1.nc);
	return (mr.sub (m1, m2));
}

RxMatd operator*(const RxMatd &m1, const RxMatd &m2)
{
	RxMatd mr(m1.nr, m2.nc);
	return (mr.mul (m1, m2));
}

RxMatd operator*(const RxMatd &m1, double a)
{
	RxMatd mr(m1.nr, m1.nc);
	return (mr.scale (a, m1));
}

RxMatd operator*(double a, const RxMatd &m1)
{
	RxMatd mr(m1.nr, m1.nc);
	return (mr.scale (a, m1));
}

RxMatd operator/(const RxMatd &m1, double a)
{
	RxMatd mr(m1.nr, m1.nc);
	return (mr.scale (1/a, m1));
}
