#include <RxMatd3x3.h>
#include <RxMathError.h>
#include <RxMatdLu.h>
#include <stdlib.h>
#include <cdefs.h>
#include <math.h>

const RxMatd3x3 RxMatd3x3::UNIT (1, 0, 0, 0, 1, 0, 0, 0, 1);
const RxMatd3x3 RxMatd3x3::ZERO (0, 0, 0, 0, 0, 0, 0, 0, 0);

const int RxMatd3x3::width = 3;
const int RxMatd3x3::nr = 3;
const int RxMatd3x3::nc = 3;

#define E00	store[0]
#define E01	store[1]
#define E02	store[2]
#define E10	store[3]
#define E11	store[4]
#define E12	store[5]
#define E20	store[6]
#define E21	store[7]
#define E22	store[8]

void RxMatd3x3::doCopy (const RxMatd3x3 &m1)
{
	E00 = m1.E00;
	E01 = m1.E01;
	E02 = m1.E02;

	E10 = m1.E10;
	E11 = m1.E11;
	E12 = m1.E12;

	E20 = m1.E20;
	E21 = m1.E21;
	E22 = m1.E22;
}

RxMatd3x3& RxMatd3x3::set (const double *dp)
{
	E00 = dp[0];
	E01 = dp[1];
	E02 = dp[2];

	E10 = dp[3];
	E11 = dp[4];
	E12 = dp[5];

	E20 = dp[6];
	E21 = dp[7];
	E22 = dp[8];
	return (*this);
}

RxMatd3x3& RxMatd3x3::set (double e00, double e01, double e02, 
			 double e10, double e11, double e12,
			 double e20, double e21, double e22)
{
	E00 = e00;
	E01 = e01;
	E02 = e02;

	E10 = e10;
	E11 = e11;
	E12 = e12;

	E20 = e20;
	E21 = e21;
	E22 = e22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::set (const RxMatdBase &m1)
{
	if (nr != m1.nrows() || nc != m1.ncols())
	 { RxMathError (RxMathError::BadSize,
			"RxMatd3x3::operator=: %K\n").raise();
	   return (*this);
	 }
	E00 = m1(0,0);
	E01 = m1(0,1);
	E02 = m1(0,2);

	E10 = m1(1,0);
	E11 = m1(1,1);
	E12 = m1(1,2);

	E20 = m1(2,0);
	E21 = m1(2,1);
	E22 = m1(2,2);
	return (*this);
}

RxMatd3x3& RxMatd3x3::setDiag (const RxVecd3 &v1)
{
	E00 = v1.vbuf[0];
	E01 = 0;
	E02 = 0;

	E10 = 0;
	E11 = v1.vbuf[1];
	E12 = 0;

	E20 = 0;
	E21 = 0;
	E22 = v1.vbuf[2];
	return (*this);
}

RxMatd3x3& RxMatd3x3::setDiag (double *buf)
{
	E00 = buf[0];
	E01 = 0;
	E02 = 0;

	E10 = 0;
	E11 = buf[1];
	E12 = 0;

	E20 = 0;
	E21 = 0;
	E22 = buf[2];
	return (*this);
}

RxMatd3x3& RxMatd3x3::setSymm (const RxVecd3 &d, const RxVecd3 &o)
{
	E00 = d.vbuf[0];
	E01 = o.vbuf[2];
	E02 = o.vbuf[1];

	E10 = o.vbuf[2];
	E11 = d.vbuf[1];
	E12 = o.vbuf[0];

	E20 = o.vbuf[1];
	E21 = o.vbuf[0];
	E22 = d.vbuf[2];
	return (*this);
}

RxMatd3x3& RxMatd3x3::setSymm (double *buf)
{
	E00 = buf[0];
	E01 = buf[5];
	E02 = buf[4];

	E10 = buf[5];
	E11 = buf[1];
	E12 = buf[3];

	E20 = buf[4];
	E21 = buf[3];
	E22 = buf[2];
	return (*this);
}

void RxMatd3x3::get (double *dp) const
{
	dp[0] = E00;
	dp[1] = E01;
	dp[2] = E02;

	dp[3] = E10;
	dp[4] = E11;
	dp[5] = E12;

	dp[6] = E20;
	dp[7] = E21;
	dp[8] = E22;
}

void RxMatd3x3::getDiag (RxVecd3 *d) const
{
	d->x() = E00;
	d->y() = E11;
	d->z() = E22;
}

void RxMatd3x3::getDiag (double *buf) const
{
	buf[0] = E00;
	buf[1] = E11;
	buf[2] = E22;
}

void RxMatd3x3::getSymm (RxVecd3 *d, RxVecd3 *o) const
{
	d->x() = E00;
	d->y() = E11;
	d->z() = E22;
	o->x() = E12;
	o->y() = E02;
	o->z() = E01;
}

void RxMatd3x3::getSymm (double *buf) const
{
	buf[0] = E00;
	buf[1] = E11;
	buf[2] = E22;
	buf[3] = E12;
	buf[4] = E02;
	buf[5] = E01;
}

RxMatd3x3& RxMatd3x3::add(const RxMatd3x3& m1, const RxMatd3x3& m2)
{
	E00 = m1.E00 + m2.E00;
	E01 = m1.E01 + m2.E01;
	E02 = m1.E02 + m2.E02;

	E10 = m1.E10 + m2.E10;
	E11 = m1.E11 + m2.E11;
	E12 = m1.E12 + m2.E12;

	E20 = m1.E20 + m2.E20;
	E21 = m1.E21 + m2.E21;
	E22 = m1.E22 + m2.E22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::add(const RxMatd3x3& m2)
{
	E00 += m2.E00;
	E01 += m2.E01;
	E02 += m2.E02;

	E10 += m2.E10;
	E11 += m2.E11;
	E12 += m2.E12;

	E20 += m2.E20;
	E21 += m2.E21;
	E22 += m2.E22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::sub(const RxMatd3x3& m1, const RxMatd3x3& m2)
{
	E00 = m1.E00 - m2.E00;
	E01 = m1.E01 - m2.E01;
	E02 = m1.E02 - m2.E02;

	E10 = m1.E10 - m2.E10;
	E11 = m1.E11 - m2.E11;
	E12 = m1.E12 - m2.E12;

	E20 = m1.E20 - m2.E20;
	E21 = m1.E21 - m2.E21;
	E22 = m1.E22 - m2.E22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::sub(const RxMatd3x3& m2)
{
	E00 -= m2.E00;
	E01 -= m2.E01;
	E02 -= m2.E02;

	E10 -= m2.E10;
	E11 -= m2.E11;
	E12 -= m2.E12;

	E20 -= m2.E20;
	E21 -= m2.E21;
	E22 -= m2.E22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::mul(const RxMatd3x3& m1, const RxMatd3x3& m2)
{
	double buf[9];

	buf[0] = m1.E00*m2.E00 + m1.E01*m2.E10 + m1.E02*m2.E20;
	buf[1] = m1.E00*m2.E01 + m1.E01*m2.E11 + m1.E02*m2.E21;
	buf[2] = m1.E00*m2.E02 + m1.E01*m2.E12 + m1.E02*m2.E22;

	buf[3] = m1.E10*m2.E00 + m1.E11*m2.E10 + m1.E12*m2.E20;
	buf[4] = m1.E10*m2.E01 + m1.E11*m2.E11 + m1.E12*m2.E21;
	buf[5] = m1.E10*m2.E02 + m1.E11*m2.E12 + m1.E12*m2.E22;

	buf[6] = m1.E20*m2.E00 + m1.E21*m2.E10 + m1.E22*m2.E20;
	buf[7] = m1.E20*m2.E01 + m1.E21*m2.E11 + m1.E22*m2.E21;
	buf[8] = m1.E20*m2.E02 + m1.E21*m2.E12 + m1.E22*m2.E22;

	E00 = buf[0];
	E01 = buf[1];
	E02 = buf[2];
	E10 = buf[3];
	E11 = buf[4];
	E12 = buf[5];
	E20 = buf[6];
	E21 = buf[7];
	E22 = buf[8];
	return (*this);
}

RxMatd3x3& RxMatd3x3::Tmul(const RxMatd3x3& m1, const RxMatd3x3& m2)
{
	double buf[9];

	buf[0] = m1.E00*m2.E00 + m1.E10*m2.E10 + m1.E20*m2.E20;
	buf[1] = m1.E00*m2.E01 + m1.E10*m2.E11 + m1.E20*m2.E21;
	buf[2] = m1.E00*m2.E02 + m1.E10*m2.E12 + m1.E20*m2.E22;

	buf[3] = m1.E01*m2.E00 + m1.E11*m2.E10 + m1.E21*m2.E20;
	buf[4] = m1.E01*m2.E01 + m1.E11*m2.E11 + m1.E21*m2.E21;
	buf[5] = m1.E01*m2.E02 + m1.E11*m2.E12 + m1.E21*m2.E22;

	buf[6] = m1.E02*m2.E00 + m1.E12*m2.E10 + m1.E22*m2.E20;
	buf[7] = m1.E02*m2.E01 + m1.E12*m2.E11 + m1.E22*m2.E21;
	buf[8] = m1.E02*m2.E02 + m1.E12*m2.E12 + m1.E22*m2.E22;

	E00 = buf[0];
	E01 = buf[1];
	E02 = buf[2];
	E10 = buf[3];
	E11 = buf[4];
	E12 = buf[5];
	E20 = buf[6];
	E21 = buf[7];
	E22 = buf[8];
	return (*this);
}

RxMatd3x3& RxMatd3x3::mulT(const RxMatd3x3& m1, const RxMatd3x3& m2)
{
	double buf[9];

	buf[0] = m1.E00*m2.E00 + m1.E01*m2.E01 + m1.E02*m2.E02;
	buf[1] = m1.E00*m2.E10 + m1.E01*m2.E11 + m1.E02*m2.E12;
	buf[2] = m1.E00*m2.E20 + m1.E01*m2.E21 + m1.E02*m2.E22;

	buf[3] = m1.E10*m2.E00 + m1.E11*m2.E01 + m1.E12*m2.E02;
	buf[4] = m1.E10*m2.E10 + m1.E11*m2.E11 + m1.E12*m2.E12;
	buf[5] = m1.E10*m2.E20 + m1.E11*m2.E21 + m1.E12*m2.E22;

	buf[6] = m1.E20*m2.E00 + m1.E21*m2.E01 + m1.E22*m2.E02;
	buf[7] = m1.E20*m2.E10 + m1.E21*m2.E11 + m1.E22*m2.E12;
	buf[8] = m1.E20*m2.E20 + m1.E21*m2.E21 + m1.E22*m2.E22;

	E00 = buf[0];
	E01 = buf[1];
	E02 = buf[2];
	E10 = buf[3];
	E11 = buf[4];
	E12 = buf[5];
	E20 = buf[6];
	E21 = buf[7];
	E22 = buf[8];
	return (*this);
}

RxMatd3x3& RxMatd3x3::scale(double a, const RxMatd3x3& m1)
{
	E00 = a*m1.E00;
	E01 = a*m1.E01;
	E02 = a*m1.E02;

	E10 = a*m1.E10;
	E11 = a*m1.E11;
	E12 = a*m1.E12;

	E20 = a*m1.E20;
	E21 = a*m1.E21;
	E22 = a*m1.E22;
	return (*this);
}

RxMatd3x3 &RxMatd3x3::scale(double a)
{
	E00 *= a;
	E01 *= a;
	E02 *= a;

	E10 *= a;
	E11 *= a;
	E12 *= a;

	E20 *= a;
	E21 *= a;
	E22 *= a;
	return (*this);
}

RxMatd3x3& RxMatd3x3::xform (const RxMatd3x3 &m1, const RxRot& R)
{
	double xx, xy, xz;
	double yx, yy, yz;
	double zx, zy, zz;   

	xx = m1.E00*R.n.x() + m1.E01*R.n.y() + m1.E02*R.n.z();
	xy = m1.E00*R.o.x() + m1.E01*R.o.y() + m1.E02*R.o.z();
	xz = m1.E00*R.a.x() + m1.E01*R.a.y() + m1.E02*R.a.z();

	yx = m1.E10*R.n.x() + m1.E11*R.n.y() + m1.E12*R.n.z();
	yy = m1.E10*R.o.x() + m1.E11*R.o.y() + m1.E12*R.o.z();
	yz = m1.E10*R.a.x() + m1.E11*R.a.y() + m1.E12*R.a.z();

	zx = m1.E20*R.n.x() + m1.E21*R.n.y() + m1.E22*R.n.z();
	zy = m1.E20*R.o.x() + m1.E21*R.o.y() + m1.E22*R.o.z();
	zz = m1.E20*R.a.x() + m1.E21*R.a.y() + m1.E22*R.a.z();
    
	E00 = xx*R.n.x() + yx*R.n.y() + zx*R.n.z();
	E01 = xy*R.n.x() + yy*R.n.y() + zy*R.n.z();
	E02 = xz*R.n.x() + yz*R.n.y() + zz*R.n.z();

	E10 = xx*R.o.x() + yx*R.o.y() + zx*R.o.z();
	E11 = xy*R.o.x() + yy*R.o.y() + zy*R.o.z();
	E12 = xz*R.o.x() + yz*R.o.y() + zz*R.o.z();

	E20 = xx*R.a.x() + yx*R.a.y() + zx*R.a.z();
	E21 = xy*R.a.x() + yy*R.a.y() + zy*R.a.z();
	E22 = xz*R.a.x() + yz*R.a.y() + zz*R.a.z();

	return *this;
}

RxMatd3x3& RxMatd3x3::Ixform (const RxMatd3x3 &m1, const RxRot& R)
{
	double xx, xy, xz;
	double yx, yy, yz;
	double zx, zy, zz;

	xx = m1.E00*R.n.x() + m1.E01*R.o.x() + m1.E02*R.a.x();
	xy = m1.E00*R.n.y() + m1.E01*R.o.y() + m1.E02*R.a.y();
	xz = m1.E00*R.n.z() + m1.E01*R.o.z() + m1.E02*R.a.z();

	yx = m1.E10*R.n.x() + m1.E11*R.o.x() + m1.E12*R.a.x();
	yy = m1.E10*R.n.y() + m1.E11*R.o.y() + m1.E12*R.a.y();
	yz = m1.E10*R.n.z() + m1.E11*R.o.z() + m1.E12*R.a.z();

	zx = m1.E20*R.n.x() + m1.E21*R.o.x() + m1.E22*R.a.x();
	zy = m1.E20*R.n.y() + m1.E21*R.o.y() + m1.E22*R.a.y();
	zz = m1.E20*R.n.z() + m1.E21*R.o.z() + m1.E22*R.a.z();

	E00 = xx*R.n.x() + yx*R.o.x() + zx*R.a.x();
	E01 = xy*R.n.x() + yy*R.o.x() + zy*R.a.x();
	E02 = xz*R.n.x() + yz*R.o.x() + zz*R.a.x();

	E10 = xx*R.n.y() + yx*R.o.y() + zx*R.a.y();
	E11 = xy*R.n.y() + yy*R.o.y() + zy*R.a.y();
	E12 = xz*R.n.y() + yz*R.o.y() + zz*R.a.y();

	E20 = xx*R.n.z() + yx*R.o.z() + zx*R.a.z();
	E21 = xy*R.n.z() + yy*R.o.z() + zy*R.a.z();
	E22 = xz*R.n.z() + yz*R.o.z() + zz*R.a.z();

	return *this;
}

RxMatd3x3& RxMatd3x3::cross (const RxVecd3 &p, const RxMatd3x3 &m1)
{
	double xx, xy, xz;
	double yx, yy, yz;
	double zx, zy, zz;   

	xx = m1.E00;
	xy = m1.E01;
	xz = m1.E02;

	yx = m1.E10;
	yy = m1.E11;
	yz = m1.E12;

	zx = m1.E20;
	zy = m1.E21;
	zz = m1.E22;

	E00 = p.y()*zx - p.z()*yx;
	E01 = p.y()*zy - p.z()*yy;
	E02 = p.y()*zz - p.z()*yz;

	E10 = p.z()*xx - p.x()*zx;
	E11 = p.z()*xy - p.x()*zy;
	E12 = p.z()*xz - p.x()*zz;

	E20 = p.x()*yx - p.y()*xx;
	E21 = p.x()*yy - p.y()*xy;
	E22 = p.x()*yz - p.y()*xz;
}

RxMatd3x3& RxMatd3x3::cross (const RxMatd3x3 &m1, const RxVecd3 &p)
{
	double xx, xy, xz;
	double yx, yy, yz;
	double zx, zy, zz;   

	xx = m1.E00;
	xy = m1.E01;
	xz = m1.E02;

	yx = m1.E10;
	yy = m1.E11;
	yz = m1.E12;

	zx = m1.E20;
	zy = m1.E21;
	zz = m1.E22;

	E00 = p.z()*xy - p.y()*xz;
	E01 = p.x()*xz - p.z()*xx;
	E02 = p.y()*xx - p.x()*xy;

	E10 = p.z()*yy - p.y()*yz;
	E11 = p.x()*yz - p.z()*yx;
	E12 = p.y()*yx - p.x()*yy;

	E20 = p.z()*zy - p.y()*zz;
	E21 = p.x()*zz - p.z()*zx;
	E22 = p.y()*zx - p.x()*zy;
}

#define MAXT	16

RxVecd3 &RxMatd3x3::Vmul (RxVecd3 *vr, const RxVecd3 &v1) const
{
	double buf[3];

	buf[0] = v1.vbuf[0]*E00 + v1.vbuf[1]*E10 + v1.vbuf[2]*E20;
	buf[1] = v1.vbuf[0]*E01 + v1.vbuf[1]*E11 + v1.vbuf[2]*E21;
	buf[2] = v1.vbuf[0]*E02 + v1.vbuf[1]*E12 + v1.vbuf[2]*E22;

	vr->vbuf[0] = buf[0];
	vr->vbuf[1] = buf[1];
	vr->vbuf[2] = buf[2];
	return (*vr);
}

RxVecd3 &RxMatd3x3::mulV (RxVecd3 *vr, const RxVecd3 &v1) const
{
	double buf[3];

	buf[0] = E00*v1.vbuf[0] + E01*v1.vbuf[1] + E02*v1.vbuf[2];
	buf[1] = E10*v1.vbuf[0] + E11*v1.vbuf[1] + E12*v1.vbuf[2];
	buf[2] = E20*v1.vbuf[0] + E21*v1.vbuf[1] + E22*v1.vbuf[2];

	vr->vbuf[0] = buf[0];
	vr->vbuf[1] = buf[1];
	vr->vbuf[2] = buf[2];
	return (*vr);
}

RxMatd3x3& RxMatd3x3::tsp(const RxMatd3x3& m1)
{
	double tmp;

	E00 = m1.E00;
	tmp = m1.E01;
	E01 = m1.E10;
	E10 = tmp;

	E11 = m1.E11;
	tmp = m1.E02;
	E02 = m1.E20;
	E20 = tmp;

	E22 = m1.E22;
	tmp = m1.E12;
	E12 = m1.E21;
	E21 = tmp;
	return (*this);
}

RxMatd3x3& RxMatd3x3::tsp()
{
	double tmp;

	tmp = E01;
	E01 = E10;
	E10 = tmp;

	tmp = E02;
	E02 = E20;
	E20 = tmp;

	tmp = E12;
	E12 = E21;
	E21 = tmp;
	return (*this);
}

RxMatd3x3& RxMatd3x3::inv(const RxMatd3x3& m1)
{
	double c00, c01, c02;
	double c10, c11, c12;
	double c20, c21, c22;
	double det;

	c00 = m1.E11*m1.E22 - m1.E21*m1.E12;
	c01 = m1.E12*m1.E20 - m1.E10*m1.E22;
	c02 = m1.E10*m1.E21 - m1.E20*m1.E11;

	c10 = m1.E21*m1.E02 - m1.E01*m1.E22;
	c11 = m1.E00*m1.E22 - m1.E20*m1.E02;
	c12 = m1.E01*m1.E20 - m1.E00*m1.E21;

	c20 = m1.E01*m1.E12 - m1.E11*m1.E02;
	c21 = m1.E02*m1.E10 - m1.E00*m1.E12;
	c22 = m1.E00*m1.E11 - m1.E10*m1.E01;

	det = m1.E00*c00 + m1.E01*c01 + m1.E02*c02;

	E00 = c00/det;
	E01 = c10/det;
	E02 = c20/det;
	
	E10 = c01/det;
	E11 = c11/det;
	E12 = c21/det;
	
	E20 = c02/det;
	E21 = c12/det;
	E22 = c22/det;

	return (*this);
}

RxMatd3x3& RxMatd3x3::neg(const RxMatd3x3& m1)
{
	E00 = -m1.E00;
	E01 = -m1.E01;
	E02 = -m1.E02;

	E10 = -m1.E10;
	E11 = -m1.E11;
	E12 = -m1.E12;

	E20 = -m1.E20;
	E21 = -m1.E21;
	E22 = -m1.E22;
	return (*this);
}

RxMatd3x3& RxMatd3x3::setZero()
{
	E00 = 0;
	E01 = 0;
	E02 = 0;

	E10 = 0;
	E11 = 0;
	E12 = 0;

	E20 = 0;
	E21 = 0;
	E22 = 0;
	return (*this);
}

RxMatd3x3& RxMatd3x3::setIdent()
{
	E00 = 1;
	E01 = 0;
	E02 = 0;

	E10 = 0;
	E11 = 1;
	E12 = 0;

	E20 = 0;
	E21 = 0;
	E22 = 1;
	return (*this);
}

double RxMatd3x3::det() const
{
	return (E00*E11*E22 + 
		E10*E21*E02 + 
		E20*E01*E12 -
		E02*E11*E20 -
		E12*E21*E00 -
		E10*E01*E22);
}

double RxMatd3x3::norm1() const
{
	double max = 0;
	double sum;

	max = ABS(E00)+ ABS(E10) + ABS(E20);
	sum = ABS(E01)+ ABS(E11) + ABS(E21);
	if (sum > max)
	 { max = sum;
	 }
	sum = ABS(E02)+ ABS(E12) + ABS(E22);
	if (sum > max)
	 { max = sum;
	 }
	return (max);	
}

double RxMatd3x3::normInf() const
{
	double max = 0;
	double sum;

	max = ABS(E00)+ ABS(E01) + ABS(E02);
	sum = ABS(E10)+ ABS(E11) + ABS(E12);
	if (sum > max)
	 { max = sum;
	 }
	sum = ABS(E20)+ ABS(E21) + ABS(E22);
	if (sum > max)
	 { max = sum;
	 }
	return (max);	
}

double RxMatd3x3::normF() const
{
	double sum = 0;
	int i;

	for (i=0; i<nr*nc; i++)
	 { sum += store[i]*store[i];
	 }
	return (sqrt(sum));
}

double RxMatd3x3::trace() const
{
	return (E00 + E11 + E22);
}

double RxMatd3x3::condest() const
{
	RxMatdLu lu;

	if (nr != nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatd3x3::condest: %K\n").raise();
	   return -1;
	 }
	lu.calc (*this);
	return (lu.condest(*this));
}

RxMatd3x3 operator+(const RxMatd3x3 &m1, const RxMatd3x3 &m2)
{
	RxMatd3x3 mr;
	return (mr.add (m1, m2));
}

RxMatd3x3 operator-(const RxMatd3x3 &m1, const RxMatd3x3 &m2)
{
	RxMatd3x3 mr;
	return (mr.sub (m1, m2));
}

RxMatd3x3 operator*(const RxMatd3x3 &m1, const RxMatd3x3 &m2)
{
	RxMatd3x3 mr;
	return (mr.mul (m1, m2));
}

RxMatd3x3 operator*(const RxMatd3x3 &m1, double a)
{
	RxMatd3x3 mr;
	return (mr.scale (a, m1));
}

RxMatd3x3 operator*(double a, const RxMatd3x3 &m1)
{
	RxMatd3x3 mr;
	return (mr.scale (a, m1));
}

RxMatd3x3 operator/(const RxMatd3x3 &m1, double a)
{
	RxMatd3x3 mr;
	return (mr.scale (1/a, m1));
}
