#include <RxMatdBaseTest.h>
#include <cdefs.h>
#include <math.h>

int testAdd (const RxMatdBase &mr, const RxMatdBase &m1, const RxMatdBase &m2,
	 double eps)
{
	double accum;
	int i, j;

	if (mr.nrows() != m1.nrows() || mr.ncols() != m1.ncols() ||
	    mr.nrows() != m2.nrows() || mr.ncols() != m2.ncols())
	 { return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { accum = m1(i,j) + m2(i,j);
	      if (ABS(accum - mr(i,j)) > eps)
	       { return (0);
	       }
	    }
	 }
	return (1);
}

int testSub (const RxMatdBase &mr, const RxMatdBase &m1, const RxMatdBase &m2,
	 double eps)
{
	double accum;
	int i, j;

	if (mr.nrows() != m1.nrows() || mr.ncols() != m1.ncols() ||
	    mr.nrows() != m2.nrows() || mr.ncols() != m2.ncols())
	 { return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { accum = m1(i,j) - m2(i,j);
	      if (ABS(accum - mr(i,j)) > eps)
	       { return (0);
	       }
	    }
	 }
	return (1);
}

int testScale (const RxMatdBase &mr, double a, const RxMatdBase &m1, double eps)
{
	double accum;
	int i, j;

	if (mr.nrows() != m1.nrows() || mr.ncols() != m1.ncols())
	 { return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { accum = a*m1(i,j);
	      if (ABS(accum - mr(i,j)) > eps)
	       { return (0);
	       }
	    }
	 }
	return (1);
}

int testMul (const RxMatdBase &mr, const RxMatdBase &m1, const RxMatdBase &m2,
	 double eps)
{
	int i, j, k;
	double accum;

	if (mr.nrows() != m1.nrows() || mr.ncols() != m2.ncols() ||
	    m1.ncols() != m2.nrows())
	 { printf ("testMul: mr is %dX%d, m1 is %dX%d, m2 is %dX%d\n",
		    mr.nrows(), mr.ncols(), m1.nrows(), m1.ncols(),
		    m2.nrows(), m2.ncols());
	   return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { accum = 0;
	      for (k=0; k<m1.ncols(); k++)
	       { accum += m1(i,k)*m2(k,j);
	       }
	      if (ABS(mr(i,j) - accum) > eps)
	       { printf ("testMul:\n");
		 mr.print ("mr = [%12.7f]\n");
		 m1.print ("m1 = [%12.7f]\n");
		 m2.print ("m2 = [%12.7f]\n");
		 return (0);
	       }
	    }
	 }
	return (1);
}

int testTsp (const RxMatdBase &mr, const RxMatdBase &m1, double eps)
{
	int i, j;

	if (mr.nrows() != m1.ncols() || mr.ncols() != m1.nrows())
	 { printf ("foo\n");
	   return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { if (ABS(mr(i,j) - m1(j,i)) > eps)
	       { printf ("here %d %d\n", i, j);
		 return (0);
	       }
	    }
	 }
	return (1);
}

int testInv (const RxMatdBase &mr, const RxMatdBase &m1, double eps)
{
	int i, j, k;
	double accum;

	if (mr.nrows() != m1.nrows() || mr.ncols() != m1.ncols() ||
	    mr.nrows() != mr.ncols())
	 { return (0);
	 }
	for (i=0; i<mr.nrows(); i++)
	 { for (j=0; j<mr.ncols(); j++)
	    { accum = 0;
	      for (k=0; k<mr.ncols(); k++)
	       { accum += mr(i,k)*m1(k,j);
	       }
	      if (i == j)
	       { accum -= 1;
	       }
	      if (ABS(accum) > eps)
	       { return (0);
	       }
	    }
	 }
	return (1);
}
