#include <RxMatd.h>
#include <RxMatdBaseTest.h>

#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-10)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	12

double testVals[][LEN] = {
	{ 1.0, 0.0, 0.0,   0.0, 1.0, 0.0,   0.0, 0.0, 1.0,  0.0, 0.0, 0.0},
	{ 1.0, 2.0, 3.0,   4.0, 5.0, 6.0,   9.0, 8.0, 1.0,  3.0, 2.0, 1.0},
	{ 0.0, 0.0, 1.0,   1.0, 2.0, 3.0,   5.0, 6.6, 7.0,  1.0, 7.0,-9.0 },
	{ 1.0, 1.0, 1.0,  -1.0,-2.0,-3.0,  -5.0, 6.6,-7.0, -8.0,-4.0, 8.0 },
	{ 3.0,-2.0, 1.0,   0.0, 1.0, 0.0,  10.0,-9.0,-5.0,  2.0, 2.0, 4.0 }
};

main()
{
	double vec[100];
	RxMatd m1(4,3);
	RxMatd m2(testVals[0], 4,3);
	RxMatd m3(testVals[1], 4,3);
	int i, j;
	int p, q;

	assert (m1.nrows() == 4);
	assert (m1.ncols() == 3);
	assert (m2.nrows() == 4);
	assert (m2.ncols() == 3);

	for (i=0; i<m3.nrows(); i++)
	 { for (j=0; j<m3.ncols(); j++)
	    { assert (EQ(testVals[1][i*m3.ncols()+j],m3(i,j)));
	      m3(i,j) = (i+1)*100+(j+1);
	      assert (EQ(m3(i,j),(double)((i+1)*100+(j+1))));
	    }
	 }

	RxMatd matA(m3);

	for (i=0; i<m3.nrows(); i++)
	 { for (j=0; j<m3.ncols(); j++)
	    { assert (EQ(m3(i,j),matA(i,j)));
	    }
	 }

	m1.setZero();
	for (i=0; i<m1.nrows(); i++)
	 { for (j=0; j<m1.ncols(); j++)
	    { assert (EQ(m1(i,j),0.0));
	    }
	 }
	assert (!m1.equal(m3,EPS));
	m1 = m3;
	assert (m1.equal(m3,EPS));

	m1.set(testVals[2]);
	for (i=0; i<m1.nrows(); i++)
	 { for (j=0; j<m1.ncols(); j++)
	    { assert (EQ(m1(i,j),testVals[2][i*m1.ncols()+j]));
	    }
	 }
	m1.get (vec);
	for (i=0; i<m1.nrows()*m1.ncols(); i++)
	 { assert (EQ(vec[i],testVals[2][i]));
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { m1.set(testVals[p]);

	   assert (testScale (m3.scale(M_PI,m1), M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3.scale(M_PI), M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3 *= M_PI, M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3 /= M_PI, 1/M_PI, m1, EPS));
	   assert (testScale (M_PI*m1, M_PI, m1, EPS));
	   assert (testScale (m1*M_PI, M_PI, m1, EPS));
	   assert (testScale (m1/M_PI, 1/M_PI, m1, EPS));

	   assert (testTsp (m3.tsp(m1), m1, EPS));
	   m3 = m1;
	   assert (testTsp (m3.tsp(), m1, EPS));
	   assert (testTsp (m1.T(), m1, EPS));

	   RxMatd mc(3,3);
	   RxMatd m1x(testVals[p],3,3);

	   assert (testTsp (mc.tsp(m1x), m1x, EPS));
	   mc = m1x;
	   assert (testTsp (mc.tsp(), m1x, EPS));
	   assert (testTsp (m1x.T(), m1x, EPS));

	   m3 = m1;
	   assert (testScale (m3.neg(m1), -1.0, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3.neg(), -1.0, m1, EPS));
	   assert (testScale (-m1, -1.0, m1, EPS));
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { for (q=0; q<ARRAY_SIZE(testVals); q++)
	    { m1.set(testVals[p]);
	      m2.set(testVals[q]);

	      assert (testAdd (m1+m2, m1, m2, EPS));
	      assert (testAdd (m3.add(m1,m2), m1, m2, EPS));
	      m3 = m1;
	      assert (testAdd (m3.add(m2), m1, m2, EPS));
	      m3 = m1;					    
	      assert (testAdd (m3 += m2, m1, m2, EPS));

	      assert (testSub (m1-m2, m1, m2, EPS));
	      assert (testSub (m3.sub(m1,m2), m1, m2, EPS));
	      m3 = m1;
	      assert (testSub (m3.sub(m2), m1, m2, EPS));
	      m3 = m1;					    
	      assert (testSub (m3 -= m2, m1, m2, EPS));

	      RxMatd mc(3,3);
	      RxMatd m2x(testVals[q],3,3);

	      assert (testMul (m1.T()*m2, m1.T(), m2, EPS));
	      assert (testMul (mc.mul(m1.T(),m2), m1.T(), m2, EPS));
	      m3 = m1;
	      assert (testMul (m3.mul(m2x), m1, m2x, EPS));
	      m3 = m1;
	      assert (testMul (m3 *= m2x, m1, m2x, EPS));
	    }
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { RxMatd ma(testVals[p],3,3);
	   RxMatd mc(3,3);

	   assert (testInv (mc.inv(ma), ma, EPS));
	   mc = ma;
	   assert (testInv (mc.inv(), ma, EPS));
	   assert (testInv (ma.I(), ma, EPS));
	 }

	RxMatd mc(5,3);

	mc.setIdent();
	for (i=0; i<mc.nrows(); i++)
	 { for (j=0; j<mc.ncols(); j++)
	    { assert (EQ(mc(i,j), (i==j) ? 1.0 : 0.0));
	    }
	 }

#define LENV	4

	// test new vector based routines

	RxVecd vtest(LENV);
	for (i=0; i<LENV; i++)
	 { vtest(i) = i;
	 }

	 { RxMatd MV(vtest);
	   assert (MV.nrows() == LENV && MV.ncols() == 1);
	   for (i=0; i<LENV; i++)
	    { assert (EQ(MV(i,0),vtest(i)));
	    }
	 }
	 { RxMatd MV(vtest, /*column=*/0);
	   assert (MV.nrows() == 1 && MV.ncols() == LENV);
	   for (i=0; i<LENV; i++)
	    { assert (EQ(MV(0,i),vtest(i)));
	    }
	 }
	 { RxMatd MV(LENV,LENV+2);

	   MV.setDiag (vtest);
	   for (i=0; i<LENV; i++)
	    { for (j=0; j<LENV+2; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	   MV.setDiag (vtest.buf());
	   for (i=0; i<LENV; i++)
	    { for (j=0; j<LENV+2; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	 }
	 { RxMatd MV(LENV+2,LENV);

	   MV.setDiag (vtest);
	   for (i=0; i<LENV+2; i++)
	    { for (j=0; j<LENV; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	   MV.setDiag (vtest.buf());
	   for (i=0; i<LENV+2; i++)
	    { for (j=0; j<LENV; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	 }

	// test vector multiply

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { for (q=0; q<ARRAY_SIZE(testVals); q++)
	    { m1.set(testVals[p]);
	      m2.set(testVals[q]);

	      for (i=0; i<m2.nrows(); i++)
	       { RxMatd rowM (m2, i, 0, 1, m2.ncols());	// ith row of m2
		 RxVecd rowV (m2.ncols());
		 RxVecd resV (m1.nrows());
		 RxMatd resM (m1.nrows(), 1);

		 rowM.get (&rowV);
		 for (j=0; j<m2.ncols(); j++)
		  { assert (EQ(rowV(j),m2(i,j)));
		  }
		 m1.mulV (&resV, rowV);
		 assert (testMul (resM.set(resV), m1, rowM.T(), EPS));
		 resV = m1*rowV;
		 assert (testMul (resM.set(resV), m1, rowM.T(), EPS));
	       }
	      for (j=0; j<m2.ncols(); j++)
	       { RxMatd colM (m2, 0, j, m2.nrows(), 1);	// ith col of m2
		 RxVecd colV (m2.nrows());
		 RxVecd resV (m1.ncols());
		 RxMatd resM (1, m1.ncols());

		 colM.get (&colV);
		 for (i=0; i<m2.nrows(); i++)
		  { assert (EQ(colV(i),m2(i,j)));
		  }
		 m1.Vmul (&resV, colV);
		 assert (testMul (resM.set(resV), colM.T(), m1, EPS));
		 resV = colV*m1;
		 assert (testMul (resM.set(resV), colM.T(), m1, EPS));
	       }
	    }
	 }
	// test out the I/O routines
	 { char buf[256];
	   int status;
	   double vals[] = { 11.10, 22.20, 33.30, 55.50, 66.60, 77.70 };

	   RxMatd mm1(vals, 3,2);
	   RxMatd mm2(3,2);

	   status = mm1.print (buf, "foo %4.1f bar\n");
	   assert (strcmp (buf, "\
foo [11.1 22.2\n\
     33.3 55.5\n\
     66.6 77.7] bar\n") == 0);
	   assert (status == strlen(buf));
	   status = mm2.scan (buf+3);
	   assert (status == 1);
	   assert (mm1.equal (mm2, EPS));

	   status = mm1.print (buf, "foo %4.1f,%5.2f bar\n");
	   assert (strcmp (buf, "\
foo [11.1,22.20\n\
     33.3,55.50\n\
     66.6,77.70] bar\n") == 0);
	   assert (status == strlen(buf));

	   status = mm1.print (buf, "foo %4.1f,%5.2f bar\n", 12);
	   assert (strcmp (buf, "foo [11.1,2") == 0);
	   assert (status == -1);

	   status = mm1.print (buf, "foo xxxx %5.2f,%5.2f bar\n");
	   assert (strcmp (buf, "\
foo xxxx [11.10,22.20\n\
	  33.30,55.50\n\
	  66.60,77.70] bar\n") == 0);
	   assert (status == strlen(buf));


	 }


	printf ("\nROUTINES OK\n\n");
	return 0;
}
