#include <RxVecd2.h>
#include <math.h>
#include <cdefs.h>

const RxVecd2 RxVecd2::ZERO(0,0);
const RxVecd2 RxVecd2::I(1,0);
const RxVecd2 RxVecd2::J(0,1);
const RxVecd2 RxVecd2::Ineg(-1,0);
const RxVecd2 RxVecd2::Jneg(0,-1);

#if 0
RxVecd2& RxVecd2::set (const double *dp)
{
	v[0] = dp[0];
	v[1] = dp[1];
	return (*this);
}

void RxVecd2::get (double *dp)
{
	dp[0] = v[0];
	dp[1] = v[1];
}
#endif

RxVecd2& RxVecd2::add(const RxVecd2 &v1, const RxVecd2 &v2)
{
	v[0] = v1.vbuf[0] + v2.vbuf[0];
	v[1] = v1.vbuf[1] + v2.vbuf[1];
	return (*this);
}

RxVecd2& RxVecd2::add(const RxVecd2 &v2)
{
	v[0] += v2.vbuf[0];
	v[1] += v2.vbuf[1];
	return (*this);
}

RxVecd2& RxVecd2::sub(const RxVecd2 &v1, const RxVecd2 &v2)
{
	v[0] = v1.vbuf[0] - v2.vbuf[0];
	v[1] = v1.vbuf[1] - v2.vbuf[1];
	return (*this);
}

RxVecd2& RxVecd2::sub(const RxVecd2 &v2)
{
	v[0] -= v2.vbuf[0];
	v[1] -= v2.vbuf[1];
	return (*this);
}

RxVecd2& RxVecd2::extrap (const RxVecd2 &v1, double a, const RxVecd2 &v2)
{
	x() = v1.x() + a*v2.x();
	y() = v1.y() + a*v2.y();
	return (*this);
}

RxVecd2& RxVecd2::extrap (double a, const RxVecd2 &v2)
{
	x() += a*v2.x();
	y() += a*v2.y();
	return (*this);
}

RxVecd2& RxVecd2::interp (const RxVecd2 &v1, double a, const RxVecd2 &v2)
{
	x() = (1-a)*v1.x() + a*v2.x();
	y() = (1-a)*v1.y() + a*v2.y();
	return (*this);
}

RxVecd2& RxVecd2::interp (double a, const RxVecd2 &v2)
{
	x() = (1-a)*x() + a*v2.x();
	y() = (1-a)*y() + a*v2.y();
	return (*this);
}

RxVecd2& RxVecd2::scale(double a, const RxVecd2 &v1)
{
	v[0] = v1.vbuf[0] * a;
	v[1] = v1.vbuf[1] * a;
	return (*this);
}

RxVecd2& RxVecd2::scale(double a)
{
	v[0] *= a;
	v[1] *= a;
	return (*this);
}

#if 0
double RxVecd2::dot(const RxVecd2 &v2) const
{
	return (v[0]*v2.vbuf[0] + v[1]*v2.vbuf[1]);
}
#endif

RxVecd2& RxVecd2::unit(const RxVecd2 &v1)
{
	double mag;
 
	if ((mag = sqrt(v1.vbuf[0]*v1.vbuf[0] + v1.vbuf[1]*v1.vbuf[1])) > 0)
	 { v[0] = v1.vbuf[0]/mag;
	   v[1] = v1.vbuf[1]/mag;
	 }
	else
	 { v[0] = v[1] = 0;
	 }
	return (*this);
}

#if 0
RxVecd2& RxVecd2::neg(const RxVecd2 &v1)
{
	v[0] = -v1.vbuf[0];
	v[1] = -v1.vbuf[1];
	return (*this);
}
#endif

RxVecd2& RxVecd2::setZero()
{
	v[0] = 0;
	v[1] = 0;
	return (*this);
}

double RxVecd2::norm1() const
{
	return (ABS(v[0]) + ABS(v[1]));
}

double RxVecd2::norm2() const
{
	return (sqrt(v[0]*v[0] + v[1]*v[1]));
}

double RxVecd2::normInf() const
{
	double xabs, yabs;
 
	xabs = ABS(v[0]);
	yabs = ABS(v[1]);
	if (xabs > yabs)
	 { return (xabs);
	 }
	else
	 { return (yabs);
	 }
}
