#include <RxVecd2.h>
#include <RxVecdBaseTest.h>
#include <RxRot.h>
#include <RxTrsf.h>
#include <RxTimeval.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

foo(RxVecd2 &vr, RxVecd2 &v3, RxVecd2 &v1)
{
	vr.x() = v3.x()+v1.x(); vr.y() = v3.y()+v1.y();
}

bar(RxVecd2 &v3, RxVecd2 &v1)
{
	foo(v3,v3,v1);
	v3.add(v3,v1);
}

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

testCross (double res, const RxVecdBase &v1, const RxVecdBase &v2,
	   double eps)
{
	double z;

	z = v1(0)*v2(1) - v1(1)*v2(0);
	if (ABS(res-z) > eps)
	 { return (0);
	 }
	return (1);
}

double testVals[][3] = {
 { 0.0, 0.0 },
 { 1.0, 2.0 },
 {-3.0, -2.0 }};

#if 0
double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};
#endif

main()
{
	double vec[3];
	RxVecd2 v1(10,20);
	RxVecd2 v2(40,50);
	int i, j, k;

	assert (EQ(v1.x(),10) && EQ(v1.y(),20));
	assert (EQ(v2.x(),40) && EQ(v2.y(),50));
	
	RxVecd2 v3(v2);

	assert (EQ(v3.x(),40) && EQ(v3.y(),50));

	assert (v3.length() == 2);

	assert (v3.equal(v2,EPS));
	assert (!v3.equal(v1,EPS));
	for (i=0;i<2;i++)
	 { v1(i) = i*10+1;
	 }
	assert (v1.equal(RxVecd2(1,11),EPS));
	
	v1 = v3;

	assert (v3.equal(v1,EPS));
	
	v1.set(11,22);

	assert (v1.equal(RxVecd2(11,22),EPS));
	
	v1.get(vec);
	assert (EQ(vec[0],11) && EQ(vec[1],22));
	vec[0] = 111; vec[1] = 222;
	assert ((v3.set(vec)).equal(RxVecd2(111,222),EPS));

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { v1.set(testVals[i]);
	      v2.set(testVals[j]);

	      assert (testAdd (v1+v2, v1, v2, EPS));
	      assert (testAdd (v3.add(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testAdd (v3.add(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testAdd (v3 += v2, v1, v2, EPS));
	      assert (testAdd (v3.addQ(v1,v2), v1, v2, EPS));

	      assert (testSub (v1-v2, v1, v2, EPS));
	      assert (testSub (v3.sub(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testSub (v3.sub(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testSub (v3 -= v2, v1, v2, EPS));
	      assert (testSub (v3.subQ(v1,v2), v1, v2, EPS));

	      assert (testCross (v3.cross(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testCross (v3.cross(v2), v1, v2, EPS));

	      assert (testExtrap (v3.extrap(v1,11.1,v2), v1, 11.1, v2, EPS));
	      v3 = v1;
	      assert (testExtrap (v3.extrap(11.1,v2), v1, 11.1, v2, EPS));
	      assert (testExtrap (v3.extrapQ(v1,11.1,v2), v1, 11.1, v2, EPS));

	      v3 = v1;
	      assert (testDot (v3.dot(v2), v1, v2, EPS));
	      assert (testDot (v3.dotQ(v2), v1, v2, EPS));
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);

	   assert (testScale (v3.scale(M_PI,v1), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.scale(M_PI), M_PI, v1, EPS));
	   assert (testScale (v3.scaleQ(M_PI,v1), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 *= M_PI, M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 /= M_PI, 1/M_PI, v1, EPS));
	   assert (testScale (M_PI*v1, M_PI, v1, EPS));
	   assert (testScale (v1*M_PI, M_PI, v1, EPS));
	   assert (testScale (v1/M_PI, 1/M_PI, v1, EPS));

	   assert (testUnit (v3.unit(v1), v1, EPS));
	   v3 = v1;
	   assert (testUnit (v3.unit(), v1, EPS));
	   assert (testScale (v3.neg(v1), -1.0, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.neg(), -1.0, v1, EPS));
	   assert (testScale (-v1, -1.0, v1, EPS));

	   assert (testNorm1 (v1.norm1(), v1, EPS));
	   assert (testNorm2 (v1.norm2(), v1, EPS));
	   assert (testNormInf (v1.normInf(), v1, EPS));
	 }
#if 0
	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);
 
	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxTrsf T;

	      T.R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      v3.xform (v1, T.R);
	      v3.Ixform (v3, T.R);
	      assert (v3.equal(v1,EPS));
	      v3 = v1;
	      v3.xform (T.R);
	      v3.Ixform (T.R);
	      assert (v3.equal(v1,EPS));

	      for (k=0; k<ARRAY_SIZE(testVals); k++)
	       { T.p.set (testVals[k]);

		 v3.xform (v1, T);
		 v3.Ixform (v3, T);
		 assert (v3.equal(v1,EPS));
		 v3 = v1;
		 v3.xform (T);
		 v3.Ixform (T);
		 assert (v3.equal(v1,EPS));
	       }
	    }
	 }
#endif
	v1.set(0,0);
	assert ((v2.setZero()).equal(v1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}
