#include <RxVecd5.h>
#include <math.h>
#include <cdefs.h>

RxVecd5& RxVecd5::set (const double *dp)
{
	v[0] = dp[0];
	v[1] = dp[1];
	v[2] = dp[2];
	v[3] = dp[3];
	v[4] = dp[4];
	return (*this);
}

void RxVecd5::get (double *dp)
{
	dp[0] = v[0];
	dp[1] = v[1];
	dp[2] = v[2];
	dp[3] = v[3];
	dp[4] = v[4];
}

RxVecd5& RxVecd5::add(const RxVecd5 &v1, const RxVecd5 &v2)
{
	v[0] = v1.v[0] + v2.v[0];
	v[1] = v1.v[1] + v2.v[1];
	v[2] = v1.v[2] + v2.v[2];
	v[3] = v1.v[3] + v2.v[3];
	v[4] = v1.v[4] + v2.v[4];
	return (*this);
}

RxVecd5& RxVecd5::add(const RxVecd5 &v2)
{
	v[0] += v2.v[0];
	v[1] += v2.v[1];
	v[2] += v2.v[2];
	v[3] += v2.v[3];
	v[4] += v2.v[4];
	return (*this);
}

RxVecd5& RxVecd5::sub(const RxVecd5 &v1, const RxVecd5 &v2)
{
	v[0] = v1.v[0] - v2.v[0];
	v[1] = v1.v[1] - v2.v[1];
	v[2] = v1.v[2] - v2.v[2];
	v[3] = v1.v[3] - v2.v[3];
	v[4] = v1.v[4] - v2.v[4];
	return (*this);
}

RxVecd5& RxVecd5::sub(const RxVecd5 &v2)
{
	v[0] -= v2.v[0];
	v[1] -= v2.v[1];
	v[2] -= v2.v[2];
	v[3] -= v2.v[3];
	v[4] -= v2.v[4];
	return (*this);
}

RxVecd5& RxVecd5::extrap(const RxVecd5 &v1, double a, const RxVecd5 &v2)
{
	v[0] = v1.v[0] + a*v2.v[0];
	v[1] = v1.v[1] + a*v2.v[1];
	v[2] = v1.v[2] + a*v2.v[2];
	v[3] = v1.v[3] + a*v2.v[3];
	v[4] = v1.v[4] + a*v2.v[4];
	return (*this);
}

RxVecd5& RxVecd5::scale(double a, const RxVecd5 &v1)
{
	v[0] = v1.v[0] * a;
	v[1] = v1.v[1] * a;
	v[2] = v1.v[2] * a;
	v[3] = v1.v[3] * a;
	v[4] = v1.v[4] * a;
	return (*this);
}

RxVecd5& RxVecd5::scale(double a)
{
	v[0] *= a;
	v[1] *= a;
	v[2] *= a;
	v[3] *= a;
	v[4] *= a;
	return (*this);
}

double RxVecd5::dot(const RxVecd5 &v2) const
{
	return (v[0]*v2.v[0] + 
		v[1]*v2.v[1] + 
		v[2]*v2.v[2] +
		v[3]*v2.v[3] + 
		v[4]*v2.v[4]
	       );
}

RxVecd5& RxVecd5::unit(const RxVecd5 &v1)
{
	double mag;
 
	if ((mag = v1.norm2()) > 0)
	 { v[0] = v1.v[0]/mag;
	   v[1] = v1.v[1]/mag;
	   v[2] = v1.v[2]/mag;
	   v[3] = v1.v[3]/mag;
	   v[4] = v1.v[4]/mag;
	 }
	else
	 { setZero();
	 }
	return (*this);
}

RxVecd5& RxVecd5::neg(const RxVecd5 &v1)
{
	v[0] = -v1.v[0];
	v[1] = -v1.v[1];
	v[2] = -v1.v[2];
	v[3] = -v1.v[3];
	v[4] = -v1.v[4];
	return (*this);
}

RxVecd5& RxVecd5::setZero()
{
	v[0] = 0;
	v[1] = 0;
	v[2] = 0;
	v[3] = 0;
	v[4] = 0;
	return (*this);
}

double RxVecd5::norm1() const
{
	return (ABS(v[0]) +
		ABS(v[1]) +
		ABS(v[2]) +
		ABS(v[3]) +
		ABS(v[4])
	       );
}

double RxVecd5::norm2() const
{
	return (sqrt(v[0]*v[0] +
		     v[1]*v[1] +
		     v[2]*v[2] +
		     v[3]*v[3] +
		     v[4]*v[4]
		    ));
}

double RxVecd5::normInf() const
{
	double tmp;
	double max;
 
	max = ABS(v[0]);
	if ((tmp = ABS(v[1])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[2])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[3])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[4])) > max)
	 { max = tmp;
	 }
	return (max);
}



