#ifndef _RXVECDN5_H_WRAP_
#define _RXVECDN5_H_WRAP_

#include <RxVecd.h>

class RxVecd5: public RxVecd
{
	double vbuf[5];

    public:

	// query 
	
	int length() const
	 { return 5;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 

	// assignment

	RxVecd5& operator=(const RxVecd5 &v2)
	 { set (v2.v);
	   return (*this);
	 }
	RxVecd5& set (const double *dp);
	void get (double *dp);

	// constructors	

	RxVecd5 () : RxVecd (5, vbuf) 
	 { 
	 }
	RxVecd5 (const RxVecd5 &v2) : RxVecd (5, vbuf)
	 { set (v2.v);
	 }
	RxVecd5 (double *dp) : RxVecd (5, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd5& add (const RxVecd5 &v1, const RxVecd5 &v2);
	RxVecd5& add (const RxVecd5 &v2);

	RxVecd5& operator+=(const RxVecd5 &v2)
	 { return (add (v2));
	 }
	friend RxVecd5 operator+(const RxVecd5 &v1, const RxVecd5 &v2);

	RxVecd5& sub (const RxVecd5 &v1, const RxVecd5 &v2);
	RxVecd5& sub (const RxVecd5 &v2);
	RxVecd5& operator-=(const RxVecd5 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd5 operator-(const RxVecd5 &v1, const RxVecd5 &v2);

	RxVecd5& extrap (const RxVecd5 &v1, double a, const RxVecd5 &v2);
	RxVecd5& extrap (double a, const RxVecd5 &v2)
	 { return (extrap (*this, a, v2));
	 }

	// T <- T,O operations

	RxVecd5& scale (double a, const RxVecd5 &v1);
	RxVecd5& scale (double a);
	RxVecd5& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd5& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd5 operator*(const RxVecd5 &v1, double a);
	friend RxVecd5 operator*(double a, const RxVecd5 &v1);
	friend RxVecd5 operator/(const RxVecd5 &v1, double a);

	// O <- T,T operations

	double dot(const RxVecd5 &v2) const;

	// T <- T operations

	RxVecd5 &unit(const RxVecd5 &v1);
	RxVecd5 &unit()
	 { return (unit(*this));
	 }
	RxVecd5 &neg(const RxVecd5 &v1);
	RxVecd5 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd5 operator-(const RxVecd5 &v1);

	// T <- O operations

	RxVecd5 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;
};

inline RxVecd5 operator+(const RxVecd5 &v1, const RxVecd5 &v2)
{
	RxVecd5 vr;
	return (vr.add (v1, v2));
}

inline RxVecd5 operator-(const RxVecd5 &v1, const RxVecd5 &v2)
{
	RxVecd5 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd5 operator*(const RxVecd5 &v1, double a)
{
	RxVecd5 vr;
	return (vr.scale (a, v1));
}

inline RxVecd5 operator*(double a, const RxVecd5 &v1)
{
	RxVecd5 vr;
	return (vr.scale (a, v1));
}

inline RxVecd5 operator/(const RxVecd5 &v1, double a)
{
	RxVecd5 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd5 operator-(const RxVecd5 &v1)
{
	RxVecd5 vr;
	return (vr.neg (v1));
}

#endif
