#ifndef _RXVECDN6_H_WRAP_
#define _RXVECDN6_H_WRAP_

#include <RxVecd.h>

class RxVecd6: public RxVecd
{
	double vbuf[6];

    public:

	int length() const
	 { return 6;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 

	// assignment

	RxVecd6& operator=(const RxVecd6 &v2)
	 { set (v2.v);
	   return (*this);
	 }
	RxVecd6& set (const double *dp);
	void get (double *dp);

	// constructors	

	RxVecd6 () : RxVecd (6, vbuf) 
	 { 
	 }
	RxVecd6 (const RxVecd6 &v2) : RxVecd (6, vbuf)
	 { set (v2.v);
	 }
	RxVecd6 (double *dp) : RxVecd (6, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd6& add (const RxVecd6 &v1, const RxVecd6 &v2);
	RxVecd6& add (const RxVecd6 &v2);

	RxVecd6& operator+=(const RxVecd6 &v2)
	 { return (add (v2));
	 }
	friend RxVecd6 operator+(const RxVecd6 &v1, const RxVecd6 &v2);

	RxVecd6& sub (const RxVecd6 &v1, const RxVecd6 &v2);
	RxVecd6& sub (const RxVecd6 &v2);
	RxVecd6& operator-=(const RxVecd6 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd6 operator-(const RxVecd6 &v1, const RxVecd6 &v2);

	RxVecd6& extrap (const RxVecd6 &v1, double a, const RxVecd6 &v2);
	RxVecd6& extrap (double a, const RxVecd6 &v2)
	 { return (extrap (*this, a, v2));
	 }

	// T <- T,O operations

	RxVecd6& scale (double a, const RxVecd6 &v1);
	RxVecd6& scale (double a);
	RxVecd6& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd6& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd6 operator*(const RxVecd6 &v1, double a);
	friend RxVecd6 operator*(double a, const RxVecd6 &v1);
	friend RxVecd6 operator/(const RxVecd6 &v1, double a);

	// O <- T,T operations

	double dot(const RxVecd6 &v2) const;

	// T <- T operations

	RxVecd6 &unit(const RxVecd6 &v1);
	RxVecd6 &unit()
	 { return (unit(*this));
	 }
	RxVecd6 &neg(const RxVecd6 &v1);
	RxVecd6 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd6 operator-(const RxVecd6 &v1);

	// T <- O operations

	RxVecd6 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;
};

inline RxVecd6 operator+(const RxVecd6 &v1, const RxVecd6 &v2)
{
	RxVecd6 vr;
	return (vr.add (v1, v2));
}

inline RxVecd6 operator-(const RxVecd6 &v1, const RxVecd6 &v2)
{
	RxVecd6 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd6 operator*(const RxVecd6 &v1, double a)
{
	RxVecd6 vr;
	return (vr.scale (a, v1));
}

inline RxVecd6 operator*(double a, const RxVecd6 &v1)
{
	RxVecd6 vr;
	return (vr.scale (a, v1));
}

inline RxVecd6 operator/(const RxVecd6 &v1, double a)
{
	RxVecd6 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd6 operator-(const RxVecd6 &v1)
{
	RxVecd6 vr;
	return (vr.neg (v1));
}

#endif
