#include <RxVecd7.h>
#include <RxVecdBaseTest.h>
#include <RxTimeval.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	7

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0},
 { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0 },
 {-3.0, -2.0, -1.0, 3.0, 7.0, 6.6, 7.7 }};

main()
{
	static double vec1[LEN] = { 11.1, 22.2, 33.3, 44.4, 55.5, 66.6, 77.7};
	static double vec2[LEN] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
	RxVecd7 v1;
	RxVecd7 v2;
	RxVecd7 vd(vec1);
	int i, j;

	for (i=0; i<LEN; i++)
	 { assert (EQ(vd(i), vec1[i]));
	 }
	RxVecd7 v3(vd);
	for (i=0; i<LEN; i++)
	 { assert (EQ(v3(i), vec1[i]));
	 }
	assert (v3.length() == LEN);

	assert (v3.equal(vd,EPS));
	assert (!v3.equal(v1,EPS));

	for (i=0;i<LEN;i++)
	 { v1(i) = vec1[i];
	 }
	assert (v1.equal(v3,EPS));

	for (i=0;i<LEN;i++)
	 { v1(i) = i*10+2;
	 }
	assert (!v2.equal(v1,EPS));
	v2 = v1;
	assert (v2.equal(v1,EPS));

	v1.set (vec1);
	for (i=0;i<LEN;i++)
	 { assert (EQ(v1(i), vec1[i]));
	 }
	v1.get (vec2);		
	for (i=0;i<LEN;i++)
	 { assert (EQ(vec1[i], vec2[i]));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { v1.set(testVals[i]);
	      v2.set(testVals[j]);

	      assert (testAdd (v1+v2, v1, v2, EPS));
	      assert (testAdd (v3.add(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testAdd (v3.add(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testAdd (v3 += v2, v1, v2, EPS));

	      assert (testSub (v1-v2, v1, v2, EPS));
	      assert (testSub (v3.sub(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testSub (v3.sub(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testSub (v3 -= v2, v1, v2, EPS));

	      assert (testExtrap (v3.extrap(v1,11.1,v2), v1, 11.1, v2, EPS));
	      v3 = v1;
	      assert (testExtrap (v3.extrap(11.1,v2), v1, 11.1, v2, EPS));

	      v3 = v1;
	      assert (testDot (v3.dot(v2), v1, v2, EPS));
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);

	   assert (testScale (v3.scale(M_PI,v1), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.scale(M_PI), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 *= M_PI, M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 /= M_PI, 1/M_PI, v1, EPS));
	   assert (testScale (M_PI*v1, M_PI, v1, EPS));
	   assert (testScale (v1*M_PI, M_PI, v1, EPS));
	   assert (testScale (v1/M_PI, 1/M_PI, v1, EPS));

	   assert (testUnit (v3.unit(v1), v1, EPS));
	   v3 = v1;
	   assert (testUnit (v3.unit(), v1, EPS));
	   assert (testScale (v3.neg(v1), -1.0, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.neg(), -1.0, v1, EPS));
	   assert (testScale (-v1, -1.0, v1, EPS));

	   assert (testNorm1 (v1.norm1(), v1, EPS));
	   assert (testNorm2 (v1.norm2(), v1, EPS));
	   assert (testNormInf (v1.normInf(), v1, EPS));
	 }

	for (i=0; i<LEN; i++)
	 { v1(i) = 0;
	 }
	assert ((v2.setZero()).equal(v1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}


