#include <RxVecdBase.h>
#include <RxGenError.h>
#include <cdefs.h>
#include <math.h>
#include <stdlib.h>
#include <RxCio.h>
#include <RxPrintSpec.h>
#include <RxParseObj.h>
#include <RxSysError.h>

void RxVecdBase::doAdd(const RxVecdBase &v1, const RxVecdBase &v2, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = v1.v[i] + v2.v[i];
	 }
}

void RxVecdBase::doAdd(const RxVecdBase &v2, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] += v2.v[i];
	 }
}

void RxVecdBase::doSub(const RxVecdBase &v1, const RxVecdBase &v2, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = v1.v[i] - v2.v[i];
	 }
}

void RxVecdBase::doSub(const RxVecdBase &v2, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] -= v2.v[i];
	 }
}

void RxVecdBase::doExtrap(const RxVecdBase &v1, double a, 
			  const RxVecdBase &v2, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = v1.v[i] + a*v2.v[i];
	 }
}

void RxVecdBase::doScale(double a, const RxVecdBase &v1, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = v1.v[i]*a;
	 }
}

void RxVecdBase::doScale(double a, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] *= a;
	 }
}

double RxVecdBase::doDot(const RxVecdBase &v2, int l) const
{
	double x;
	int i;

	x = 0;
	for (i=0; i<l; i++)
	 { x += v[i]*v2.v[i];
	 }
	return (x);
}

void RxVecdBase::doUnit(const RxVecdBase &v1, int l)
{
	double mag;
	int i;

	mag = 0;
	for (i=0; i<l; i++)
	 { mag += v1.v[i]*v1.v[i];
	 }
	mag = sqrt(mag);
	if (mag > 0)
	 { for (i=0; i<l; i++)
	    { v[i] = v1.v[i] / mag;
	    }
	 }
}

void RxVecdBase::doNeg(const RxVecdBase &v1, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = -v1.v[i];
	 }
}

void RxVecdBase::doZero(int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = 0;
	 }
}

double RxVecdBase::doNorm1(int l) const
{
	double x;
	int i;

	x = 0;
	for (i=0; i<l; i++)
	 { x += ABS(v[i]);
	 }
	return (x);
}

double RxVecdBase::doNorm2(int l) const
{
	double x;
	int i;

	x = 0;
	for (i=0; i<l; i++)
	 { x += v[i]*v[i];
	 }
	return (sqrt(x));
}

double RxVecdBase::doNormInf(int l) const
{
	double x, max;
	int i;

	max = 0;
	for (i=0; i<l; i++)
	 { if ((x = ABS(v[i])) > max)
	    { max = x;
	    }
	 }
	return (max);
}

void RxVecdBase::doCopy(const RxVecdBase &v1, int l)
{
	int i;

	for (i=0; i<l; i++)
	 { v[i] = v1.v[i];
	 }
}

double& RxVecdBase::elem(int i)
{
	static double dum = 0;
	if (i < 0 || i >= length())
	 { RxGenError(RxGenError::BadIndex, 
		     "RxVecdBase::elem %K (%d)\n", i).raise();
	   return (dum);
	 }
	else
	 { return v[i];
	 }
}

double RxVecdBase::elem(int i) const
{
	if (i < 0 || i >= length())
	 { RxGenError(RxGenError::BadIndex, 
		     "RxVecdBase::elem %K (%d)\n", i).raise();
	   return (0);
	 }
	else
	 { return v[i];
	 }
}

#if 0
int RxVecdBase::scan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/0);
	return scan (&ci);
}

int RxVecdBase::xscan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/1);
	return scan (&ci);
}

int RxVecdBase::scan (const char *buf, int *cntp)
{
	RxCinput ci(buf, cntp);
	return scan (&ci);
}

istream &RxVecdBase::scan(istream &is)
{
	RxCinput ci(&is);
	int status;

	status = scan (&ci);
	if (status == EOF)
	 { is.clear (ios::eofbit);
	 }
	else if (status == 0)
	 { is.clear (ios::failbit);
	 }
	return is;
}

int RxVecdBase::scan (RxCinput *ci)
{
	RxParseObj *obj;

	obj = RxParseObj::scan (ci);
	if (obj == RxParseObj::eof)
	 { return EOF;
	 }
	else if (obj == NULL)
	 { return 0;
	 }
	else
	 { return oscan (obj);
	 }
}
#endif

int RxVecdBase::oscan (RxParseObj *obj)
{
	RxParseObj *op;
	int i;

	if (!obj->isNumberList() || obj->getListLength() != length())
	 { RxSysError (RxSysError::ParsingError,
"RxVecdBase::oscan: not a list of %d numbers\n", length()).post();
	   return 0;
	 }
	for (i=0, op=obj->firstChild(); op; op=obj->nextChild(), i++)
	 { (*this)(i) = op->getNumberValue();
	 }
	return 1;
}

#if 0
int RxVecdBase::print (FILE *fp, const char *s) const
{
	RxCoutput co(fp);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

int RxVecdBase::print (char *buf, const char *s, int n) const
{
	RxCoutput co(buf, n);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

ostream &RxVecdBase::print(ostream &os) const
{
	RxCoutput co(&os);
	RxPrintSpec fs("%g");	// fake print spec -- format will be ignored
	cprint (&co, &fs);
	return os;
}
#endif

int RxVecdBase::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	RxFmtNode *cf, *cf0, *cf1;
	int i;

	co->output_str (ps->prefix());
	if (ps->numConvs() == 0) // then allocate our our format node
	 { cf0 = new RxFmtNode ("%g", 0, RxFmtNode::CONV);
	 }
	else
	 { cf0 = (RxFmtNode *)ps->convNode(0);
	 }
	cf1 = cf0->next();
	if (cf1 == ps->suffixNode())
	 { cf1 = NULL;
	 }
       	cf = cf0;
	co->output_str ("[");
	for (i=0; i<length(); i++)
	 { co->output_double (cf->fmt(), (*this)(i));
	   if ((cf = cf->next()) == ps->suffixNode())
	    { cf = (cf1 ? cf1 : cf0);
	    }
	   if (i < length()-1)	// not last element 
	    { if (cf && cf->type() == RxFmtNode::STR)
	       { co->output_str (cf->fmt());
		 cf = cf->next();
	       }
	      else
	       { co->output_str (" ");
	       }
	    }
	 }
	co->output_str ("]");
	co->output_str (ps->suffix());
	if (ps->numConvs() == 0) // then cf0 was allocated here 
	 { delete cf0;
	 }
	return co->nchars();
}

int RxVecdBase::equal (const RxVecdBase& v2, double eps) const
{
	double diff;
	int i;

	for (i=0; i<length(); i++)
	 { diff = v2(i) - (*this)(i);
	   if (ABS(diff) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

RxVecdBase& RxVecdBase::setArray(double *dp)
{
	int i;

	for (i=0; i<length(); i++)
	 { (*this)(i) = dp[i];
	 }
	return (*this);
}

void RxVecdBase::getArray(double *dp)
{
	int i;

	for (i=0; i<length(); i++)
	 { dp[i] = (*this)(i);
	 }
}
