#include <RxVecdBaseTest.h>
#include <cdefs.h>
#include <math.h>

int testAdd (const RxVecdBase &vr, const RxVecdBase &v1, const RxVecdBase &v2, 
	 double eps)
{
	int i;

	for (i=0; i<vr.length(); i++)
	 { if (ABS(vr(i)-v1(i)-v2(i)) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

int testSub (const RxVecdBase &vr, const RxVecdBase &v1, const RxVecdBase &v2,
	 double eps)
{
	int i;

	for (i=0; i<vr.length(); i++)
	 { if (ABS(vr(i)-v1(i)+v2(i)) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

int testExtrap (const RxVecdBase &vr, const RxVecdBase &v1, double a,
	    const RxVecdBase &v2, double eps)
{
	int i;

	for (i=0; i<vr.length(); i++)
	 { if (ABS(vr(i)-v1(i)-a*v2(i)) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

int testScale (const RxVecdBase &vr, double a, const RxVecdBase &v1, double eps)
{
	int i;

	for (i=0; i<vr.length(); i++)
	 { if (ABS(vr(i)-a*v1(i)) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

int testDot (double a, const RxVecdBase &v1, const RxVecdBase &v2, double eps)
{
	double dot;
	int i;

	dot = 0;
	for (i=0; i<v1.length(); i++)
	 { dot += v1(i)*v2(i);
	 }
	if (ABS(dot-a) > eps)
	 { return (0);
	 }
	return (1);
}

int testUnit (const RxVecdBase &vr, const RxVecdBase &v1, double eps)
{
	double mag;
	int i;

	mag = 0;
	for (i=0; i<v1.length(); i++)
	 { mag += v1(i)*v1(i);
	 }
	mag = sqrt(mag);
	for (i=0; i<v1.length(); i++)
	 { if (ABS(vr(i)-v1(i)/mag) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

int testNorm2 (double a, const RxVecdBase &v1, double eps)
{
	double dot;
	int i;

	dot = 0;
	for (i=0; i<v1.length(); i++)
	 { dot += v1(i)*v1(i);
	 }
	if (ABS(sqrt(dot)-a) > eps)
	 { return (0);
	 }
	return (1);
}

int testNorm1 (double a, const RxVecdBase &v1, double eps)
{
	double sum;
	int i;

	sum = 0;
	for (i=0; i<v1.length(); i++)
	 { sum += ABS(v1(i));
	 }
	if (ABS(sum-a) > eps)
	 { return (0);
	 }
	return (1);
}

int testNormInf (double a, const RxVecdBase &v1, double eps)
{
	double max;
	int i;

	max = 0;
	for (i=0; i<v1.length(); i++)
	 { if (ABS(v1(i)) > max)
	    { max = ABS(v1(i));
	    }
	 }
	if (ABS(max-a) > eps)
	 { return (0);
	 }
	return (1);
}
