#define TOKEN_ERR(fmt,arg) \
	{ char buf[256]; \
	  sprintf (buf, (fmt), (arg)); \
	  yylvalp->u.s = strdup(buf); \
	  token = ERR; \
	}

#if 0
static yylexinit(fp)
FILE *fp;
{
	inputFd = (fp ? fp : stdin);
	inputBuf[0] = '\0';
	inputBufPtr = inputBuf;
}
#endif

int strtos (char *str, char *buf, int max, char **ptr)
{
	/* Converts a string delimited by quotes ("") into an internal 
	   binary string, applying the conversion rules used by C,
	   and return an optional pointer to the place in the string where the
	   conversion ended. The string is terminated by a null character
	   ('\0'). 
	
	   str -	the string to convert.

	   buf -	the buffer to place the binary string in.

	   max -	the maximum size of the buffer, including space
			for the terminating null character.

	   ptr -	an optional pointer used to return the place in the
			string where the conversion ended.

	   The routine returns the size of the string. It returns -1
	   if a valid converison can't be performed, or the string is larger
	   than the buffer size.
       
	   The string is delimited by quotes (""). Quotes within the string
	   are escaped using a backslash (\"). Other escape sequences are
	   used to delimit escape itself (\\), newline (\n), tab (\t),
	   formfeed (\f), carriage return (\r), and backspace (\b).
	   The C convention for converted \<octalNumber> into the 
	   corresponding character is also observed. 

	   Strings may not continue across new lines unless the new line
	   itself is escaped.
	 */

	int cnt;
	register char *cp;

	if (ptr)
	 { *ptr = str;
	 }
	cp = str;
	cnt = -1;

	while(isspace(*cp))
	 { cp++;
	 }
	if (*cp++ != '"')
	 { return (-1);
	 }
	while (*cp != '"' && *cp != '\n' && *cp != '\0')
	 { if (cp[0] == '\\' && cp[1] == '\n')
	    { cp += 2;
	      continue;
	    }
	   if (++cnt >= max-1)
	    { return (-1);
	    }
	   if (cp[0] == '\\')
	    { if (cp[1] == '"')
	       { buf[cnt] = '"';
		 cp += 2;
	       }
	      else if (cp[1] == '\\')
	       { buf[cnt] = '\\';
		 cp += 2;
	       }
	      else if (cp[1] == 'n')
	       { buf[cnt] = '\n';
		 cp += 2;
	       }
	      else if (cp[1] == 't')
	       { buf[cnt] = '\t';
		 cp += 2;
	       }
	      else if (cp[1] == 'f')
	       { buf[cnt] = '\f';
		 cp += 2;
	       }
	      else if (cp[1] == 'r')
	       { buf[cnt] = '\r';
		 cp += 2;
	       }
	      else if (cp[1] == 'b')
	       { buf[cnt] = '\b';
		 cp += 2;
	       }
	      else if (isdigit(cp[1]) && cp[1] < 56)
	       { int v, i;
		 v = cp[1] - '0';
		 cp += 2;
		 for (i=0; isdigit(*cp) && *cp < 56 && i<2; i++)
		  { v = v*8 + *cp++ - '0';
		  }
		 buf[cnt] = MIN(v,255);
	       }
	      else
	       { buf[cnt] = cp[1];
		 cp += 2;
	       }
	    }
	   else
	    { buf[cnt] = *cp++;
	    }
	 }	   

	if (*cp != '"')
	 { return (-1);
	 }
	if (ptr)
	 { *ptr = cp+1;
	 }
	buf[cnt+1] = '\0';
	return (cnt+1);
}

static int validDigit (int c, int base)
{
	int i;

	if (isdigit(c))
	 { i = c - '0';
	 }
	else if (strchr ("abcdef", c))
	 { i = c - 'a' + 10;
	 }
	else if (strchr ("ABCDEF", c))
	 { i = c - 'A' + 10;
	 }
	else
	 { i = -1;
	 }
	return (0 <= i && i < base);
}

static int parseDigits (char *buf, char **epp, int base, 
			int c0, RxCinput &cin)
{
	char *cp;
	int c;

	cp = buf;
	if (validDigit ((c = c0), base))
	 { *cp++ = c0;
	   while (validDigit ((c = cin.getChar()), base))
	    { *cp++ = c;
	    }
	 }
	*cp = '\0';
	*epp = cp;
	return c;	
}

static int matchNumber (char *buf, int *basep, RxCinput &cin, int c0)
{
	int radixCnt = 0;
	int base = 0;
	int token;
	char *cp, *ep;
	int c, c1;

	cp = buf;
	*cp++ = c0;
	c1 = cin.getChar();
	if (c0 == '0')
	 { if (c1 == 'x' || c1 == 'X')
	    { base = 16;
	    }
	   else if (c1 == 'o' || c1 == 'O')
	    { base = 8;
	    }
	   else if (c1 == 'b' || c1 == 'B')
	    { base = 2;
	    }
	   if (base)
	    { c = parseDigits (cp, &ep, base, cin.getChar(), cin);
	      if (cp == ep)
	       { token = ERR;		// unget
	       }
	      else
	       { cp = ep;
		 token = INTEGER_LIT;	// unget
	       }
	    }
	 }
	else if (c0 == '.')
	 { radixCnt = 1;
	 }
	if (base == 0)
	 { 
	   c = parseDigits (cp, &ep, 10, c1, cin);
	   cp = ep;
	   if (c == '.' && radixCnt == 0)
	    { *cp++ = '.';
	      radixCnt = 1;
	      c = parseDigits (cp, &ep, 10, cin.getChar(), cin);
	      cp = ep;
	    }
	   if (cp-buf == 1 && *buf == '.')
	    { token = ERR;
	    }
	   else if (c == 'e' || c == 'E')
	    { *cp++ = c;
	      c = cin.getChar();
	      if (c == '+' || c == '-')
	       { *cp++ = c;
		 c = cin.getChar();
	       }
	      c = parseDigits (cp, &ep, 10, c, cin);
	      if (cp == ep)
	       { token = ERR;
	       }
	      else
	       { cp = ep;
		 token = DOUBLE_LIT;
	       }
	    }
	   else
	    { if (radixCnt)
	       { token = DOUBLE_LIT;
	       }
	      else
	       { token = INTEGER_LIT;
		 base = 10;
	       }
	    }
	 }
	if ((isalnum(c) || c == '_') && *(cp-1) != '.')
	 { while (isalnum(c))
	    { c = cin.getChar();
	    }
	   token = ERR;
	 }
	cin.ungetChar(c);
	*basep = base;
	return token;
}

int rxParseLexer (YYSTYPE *yylvalp, ParseArg *lexarg)
{
#define BUFSIZE	1056
	char ibuf[BUFSIZE];
	RxCinput &cin = *lexarg->cinp;
	char *cp;
	int token;
	int c;

	// start by skipping white space
	do
	 { c = cin.getChar();
	   if (c == '\n')
	    { lexarg->lineno++;
	    }
	 }
	while (isspace(c));

	if (c == EOF)
	 { yylvalp->lineno = lexarg->lineno;
	   return EOF_TOK;
	 }
	else if (c == '@')
	 { cp = ibuf;
	   do
	    { *cp++ = (c = cin.getChar());
	    }
	   while (isalnum(c) || c == '_');
	   if (cp == ibuf+1)
	    { TOKEN_ERR ("expected alphanumeric after '@'\n", 0);
	    }
	   else
	    { *(cp-1) = '\0';
	      yylvalp->u.s = strdup(ibuf);
	      yylvalp->s_raw = (char*)malloc(strlen(ibuf)+2);
	      strcpy (yylvalp->s_raw, "@");
	      strcat (yylvalp->s_raw, ibuf);		
	      token = REFERENCE;
	      cin.ungetChar (c);
	    }
	 }
	else if (isalpha(c) || c == '_')
	 { ibuf[0] = c;
	   cp = ibuf + 1;
	   do
	    { *cp++ = (c = cin.getChar());
	    }
	   while (isalnum(c) || c == '_');
	   *(cp-1) = '\0';
	   if (c == ':' && cin.peekChar() != '=')
	    { yylvalp->u.s = strdup(ibuf);
	      yylvalp->s_raw = (char*)malloc(strlen(ibuf)+2);
	      strcpy (yylvalp->s_raw, ibuf);
	      strcat (yylvalp->s_raw, ":");
	      token = TYPENAME;
	    }
	   else if (strcmp(ibuf, "true") == 0)
	    { yylvalp->u.i = 1;
	      yylvalp->s_raw = "true";
	      token = BOOLEAN_LIT;
	    }
	   else if (strcmp(ibuf, "false") == 0)
	    { yylvalp->u.i = 0;
	      yylvalp->s_raw = "false";
	      token = BOOLEAN_LIT;
	    }
	   else
	    { yylvalp->u.s = strdup(ibuf);
	      yylvalp->s_raw = yylvalp->u.s;
	      token = IDENTIFIER_TOK;
	      cin.ungetChar (c);
	    }
	 }
	else if (c == '"')
	 { int lastc;
	   ibuf[0] = '"';
	   cp = ibuf + 1;
	   do
	    { lastc = c;
	      *cp++ = (c = cin.getChar());
	    }
	   while ((c != '"' || lastc == '\\') && c != '\n');
	   if (c == '\n')
	    { TOKEN_ERR ("new line in character literal\n", 0);
	      lexarg->lineno++;
	    }
	   else
	    { *cp = '\0';
	      token = STRING_LIT;
	      yylvalp->u.s = (char*)malloc(cp-ibuf);
	      strtos (ibuf, yylvalp->u.s, cp-ibuf, (char**)0);
	      yylvalp->s_raw = strdup(ibuf);
	    }
	 }
	else if (isdigit(c) || c == '.')
	 { int base;
	   char *ep = 0;
	   
	   token = matchNumber (ibuf, &base, cin, c);
	   switch (token)
	    { case INTEGER_LIT:
	       { yylvalp->u.i = strtol (ibuf, &ep, base);
		 *ep = '\0';
		 yylvalp->s_raw = strdup(ibuf);
		 break;
	       }
	      case DOUBLE_LIT:
	       { yylvalp->u.d = strtod (ibuf, &ep);
		 *ep = '\0';
		 yylvalp->s_raw = strdup(ibuf);
		 break;
	       }
	      case ERR:
	       { TOKEN_ERR (" malformed number %s\n", ibuf);
		 break;
	       }
	    }
	   if (ep == ibuf)
	    { fprintf (stderr,
"rxParseLexer(): can't parse number %s\n", ibuf);
	      exit (1);
	    }
	 }
	else
	 { switch (c)
	    { case ',':
	       { token = COMMA;
		 break;
	       }
	      case '[':
	       { token = LEFTBRACKET;
		 lexarg->level++;
		 break;
	       }
	      case ']':
	       { token = RIGHTBRACKET;
		 lexarg->level--;
		 break;
	       }
	      case '(':
	       { token = LEFTPAREN;
		 lexarg->level++;
		 break;
	       }
	      case ')':
	       { token = RIGHTPAREN;
		 lexarg->level--;
		 break;
	       }
	      case '=':
	       { token = EQUAL;
		 break;
	       }
#if 0
	      case '?':
	       { token = QUESTION_MARK;
		 break;
	       }
#endif
	      case '+':
	       { token = PLUS;
		 break;
	       }
	      case '-':
	       { token = MINUS;
		 break;
	       }
	      case '*':
	       { token = TIMES;
		 break;
	       }
	      case '/':
	       { token = DIVIDE;
		 break;
	       }
	      case ':':
	       { c = cin.getChar();
		 if (c == '=')
		  { token = ASSIGN;
		    break;
		  }
	       }
	      default:
	       { TOKEN_ERR ("illegal character '%c'\n", c);
	       }
	    }
	 }
	yylvalp->lineno = lexarg->lineno;
	return (token);
}

static struct {
	int token;
	char* name; } tokenStrings[] = 
{
	INTEGER_LIT,	"INTEGER",
	DOUBLE_LIT,	"DOUBLE",
	STRING_LIT,	"STRING",
	BOOLEAN_LIT,    "BOOLEAN",
	LEFTBRACKET,	"[",
	RIGHTBRACKET, 	"]",
	LEFTPAREN,	"(",
	RIGHTPAREN, 	")",
	EQUAL,		"=",
	IDENTIFIER_TOK,	"IDENTIFIER",
	REFERENCE,	"REFERENCE",
	TYPENAME,	"TYPENAME",
	COMMA,		",",
	ERR,		"ERR",
	QUESTION_MARK,  "?",
	ASSIGN,      	":=",
	PLUS,		"+",
	MINUS,		"-",
	TIMES,		"*",
	DIVIDE,		"/",
	EOF_TOK,	"EOF",
};

static char* tokenToStr(int token)
{
	int k;

	for (k=0; k<ARRAY_SIZE(tokenStrings); k++)
	 { if (token == tokenStrings[k].token)
	    { return (tokenStrings[k].name);
	    }
	 }
	return ("UNKNOWN");
}

static int strToToken (char *str)
{
	int k;

	for (k=0; k<ARRAY_SIZE(tokenStrings); k++)
	 { if (strcmp(tokenStrings[k].name, str) == 0)
	    { return (tokenStrings[k].token);
	    }
	 }
	return (-1);
}
