#include <RxP2RKin.h>
#include <stdlib.h>
#include <mathx.h>
#include <assert.h>
#include <RxError.h>
#include <cdefs.h>

#define SING_ELBOW	RX_PLANAR2RKIN_SING_ELBOW

typedef struct {
	double jnts[2];
	int sing;
} TestJnts;

TestJnts testJnts[] = { 
	0.0,    -45.0,		0,
       90.0,     45.0,		0,
	0.0,      0.0,          SING_ELBOW,
       45.0,     30.0,		0,
       45.0,    -30.0,		0,
};

#define EPS	(1e-10)

#define NUMJ	2

main()
{
	RxMatd Jplanar(6,NUMJ);
	RxMatd Jsub(NUMJ,NUMJ);
	RxMatd JsubI(NUMJ,NUMJ);
	RxMatd JchkI1(NUMJ,NUMJ);
	RxMatd JchkI2(NUMJ,NUMJ);
	RxMatd Jchk1(6,NUMJ);
	RxMatd Jchk2(6,NUMJ);
	RxTrsf T6planar;
	RxTrsf T6genk;
	int i, j, k;
	int conf;

	putenv ("RX_RBT_PATH=.");

	RxP2RKin planar("p2R");
	RxGenKin genk("p2R");

	assert (!RxErrCtx::error());
	assert (planar.numJnts() == NUMJ);

	for (k=0; k<ARRAY_SIZE(testJnts); k++)
	 { RxVecd2 jnts;
	   RxTwist tw;
	   RxVecd2 djnts;
	   RxVecd2 tor;
	   RxWrench wr;

	   for (j=0; j<NUMJ; j++)
	    { testJnts[k].jnts[j] *= DTOR;
	    }
	   jnts.set (testJnts[k].jnts);

	   planar.setJnts (jnts);
	   genk.setJnts (jnts);

	   planar.fwdK (&T6planar, &conf);
	   genk.fwdK (&T6genk, (int*)0);

	   if (!T6planar.equal (T6genk, EPS))
	    { printf ("k=%d\n", k);
	      T6planar.print ("T6planar");
	      T6genk.print ("T6genk");
	      (T6planar-T6genk).print ("diff");
	      exit (1);
	    }

	   planar.Jset (&Jplanar);
	   genk.Jset (&Jchk1);
	   Jchk1(5,0) = 0;
	   Jchk1(5,1) = 0;

	   if (!Jplanar.equal (Jchk1, EPS))
	    { printf ("Jplanar != genk.Jset, k=%d\n", k);
	      Jplanar.print ("Jplanar %12.7f\n\n");
	      Jchk1.print ("Jgenk %12.7f\n\n");
	      exit (1);
	    }

	   for (j=0; j<NUMJ; j++)
	    { djnts.setZero();
	      djnts(j) = 1.0;

	      // test planar.Jmul and genk.Jmul by constructing a Jacobian

	      planar.Jmul (&tw, djnts);
	      for (i=0; i<6; i++)
	       { Jchk1(i,j) = tw(i);
	       }
	      genk.Jmul (&tw, djnts);
	      for (i=0; i<6; i++)
	       { Jchk2(i,j) = tw(i);
	       }
	      Jchk2(5,0) = 0;
	      Jchk2(5,1) = 0;
	    }
	   if (!Jplanar.equal (Jchk1, EPS))
	    { printf ("Jplanar != planar.Jmul(I), k=%d\n", k);
	      Jplanar.print ("Jplanar %12.7f\n\n");
	      Jchk1.print ("Jpmul %12.7f\n\n");
	      exit (1);
	    }
	   if (!Jplanar.equal (Jchk2, EPS))
	    { printf ("Jplanar != genk.Jmul(I), k=%d\n", k);
	      Jplanar.print ("Jplanar %12.7f\n\n");
	      Jchk2.print ("Jgenk %12.7f\n\n");
	      exit (1);
	    }

	   Jchk1.tsp();
	   Jchk2.tsp();

	   for (i=0; i<6; i++)
	    { wr.setZero();
	      wr(i) = 1.0;

	      // test planar.JTmul and genk.JTmul 
	      // by constructing transpose Jacobian
	      
	      planar.JTmul (&tor, wr);
	      for (j=0; j<NUMJ; j++)
	       { Jchk1(j,i) = tor(j);
	       }
	      genk.JTmul (&tor, wr);
	      for (j=0; j<NUMJ; j++)
	       { Jchk2(j,i) = tor(j);
	       }
	      Jchk2(0,5) = 0;
	      Jchk2(1,5) = 0;
	    }
	   if (!Jplanar.T().equal (Jchk1, EPS))
	    { printf ("Jplanar^T != planar.JTmul(I), k=%d\n", k);
	      Jplanar.T().print ("Jplanar^T %12.7f\n\n");
	      Jchk1.print ("JpTmul  %12.7f\n\n");
	      exit (1);
	    }
	   if (!Jplanar.T().equal (Jchk2, EPS))
	    { printf ("Jplanar^T != genk.JTmul(I), k=%d\n", k);
	      Jplanar.T().print ("Jplanar^T %12.7f\n\n");
	      Jchk2.print ("JgTmul  %12.7f\n\n");
	      exit (1);
	    }

	   Jsub = RxMatd (Jplanar, 0, 0, 2, 2);

	   if (Jsub.condest() < 1e6)
	    { JsubI.inv (Jsub);
	      
	      // test planar.Jsolve by constructing an inverse Jacobian

	      for (i=0; i<2; i++)
	       { tw.setZero();
		 tw(i) = 1.0;

		 planar.Jsolve (&djnts, tw);
		 for (j=0; j<NUMJ; j++)
		  { JchkI1(j,i) = djnts(j);
		  }
		 genk.Jsolve (&djnts, tw);
		 for (j=0; j<NUMJ; j++)
		  { JchkI2(j,i) = djnts(j);
		  }
	       }
	      if (!JsubI.equal (JchkI1, EPS))
	       { printf ("JsubI != planar.Jsolve(I), k=%d\n", k);
		 JsubI.print ("JsubI %12.7f\n\n");
		 JchkI1.print ("Jsolve %12.7f\n\n");
		 exit (1);
	       }
#if 0
	      if (!JsubI.equal (JchkI2, EPS))
	       { printf ("JsubI != genk.Jsolve(I), k=%d\n", k);
		 JsubI.print ("JsubI %12.7f\n\n");
		 JchkI2.print ("Jsolve %12.7f\n\n");
		 exit (1);
	       }
#endif
	      // test planar.JTsolve by constructing an inverse Jacobian transpose

	      for (j=0; j<NUMJ; j++)
	       { tor.setZero();
		 tor(j) = 1.0;

		 planar.JTsolve (&wr, tor);
		 for (i=0; i<2; i++)
		  { JchkI1(i,j) = wr(i);
		  }
		 genk.JTsolve (&wr, tor);
		 for (i=0; i<2; i++)
		  { JchkI2(i,j) = wr(i);
		  }
	       }
	      if (!JsubI.T().equal (JchkI1, EPS))
	       { printf ("JsubI^T != planar.JTsolve(I), k=%d\n", k);
		 JsubI.T().print ("JsubI^T %12.7f\n\n");
		 JchkI1.print ("JTsolve  %12.7f\n\n");
		 exit (1);
	       }
#if 0
	      if (!JsubI.T().equal (JchkI2, EPS))
	       { printf ("JsubI^T != genk.JTsolve(I), k=%d\n", k);
		 JsubI.T().print ("JsubI^T %12.7f\n\n");
		 JchkI2.print ("JTsolve  %12.7f\n\n");
		 exit (1);
	       }
#endif
	    }
	   else
	    { printf (
"notice: (near) singular Jacobian at k=%d, condest=%g\n", k, Jsub.condest());
	    }
	   Jchk1.tsp();
	   Jchk2.tsp();
	 }

	printf ("\nROUTINES OK\n\n");
	exit (0);
}
