#include <RxPumaKin.h>
#include <math.h>
#include <mathx.h>
#include <cdefs.h>

#define J1	0
#define J2	1
#define J3	2
#define J4	3
#define J5	4
#define J6	5

#define A2	dhpSet[J2].a
#define D3	dhpSet[J3].d
#define D4	dhpSet[J4].d
#define A3	dhpSet[J3].a

#define S1	dhpSet[J1].sT
#define S2	dhpSet[J2].sT
#define S23	s23
#define S3	dhpSet[J3].sT
#define S4	dhpSet[J4].sT
#define S5	dhpSet[J5].sT
#define S6	dhpSet[J6].sT

#define C1	dhpSet[J1].cT
#define C2	dhpSet[J2].cT
#define C23	c23
#define C3	dhpSet[J3].cT
#define C4	dhpSet[J4].cT
#define C5	dhpSet[J5].cT
#define C6	dhpSet[J6].cT

#define EPSILONF	(1e-13)
#define SING_TOL	(0.01*DTOR)

void RxPumaKin::setDefaultPuma ()
{
	DHParams dhp[6];
	int i;

	for (i=0; i<6; i++)
	 { dhp[i].type = 'R';
	   dhp[i].theta = 0;
	   dhp[i].alpha = 0;
	   dhp[i].d = 0;
	   dhp[i].a = 0;
	 }
	dhp[0].alpha = -M_PI/2;
	dhp[1].a     =  431.80;
	dhp[2].a     = -20.32;
	dhp[2].d     =  149.09;
	dhp[2].alpha =  M_PI/2;
	dhp[3].d     =  433.07;
	dhp[3].alpha = -M_PI/2;
	dhp[4].alpha =  M_PI/2;
	set (dhp, 6);
}

#define SINCOS(s,c,ang)	s = sin(ang); c = cos(ang);
void RxPumaKin::pumaInit()
{
	d32 = D3*D3;
	l4 = sqrt(D4*D4+A3*A3);
	a22_a32d42 = A2*A2 - A3*A3 - D4*D4;
	atana3d4 = atan2(A3, D4);
	sqrt_a32d42 = sqrt(A3*A3+D4*D4);
	d42a22a32 = D4*D4 + A2*A2 + A3*A3;
	d32a222d422a32 = D3*D3 + A2*A2 + 2*D4*D4 + 2*A3*A3;

	epsilon_a2d4 = (A2+D4)/2.0*EPSILONF;
	epsilon_d32 = D3*D3*EPSILONF;

	s23 = sin(dhpSet[J2].theta + dhpSet[J3].theta);
	c23 = cos(dhpSet[J2].theta + dhpSet[J3].theta);
}

int RxPumaKin::setPumaJnts (const RxVecd6 &jnts)
{
	double *jvec = jnts.buf();
#if 1
	dhpSet[J1].theta = jvec[J1];
	SINCOS (S1, C1, jvec[J1]);
	dhpSet[J2].theta = jvec[J2];
	SINCOS (S2, C2, jvec[J2]);
	dhpSet[J3].theta = jvec[J3];
	SINCOS (S3, C3, jvec[J3]);
	dhpSet[J4].theta = jvec[J4];
	SINCOS (S4, C4, jvec[J4]);
	dhpSet[J5].theta = jvec[J5];
	SINCOS (S5, C5, jvec[J5]);
	dhpSet[J6].theta = jvec[J6];
	SINCOS (S6, C6, jvec[J6]);

	s23 = sin(jvec[J2]+jvec[J3]);
	c23 = cos(jvec[J2]+jvec[J3]);
#endif
	flags = 0;
	return (1);
}

int RxPumaKin::fwdK (RxTrsf *Tr, int *conf)
{
	int rcode = 0;
	register double k2, k3, k4;

	if (Tr)
	 { 
	   k2 = C4*C5*S6 + S4*C6;
	   k3 = - C23*k2 + S23*S5*S6;
	   k4 = - S4*C5*S6 + C4*C6;

	   Tr->R.o.x() = C1*k3 - S1*k4;
	   Tr->R.o.y() = S1*k3 + C1*k4;
	   Tr->R.o.z() = S23*k2 + C23*S5*S6;

	   k2 = S4*S5;
	   k3 = C23*C4*S5 + S23*C5;

	   Tr->R.a.x() = C1*k3 - S1*k2;
	   Tr->R.a.y() = S1*k3 + C1*k2;
	   Tr->R.a.z() = - S23*C4*S5 + C23*C5;

	   k2 = D4*S23 + A3*C23 + A2*C2;

	   Tr->p.x() = C1*k2 - S1*D3;
	   Tr->p.y() = S1*k2 + C1*D3;
	   Tr->p.z() = D4*C23 - A2*S2 - A3*S23;

	   /* Get n from n = o X a, to make matrix more normal 
	      and save time */

	   Tr->R.n.x() = Tr->R.o.y()*Tr->R.a.z() - Tr->R.o.z()*Tr->R.a.y();
	   Tr->R.n.y() = Tr->R.o.z()*Tr->R.a.x() - Tr->R.o.x()*Tr->R.a.z();
	   Tr->R.n.z() = Tr->R.o.x()*Tr->R.a.y() - Tr->R.o.y()*Tr->R.a.x();
	 }

	if (conf != NULL)
	 { double ang;
	   *conf = 0;

	   /* Determine the configurations: */
	   /*  Righty - Lefty */

	   if (Tr == NULL)
	    { double px, py;

	      k2 = D4*S23 + A3*C23 + A2*C2;
	      px = C1*k2 - S1*D3;
	      py = S1*k2 + C1*D3;
	      ang = rangePi(dhpSet[J1].theta + atan2(py, -px));
	    }
	   else
	    { ang = rangePi(dhpSet[J1].theta + atan2(Tr->p.y(), -Tr->p.x()));
	    }
   	   ang = ABS(ang);
	   if (ang <= M_PI/2)
	    { *conf |= RX_PUMAKIN_CONF_RIGHT;
	    }
	   if (ABS(ang-M_PI/2) <= SING_TOL)
	    { rcode |= RX_PUMAKIN_SING_SHOULDER;
	    }
   
	   /*  Up - Down  */
   
	   ang = rangePi(atana3d4 + dhpSet[J3].theta);
	   ang = ABS(ang);
	   if (ang <= M_PI/2)
	    { *conf |= RX_PUMAKIN_CONF_DOWN;
	    }	
	   if (ABS(ang-M_PI/2) <= SING_TOL)
	    { rcode |= RX_PUMAKIN_SING_ELBOW;
	    }

	   ang = rangePi(dhpSet[J5].theta);
	   if (ang >= 0)
	    { *conf |= RX_PUMAKIN_CONF_FLIP;
	    }
	   if (ABS(ang) <= SING_TOL || 
	       ang >= M_PI-SING_TOL || ang < -M_PI+SING_TOL)
	    { rcode |= RX_PUMAKIN_SING_WRIST;
	    }
	 }
	return (rcode);
}

void RxPumaKin::buildJacobian()
{
	RxTrsf Um, Am;

	// If the Jacobian is written as
	//
	//     [ J11	 0  ]
	// J = [	    ]
	//     [ J21    J22 ]
	//
	// then in link 4, using the coordinates convention of (Paul 1981),
	// we have
	//
	//  	 [ h si4 - co23 co4 dd3   aa2 co4 si3    co4 dd4  ]
	//	 [                                                ]
	// J11 = [       dd3 si23           aa2 co3        aa3    ]
	//	 [                                                ]
	//	 [ co23 dd3 si4 + co4 h  - aa2 si3 si4  - dd4 si4 ]
 	//
	//	   h   =   dd4 si23 + aa3 co23 + aa2 co2
 	//
	//		       [  0   0   si5  ]
	//		       [               ]
	// J22 =               [ - 1  0  - co5 ]
	//		       [               ]
	//		       [  0   1    0   ]
 	//
	//		    [ - co4 si23  0  si4 ]
	//		    [                    ]
	// J21 =            [   - co23    0   0  ]
	//		    [                    ]
	//		    [  si23 si4   0  co4 ]
 	//
	//
	// To map from link 6 to link 4, we use the matrix U5 = A5 A6.
	// Since this is a pure rotation, we can use only the rotational
	// matrix part, which we call R5.
	//                                    T
	// 4     [ R5  0  ] 6        6    [ R5   0  ] 4
	//  dx = [        ]  dx  ==>  J = [       T ]  J
	//	 [ 0   R5 ]		  [  0  R5  ]
	//

	h2 = c23*A3 + s23*D4;
	h1 = h2 + C2*A2;
	h3 = S3*A3 - C3*D4;
	h1_norm = D4 + A2;

	j00 =  h1*S4 - D3*C23*C4;
	j10 =  S23*D3;
	j20 =  h1*C4 + D3*C23*S4;
	j01 =  A2*S3*C4;
	j11 =  A2*C3;
	j21 = -A2*S3*S4;
	j02 =  C4*D4;
     // j12 =  A3;
	j22 = -S4*D4;
	j30 = -S23*C4;
     // j40 = -C23;
	j50 =  S23*S4;
     // j31 = 0;
     // j41 = 0;
     // j51 = 0;
     // j32 = S4;
     // j42 = 0;
     // j52 = C4;
     // j33 = 0;
     // j43 = -1;
     // j53 = 0;
     // j34 = 0;
     // j44 = 0;
     // j54 = 1;
     // j35 = S5;
     // j45 = -C5;
     // j55 = 0;

	R5.n.x() =  C5*C6;
	R5.n.y() =  S5*C6;
	R5.n.z() =  S6;
	R5.o.x() = -C5*S6;
	R5.o.y() = -S5*S6;
	R5.o.z() =  C6;
	R5.a.x() =  S5;
	R5.a.y() = -C5;
	R5.a.z() =  0.;

	flags |= RX_GENKIN_JACOBIAN_VALID;
}

RxMatd& RxPumaKin::Jset (RxMatd *Jr)
{
	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	(*Jr)(3,0) = R5.n.x()*j30 - R5.n.y()*C23 + R5.n.z()*j50;
	(*Jr)(4,0) = R5.o.x()*j30 - R5.o.y()*C23 + R5.o.z()*j50;
	(*Jr)(5,0) = R5.a.x()*j30 - R5.a.y()*C23 + R5.a.z()*j50;
	(*Jr)(3,2) = R5.n.x()*S4 		     + R5.n.z()*C4 ;
	(*Jr)(4,2) = R5.o.x()*S4 		     + R5.o.z()*C4 ;
	(*Jr)(5,2) = R5.a.x()*S4               + R5.a.z()*C4 ;
	(*Jr)(3,1) = (*Jr)(3,2);
	(*Jr)(4,1) = (*Jr)(4,2);
	(*Jr)(5,1) = (*Jr)(5,2);

	(*Jr)(0,0) = R5.n.x()*j00 + R5.n.y()*j10 + R5.n.z()*j20;
	(*Jr)(1,0) = R5.o.x()*j00 + R5.o.y()*j10 + R5.o.z()*j20;
	(*Jr)(2,0) = R5.a.x()*j00 + R5.a.y()*j10 + R5.a.z()*j20;
	(*Jr)(0,2) = R5.n.x()*j02 + R5.n.y()*A3  + R5.n.z()*j22;
	(*Jr)(1,2) = R5.o.x()*j02 + R5.o.y()*A3  + R5.o.z()*j22;
	(*Jr)(2,2) = R5.a.x()*j02 + R5.a.y()*A3  + R5.a.z()*j22;
	(*Jr)(0,1) = R5.n.x()*j01 + R5.n.y()*j11 + R5.n.z()*j21 + (*Jr)(0,2);
	(*Jr)(1,1) = R5.o.x()*j01 + R5.o.y()*j11 + R5.o.z()*j21 + (*Jr)(1,2);
	(*Jr)(2,1) = R5.a.x()*j01 + R5.a.y()*j11 + R5.a.z()*j21 + (*Jr)(2,2);

	(*Jr)(3,3) =            - R5.n.y();
	(*Jr)(4,3) =            - R5.o.y();
	(*Jr)(5,3) =            - R5.a.y();
	(*Jr)(3,4) =                           R5.n.z();
	(*Jr)(4,4) = 			       R5.o.z();
	(*Jr)(5,4) =                           R5.a.z();
	(*Jr)(3,5) = R5.n.x()*S5  - R5.n.y()*C5;
	(*Jr)(4,5) = R5.o.x()*S5  - R5.o.y()*C5;
	(*Jr)(5,5) = R5.a.x()*S5  - R5.a.y()*C5;

	(*Jr)(0,3) = 0;
	(*Jr)(1,3) = 0;
	(*Jr)(2,3) = 0;
	(*Jr)(0,4) = 0;
	(*Jr)(1,4) = 0;
	(*Jr)(2,4) = 0;
	(*Jr)(0,5) = 0;
	(*Jr)(1,5) = 0;
	(*Jr)(2,5) = 0;

	return (*Jr);
}

RxTwist& RxPumaKin::Jmul (RxTwist *tw, const RxVecd &djnts)
{
	double q23;
	RxTwist tw4;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }

	q23 = djnts(J2) + djnts(J3);
 
	tw4.v().x() = j00*djnts(J1) + j01*djnts(J2) + j02*q23;
	tw4.v().y() = j10*djnts(J1) + j11*djnts(J2) + A3*q23;
	tw4.v().z() = j20*djnts(J1) + j21*djnts(J2) + j22*q23;
	tw4.w().x() = j30*djnts(J1) + S4*q23 + S5*djnts(J6);
	tw4.w().y() = -C23*djnts(J1) - djnts(J4) - C5*djnts(J6);
	tw4.w().z() = j50*djnts(J1) + C4*q23 + djnts(J5);

	tw->v().x() = R5.n.x()*tw4.v().x() + R5.n.y()*tw4.v().y() + R5.n.z()*tw4.v().z();
	tw->v().y() = R5.o.x()*tw4.v().x() + R5.o.y()*tw4.v().y() + R5.o.z()*tw4.v().z();
	tw->v().z() = R5.a.x()*tw4.v().x() + R5.a.y()*tw4.v().y() + R5.a.z()*tw4.v().z();
	tw->w().x() = R5.n.x()*tw4.w().x() + R5.n.y()*tw4.w().y() + R5.n.z()*tw4.w().z();
	tw->w().y() = R5.o.x()*tw4.w().x() + R5.o.y()*tw4.w().y() + R5.o.z()*tw4.w().z();
	tw->w().z() = R5.a.x()*tw4.w().x() + R5.a.y()*tw4.w().y() + R5.a.z()*tw4.w().z();
	
	return (*tw);
}

RxVecd& RxPumaKin::JTmul (RxVecd *tor, const RxWrench &wr)
{
	RxWrench wr4;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	wr4.f().x() = R5.n.x()*wr.f().x() + R5.o.x()*wr.f().y() + R5.a.x()*wr.f().z();
	wr4.f().y() = R5.n.y()*wr.f().x() + R5.o.y()*wr.f().y() + R5.a.y()*wr.f().z();
	wr4.f().z() = R5.n.z()*wr.f().x() + R5.o.z()*wr.f().y() + R5.a.z()*wr.f().z();
	wr4.m().x() = R5.n.x()*wr.m().x() + R5.o.x()*wr.m().y() + R5.a.x()*wr.m().z();
	wr4.m().y() = R5.n.y()*wr.m().x() + R5.o.y()*wr.m().y() + R5.a.y()*wr.m().z();
	wr4.m().z() = R5.n.z()*wr.m().x() + R5.o.z()*wr.m().y() + R5.a.z()*wr.m().z();

	(*tor)(J1) = j00*wr4.f().x() + j10*wr4.f().y() +
		     j20*wr4.f().z() + j30*wr4.m().x() -
		     C23*wr4.m().y() + j50*wr4.m().z();

	(*tor)(J2) = j01*wr4.f().x() + j11*wr4.f().y() +
		     j21*wr4.f().z();
   
	(*tor)(J3) = j02*wr4.f().x() + A3*wr4.f().y() +
		     j22*wr4.f().z() + S4*wr4.m().x() + C4*wr4.m().z();
   
	(*tor)(J2) += (*tor)(J3);
   
	(*tor)(J4) = - wr4.m().y();
   
	(*tor)(J5) = wr4.m().z();

	(*tor)(J6) = S5*wr4.m().x() - C5*wr4.m().y();

	return (*tor);
}

int RxPumaKin::invK (RxVecd *jnts, const RxTrsf &T, int conf,
		     const RxVecd *jref, int update)
{
	int code = 0;

	register double r1, r2, r3;
	double ang3p, s3p, c3p;
	double s1, c1, s23, c23, s3, c3, s4, c4, s5, c5, s6, c6;

	double *jvec = jnts->buf();
	double fxnArg;			/* argument to square root */

	/* theta1 */

	fxnArg = T.p.x()*T.p.x() + T.p.y()*T.p.y() - d32;
	if (fxnArg <= epsilon_d32)
	 { if (fxnArg < 0)
	    { code |= RX_PUMAKIN_CLIP_SHOULDER;		/* too far away */
	    }
	   fxnArg = 0.0;
	   code |= RX_PUMAKIN_SING_SHOULDER;
	 }
	r1 = ((conf & RX_PUMAKIN_CONF_RIGHT) ? -1:1)*sqrt(fxnArg);
	if (ABS(T.p.y()) <= EPSILONF && ABS(T.p.x()) <= EPSILONF)
	 { jvec[J1] = (jref ? (*jref)(J1) : dhpSet[J1].theta);
	   code |= RX_GENKIN_INDETERMINATE;
	 }
	else
	 { jvec[J1] = atan2(T.p.y(),T.p.x()) - atan2(D3, r1);
	 }
	SINCOS (s1, c1, jvec[J1]);
#if 1
	/* theta3 */

	fxnArg = (r1*r1 + T.p.z()*T.p.z() - A2*A2 - l4*l4)/(2*A2*l4);
	if (fxnArg >= 1.0-EPSILONF)
	 { if (fxnArg > 1.0)
	    { code |= RX_PUMAKIN_CLIP_OUTER;
	    }
	   fxnArg = 1.0;
	   code |= RX_PUMAKIN_SING_ELBOW;
	 }
	else if (fxnArg <= -1.0+EPSILONF)
	 { if (fxnArg < -1.0)
	    { code |= RX_PUMAKIN_CLIP_INNER;
	    }
	   fxnArg = -1.0;
	   code |= RX_PUMAKIN_SING_ELBOW;
	 }
	if (conf & RX_PUMAKIN_CONF_DOWN)
	 { ang3p = acos(fxnArg);
	   c3p = fxnArg;
	   s3p = sqrt(1 - fxnArg*fxnArg);
	 }
	else
	 { ang3p = -acos(fxnArg);
	   c3p = fxnArg;
	   s3p = -sqrt(1 - fxnArg*fxnArg);
	 }
	jvec[J3] = M_PI/2 - atana3d4 - ang3p;
	jvec[J2] = atan2(l4*s3p, A2+l4*c3p) - atan2(T.p.z(),r1);
#endif
	s23 = sin(jvec[J3]+jvec[J2]);
	c23 = cos(jvec[J3]+jvec[J2]);
	
	/* theta4 */

	r1 = - s1*T.R.a.x() + c1*T.R.a.y();
	r2 =   c1*T.R.a.x() + s1*T.R.a.y();
	r3 = c23*r2 - s23*T.R.a.z();

	if (ABS (r1) < EPSILONF && ABS (r3) < EPSILONF)
	 { jvec[J4] = (jref ? (*jref)(J4) : dhpSet[J4].theta);
	   code |= (RX_GENKIN_INDETERMINATE | RX_PUMAKIN_SING_WRIST);
	 }
	else
	 { jvec[J4] = (conf & RX_PUMAKIN_CONF_FLIP) ? atan2(r1,  r3) : 
						      atan2(-r1, -r3);
	 }

	SINCOS (s4, c4, jvec[J4]);

	/* theta5 */

	s5 = c4*r3 + s4*r1 ;
	c5 = s23*r2 + c23*T.R.a.z();
	jvec[J5] = atan2(s5, c5);

	/* theta6 */

	r1 =   c1*T.R.o.x() + s1*T.R.o.y();
	r3 = - s1*T.R.o.x() + c1*T.R.o.y();
	s6 = - c5*(c4*(c23*r1 - s23*T.R.o.z()) + s4*r3)
	     + s5*(s23*r1 + c23*T.R.o.z());
	c6 = - s4*(c23*r1 - s23*T.R.o.z()) + c4*r3;

	jvec[J6] = atan2(s6, c6);

	if (update)	/* Update sines and cosines if needed */
	 { S1 = s1;
	   C1 = c1;
	   dhpSet[J1].theta = jvec[J1];
	   S2 = sin(jvec[J2]);
	   C2 = cos(jvec[J2]);
	   dhpSet[J2].theta = jvec[J2];
	   S23 = s23;
	   C23 = c23;
	   S3 = s3;
	   C3 = c3;
	   dhpSet[J3].theta = jvec[J3];
	   S4 = s4;
	   C4 = c4;
	   dhpSet[J4].theta = jvec[J4];
	   S5 = s5;
	   C5 = c5;
	   dhpSet[J5].theta = jvec[J5];
	   S6 = s6;
	   C6 = c6;
	   dhpSet[J6].theta = jvec[J6];
	 }
	return(code);
}

int RxPumaKin::Jsolve (RxVecd *dj, const RxTwist &tw)
{
	RxTwist tw4;
	int code = 0;
	register double threshold;
	double q23, vx;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	tw4.v().x() = R5.n.x()*tw.v().x() + R5.o.x()*tw.v().y() + R5.a.x()*tw.v().z();
	tw4.v().y() = R5.n.y()*tw.v().x() + R5.o.y()*tw.v().y() + R5.a.y()*tw.v().z();
	tw4.v().z() = R5.n.z()*tw.v().x() + R5.o.z()*tw.v().y() + R5.a.z()*tw.v().z();
	tw4.w().x() = R5.n.x()*tw.w().x() + R5.o.x()*tw.w().y() + R5.a.x()*tw.w().z();
	tw4.w().y() = R5.n.y()*tw.w().x() + R5.o.y()*tw.w().y() + R5.a.y()*tw.w().z();
	tw4.w().z() = R5.n.z()*tw.w().x() + R5.o.z()*tw.w().y() + R5.a.z()*tw.w().z();

	/* theta1 */

	threshold = tol*h1_norm;
	if (FABS(h1) < threshold) 
	 { code |= RX_PUMAKIN_SING_SHOULDER;
	   (*dj)(J1) = (S4*tw4.v().x() + C4*tw4.v().z())*h1 /
		    (threshold*threshold);
	 }
	else
	 { (*dj)(J1) = (S4*tw4.v().x() + C4*tw4.v().z()) / h1;
	 }

	/* theta2 */

	vx = C4*tw4.v().x() - S4*tw4.v().z();

	threshold = tol*sqrt_a32d42;
	if (FABS(h3) < threshold)
	 { code |= RX_PUMAKIN_SING_ELBOW;
	   (*dj)(J2) =   (- D4*tw4.v().y() + A3*vx + (*dj)(J1)*D3*h2 )
		   *(h3) / (A2*threshold*threshold);
	 }
	else
	 { (*dj)(J2) =   (- D4*tw4.v().y() + A3*vx + (*dj)(J1)*D3*h2 )
		    / (A2*h3);
	 }

	/* theta3 */

	q23 = ( vx - A2*S3*(*dj)(J2) + D3*C23*(*dj)(J1) ) / D4;
	(*dj)(J3) = q23 - (*dj)(J2);

	/* theta6 */

	threshold = tol;
	if (FABS(S5) < threshold)
	 { code |= RX_PUMAKIN_SING_WRIST;
	   (*dj)(J6) = (tw4.w().x() + S23*C4*(*dj)(J1) -
		    S4*q23)*S5 / (threshold*threshold);
	 }
	else
	 { (*dj)(J6) = (tw4.w().x() + S23*C4*(*dj)(J1) -
		       S4*q23) / S5;
	 }

	/* theta4 */

	(*dj)(J4) = - tw4.w().y() - C23*(*dj)(J1) - C5*(*dj)(J6);

	/* theta5 */

	(*dj)(J5) = tw4.w().z() - S23*S4*(*dj)(J1) - C4*q23;

	return (code);
}

int RxPumaKin::JTsolve (RxWrench *wr, const RxVecd &torIn)
{
	RxVecd6 tor;
	RxWrench wr4;
	int code = 0;
	double tor2, threshold, vx, vz, wx, wy, wz;

	/* copy the const torIn into a local tor because of a compiler bug!! */
	tor(J1) = torIn(J1);
	tor(J2) = torIn(J2);
	tor(J3) = torIn(J3);
	tor(J4) = torIn(J4);
	tor(J5) = torIn(J5);
	tor(J6) = torIn(J6);

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { RxGenKin::buildJacobian();
	 }

	tor2 = tor(J2) - tor(J3);

	wr4.m().z() = tor(J5);

	wr4.m().y() = -tor(J4);

	threshold = tol;
	if (FABS(S5) < threshold)
	 { code |= RX_PUMAKIN_SING_WRIST;
	   wr4.m().x() = (- C5*tor(J4) + tor(J6))*S5 / 
		     (threshold*threshold);
	 }
	else
	 { wr4.m().x() = (- C5*tor(J4) + tor(J6)) / S5;
	 }

	vx = tor(J1) + C23*wr4.m().y() + 
		     S23*(C4*wr4.m().x() - S4*wr4.m().z());
	vz = tor(J3) - wr4.m().x()*S4 - wr4.m().z()*C4;

	threshold = tol*sqrt_a32d42;
	if (FABS(h3) < threshold)
	 { code |= RX_PUMAKIN_SING_ELBOW;
	   wz = (tor2*D4 - A2*S3*vz) 
	   *(h3) / (A2*threshold*threshold);
	 }
	else
	 { wz = (tor2*D4 - A2*S3*vz) / (A2*h3);
	 }

	wx = (A3*wz + vz) / D4;

	threshold = tol*h1_norm;
	if (FABS(h1) < threshold)
	 { code |= RX_PUMAKIN_SING_SHOULDER;
	   wy =  (D3*(S23*wz + C23*wx) + vx)
		 *h1 / (threshold*threshold);
	 }
	else
	 { wy =  (D3*(S23*wz + C23*wx) + vx)
		  / h1;
	 }

	wr4.f().y() = -wz;

	wr4.f().x() =   C4*wx + S4*wy;

	wr4.f().z() = - S4*wx + C4*wy;

	wr->m().x() = R5.n.x()*wr4.m().x() + R5.n.y()*wr4.m().y() + R5.n.z()*wr4.m().z();
	wr->m().y() = R5.o.x()*wr4.m().x() + R5.o.y()*wr4.m().y() + R5.o.z()*wr4.m().z();
	wr->m().z() = R5.a.x()*wr4.m().x() + R5.a.y()*wr4.m().y() + R5.a.z()*wr4.m().z();
	wr->f().x() = R5.n.x()*wr4.f().x() + R5.n.y()*wr4.f().y() + R5.n.z()*wr4.f().z();
	wr->f().y() = R5.o.x()*wr4.f().x() + R5.o.y()*wr4.f().y() + R5.o.z()*wr4.f().z();
	wr->f().z() = R5.a.x()*wr4.f().x() + R5.a.y()*wr4.f().y() + R5.a.z()*wr4.f().z();

	return code;
}
