#ifndef _RXPUMAKIN_H_WRAP_
#define _RXPUMAKIN_H_WRAP_

#include <RxGenKin.h>
#include <RxVecd6.h>

#define RX_PUMAKIN_SING_SHOULDER 0x0100
#define RX_PUMAKIN_SING_ELBOW	 0x0200
#define RX_PUMAKIN_SING_WRIST	 0x0400

#define RX_PUMAKIN_CONF_RIGHT	 0x0001
#define RX_PUMAKIN_CONF_DOWN	 0x0002
#define RX_PUMAKIN_CONF_FLIP	 0x0004

#define RX_PUMAKIN_CLIP_OUTER	 0x0001
#define RX_PUMAKIN_CLIP_INNER	 0x0002
#define RX_PUMAKIN_CLIP_SHOULDER 0x0004

class RxPumaKin : public RxGenKin
{
	void pumaInit();
	int setPumaJnts (const RxVecd6 &jnts);

	double d32;
	double l4;
	double a22_a32d42;
	double atana3d4;
	double sqrt_a32d42;
	double d42a22a32;
	double d32a222d422a32;
	double epsilon_a2d4;
	double epsilon_d32;	

	double s23;
	double c23;

	// things for the Jacobian computation:

	double h1;
	double h2;
	double h3;
	double h1_norm;
	RxRot R5;
	double j00, j10, j20;
	double j01, j11, j21;
	double j02, j22;
	double j30, j50;

	void buildJacobian();
	void setDefaultPuma ();

public:
	RxPumaKin () : RxGenKin()
	 { setDefaultPuma ();
	   pumaInit();
	 }
	RxPumaKin (const char *fileName) : RxGenKin (fileName)
	 { pumaInit();
	 }
	RxPumaKin (const DHParams *params, int n) : RxGenKin (params, n)
	 { pumaInit();
	 }
	RxPumaKin (const RxPumaKin &kin) : RxGenKin (kin)
	 { pumaInit();
	 }

	int setJnts (const RxVecd &jnts)
	 { return (setPumaJnts (*(const RxVecd6 *)&jnts));
	 }

	int fwdK (RxTrsf *T, int *conf);
	int fwdK (RxTrsf *T, int *conf, const RxVecd &jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (fwdK (T, conf));
	 }

	RxMatd& Jset (RxMatd *J);
	RxMatd& Jset (RxMatd *J, const RxVecd jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (Jset (J));
	 }
	RxTwist& Jmul (RxTwist *tw, const RxVecd &dj);
	RxTwist& Jmul (RxTwist *tw, const RxVecd &dj, 
			       const RxVecd jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (Jmul (tw, dj));
	 }
	RxVecd& JTmul (RxVecd *tor, const RxWrench &wr);
	RxVecd& JTmul (RxVecd *tor, const RxWrench &wr, 
			       const RxVecd jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (JTmul (tor, wr));
	 }

	// inverse kinematics

	int invK (RxVecd *jnts, const RxTrsf &T, int conf, 
			  const RxVecd *jref=NULL, int update=0);

	int Jsolve (RxVecd *dj, const RxTwist &tw);
	int Jsolve (RxVecd *dj, const RxTwist &tw,
		   const RxVecd jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (Jsolve (dj, tw));
	 }
	int JTsolve (RxWrench *wr, const RxVecd &tor);
	int JTsolve (RxWrench *wr, const RxVecd &tor,
		     const RxVecd jnts)
	 { setPumaJnts (*(const RxVecd6 *)&jnts);
	   return (JTsolve (wr, tor));
	 }
};

#endif
