#include <RxPumaKin.h>
#include <RxTimeval.h>
#include <stdlib.h>
#include <mathx.h>
#include <assert.h>
#include <RxError.h>
#include <cdefs.h>

#define SING_WRIST	RX_PUMAKIN_SING_WRIST
#define SING_ELBOW	RX_PUMAKIN_SING_ELBOW
#define SING_SHOULDER	RX_PUMAKIN_SING_SHOULDER

typedef struct {
	double jnts[6];
	int sing;
} TestJnts;

TestJnts testJnts[] = { 
       10.0, 	 20.0,  30.0,  40.0,   50.0,   60.0,    0,
	0.0,	  0.0,   0.0,   0.0,   45.0,    0.0,	0,
	0.0,	-90.0,  10.0,  60.0,   90.0,   60.0,	0,
	0.0,	 10.0, -45.0, -60.0,  -90.0,   60.0,	0,
       90.0,   -180.0, 180.0,   0.0,  -45.0,    0.0,	0,
       90.0,   -180.0, 180.0,  60.0,  -45.0,   30.0,	0,
       90.0,   -180.0, 180.0, -60.0,   45.0,  -30.0,	0,
      -90.0,   -135.0,   0.0,   0.0,  -90.0,    0.0, 	0, /* rcclpark */
	0.0,	-90.0,	90.0,	0.0,	0.0,    0.0,	SING_WRIST,
};

#define EPS	(1e-10)

int compareJnts (RxVecd &ja, RxVecd &jb, double eps)
{
	int j;

	for (j=0; j<ja.length(); j++)
	 { if (rangePi(ja(j)-jb(j)) > eps)
	    { return (0);
	    }
	 }
	return (1);
}

main()
{
	RxMatd Jpuma(6,6);
	RxMatd JpumaI(6,6);
	RxMatd Jchk1(6,6);
	RxMatd Jchk2(6,6);
	RxTrsf T6puma;
	RxTrsf T6genk;
	int i, j, k;
	int conf;

	putenv ("RX_RBT_PATH=.");

	RxPumaKin puma("puma560");
	RxGenKin genk("puma560");

	assert (!RxErrCtx::firstErrCur());
	assert (puma.numJnts() == 6);

	for (k=0; k<ARRAY_SIZE(testJnts); k++)
	 { RxVecd6 jnts;
	   RxTwist tw;
	   RxVecd6 djnts;
	   RxVecd6 tor;
	   RxWrench wr;
	   RxVecd ang(6);

	   for (j=0; j<6; j++)
	    { testJnts[k].jnts[j] *= DTOR;
	    }
	   jnts.set (testJnts[k].jnts);

	   if (k==0)
	    { double s, c;
	      TIMEIT (puma.fwdK (&T6puma, &conf, jnts), 10000);
	      TIMEIT (puma.invK (&ang, T6puma, conf, &jnts), 10000);
	      TIMEIT (puma.setJnts (jnts), 10000);
	      TIMEIT ((s = sin(jnts(0)), c = cos(jnts(0))), 1000);
	      if (!compareJnts (ang,jnts, EPS))
	       { printf ("ang != jnts at time test\n");
		 exit (1);
	       }
	    }

	   puma.setJnts (jnts);
	   genk.setJnts (jnts);

	   puma.fwdK (&T6puma, &conf);
	   genk.fwdK (&T6genk, (int*)0);

	   if (!T6puma.equal (T6genk, EPS))
	    { printf ("k=%d\n", k);
	      T6puma.print ("T6puma");
	      T6genk.print ("T6genk");
	      (T6puma-T6genk).print ("diff");
	      exit (1);
	    }

	   puma.invK (&ang, T6puma, conf, &jnts);
	   if (!compareJnts (ang,jnts, EPS))
	    { printf ("ang != jnts k=%d\n", k);
	      ang.print (" ang %8.4f\n");
	      jnts.print ("jnts %8.4f\n");
	      exit (1);
	    }	   

	   puma.Jset (&Jpuma);
	   genk.Jset (&Jchk1);

	   if (!Jpuma.equal (Jchk1, EPS))
	    { printf ("Jpuma != genk.Jset, k=%d\n", k);
	      Jpuma.print ("Jpuma %12.7f\n\n");
	      Jchk1.print ("Jgenk %12.7f\n\n");
	      exit (1);
	    }

	   for (j=0; j<6; j++)
	    { djnts.setZero();
	      djnts(j) = 1.0;

	      // test puma.Jmul and genk.Jmul by constructing a Jacobian

	      puma.Jmul (&tw, djnts);
	      for (i=0; i<6; i++)
	       { Jchk1(i,j) = tw(i);
	       }
	      genk.Jmul (&tw, djnts);
	      for (i=0; i<6; i++)
	       { Jchk2(i,j) = tw(i);
	       }
	    }
	   if (!Jpuma.equal (Jchk1, EPS))
	    { printf ("Jpuma != puma.Jmul(I), k=%d\n", k);
	      Jpuma.print ("Jpuma %12.7f\n\n");
	      Jchk1.print ("Jpmul %12.7f\n\n");
	      exit (1);
	    }
	   if (!Jpuma.equal (Jchk2, EPS))
	    { printf ("Jpuma != genk.Jmul(I), k=%d\n", k);
	      Jpuma.print ("Jpuma %12.7f\n\n");
	      Jchk2.print ("Jgenk %12.7f\n\n");
	      exit (1);
	    }

	   for (i=0; i<6; i++)
	    { wr.setZero();
	      wr(i) = 1.0;

	      // test puma.JTmul and genk.JTmul 
	      // by constructing transpose Jacobian
	      
	      puma.JTmul (&tor, wr);
	      for (j=0; j<6; j++)
	       { Jchk1(j,i) = tor(j);
	       }
	      genk.JTmul (&tor, wr);
	      for (j=0; j<6; j++)
	       { Jchk2(j,i) = tor(j);
	       }
	    }
	   if (!Jpuma.T().equal (Jchk1, EPS))
	    { printf ("Jpuma^T != puma.JTmul(I), k=%d\n", k);
	      Jpuma.T().print ("Jpuma^T %12.7f\n\n");
	      Jchk1.print ("JpTmul  %12.7f\n\n");
	      exit (1);
	    }
	   if (!Jpuma.T().equal (Jchk2, EPS))
	    { printf ("Jpuma^T != genk.JTmul(I), k=%d\n", k);
	      Jpuma.T().print ("Jpuma^T %12.7f\n\n");
	      Jchk2.print ("JgTmul  %12.7f\n\n");
	      exit (1);
	    }

	   if (Jpuma.condest() < 1e6)
	    { JpumaI.inv (Jpuma);
	      
	      // test puma.Jsolve by constructing an inverse Jacobian

	      for (i=0; i<6; i++)
	       { tw.setZero();
		 tw(i) = 1.0;

		 puma.Jsolve (&djnts, tw);
		 for (j=0; j<6; j++)
		  { Jchk1(j,i) = djnts(j);
		  }
		 genk.Jsolve (&djnts, tw);
		 for (j=0; j<6; j++)
		  { Jchk2(j,i) = djnts(j);
		  }
	       }
	      if (!JpumaI.equal (Jchk1, EPS))
	       { printf ("JpumaI != puma.Jsolve(I), k=%d\n", k);
		 JpumaI.print ("JpumaI %12.7f\n\n");
		 Jchk1.print ("Jsolve %12.7f\n\n");
		 exit (1);
	       }
	      if (!JpumaI.equal (Jchk2, EPS))
	       { printf ("JpumaI != genk.Jsolve(I), k=%d\n", k);
		 JpumaI.print ("JpumaI %12.7f\n\n");
		 Jchk2.print ("Jsolve %12.7f\n\n");
		 exit (1);
	       }

	      // test puma.JTsolve by constructing an inverse Jacobian transpose

	      for (j=0; j<6; j++)
	       { tor.setZero();
		 tor(j) = 1.0;

		 puma.JTsolve (&wr, tor);
		 for (i=0; i<6; i++)
		  { Jchk1(i,j) = wr(i);
		  }
		 genk.JTsolve (&wr, tor);
		 for (i=0; i<6; i++)
		  { Jchk2(i,j) = wr(i);
		  }
	       }
	      if (!JpumaI.T().equal (Jchk1, EPS))
	       { printf ("JpumaI^T != puma.JTsolve(I), k=%d\n", k);
		 JpumaI.T().print ("JpumaI^T %12.7f\n\n");
		 Jchk1.print ("JTsolve  %12.7f\n\n");
		 exit (1);
	       }
	      if (!JpumaI.T().equal (Jchk2, EPS))
	       { printf ("JpumaI^T != genk.JTsolve(I), k=%d\n", k);
		 JpumaI.T().print ("JpumaI^T %12.7f\n\n");
		 Jchk2.print ("JTsolve  %12.7f\n\n");
		 exit (1);
	       }
	    }
	   else
	    { printf (
"notice: (near) singular Jacobian at k=%d, condest=%g\n", k, Jpuma.condest());
	    }
	 }
	printf ("\nROUTINES OK\n\n");
	exit (0);
}
