#include <RxS3RKin.h>
#include <math.h>
#include <mathx.h>
#include <cdefs.h>

#include <roots.h>

#define J1	0
#define J2	1
#define J3	2

#define A1	dhpSet[J1].a
#define A2	dhpSet[J2].a
#define A3	dhpSet[J3].a

#define D1	dhpSet[J1].d
#define D2	dhpSet[J2].d
#define D3	dhpSet[J3].d
#define D4	0

#define S1	dhpSet[J1].sT
#define S2	dhpSet[J2].sT
#define S3	dhpSet[J3].sT

#define C1	dhpSet[J1].cT
#define C2	dhpSet[J2].cT
#define C3	dhpSet[J3].cT

#define SA1	dhpSet[J1].sA
#define SA2	dhpSet[J2].sA
#define SA3	dhpSet[J3].sA

#define CA1	dhpSet[J1].cA
#define CA2	dhpSet[J2].cA
#define CA3	dhpSet[J3].cA

#define EPSILONF	(1e-13)
#define SING_TOL	(0.0001*DTOR)

#define SINCOS(s,c,ang)	s = sin(ang); c = cos(ang);
void RxS3RKin::s3Rinit()
{
	g10 = A2;
	g1s = D4*SA3;
	g1c = A3;

	g20 = D3*SA2 + CA3*D4*SA2;
	g2s = -A3*CA2;
	g2c = CA2*D4*SA3;
 
	g30 = A1*A1 + A2*A2 + A3*A3 + D2*D2 + 2*CA2*D2*D3 + D3*D3 + 
	      2*CA2*CA3*D2*D4 + 2*CA3*D3*D4 + D4*D4;
	g3s = 2*(A3*D2*SA2 + A2*D4*SA3);
	g3c = 2*(A2*A3 - D2*D4*SA2*SA3);
 
	g40 = CA1*D2 + CA1*CA2*D3 + CA1*CA2*CA3*D4;
	g4s = A3*CA1*SA2;
	g4c = -CA1*D4*SA2*SA3;

	g50 = (D2 + CA2*D3 + CA2*CA3*D4)*SA1;
	g5c = -(D4*SA1*SA2*SA3);
	g5s = A3*SA1*SA2;	
}

void RxS3RKin::setDefaultS3R ()
{
	DHParams dhp[3];
	int i;

	for (i=0; i<3; i++)
	 { dhp[i].type = 'R';
	   dhp[i].theta = 0;
	   dhp[i].alpha = 0;
	   dhp[i].d = 0;
	   dhp[i].a = 0;
	 }
	dhp[J1].a     =  1;
	dhp[J1].alpha = -M_PI/2;
	dhp[J2].a     =  2;
	dhp[J2].d     =  1;
	dhp[J2].alpha =  M_PI/2;
	dhp[J3].a     =  1.5;
	set (dhp, 3);
}

int RxS3RKin::setS3RJnts (const RxVecd3 &jnts)
{
	double *jvec = jnts.buf();

	dhpSet[J1].theta = jvec[J1];
	SINCOS (S1, C1, jvec[J1]);
	dhpSet[J2].theta = jvec[J2];
	SINCOS (S2, C2, jvec[J2]);
	dhpSet[J3].theta = jvec[J3];
	SINCOS (S3, C3, jvec[J3]);
	flags = 0;
	return (1);
}

int RxS3RKin::invK (RxVecd *jnts, const RxTrsf &T, int conf,
		     const RxVecd *jref, int update)
{
	dcomplex coefs[5];
	dcomplex roots[4];
	double R3;
	double z;
	int status;
	int i, j, nr;
	double s3, c3, z3;
	double s2, c2;
	double s1, c1;
	double g1, g2;

	z = T.p.z() - D1;
	R3 = T.p.x()*T.p.x() + T.p.y()*T.p.y() + z*z;

	for (i=0; i<5; i++)
	 { coefs[i].i = 0;
	 }
	coefs[0].r = square(g30 + g3c - R3)*SA1*SA1 + 
    4*A1*A1*(g40*g40 + g4c*g4c - g10*g10*SA1*SA1 - 
       2*g10*g1c*SA1*SA1 - g1c*g1c*SA1*SA1 - 
       g20*g20*SA1*SA1 - 2*g20*g2c*SA1*SA1 - 
       g2c*g2c*SA1*SA1 + 2*g40*(g4c - z) - 2*g4c*z + z*z);


	coefs[1].r = 4*(g3s*(g30 + g3c - R3)*SA1*SA1 + 
      4*A1*A1*(g40*g4s + g4c*g4s - g10*g1s*SA1*SA1 - 
	 g1c*g1s*SA1*SA1 - g20*g2s*SA1*SA1 - g2c*g2s*SA1*SA1 - 
	 g4s*z));

	coefs[2].r = 2*((g30*g30 - g3c*g3c + 2*g3s*g3s -
2*g30*R3 + R3*R3)*
       SA1*SA1 + 4*A1*A1*
       (g40*g40 - g4c*g4c + 2*g4s*g4s -
	 g10*g10*SA1*SA1 + g1c*g1c*SA1*SA1 - 
	 2*g1s*g1s*SA1*SA1 - g20*g20*SA1*SA1 + 
	 g2c*g2c*SA1*SA1 - 2*g2s*g2s*SA1*SA1 - 2*g40*z + 
	 z*z));

	coefs[3].r = -4*(g3s*(-g30 + g3c + R3)*SA1*SA1 - 
      4*A1*A1*(g40*g4s - g4c*g4s - g10*g1s*SA1*SA1 + 
	 g1c*g1s*SA1*SA1 - g20*g2s*SA1*SA1 + g2c*g2s*SA1*SA1 - 
	 g4s*z));

	coefs[4].r = square(-g30 + g3c + R3)*SA1*SA1 + 
    4*A1*A1*(g40*g40 + g4c*g4c - g10*g10*SA1*SA1 + 
       2*g10*g1c*SA1*SA1 - g1c*g1c*SA1*SA1 - 
       g20*g20*SA1*SA1 + 2*g20*g2c*SA1*SA1 - 
       g2c*g2c*SA1*SA1 + 2*g4c*z + z*z - 2*g40*(g4c + z));

	if ((nr = zroots (roots, coefs, 4, /*polish=*/1)) < 0)
	 { return -1;
	 }
	if (conf >= nr)
	 { return -1;
	 }
#if 0
	printf ("nr = %d\n", nr);
	for (i=0; i<4; i++)
	 { printf ("%g+%gi\n", roots[i].r, roots[i].i);
	 }
#endif
	z3 = roots[conf].r;
	(*jnts)(J3) = 2*atan(z3);
	s3 = sin((*jnts)(J3));
	c3 = cos((*jnts)(J3));

	s2 = -((((g20 + c3*g2c + g2s*s3)*(g30 + c3*g3c - R3 +
		g3s*s3))/(2.*A1) + 
	((g10 + c3*g1c + g1s*s3)*(g40 + c3*g4c + g4s*s3 - z))/SA1)/
	(square(g10 + c3*g1c + g1s*s3) + square(g20 + c3*g2c + g2s*s3)));

	c2 =    (-((g10 + c3*g1c + g1s*s3)*(g30 + c3*g3c - R3 +
		g3s*s3))/(2.*A1) + 
      ((g20 + c3*g2c + g2s*s3)*(g40 + c3*g4c + g4s*s3 - z))/SA1)/
	(square(g10 + c3*g1c + g1s*s3) + square(g20 + c3*g2c + g2s*s3));
	(*jnts)(J2) = atan2(s2,c2);

	g1 = A1 + c2*(g10 + c3*g1c + g1s*s3) + s2*(g20 + c3*g2c + g2s*s3);
	g2 = g50 + c3*g5c + g5s*s3 - CA1*s2*(g10 + c3*g1c + g1s*s3) + 
	     c2*CA1*(g20 + c3*g2c + g2s*s3);
	
	(*jnts)(J1) = atan2(g2*T.p.x()+g1*T.p.y(), g1*T.p.x() - g2*T.p.y());

	return(nr);
}
