#include <math.h>
#include <cdefs.h>
#include <roots.h>
#include <dcomplex.h>
#include <math.h>

#define EPSS 1.0e-14
#define MR 8
#define MT 10
#define MAXIT (MT*MR)

int laguer(dcomplex a[], int m, dcomplex *x, int *its)
{
	int iter,j;
	double abx,abp,abm,err;
	dcomplex dx,x1,b,d,f,g,h,sq,gp,gm,g2;
	static double frac[MR+1] = {0.0,0.5,0.25,0.75,0.13,0.38,0.62,0.88,1.0};

	for (iter=1;iter<=MAXIT;iter++) {
		*its=iter;
		b=a[m];
		err=Cabs(b);
		d=f=Complex(0.0,0.0);
		abx=Cabs(*x);
		for (j=m-1;j>=0;j--) {
			f=Cadd(Cmul(*x,f),d);
			d=Cadd(Cmul(*x,d),b);
			b=Cadd(Cmul(*x,b),a[j]);
			err=Cabs(b)+abx*err;
		}
		err *= EPSS;
		if (Cabs(b) <= err) return 1;
		g=Cdiv(d,b);
		g2=Cmul(g,g);
		h=Csub(g2,RCmul(2.0,Cdiv(f,b)));
		sq=Csqrt(RCmul((double) (m-1),Csub(RCmul((double) m,h),g2)));
		gp=Cadd(g,sq);
		gm=Csub(g,sq);
		abp=Cabs(gp);
		abm=Cabs(gm);
		if (abp < abm) gp=gm;
		dx=((MAX(abp,abm) > 0.0 ? Cdiv(Complex((double) m,0.0),gp)
			: RCmul(exp(log(1+abx)),Complex(cos((double)iter),sin((double)iter)))));
		x1=Csub(*x,dx);
		if (x->r == x1.r && x->i == x1.i) return 1;
		if (iter % MT) *x=x1;
		else *x=Csub(*x,RCmul(frac[iter/MT],dx));
	}
	return -1;
}
#undef EPSS
#undef MR
#undef MT
#undef MAXIT
#undef NRANSI

#define EPS 2.0e-13
#define MAXM 100

int closestConj (double *err, dcomplex x, dcomplex *c, int n)
{
	dcomplex xconj;
	dcomplex diff;
	int mini;
	double minm, dmag;
	int i;

	xconj = Conjg(x);
	mini = -1;
	for (i=0; i<n; i++)
	 { if (c[i].i && c[i].i != x.i && c[i].r != x.r)
	    { diff = Csub(xconj, c[i]);
	      dmag = Cabs(diff);
	      if (mini == -1 || dmag < minm)
	       { mini = i;
		 minm = dmag;
	       }
	    }
	 }
	*err = minm;
	return (mini);
}

int zroots(dcomplex roots[], dcomplex a[], int m, int polish)
{
	int i,its,j,jj,nr;
	dcomplex x,b,c,ad[MAXM];
	int realcoefs = 1;

	for (j=0;j<=m;j++)
	 { ad[j]=a[j];
	   if (ad[j].i)
	    { realcoefs = 0;
	    }
	 }
	for (j=m;j>=1;j--) {
		x=Complex(0.0,0.0);
		if (laguer(ad,j,&x,&its) < 0)
		 { return -1;
		 }
		if (fabs(x.i) <= 2.0*EPS*fabs(x.r)) x.i=0.0;
		roots[j-1]=x;
		b=ad[j];
		for (jj=j-1;jj>=0;jj--) {
			c=ad[jj];
			ad[jj]=b;
			b=Cadd(Cmul(x,b),c);
		}
	}
	if (polish)
		for (j=1;j<=m;j++)
			if (laguer(a,m,&roots[j-1],&its) < 0)
			 { return -1;
			 }
	/* sort roots according to their real parts */
	for (j=2;j<=m;j++) {
		x=roots[j-1];
		for (i=j-1;i>=1;i--) {
			if (roots[i-1].r <= x.r) break;
			roots[i]=roots[i-1];
		}
		roots[i]=x;
	}

	nr = 0;
	for (j=0; j<m; j++)
	 { if (roots[j].i == 0)
	    { nr++;
	    }
	 }
#if 0
	printf ("nr=%d\n", nr);
	for (j=0; j<m; j++)
	 { printf ("% %g+%g\n", roots[j].r, roots[j].i);
	 }
#endif
	/* for real coefs, make sure complex roots have conjugates
	   and resort roots so that the real ones come first */
	if (realcoefs && (nr%2) != 0)
	 { int maxj = -1;
	   double maxe;
	   double err;
	      
	   for (j=0; j<m; j++)
	    { if (roots[j].i != 0)
	       { i = closestConj (&err, roots[j], roots, m);
		 if (i == -1)
		  { roots[j].i = 0;
		    return m;
		  }
		 if (maxj == -1 || err > maxe)
		  { maxj = j;
		    maxe = err;
		  }
	       }
	    }
	   roots[maxj].i = 0;
	   nr--;
	 }

	if (realcoefs)
	 { i = -1;
	   nr = 0;
	   for (j=0; j<m; j++)
	    { if (roots[j].i == 0)
	       { if (i != -1)
		  { x = roots[j];
		    roots[j] = roots[i];
		    roots[i] = x;
		    i++;
		  }
		 nr++;
	       }
	      else
	       { if (i == -1)
		  { i = j;
		  }
	       }
	    }
	 }
			
	return nr;
}

#undef EPS
#undef MAXM
