#include <stdio.h>
#include <cdefs.h>

#include "postScript.h"

FILE *psFp;

void psFileSet (fp)
FILE *fp;
{
	psFp = fp;
}
       
void psMoveto (x, y)
double x, y;
{
	fprintf (psFp, "%9.4f %9.4f moveto\n", x, y);
}

void psLineto (x, y)
double x, y;
{
	fprintf (psFp, "%9.4f %9.4f lineto\n", x, y);
}

void psCircle (x, y, r)
double x, y, r;
{
	fprintf (psFp, "newpath %9.4f %9.4f %9.4f 0 360 arc stroke\n", x, y, r);
}

void psLineSeg (x1, y1, x2, y2)
double x1, x2, y1, y2;
{
	fprintf (psFp, "%9.4f %9.4f moveto\n", x1, y1);
	fprintf (psFp, "%9.4f %9.4f lineto\n", x2, y2);
}

void psPutStrR (s, x, y)
char *s;
double x, y;
{
	fprintf (psFp, "(%s) %9.4f %9.4f putStrR\n", s, x, y);
}

void psPutStrL (s, x, y)
char *s;
double x, y;
{
	fprintf (psFp, "(%s) %9.4f %9.4f putStrL\n", s, x, y);
}

void psFont (size, s)
double size;
char *s;
{
	fprintf (psFp, "%9.4f %sFont\n", size, s);
}

void psTranslate (x, y)
double x, y;
{
	fprintf (psFp, "%9.4f %9.4f translate\n", x, y);
}

void psNewpath()
{
	fprintf (psFp, "newpath\n");
}

void psStroke()
{
	fprintf (psFp, "stroke\n");
}

void psShowpage()
{
	fprintf (psFp, "showpage\n");
}

void psPathdot(x, y, r, grey)
double x, y, r, grey;
{
	fprintf (psFp, "%9.4f %9.4f %9.4f pathdot\n", x, y, r, grey);
}

void psSetlinewidth(w)
double w;
{
	fprintf (psFp, "%9.4f setlinewidth\n", w);
}

void psSetclipBox (x1, y1, x2, y2)
double x1, y1, x2, y2;
{
	psCommand ("newpath");
	psMoveto (x1, y1);
	psLineto (x2, y1);
	psLineto (x2, y2);
	psLineto (x1, y2);
	psCommand ("closepath clip");
}

void psCommand (s)
char *s;
{
	fprintf (psFp, "%s\n", s);
}

/* ----- separate routines for idraw ----- */

#define IDRAW_SCALE	1.54166583
#define IDRAW_MAG	100.0

typedef struct {
	double scaleX;
	double scaleY;
	double baseX;
	double baseY;
	double curX;
	double curY;
	int lineWidth;
	int dashIdx;
} IdrawCtx;

#define CTX_STACK_SIZE	32

static double scale = IDRAW_SCALE*IDRAW_MAG;

static IdrawCtx ctx;
static IdrawCtx ctxStack[CTX_STACK_SIZE];
static ctxPtr;

static char* patterns[] = { 
	"none", 
	"0",
	"1",
	".75",
	".5",
	".25",
	"< 11 22 44 88 11 22 44 88 > -1",
	"< 88 44 22 11 88 44 22 11 > -1",
	"< ff 00 00 00 ff 00 00 00 > -1",
	"< 88 88 88 88 88 88 88 88 > -1",
	"< ff 88 88 88 ff 88 88 88 > -1",
	"< 88 55 22 55 88 55 22 55 > -1",
	"< cc cc 33 33 cc cc 33 33 > -1",
	"< 77 bb ee dd 77 bb ee dd > -1"
};

static int dashesCode[] = {
	0xffff,
	0x00ff,
	0x0f0f,
	0x3333
};	

static char *dashes[] = { 
	"[] 0",
	"[8 8] 9",
	"[4 4 4 4] 13",
	"[2 2 2 2 2 2 2 2] 15"
};

#define MLINEBUFSIZE	1024

static int mlineCnt;
static int mlineXbuf[MLINEBUFSIZE];
static int mlineYbuf[MLINEBUFSIZE];

static int xout(x)
double x;
{
	return ((ctx.baseX + x*ctx.scaleX)*scale);
}

static int yout(y)
double y;
{
	return ((ctx.baseY + y*ctx.scaleY)*scale);
}

static int printMatrix ()
{
	fprintf (psFp, "%%I t\n");
	fprintf (psFp, "[ %6.4f 0 0 %6.4f 0 0 ] concat\n",
			1/IDRAW_MAG, 1/IDRAW_MAG);
}

static int printPattern (idx)
int idx;
{
	if (idx == IDRAW_PAT_NONE)
	 { fprintf (psFp, "none SetP %%I p n\n");
	 }
	else
	 { fprintf (psFp, "%%I p\n");
	   fprintf (psFp, "%s SetP\n", patterns[idx]);
	 }
}

static int printLineInfo (doArrow)
int doArrow;
{
	fprintf (psFp, "%%I b %d\n", dashesCode[ctx.dashIdx]);
	fprintf (psFp, "%d 0 %d %s SetB\n", ctx.lineWidth, doArrow, 
					    dashes[ctx.dashIdx]);
}

void IdrawSave()
{
	if (ctxPtr >= CTX_STACK_SIZE)
	 { fprintf (stderr, "IdrawSave(): stack overflow\n");
	   exit (1);
	 }
	ctxStack[ctxPtr] = ctx;
	ctxPtr++;
}

void IdrawRestore()
{
	if (ctxPtr > 0)
	 { ctxPtr--;
	   ctx = ctxStack[ctxPtr];
	 }
}

void IdrawInit(fp)
FILE *fp;
{
	psFp = fp;

	ctx.scaleX = 1.0;
	ctx.scaleY = 1.0;

	ctx.lineWidth = IDRAW_WIDTH_1;
	ctx.dashIdx = IDRAW_SOLID;

	fprintf (psFp, "%%I Idraw 9 Grid 8\n\n");
	fprintf (psFp, "%%%%Page: 1 1\n\n");
	fprintf (psFp, "Begin\n");
	fprintf (psFp, "%%I b u\n");
	fprintf (psFp, "%%I cfg u\n");
	fprintf (psFp, "%%I cbg u\n");
	fprintf (psFp, "%%I f u\n");
	fprintf (psFp, "%%I p u\n");
	fprintf (psFp, "%%I t\n");
	fprintf (psFp, "[ %8.6f 0 0 %8.6f 0 0 ] concat\n",
			1/IDRAW_SCALE, 1/IDRAW_SCALE);
	fprintf (psFp, "/originalCTM matrix currentmatrix def\n");
}

int IdrawLineWidth (w)
int w;
{
	int old;

	old = ctx.lineWidth;
	ctx.lineWidth = w;
	return (old);
}

int IdrawDashes (idx)
int idx;
{
	int old;

	old = ctx.dashIdx;
	ctx.dashIdx = idx;
	return (old);
}

void IdrawBaseTrans (x, y)
double x, y;
{
	ctx.baseX += x;
	ctx.baseY += y;
}

void IdrawBaseSet (x, y)
double x, y;
{
	ctx.baseX = x;
	ctx.baseY = y;
}

void IdrawScaleSet (xs, ys)
double xs, ys;
{
	ctx.scaleX = xs;
	ctx.scaleY = ys;
}

void IdrawMove(x, y)
double x, y;
{
	ctx.curX = x;
	ctx.curY = y;
}

void IdrawCircle(r, patIdx)
double r;
int patIdx;
{
	IdrawEllipse(r, r, patIdx);
}

void IdrawEllipse(a, b, patIdx)
double a, b;
int patIdx;
{
	fprintf (psFp, "\nBegin %%I Elli\n");
	printLineInfo (/*doArrow=*/0);
	fprintf (psFp, "%%I cfg Black\n");
	fprintf (psFp, "0 0 0 SetCFg\n");
	fprintf (psFp, "%%I cbg White\n");
	fprintf (psFp, "1 1 1 SetCBg\n");
	printPattern (patIdx);
	printMatrix ();
	fprintf (psFp, "%%I\n");
	fprintf (psFp, "%d %d %d %d Elli\n", 
			xout(ctx.curX), yout(ctx.curY),
			(int)(a*scale*ctx.scaleX), (int)(b*scale*ctx.scaleY));
	fprintf (psFp, "End\n");
}

void IdrawLine (x, y)
double x, y;
{
	fprintf (psFp, "\nBegin %%I Line\n");
	printLineInfo (/*doArrow=*/0);
	fprintf (psFp, "%%I cfg Black\n");
	fprintf (psFp, "0 0 0 SetCFg\n");
	fprintf (psFp, "%%I cbg White\n");
	fprintf (psFp, "1 1 1 SetCBg\n");
	printPattern (IDRAW_PAT_NONE);
	printMatrix ();
	fprintf (psFp, "%%I\n");
	fprintf (psFp, "%d %d %d %d Line\n", 
			xout(ctx.curX), yout(ctx.curY), 
			xout(x), yout(y));
	fprintf (psFp, "End\n");

	ctx.curX = x;
	ctx.curY = y;
}

void IdrawLineSeg (x1, y1, x2, y2)
double x1, y1;
double x2, y2;
{
	IdrawMove (x1, y1);
	IdrawLine (x2, y2);
}

void IdrawArrowLine (x, y)
double x, y;
{
	fprintf (psFp, "\nBegin %%I Line\n");
	printLineInfo (/*doArrow=*/1);
	fprintf (psFp, "%%I cfg Black\n");
	fprintf (psFp, "0 0 0 SetCFg\n");
	fprintf (psFp, "%%I cbg White\n");
	fprintf (psFp, "1 1 1 SetCBg\n");
	printPattern (IDRAW_PAT_BLACK);
	printMatrix ();
	fprintf (psFp, "%%I\n");
	fprintf (psFp, "%d %d %d %d Line\n", 
			xout(ctx.curX), yout(ctx.curY), 
			xout(x), yout(y));
	fprintf (psFp, "End\n");

	ctx.curX = x;
	ctx.curY = y;
}

void IdrawMlineBegin (patIdx)
int patIdx;
{
	mlineCnt = 0;

	fprintf (psFp, "\nBegin %%I MLine\n");
	printLineInfo (/*doArrow=*/0);
	fprintf (psFp, "%%I cfg Black\n");
	fprintf (psFp, "0 0 0 SetCFg\n");
	fprintf (psFp, "%%I cbg White\n");
	fprintf (psFp, "1 1 1 SetCBg\n");
	printPattern (patIdx);
	printMatrix ();
}

void IdrawMlinePut (x, y)
double x, y;
{
	if (mlineCnt >= MLINEBUFSIZE)
	 { fprintf (stderr, "IdrawMlinePut: Overflow. Increase buffer\n");
	   exit (1);
	 }
	mlineXbuf[mlineCnt] = xout(x);
	mlineYbuf[mlineCnt] = yout(y);
	mlineCnt++;
}

void IdrawMlineEnd ()
{ 
	int i;

	fprintf (psFp, "%%I %d\n", mlineCnt);
	for (i=0; i<mlineCnt; i++)
	 { fprintf (psFp, "%d %d\n", mlineXbuf[i], mlineYbuf[i]);
	 }
	fprintf (psFp, "%d MLine\n", mlineCnt);
	fprintf (psFp, "End\n");
}

void IdrawClose()
{
	fprintf (psFp, "\nEnd %%I eop\n\n");
}
