#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mathx.h>
#include <argDef.h>
#ifdef HAS_GRF4
#include <grf4.h>
#endif
#include <RxTimeval.h>
#include <RxDlist.h>
#include <RxVecd.h>

#include <RcPathTrajectory.h>
#include <RcPathKnots.h>

#include <RcCurveX.h>
#include <postScript.h>
#include <getWord.h>

#include <RxPumaKin.h>
#include <RxP2RKin.h>

#define NUM_COLUMNS	80
#define CRV_UNDEF	HUGE_VAL

int noGraphics = 0;
int plotWidth = 512;
int plotHeightFxn = 256;
int plotHeightVel = 128;
char *plotItems = "01s";
char *validPlotItems = "012345rps";
double maxJntDel = -1;
double maxJntVel = 180.0;
double maxJntAcc = 500.0;
double maxPosVel = -1;	/* desired translational velocity, mm/sec. */
double maxPosAcc = -1;	/* desired translational accel, mm/sec^2 */
double maxRotVel = -1;		/* desired rotational velocity, rad/sec. */
double maxRotAcc = -1;		/* desired rotational accel, rad/sec^2 */
int clipAccel = 0;
int clipE = 0;
int usePuma = 0;
int useP2R = 0;
int useCRS = 0;
int printSplits = 0;
int printKnots = 0;
int printE = 0;
int printJnts = 0;
int printTraj = 0;
int printPathError = 0;
int usePathLen = 0;
int timeInfo = 0;
int foo = 0;
int doRaw = 0;

double minPathDel = (1e-6);
double charDist = -1;
double maxAccErrRatio = 0.5;

double l1 = 10;
double l2 = 10;
int conf1 = -1;
int conf2 = -1;
RxGenKin *kin;
int numj;

double posTol = -1;
double rotTol = 0.1;
double interval = 0.05;

double psXoff = 0;
double psYoff = 0;
double psScale = 1.0;
double psWidth = 4.75;          /* width of PostScript output */   
double psHeight = 5.0;          /* height of PostScript output */
double psPlotHeight = -1;      /* height of PostScript plot window */
double psPlotWidth = 4.75;      /* width of PostScript plot window */
double psViewAng = 0.0;         /* angle of inclination for PostScript plot */
double psPlane = 1e6;           /* position to draw psPlane in */
char *psFileName = NULL;        /* file name for postScript plot */

#define MAXSEGS	10

double crvTime = -1.0;
RxVecd3 crv[MAXSEGS*4];
RxVecd3 rpy[MAXSEGS*2];
int numRpts = 1;
double segLens[MAXSEGS];

#define FXN_MARGIN	5
#define TICK_HEIGHT	4

int winw;
int winh;

ArgDefList *argDefs;
char **argvec;
int printHelp;

RcPathTrajectory *pathTraj;

// linked list of plot items

class PlotItem : public RxDnodeBase
{
public:
	char desc;
	RX_DNODE_DECLS (PlotItem);
};

typedef RxDlist<PlotItem> PlotList;

// linked list of plot values

extern int numj;

class TimeSamp : public RxDnodeBase, public RxVecd 
{
public:
	TimeSamp () : RxVecd(numj+3)
	 {
	 }
	double t;
	RX_DNODE_DECLS (TimeSamp);
};

typedef RxDlist<TimeSamp> TimeList;

PlotList plotList;
TimeList tlist;

int fwdKinFxn (RxTrsf *X, const RxVecd &jnts)
{
	kin->fwdK (X, (int*)0, jnts);
	return 1;
}

int invKinFxn (RxVecd *jnts, RxTrsf *X, RxVecd *old, double s)
{
	RcCurveXPoint cpnt;
	int status;
	int rcode = 0;
	int conf;

	cpnt.set (s);
	conf = (cpnt.seg == 0 ? conf1 : conf2);

	if (X)
	 { *X = cpnt.X;
	 }
	if (jnts)
	 { status = kin->invK (jnts, cpnt.X, conf, old);
	   if (status & RX_GENKIN_INDETERMINATE)
	    { rcode |= RC_PATHKIN_INDETERMINATE;
	    }
	   if (status & RX_GENKIN_BOUNDARY_LIMIT)
	    { rcode |= RC_PATHKIN_BOUNDARY_LIMIT;
	    }
	 }
	return (rcode);
}

void parsePlotItems (char *str)
{
	PlotItem *item;
	char *cp;

	plotList.init();
	for (cp=str; *cp != '\0'; cp++)
	 { if (strchr (validPlotItems, *cp) == NULL)
	    { fprintf (stderr, "%s: plotItems: bogus item %c\n", 
			       *argvec, *cp);
	      exit (1);
	    }
	   item = new PlotItem();
	   if (*cp == 'p')
	    { item->desc = '0'+numj;
	    }
	   else if (*cp == 'r')
	    { item->desc = '0'+numj+1;
	    }
	   else if (*cp == 's')
	    { item->desc = '0'+numj+2;
	    }
	   else
	    { item->desc = *cp;
	    }
	   plotList.addTail (item);
	 }
}

#ifdef HAS_GRF4

void onResize (Grf4Event *ev, int)
{
	winw = ev->val1;
	winh = ev->val2;
	grf4Ortho (0, winw, 0, winh, -1.0, 1.0, /*aspectRatio=*/1);
}

void drawFxns(int plothFxn)
{
	RcPathKnot *knot, *tail;

	PlotItem *item;
	double scaleY;
	double scaleX;
	double x;
	int idx;
	int jx;

	tail = pathTraj->lastKnot();

	scaleY = plothFxn/(4*M_PI);
	scaleX = (winw-2*FXN_MARGIN)/tail->coordx(JS);

	grf4Move2 (FXN_MARGIN,0);
	grf4Draw2 (winw-FXN_MARGIN, 0);

	for (knot=pathTraj->firstKnot(); knot; knot=knot->next())
	 { x = FXN_MARGIN + scaleX*knot->coordx(JS);
	   grf4Move2 (x, -TICK_HEIGHT);
	   grf4Draw2 (x,  TICK_HEIGHT);
	 }

	for (item=plotList.head(); item; item=item->next())
	 { jx = item->desc - '0';
	   if (jx < numj)
	    { for (knot=pathTraj->firstKnot(); knot; knot=knot->next())
	       { x = FXN_MARGIN + scaleX*knot->coordx(JS);
		 if (knot==pathTraj->firstKnot())
		  { grf4Move2 (x, scaleY*knot->coordx(jx));
		  }
		 else
		  { grf4Draw2 (x, scaleY*knot->coordx(jx));
		  }
	       }
	    }
	 }
}

void drawItem (PlotItem *item, int plothVel)
{
	double halfheight;	// no. pixels for top or bottom half of plot
	double scaleY;
	double scaleX;
	double x;

	grf4Move2 (FXN_MARGIN,0);
	grf4Draw2 (winw-FXN_MARGIN, 0);


	if ('0' <= item->desc && item->desc <= '0'+numj)
	 { TimeSamp *samp;
	   TimeSamp *prev;
	   int jx = item->desc-'0';

	   halfheight = (plothVel/2-FXN_MARGIN);
#if 0	   
	   scaleY = halfheight/(jx==JP ? maxPathVel : maxJntVel);
	   scaleX = scaleY*(jx==JP ? maxPathAcc : maxJntAcc);
#else
	   if (jx != JP)
	    { scaleY = halfheight/maxJntVel;
	      scaleX = scaleY*(maxJntAcc);
	    }
	   else
	    { scaleX = halfheight*maxJntAcc/maxJntVel;
	      scaleY = scaleX/maxPosAcc;
	    }
#endif
	   for (x=scaleX*interval+FXN_MARGIN; x<=winw-FXN_MARGIN; 
		x+=scaleX*interval)
	    { grf4Move2 (x, -TICK_HEIGHT);
	      grf4Draw2 (x,  TICK_HEIGHT);
	    }
	   if (tlist.head())
	    { grf4Move2 (FXN_MARGIN,0);
	      prev = tlist.head();
	      for (samp=prev->next(); samp; samp=samp->next())
	       { if (scaleX*samp->t > winw-FXN_MARGIN)
		  { break;
		  }
		 grf4Draw2 (scaleX*samp->t+FXN_MARGIN,
			    scaleY*((*samp)(jx)-(*prev)(jx))/interval);
		 prev = samp;
	       }
	    }
	 }
}

void doDraw(void *arg)
{
	int plothFxn;
	int plothVel;
	int origHeight;
	int n;
	PlotItem *item;

	origHeight = plotHeightFxn+plotHeightVel*plotList.length();
	plothFxn = (int)(plotHeightFxn/(double)(origHeight) * winh);
	plothVel = (int)(plotHeightVel/(double)(origHeight) * winh);

	grf4Clear();
	grf4PushTrans();
	n = plothFxn/2+plothVel*plotList.length();
	grf4Translate (0, n, 0);

	drawFxns(plothFxn);
	grf4Translate (0, -(plothFxn+plothVel)/2, 0);
	for (item=plotList.head(); item; item=item->next())
	 { drawItem(item, plothVel);
	   grf4Translate (0, -plothVel, 0);
	 }
	grf4PopTrans();
	grf4Flush();
}

#endif

void drawIdrawPlots (int, double, double height, double trackHeight)
{
	PlotItem *item;

	double scaleX;
	double scaleY;
	TimeSamp *samp;
	TimeSamp *prev;
	int track = 0;
	int k;

	scaleY = trackHeight/maxJntVel;
	scaleX = scaleY*(maxJntAcc);	

	if (tlist.head())				   
	 { for (item=plotList.head(); item; item=item->next())
	    { double yval, yoff;
	      int jx = item->desc-'0';

	      if (jx != JP)
	       { scaleY = trackHeight/maxJntVel;
		 scaleX = scaleY*(maxJntAcc);
	       }
	      else
	       { scaleX = trackHeight*maxJntAcc/maxJntVel;
		 scaleY = scaleX/maxPosAcc;
	       }

	      yoff = -(track+0.5)*height/plotList.length();
	      IdrawMlineBegin (IDRAW_PAT_NONE);
	      IdrawMlinePut (0, yoff);
	      prev = tlist.head();
	      for (samp=prev->next(); samp; samp=samp->next())
	       { yval = ((*samp)(jx)-(*prev)(jx))/interval;

		 IdrawMlinePut (scaleX*samp->t, 
				scaleY*yval + yoff);
		 k++;
		 prev = samp;
	       }
	      IdrawMlineEnd ();

	      track++;
	    }
	 }
}

void argListAdd (int *argcp, char ***argvp, char *name)
{
	char ibuf[256];
	char *cppArgVec[256];
	int cppN;
	int ac, ac0;
	char **av, **av0;
	char **ap;

	if (!name)
	 { return;
	 }
	ac0 = *argcp;
	av0 = *argvp;

	cppN = 0;
	while (ac0 > 0)
	 { if (strncmp (*av0, "-D", 2) == 0)
	    { cppArgVec[cppN++] = *av0;
	    }
	   ac0--; av0++;
	 }
	cppArgVec[cppN] = 0;

	ac0 = *argcp;
	av0 = *argvp;

	getWordOpen(name, cppArgVec);
	for (ac=ac0-cppN; getWord (ibuf)==1; ac++)
	   ;
	av = (char**)malloc(ac*sizeof(char**));
	getWordReopen();
	for (ap=av; getWord (ibuf)==1; )
	 { *ap++ = strcpy ((char*)malloc(strlen(ibuf)+1), ibuf);
	 }
	getWordClose ();
	while (ac0 > 0)
	 { if (strncmp (*av0, "-D", 2) != 0)
	    { *ap++ = *av0;
	    }
	   ac0--; av0++;
	 }
	*argcp = ac;
	*argvp = av;
}

void calcPathError (double &maxPosErr, double &maxRotErr, TimeSamp *samp)
{
	RxTrsf Xd;
	RxTrsf X;
	RxVecd Jn(6);
	int status;

	double s = (*samp)(JS);
	double rdist;
	status = invKinFxn (&Jn, &Xd, tlist.tail(), s);
	if (status & RC_PATHKIN_BOUNDARY_LIMIT)
	 { fwdKinFxn (&Xd, Jn);
	 }
	fwdKinFxn (&X, *samp);
	Xd.mulI (Xd, X);
// printf ("%g\n", Xd.p.norm2());
	if (Xd.p.norm2() > maxPosErr)
	 { maxPosErr = Xd.p.norm2();
	 }
	rdist = Xd.R.o.x()*Xd.R.o.x() + Xd.R.a.x()*Xd.R.a.x() +
		Xd.R.a.y()*Xd.R.a.y();
	if (rdist > maxRotErr)
	 { maxRotErr = rdist;
	 }
}

void doComputations ()
{
	RxTrsf Xp[MAXSEGS*4];
	int numSegs;
	int i, k;

	if (segLens[0] == 1.0)
	 { numSegs = 1;
	 }
	else
	 { segLens[1] = 1.0;
	   numSegs = 2;
	 }
	for (i=0; i<numSegs; i++)
	 { if (crv[i*4+0].x() != CRV_UNDEF && crv[i*4+3].x() != CRV_UNDEF &&
	       crv[i*4+1].x() == CRV_UNDEF && crv[i*4+2].x() == CRV_UNDEF)
	    { crv[i*4+1] = crv[i*4+0] + (1.0/3.0)*(crv[i*4+3]-crv[i*4+0]);
	      crv[i*4+2] = crv[i*4+0] + (2.0/3.0)*(crv[i*4+3]-crv[i*4+0]);
	    }
	   for (k=0; k<4; k++)
	    { if (crv[i*4+k].x() == CRV_UNDEF)
	       { Xp[i*4+k].p.setZero();
	       }
	      else
	       { Xp[i*4+k].p = crv[i*4+k];
	       }
	    }
	   for (k=0; k<numRpts; k++)
	    { if (rpy[i*4+k].x() == CRV_UNDEF)
	       { rpy[i*4+k].setZero();
	       }
	      Xp[i*4+k].R.setRpyD (rpy[i*4+k].x(), rpy[i*4+k].y(),
				   rpy[i*4+k].z());
	    }
	 }

	RcCurveXPoint::setSegDefs (Xp, segLens, numSegs, numRpts);
	RcPathKnot::reserve (numj, 1000);

	pathTraj->initPath (0, 1, 20);

	if (timeInfo)
	 { TIMEIT (pathTraj->dividePath(), 1);
	 }
	else
	 { pathTraj->dividePath();
	 }
	pathTraj->calcPathLength();
	if (!printTraj)
	 { printf ("num knots = %d\n", pathTraj->length());
	 }

	if (timeInfo)
	 { TIMEIT (pathTraj->calcElimits(); pathTraj->calcEnergies(), 1);
	 }
	else if (!doRaw)
	 { pathTraj->calcElimits();
	   pathTraj->calcEnergies();
	 }
	else
	 { pathTraj->calcConstSpeedEnergies();
	 }
#if 1
	 { double lam, len;
	   RxVecd3 pos;
	   TimeSamp *samp;
	   int status;
	   int j;

	   tlist.init();

	   samp = new TimeSamp;
	   pathTraj->integrateInit (samp);
	   samp->t = 0;
	   tlist.addTail (samp);

	   double maxPosErr = 0;
	   double maxRotErr = 0;

	   do
	    { 
	      samp = new TimeSamp;
	      status = pathTraj->integrate (samp, interval);
	      samp->t = tlist.tail()->t + interval;

	      calcPathError (maxPosErr, maxRotErr, samp);
	      tlist.addTail (samp);
	    }
	   while (status);

	   if (printPathError)
	    { printf ("maxPosErr=%g maxRotErr=%g\n", maxPosErr, maxRotErr);
	    }
	 }
#endif
#if 0
	if (timeInfo)
	 { RcPathKnot *next;
	   RcPathKnot *newk;
	   KnotList testlist;
	   RxVecd ang(6);
	   RxTrsf T6;
	   int conf;
	   double *store;
	   RcCurveXPoint *cpntp;

	   testlist.init();
	   ang.setZero();
	   ang(1) = DTOR*17;
	   ang(2) = DTOR*21;
	   ang(3) = DTOR*33;
	   ang(4) = 1.0;

	   knot = new RcPathKnot (*kin, 0, (RxVecd*)0);
	   next = new RcPathKnot (*kin, 0.01, (RxVecd*)0);
	   TIMEIT (newk = new RcPathKnot (*kin, 0.005), 500);
	   TIMEIT (knot->setPair (newk); newk->setPair (next), 1000);
	   testlist.addTail (knot);
	   testlist.addTail (newk);
	   TIMEIT (knot->posTolOK (newk), 1000);
	   TIMEIT (kin->fwdK (&T6, &conf, ang), 1000);
	   TIMEIT (kin->invK (&ang, T6, conf, &ang), 1000);
	   TIMEIT (knot->cpnt.set(0), 1000);
	   TIMEIT (RcPathKnot kk, 1000);
	   TIMEIT (newk = RcPathKnot::allocK(), 500);
	   TIMEIT (newk = new RcPathKnot(1), 500);
	   TIMEIT (store = (double*)malloc(8*sizeof(double)), 1000);
	   TIMEIT (ang.init (6, store), 10000);
	   TIMEIT (Polytope2D E, 10000);
	   TIMEIT (RxVecd dcxdsR, 10000)
	   TIMEIT (RxTrsf X, 10000);
	   TIMEIT (cpntp = new RcCurveXPoint, 10000);
	 }
#endif
	fflush (stdout);
}

void drawIdraw (FILE *fp, double w, double h, double pw, double ph)
{
	double sAng, cAng;
	int i;

	IdrawInit (fp);

	w *= PPI;
	h *= PPI;
	ph *= PPI;
	pw *= PPI;

	sAng = sin(psViewAng);
	cAng = cos(psViewAng);

	IdrawDashes (IDRAW_SOLID);

	 { double plotTrackH;
	   double plotMarginX;
	   double plotMarginY;
	   double y0;

	   IdrawBaseSet ((PAGE_WIDTH-pw)/2, PAGE_LENGTH-h);
	   IdrawScaleSet (1, 1);

	   plotTrackH = ph/(2*1.2*plotList.length());
	   plotMarginY = 0.05*plotTrackH;
	   plotMarginX = 0.0;

	   y0 = -ph/(2*plotList.length());
	   for (i=0; i<plotList.length(); i++)
	    { IdrawMove (plotMarginX, y0-plotTrackH);
	      IdrawArrowLine (plotMarginX, y0+plotTrackH);
	      IdrawMove (plotMarginX, y0);
	      IdrawArrowLine (pw, y0);
	      y0 -= ph/plotList.length();
	    }
	   IdrawBaseTrans (plotMarginX, 0.0);
	   IdrawLineWidth (IDRAW_WIDTH_2);
	   drawIdrawPlots (0, pw-plotMarginX, ph, plotTrackH);
	 }
	IdrawClose ();	
}

void psWriteOutput (char *)
{
	FILE *fp;

	if ((fp = fopen(psFileName, "w")) == NULL)
	 { fprintf (stderr, "Can't open %s\n", psFileName);
	   perror ("fopen");
	   exit (1);
	 }
	printf ("Drawing IDRAW PostScript ... \n");
	drawIdraw (fp, psWidth, psHeight, psPlotWidth, psPlotHeight);
	printf ("Done.\n");
	fclose (fp);
}

main(int argc, char **argv)
{
	int opts = 0;
	int i;

	for (i=0; i<4*MAXSEGS; i++)
	 { crv[i].x() = CRV_UNDEF;
	 }
	for (i=0; i<2*MAXSEGS; i++)
	 { rpy[i].x() = CRV_UNDEF;
	 }
	segLens[0] = 1.0;

	argDefs = argDefCreate();
	if (argDefAdd (argDefs,
"-help %v,1", &printHelp,
"-raw %v,1,		\"raw energies with no algorithm applied\"", &doRaw,
"-noGraphics %v,1, 	\"don't do any graphic plotting\"", &noGraphics,
"-clipAccel %v,1, 	\"clip accelerations\"", &clipAccel,
"-clipE %v,1, 	\"clip energy\"", &clipE,
"-printKnots %v,1, 	\"print knot info\"", &printKnots,
"-printSplits %v,1, 	\"print knot info\"", &printSplits,
"-printE %v,1, 		\"print energy info\"", &printE,
"-printJnts %v,1, 	\"print joint info\"", &printJnts,
"-printTraj %v,1, 	\"print trajectory\"", &printTraj,
"-printError %v,1, 	\"print path error\"", &printPathError,
"-foo %i,, 		\"foo (used for debugging)\"", &foo,
"-usePathLen %v,1, 	\"use path len as main coordinate\"", &usePathLen,
"-timeInfo %v,1,        \"determine timing information\"", &timeInfo,
"-usePuma %v,1, 	\"use PUMA robot\"", &usePuma,
"-useCRS %v,1, 		\"use CRS robot\"", &useCRS,
"-useP2R %v,1, 		\"use Planar 2R robot\"", &useP2R,
"-plotWidth %i,[0,),	\"width of plot window (pixels)\"", &plotWidth,
"-plotHeightFxn %i,[0,), \"height of kinematic fxn plot (pixels)\"", 
		&plotHeightFxn,
"-plotHeightVel %i,[0,), \"height of velocity plots (pixels)\"", 
		&plotHeightVel,
"-plotItems %s,	\"which joints to plot\"", &plotItems,
"-crvA0 %3lf,,		\"curve vector Ap0\"", &crv[0].x(),
"-crvA1 %3lf,,		\"curve vector Ap1\"", &crv[1].x(),
"-crvA2 %3lf,,		\"curve vector Ap2\"", &crv[2].x(),
"-crvA3 %3lf,,		\"curve vector Ap3\"", &crv[3].x(),
"-rpyA0 %3lf,,		\"curve vector Ap0\"", &rpy[0].x(),
"-rpyA1 %3lf,,		\"curve vector Ap1\"", &rpy[1].x(),
"-crvSA %lf,,		\"curve segment A length\"", &segLens[0],
"-crvB0 %3lf,,		\"curve vector Bp0\"", &crv[4+0].x(),
"-crvB1 %3lf,,		\"curve vector Bp1\"", &crv[4+1].x(),
"-crvB2 %3lf,,		\"curve vector Bp2\"", &crv[4+2].x(),
"-crvB3 %3lf,,		\"curve vector Bp3\"", &crv[4+3].x(),
"-rpyB0 %3lf,,		\"curve vector Ap0\"", &rpy[4+0].x(),
"-rpyB1 %3lf,,		\"curve vector Ap1\"", &rpy[4+1].x(),
"-numRpts %i,[1,2],	\"number of rotation points in curve spec\"", &numRpts,
"-crvTime %lf,,		\"nominal curve travel time\"", &crvTime,
"-interval %lf,,	\"sample interval (msec)\"", &interval,
"-l1 %lf,(0,),		\"length of link 1\"", &l1,
"-l2 %lf,(0,),		\"length of link 2\"", &l2,
"-conf1 %i,,		\"configuration\"", &conf1,
"-conf2 %i,,		\"configuration on 2nd path seg\"", &conf2,
"-posTol %lf,(0,),	\"path position tolerence (Taylor)\"", &posTol,
"-charDist %lf,(0,),	\"characteristic distance\"", &charDist,
"-maxJntDel %lf,(0,),	\"max joint diff between knots (deg)\"", &maxJntDel,
"-maxJntVel %lf,(0,),	\"max joint velocity (deg/sec)\"", &maxJntVel,
"-maxJntAcc %lf,(0,),	\"max joint acceleration (deg/sec^2)\"", &maxJntAcc,
"-maxPosVel %lf,(0,),	\"max path velocity (mm/sec)\"", &maxPosVel,
"-maxPosAcc %lf,(0,),	\"max path acceleration (mm/sec^2)\"", &maxPosAcc,
"-maxRotVel %lf,(0,),	\"max path velocity (deg/sec)\"", &maxRotVel,
"-maxRotAcc %lf,(0,),	\"max path acceleration (deg/sec^2)\"", &maxRotAcc,
"-maxAccErrRatio %lf,(0,), \"max acceleration error (0-1)\"", &maxAccErrRatio,
"-minPathDel %lf,(0,),	\"min diff for path parameter (s)\"", &minPathDel,
"-psFileName %s,	\"file name for PostScript plotting\"", &psFileName,
"-psWidth %lf,(0,),	\"width of PostScript output\"", &psWidth,
"-psHeight %lf,(0,),	\"height of PostScript output\"", &psHeight,
"-psXoff %lf,,		\"robot x display offset for PostScript\"", &psXoff,
"-psYoff %lf,,		\"robot y display offset for PostScript\"", &psYoff,
"-psScale %lf,(0,),	\"robot display scale for PostScript\"", &psScale,
"-psPlotHeight %lf,(0,),\"height of PostScript plot window\"", &psPlotHeight,
"-psPlotWidth %lf,(0,), \"width of PostScript plot window\"", &psPlotWidth,
"-psViewAng %lf,, 	\"angle of inclination for PostScript plot\"", 
			&psViewAng,
"-psPlane %lf,, 	\"height of plane in PostScript figure\"", &psPlane,
	   		0) < 0)
	 { fprintf (stderr, "%s\n", argDefErrorStr());
	   exit (1);
	 }

	argvec = argv;
	argc--; argv++;

	argListAdd (&argc, &argv, "testbed.opt");
	while (argc > 0)
	 { int n;
	   if ((n = argDefMatch (argDefs, argc, argv)) <= 0)
	    { if (n < 0)
	       { fprintf (stderr, "%s\n", argDefErrorStr());
	       }
	      else
	       { argDefPrintUsage (stderr, argDefs, NUM_COLUMNS);
	       }
	      exit (1);
	    }
	   argc -= n; argv += n;
	 }

	if (printHelp)
	 { argDefPrintHelp (stdout, argDefs);
	   exit (0);
	 }

	if (crvTime == -1.0)
	 { crvTime = 1.0;
	 }

	if (psPlotHeight == -1)
	 { psPlotHeight = psHeight - psWidth;
	 }
	if (usePuma)
	 { kin = new RxPumaKin ();
	   numj = 6;

	   // default configuration is luf 
	   if (conf1 == -1) conf1 = RX_PUMAKIN_CONF_FLIP;
	   if (conf2 == -1) conf2 = RX_PUMAKIN_CONF_FLIP;
	   if (charDist == -1) charDist = 1000;
	   if (maxPosVel == -1) maxPosVel = 400;
	   if (maxPosAcc == -1) maxPosAcc = 2500;
	   if (maxRotVel == -1) maxRotVel = 180;
	   if (maxRotAcc == -1) maxRotAcc = 500;
	   if (posTol == -1) posTol = charDist/10000.0;
	 }
	else if (useCRS)
	 { kin = new RxPumaKin ("crsA460");
	   numj = 6;

	   // default configuration is luf 
	   if (conf1 == -1) conf1 = RX_PUMAKIN_CONF_FLIP;
	   if (conf2 == -1) conf2 = RX_PUMAKIN_CONF_FLIP;
	   if (charDist == -1) charDist = 1000;
	   if (maxPosVel == -1) maxPosVel = 400;
	   if (maxPosAcc == -1) maxPosAcc = 2500;
	   if (maxRotVel == -1) maxRotVel = 180;
	   if (maxRotAcc == -1) maxRotAcc = 500;
	   if (posTol == -1) posTol = charDist/10000.0;
	 }
	else if (useP2R)
	 { DHParams dhp[2];

	   kin = new RxP2RKin();
	   numj = 2;
	   kin->get (dhp);
	   dhp[0].a = l1;
	   dhp[1].a = l2;
	   kin->set (dhp, numj);

	   if (conf1 == -1) conf1 = RX_PLANAR2RKIN_CONF_DOWN;
	   if (conf2 == -1) conf2 = RX_PLANAR2RKIN_CONF_DOWN;
	   if (charDist == -1) charDist = l1+l2;
	   if (maxPosVel == -1) maxPosVel = 10;
	   if (maxPosAcc == -1) maxPosAcc = 100;
	   if (maxRotVel == -1) maxRotVel = 180;
	   if (maxRotAcc == -1) maxRotAcc = 500;
	   if (posTol == -1) posTol = charDist/10000.0;
	 }
	else
	 { fprintf (stderr, "No robot type specified\n");
	   exit (1);
	 }
	psViewAng *= DTOR;

	rotTol *= DTOR;
	maxJntVel *= DTOR;
	maxJntAcc *= DTOR;
	maxJntDel *= DTOR;
	maxRotVel *= DTOR;
	maxRotAcc *= DTOR;

	printf ("posTol=%g\n", posTol);

	pathTraj = new RcPathTrajectory (numj);
	pathTraj->init (numj);

	pathTraj->setFwdKinFxn (fwdKinFxn);
	pathTraj->setInvKinFxn (invKinFxn);

	pathTraj->setMaxCartVel (maxPosVel, maxRotVel);
	pathTraj->setMaxCartAcc (maxPosAcc, maxRotAcc);
	pathTraj->setMaxJntVel (maxJntVel, maxPosVel);
	pathTraj->setMaxJntAcc (maxJntAcc, maxPosAcc);

	pathTraj->setMinPathDel (minPathDel);
	pathTraj->setPathTol (posTol, rotTol);
	pathTraj->setMaxAccErrRatio (maxAccErrRatio);
	pathTraj->setCharDist (charDist);

	if (maxJntDel > 0)
	 { RxVecd maxDel(numj+3);
	   int j;

	   for (j=0; j<numj; j++)
	    { maxDel(j) = maxJntDel;
	    }
	   for (j=numj; j<numj+3; j++)
	    { maxDel(j) = -1;
	    }
	   pathTraj->setMaxDel (maxDel);
	 }
	else
	 { RxVecd maxDel(numj+3);
	   int j;

	   pathTraj->getMaxDel (&maxDel);
#if 0
	   printf ("MaxDel:\n");
	   for (j=0; j<numj+3; j++)
	    { printf (" %7.2f", (j<numj ? RTOD*maxDel(j) : maxDel(j)));
	    }
	   printf ("\n");
#endif
	 }

	if (clipAccel)
	 { opts |= RC_PATHTRAJ_CLIP_ACCEL;
	 }
	if (clipE)
	 { opts |= RC_PATHTRAJ_CLIP_E;
	 }
	if (printSplits)
	 { opts |= RC_PATHTRAJ_PRINT_SPLITS;
	 }
	if (printKnots)
	 { opts |= RC_PATHTRAJ_PRINT_KNOTS;
	 }
	if (printE)
	 { opts |= RC_PATHTRAJ_PRINT_E;
	 }
	if (printJnts)
	 { opts |= RC_PATHTRAJ_PRINT_JNTS;
	 }
	if (usePathLen)
	 { opts |= RC_PATHTRAJ_USE_PATH_LEN;
	 }
	pathTraj->setOptions (opts);

	parsePlotItems (plotItems);
	winw = plotWidth;
	winh = plotHeightFxn + plotHeightVel*plotList.length();

#ifdef HAS_GRF4
	if (!noGraphics)
	 { grf4PlaceWindow (200, 400, winw, winh);
	   grf4Init ("DAO", GRF_WHITE, GRF_BLACK, GRF_2DONLY | GRF_X11);

	   grf4AttachEvent (GRF_RESIZE_EVENT, onResize, 0);

	   grf4Ortho (0, winw, 0, winh, -1.0, 1.0, /*aspectRatio=*/1);
	   grf4SetDrawFxn (doDraw, (void*)0);
	 }
#endif
	doComputations();
	fflush(stdout);

	if (printTraj)
	 { TimeSamp *samp;
	   if (tlist.head())
	    { for (samp=tlist.head(); samp; samp=samp->next())
	       { RxVecd deg(numj);
		 deg = (*samp)*RTOD;
		 if (foo)
		  { deg(2) += 360.0;
		  }
		 deg.print ("%8.3f\n");
	       }
	    }
	   fflush (stdout);
	 }
#ifdef HAS_GRF4
	if (!noGraphics)
	 { doDraw(0);
	   grf4Wait (-1);
	 }
#endif
	if (psFileName)
	 { psWriteOutput (psFileName);
	 }
}
